// C++ program file box_utils.cpp

/*
  By Nathan C. Hearn
     July 31, 2008

  Box-related utilities.
*/


#include "box_utils.hpp"
#include <cmath>
#include <vector>


bool sub_region_enclosed(const std::vector<double> & main_region_min_coords,
			 const std::vector<double> & main_region_max_coords,
			 const std::vector<double> & sub_region_min_coords,
			 const std::vector<double> & sub_region_max_coords)
  {
  bool ret_val = false;

  // ERROR CHECKING???

  const unsigned int dims = main_region_min_coords.size();

  if (dims > 0)
    {
    ret_val = true;  // Until it isn't ...

    for (unsigned int i = 0; i < dims; i++)
      {
      const double main_min_coord = main_region_min_coords[i];
      const double main_max_coord = main_region_max_coords[i];

      const double sub_min_coord = sub_region_min_coords[i];
      const double sub_max_coord = sub_region_max_coords[i];

      if ((sub_min_coord < main_min_coord) || (sub_max_coord > main_max_coord))
	ret_val = false;
      }
    }

  return ret_val;
  }


bool sub_region_enclosed(const std::vector<bool> & periodic_axis_flags,
			 const std::vector<double> & main_region_min_coords,
			 const std::vector<double> & main_region_max_coords,
			 const std::vector<double> & sub_region_min_coords,
			 const std::vector<double> & sub_region_max_coords,
			 const std::vector<double> & periodic_min_coords,
			 const std::vector<double> & periodic_max_coords)
  {
  // Version of sub_region_enclosed using periodic boundaries

  bool ret_val = false;

  // ERROR CHECKING???

  const unsigned int dims = periodic_axis_flags.size();

  if (dims > 0)
    {
    ret_val = true;  // Until it isn't ...

    unsigned int i = 0;

    while (ret_val && (i < dims))
      {
      if (periodic_axis_flags[i])
	{
	const double main_min_coord = main_region_min_coords[i];
	const double main_max_coord = main_region_max_coords[i];

	const double main_width = main_max_coord - main_min_coord;

	const double sub_min_coord = sub_region_min_coords[i];
	const double sub_max_coord = sub_region_max_coords[i];

	const double sub_width = sub_max_coord - sub_min_coord;

	const double vol_min_coord = periodic_min_coords[i];
	const double vol_max_coord = periodic_max_coords[i];

	const double vol_width = vol_max_coord - vol_min_coord;

	const double oneOver_vol_width = 1.0 / vol_width;

	// ERROR CHECKING???

	// Put both regions on same "side" of (and closest to) vol_min_coord

	const double main_min_vol_min_delta = main_min_coord - vol_min_coord;

	const double main_shift = -1.0 * vol_width
	  * std::floor(main_min_vol_min_delta * oneOver_vol_width);

	double main_adj_min_coord = main_min_coord + main_shift;

	if (main_adj_min_coord < vol_min_coord)
	  main_adj_min_coord += vol_width;
	else if (!(main_adj_min_coord < vol_max_coord))
	  main_adj_min_coord -= vol_width;

	const double main_adj_max_coord = main_adj_min_coord + main_width;
      
	const double sub_min_vol_min_delta = sub_min_coord - vol_min_coord;

	const double sub_shift = -1.0 * vol_width
	  * std::floor(sub_min_vol_min_delta * oneOver_vol_width);

	double sub_adj_min_coord = sub_min_coord + sub_shift;

	if (sub_adj_min_coord < vol_min_coord)
	  sub_adj_min_coord += vol_width;
	else if (!(sub_adj_min_coord < vol_max_coord))
	  sub_adj_min_coord -= vol_width;

	const double sub_adj_max_coord = sub_adj_min_coord + sub_width;

	if ((sub_adj_min_coord < main_adj_min_coord) 
	    || (sub_adj_max_coord > main_adj_max_coord))
	  ret_val = false;
	}
      else
	{
	const double main_min_coord = main_region_min_coords[i];
	const double main_max_coord = main_region_max_coords[i];

	const double sub_min_coord = sub_region_min_coords[i];
	const double sub_max_coord = sub_region_max_coords[i];

	if ((sub_min_coord < main_min_coord) 
	    || (sub_max_coord > main_max_coord))
	  ret_val = false;
	}

      i++;  // Next element
      }
    }

  return ret_val;
  }


bool regions_exclusive(const std::vector<double> & region1_min_coords,
		       const std::vector<double> & region1_max_coords,
		       const std::vector<double> & region2_min_coords,
		       const std::vector<double> & region2_max_coords)
  {
  bool ret_val = false;

  // NOTE: If borders are shared, function will return false

  // ERROR CHECKING???

  const unsigned int dims = region1_min_coords.size();

  if (dims > 0)
    {
    for (unsigned int i = 0; i < dims; i++)
      {
      const double r1_min_coord = region1_min_coords[i];
      const double r1_max_coord = region1_max_coords[i];

      const double r2_min_coord = region2_min_coords[i];
      const double r2_max_coord = region2_max_coords[i];

      if ((r1_max_coord < r2_min_coord) || (r1_min_coord > r2_max_coord))
	ret_val = true;  // Regions do not overlap or touch at border
      }
    }

  return ret_val;
  }


bool regions_exclusive(const std::vector<bool> & periodic_axis_flags,
		       const std::vector<double> & region1_min_coords,
		       const std::vector<double> & region1_max_coords,
		       const std::vector<double> & region2_min_coords,
		       const std::vector<double> & region2_max_coords,
		       const std::vector<double> & periodic_min_coords,
		       const std::vector<double> & periodic_max_coords)
  {
  // Version of regions_exclusive using periodic boundaries

  bool ret_val = false;

  // NOTE: If borders are shared, function will return false

  // ERROR CHECKING???

  const unsigned int dims = periodic_axis_flags.size();

  if (dims > 0)
    {
    unsigned int i = 0;

    while ((!ret_val) && (i < dims))
      {
      if (periodic_axis_flags[i])
	{
	const double r1_min_coord = region1_min_coords[i];
	const double r1_max_coord = region1_max_coords[i];

	const double r1_width = r1_max_coord - r1_min_coord;

	const double r2_min_coord = region2_min_coords[i];
	const double r2_max_coord = region2_max_coords[i];

	const double r2_width = r2_max_coord - r2_min_coord;

	const double vol_min_coord = periodic_min_coords[i];
	const double vol_max_coord = periodic_max_coords[i];

	const double vol_width = vol_max_coord - vol_min_coord;

	const double oneOver_vol_width = 1.0 / vol_width;

	// ERROR CHECKING???

	// Put both regions on same "side" of (and closest to) vol_min_coord

	const double r1_min_vol_min_delta = r1_min_coord - vol_min_coord;

	const double r1_shift = -1.0 * vol_width
	  * std::floor(r1_min_vol_min_delta * oneOver_vol_width);

	double r1_adj_min_coord = r1_min_coord + r1_shift;

	if (r1_adj_min_coord < vol_min_coord)
	  r1_adj_min_coord += vol_width;
	else if (!(r1_adj_min_coord < vol_max_coord))
	  r1_adj_min_coord -= vol_width;

	const double r1_adj_max_coord = r1_adj_min_coord + r1_width;
      
	const double r2_min_vol_min_delta = r2_min_coord - vol_min_coord;

	const double r2_shift = -1.0 * vol_width
	  * std::floor(r2_min_vol_min_delta * oneOver_vol_width);

	double r2_adj_min_coord = r2_min_coord + r2_shift;

	if (r2_adj_min_coord < vol_min_coord)
	  r2_adj_min_coord += vol_width;
	else if (!(r2_adj_min_coord < vol_max_coord))
	  r2_adj_min_coord -= vol_width;

	const double r2_adj_max_coord = r2_adj_min_coord + r2_width;
     
	if ((r1_adj_max_coord < r2_adj_min_coord) 
	    || (r1_adj_min_coord > r2_adj_max_coord))
	  ret_val = true;  // Regions do not overlap or touch at border
	}
      else
	{
	const double r1_min_coord = region1_min_coords[i];
	const double r1_max_coord = region1_max_coords[i];

	const double r2_min_coord = region2_min_coords[i];
	const double r2_max_coord = region2_max_coords[i];

	if ((r1_max_coord < r2_min_coord) || (r1_min_coord > r2_max_coord))
	  ret_val = true;  // Regions do not overlap or touch at border
	}

      i++;  // Next element
      }
    }

  return ret_val;
  }


bool regions_overlap(const std::vector<double> & region1_min_coords,
		     const std::vector<double> & region1_max_coords,
		     const std::vector<double> & region2_min_coords,
		     const std::vector<double> & region2_max_coords)
  {
  bool ret_val = false;

  // Returns true if the area of overlap is non-zero (boundary overlap ignored)

  // ERROR CHECKING???

  const unsigned int dims = region1_min_coords.size();

  if (dims > 0)
    {
    ret_val = true;  // Until it isn't ...

    unsigned int i = 0;

    while (ret_val && (i < dims))
      {
      const double r1_min_coord = region1_min_coords[i];
      const double r1_max_coord = region1_max_coords[i];

      const double r2_min_coord = region2_min_coords[i];
      const double r2_max_coord = region2_max_coords[i];

      if (!((r1_max_coord > r2_min_coord) && (r2_max_coord > r1_min_coord)))
	ret_val = false;  // Regions do not overlap (may touch at border)

      i++;
      }
    }

  return ret_val;
  }


bool regions_overlap(const std::vector<bool> & periodic_axis_flags,
		     const std::vector<double> & region1_min_coords,
		     const std::vector<double> & region1_max_coords,
		     const std::vector<double> & region2_min_coords,
		     const std::vector<double> & region2_max_coords,
		     const std::vector<double> & periodic_min_coords,
		     const std::vector<double> & periodic_max_coords)
  {
  // Version of regions_overlap using periodic boundaries

  bool ret_val = false;

  // NOTE: The area of overlap must be non-zero for the return value to be true

  // ERROR CHECKING???

  const unsigned int dims = periodic_axis_flags.size();

  if (dims > 0)
    {
    ret_val = true;  // Until it isn't ...

    unsigned int i = 0;

    while (ret_val && (i < dims))
      {
      if (periodic_axis_flags[i])
	{
	const double r1_min_coord = region1_min_coords[i];
	const double r1_max_coord = region1_max_coords[i];

	const double r1_width = r1_max_coord - r1_min_coord;

	const double r2_min_coord = region2_min_coords[i];
	const double r2_max_coord = region2_max_coords[i];

	const double r2_width = r2_max_coord - r2_min_coord;

	const double vol_min_coord = periodic_min_coords[i];
	const double vol_max_coord = periodic_max_coords[i];

	const double vol_width = vol_max_coord - vol_min_coord;

	const double oneOver_vol_width = 1.0 / vol_width;

	// ERROR CHECKING???

	// Put both regions on same "side" of (and closest to) vol_min_coord

	const double r1_min_vol_min_delta = r1_min_coord - vol_min_coord;

	const double r1_shift = -1.0 * vol_width
	  * std::floor(r1_min_vol_min_delta * oneOver_vol_width);

	double r1_adj_min_coord = r1_min_coord + r1_shift;

	if (r1_adj_min_coord < vol_min_coord)
	  r1_adj_min_coord += vol_width;
	else if (!(r1_adj_min_coord < vol_max_coord))
	  r1_adj_min_coord -= vol_width;

	const double r1_adj_max_coord = r1_adj_min_coord + r1_width;
      
	const double r2_min_vol_min_delta = r2_min_coord - vol_min_coord;

	const double r2_shift = -1.0 * vol_width
	  * std::floor(r2_min_vol_min_delta * oneOver_vol_width);

	double r2_adj_min_coord = r2_min_coord + r2_shift;

	if (r2_adj_min_coord < vol_min_coord)
	  r2_adj_min_coord += vol_width;
	else if (!(r2_adj_min_coord < vol_max_coord))
	  r2_adj_min_coord -= vol_width;

	const double r2_adj_max_coord = r2_adj_min_coord + r2_width;
     
	if (!((r1_adj_max_coord > r2_adj_min_coord) 
	      && (r2_adj_max_coord > r1_adj_min_coord)))
	  ret_val = false;  // Regions do not overlap (may touch at border)
	}
      else
	{
	const double r1_min_coord = region1_min_coords[i];
	const double r1_max_coord = region1_max_coords[i];

	const double r2_min_coord = region2_min_coords[i];
	const double r2_max_coord = region2_max_coords[i];

	if (!((r1_max_coord > r2_min_coord) && (r2_max_coord > r1_min_coord)))
	  ret_val = false;  // Regions do not overlap (may touch at border)
	}

      i++;  // Next element
      }
    }

  return ret_val;
  }
