#!/usr/bin/env python

# generate_boxes_3d.py

# By Nathan C. Hearn
#    May 13, 2008
#
# Builds 3D Wavefront OBJ file description of transition boxes


try :
    import psyco
except ImportError :
    pass  ## No runtime optimizations for me...


dims = int(3)

## Face vertexes, based on cube with z changing fastest, and x slowest

face_verts = [ [ 0, 1, 3, 2 ],
               [ 1, 5, 3, 7 ],
               [ 7, 5, 4, 6 ],
               [ 0, 4, 5, 1 ],
               [ 0, 2, 6, 4 ],
               [ 3, 7, 6, 2 ] ]


## Bounding box defaults

default_bound_box_width_fract = 0.01


def get_box_draw_commands(min_coords, max_coords) :

    cmd_list = list()

    if (len(min_coords) != dims) :
        raise RuntimeError("Improper min coords vector")

    if (len(max_coords) != dims) :
        raise RuntimeError("Improper max coords vector")

    ## Produce the vertex commands

    for i in range(2) :

        x = None

        if (i == 0) :
            x = min_coords[0]
        else :
            x = max_coords[0]
        
        for j in range(2) :

            y = None

            if (j == 0) :
                y = min_coords[1]
            else :
                y = max_coords[1]
                
            for k in range(2) :

                z = None

                if (k == 0) :
                    z = min_coords[2]
                else :
                    z = max_coords[2]
                    
                vertex_str = "v " + str(x) + " " + str(y) + " " + str(z)

                cmd_list.append(vertex_str)

    ## Construct the face commands

    for face in face_verts :

        face_str = "f"

        for corner_index in face :
            face_str += " " + str(corner_index - 8)

        cmd_list.append(face_str)
        
    return cmd_list


def get_vector(vector_str) :

    ret_vect = list()

    vector_words = vector_str.split(",")

    for word in vector_words :
        ret_vect.append(float(word))

    return ret_vect


def get_vector_str(vector_elems) :

    ret_str = str()

    num_elems = len(vector_elems)

    if (num_elems > 0) :

        ret_str = str(vector_elems[0])

        for index in range(1, num_elems) :
            ret_str += "," + str(vector_elems[index])

    return ret_str


def main() :

    from sys import exit, stdout, stderr
    from optparse import OptionParser
    from os import path

    ## Set up the command line parser

    usage = str("%prog [options] box_list_file1 [ box_list_file2 ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_prefix_str = str("output-prefix")
    argparser.add_option("--" + output_prefix_str, type="string",
                         dest="output_prefix", help="Output file prefix")
    argparser.set_defaults(output_prefix=None)

    obj_material_str = str("obj-material")
    argparser.add_option("--" + obj_material_str, type="string",
                         dest="obj_material",
                         metavar="FILE_NAME,MATERIAL_NAME",
                         help="Comma-separated OBJ material file and name")
    argparser.set_defaults(obj_material=None)

    bound_box_str = str("bound-box")
    argparser.add_option("--" + bound_box_str, action="store_true",
                         dest="bound_box", help="Write bounding box axes file")
    argparser.set_defaults(bound_box=False)

    bound_box_width_str = str("bound-box-width")
    argparser.add_option("--" + bound_box_width_str, type="float",
                         dest="bound_box_width", metavar="WIDTH",
                         help="Bounding box line width")
    argparser.set_defaults(bound_box_width=None)

    bound_box_type_str = str("bound-box-type")
    bound_box_type_choices = ("box", "min-corner", "max-corner")
    argparser.add_option("--" + bound_box_type_str, type="choice",
                         dest="bound_box_type",
                         help="Bounding box type (box|min-corner|max-corner)",
                         choices=bound_box_type_choices)
    argparser.set_defaults(bound_box_type=None)


    ## Process the command line input

    opts, args = argparser.parse_args()

    argc = len(args)

    min_args = int(1)

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n")
        exit(1)

    argptr = int(0)

    box_file_list = list()

    while (argptr < argc) :
        box_file_list.append(args[argptr])
        argptr += 1

    output_prefix = ""

    if (opts.output_prefix is not None) :
        
        output_prefix = opts.output_prefix

        file_path, file_name = path.split(output_prefix)

        if (file_name != "") :
            if (file_name[-1] != "_") :
                output_prefix += "_"

    material_file = None
    material_name = None

    if (opts.obj_material is not None) :

        material_words = opts.obj_material.split(",")

        if (len(material_words) != 2) :
            stderr.write("\nError: OBJ material option requires "
                         + "comma-separated file and material names\n\n")
            exit(-1)

        material_file = material_words[0]
        material_name = material_words[1]

        if (material_file == "") :
            stderr.write("\nError: OBJ material file name missing\n\n")
            exit(-1)

        if (material_name == "") :
            stderr.write("\nError: OBJ material name missing\n\n")
            exit(-1)

    save_bound_box = opts.bound_box

    bound_box_width = None

    if (opts.bound_box_width is not None) :

        if (not (opts.bound_box_width > 0.0)) :
            stderr.write("\nError: Bounding box axis width [ "
                         + str(opts.bound_box_width)
                         + " ] must be positive\n\n")
            exit(-1)

        bound_box_width = opts.bound_box_width
        
        save_bound_box = True

    save_bound_box_min = False
    save_bound_box_max = False

    if (save_bound_box) :
        save_bound_box_min = True
        save_bound_box_max = True

    if (opts.bound_box_type is not None) :
        
        if (opts.bound_box_type == "box") :
            save_bound_box_min = True
            save_bound_box_max = True
        elif (opts.bound_box_type == "min-corner") :
            save_bound_box_min = True
            save_bound_box_max = False
        elif (opts.bound_box_type == "max-corner") :
            save_bound_box_min = False
            save_bound_box_max = True
        else :
            stderr.write("\nError: Bound box type [ " + opts.bound_box
                         + " ] not recognized\n\n")
            exit(-1)

        save_bound_box = True


    ## Build the box list

    for box_file_name in box_file_list :

        ## Open the output file

        output_filename = output_prefix + box_file_name + ".obj"

        f = open(output_filename, "w")

        f.write("# OBJ file for box list file " + box_file_name + "\n\n")

        f.write("# BoxData\n\n")

        if (material_file is not None) :
            f.write("mtllib " + material_file + "\n\n")
            f.write("usemtl " + material_name + "\n\n")

        ## Open the box file and process the boxes

        box_file = open(box_file_name, "r")

        box_file_finished = False

        min_vol_bounds = None
        max_vol_bounds = None

        while (not box_file_finished) :

            box_file_line = box_file.readline()

            if (box_file_line == "") :
                
                box_file_finished = True

            else :

                line_words = box_file_line.split("#")[0].split()

                num_words = len(line_words)

                if (num_words == 3) :
                    if (line_words[0] == "TransitionBox") :

                        min_coords = get_vector(line_words[1])
                        max_coords = get_vector(line_words[2])

                        box_cmd_list = get_box_draw_commands(min_coords,
                                                             max_coords)

                        for box_cmd in box_cmd_list :
                            f.write(box_cmd + "\n")

                        f.write("\n")

                        if (min_vol_bounds is None) :
                            
                            min_vol_bounds = min_coords
                            
                        else :
                            
                            for i in range(dims) :

                                axis_min_coord = min_coords[i]
                                
                                if (axis_min_coord < min_vol_bounds[i]) :
                                    min_vol_bounds[i] = axis_min_coord

                        if (max_vol_bounds is None) :
                            
                            max_vol_bounds = max_coords
                            
                        else :
                            
                            for i in range(dims) :

                                axis_max_coord = max_coords[i]
                                
                                if (axis_max_coord > max_vol_bounds[i]) :
                                    max_vol_bounds[i] = axis_max_coord

        if (save_bound_box) :

            ## Open the output file

            output_filename = output_prefix + box_file_name + "_box_bounds.obj"

            bbf = open(output_filename, "w")

            bbf.write("## OBJ file for box list file " + box_file_name
                      + "\n\n")

            bbf.write("## Bounding box\n\n")

            if ((min_vol_bounds is not None)
                and (max_vol_bounds is not None)) :

                write_bounds = True  ## Set to false if bounds are not valid

                bbf.write("# MinBounds " + get_vector_str(min_vol_bounds)
                          + "\n")
                
                bbf.write("# MaxBounds " + get_vector_str(max_vol_bounds)
                          + "\n")
                
                bbf.write("\n")

                ## Determine the width

                current_width = bound_box_width  ## Keep the original safe

                if (current_width is None) :

                    vol_width = list()

                    for i in range(dims) :
                        vol_width.append(max_vol_bounds[i] - min_vol_bounds[i])

                    min_width = min(vol_width)

                    if (not (min_width > 0.0)) :
                        write_bounds = False

                    current_width = min_width * default_bound_box_width_fract

                current_half_width = 0.5 * current_width
                    
                if (save_bound_box_min) :

                    bbf.write("## Min boundary\n\n")
                    
                    for i in range(dims) :

                        min_axes_coords = list()

                        for j in range(dims) :
                            min_axes_coords.append(min_vol_bounds[j]
                                                   - current_half_width)

                        for k in range(dims) :

                            max_axes_coords = list()

                            for j in range(dims) :

                                max_coord = 0.0
                                
                                if (j != k) :
                                    max_coord = min_axes_coords[j] \
                                                + current_width
                                else :
                                    max_coord = max_vol_bounds[j] \
                                                + current_half_width

                                max_axes_coords.append(max_coord)

                            axis_cmds = get_box_draw_commands(min_axes_coords,
                                                              max_axes_coords)

                            bbf.write("##   Axis " + str(k) + "\n\n")

                            for cmd in axis_cmds :
                                bbf.write(cmd + "\n")

                            bbf.write("\n")

                if (save_bound_box_max) :

                    bbf.write("## Max boundary\n\n")
                    
                    for i in range(dims) :

                        max_axes_coords = list()

                        for j in range(dims) :
                            max_axes_coords.append(max_vol_bounds[j]
                                                   + current_half_width)

                        for k in range(dims) :

                            min_axes_coords = list()

                            for j in range(dims) :

                                min_coord = 0.0
                                
                                if (j != k) :
                                    min_coord = max_axes_coords[j] \
                                                - current_width
                                else :
                                    min_coord = min_vol_bounds[j] \
                                                - current_half_width

                                min_axes_coords.append(min_coord)

                            axis_cmds = get_box_draw_commands(min_axes_coords,
                                                              max_axes_coords)

                            bbf.write("#   Axis " + str(k) + "\n\n")

                            for cmd in axis_cmds :
                                bbf.write(cmd + "\n")

                            bbf.write("\n")

            else :

                bbf.write("# BOUNDS NOT FOUND\n")


if (__name__ == "__main__") :
    main()
