// C++ program file demo.cpp

/*
  By Nathan C. Hearn
     December 11, 2006

  QuickFlash library demonstration program.
*/


#include <hdf5.h>
#include <string>
#include <vector>
#include <iostream>
#include <stdlib.h>
#include "quickflash_file_datafile.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_file_meshinfo.hpp"


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  if (argc < 4)
    {
    cerr << "Usage: " << argv[0] << " filename dataset buffer_size " 
	 << "[ cache_size ]" << endl
	 << endl;
    cerr << "Sums over all cells, multiplying dataset values by cell volumes" 
	 << endl << endl;
    return 1;
    }

  int argPtr = 1;

  const std::string file_name = argv[argPtr++];
  const std::string dataset_name = argv[argPtr++];

  const unsigned int buffer_size = atoi(argv[argPtr++]);

  unsigned int cache_size = 0;

  if (argPtr < argc)
    cache_size = atoi(argv[argPtr++]);

  // Open the file

  QuickFlash::File::DataFile dfile(file_name);

  // Get mesh info

  const QuickFlash::File::MeshInfo & meshinfo = dfile.get_mesh_info();

  const unsigned int num_blocks = meshinfo.get_num_blocks();
  const unsigned int dims = meshinfo.get_dims();

  cerr << "Dims [ " << dims << " ]" << endl;
  cerr << "Num blocks [ " << num_blocks << " ]" << endl;

  // Get information

  const std::vector<unsigned int> & block_dims = meshinfo.get_block_dims();

  cerr << "Block dimensions [ ";

  for (unsigned int i = 0; i < dims; i++)
    cerr << block_dims[i] << " ";

  cerr << "]" << endl;

  const std::vector<unsigned int> & base_block_dims 
    = meshinfo.get_base_block_dims();

  cerr << "Base block dimensions [ ";

  for (unsigned int i = 0; i < dims; i++)
    cerr << base_block_dims[i] << " ";

  cerr << "]" << endl;

  // Open a dataset

  const QuickFlash::File::Dataset & dset 
    = dfile.get_dataset(dataset_name, buffer_size);

  dset.set_cache_size(cache_size);
  dset.set_report_stats();

  cerr << "Dataset loaded" << endl;

  double dataset_vol_sum = 0.0;

  // Set up temporary storage for cell data

  QuickFlash::Block::BlockData<double> block_data;

  // Run through the leaf nodes

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    if(meshinfo.is_leaf(block_index))
      {
      // Get the block

      const QuickFlash::Block::BlockInfo & block_info 
	= meshinfo.get_block_info(block_index);

      dset.get_block_data(block_index, block_data);

      const unsigned int num_cells = block_info.get_num_cells();
      const double cell_vol = block_info.get_cell_volume();

      for (unsigned int index = 0; index < num_cells; index++)
	dataset_vol_sum += cell_vol * block_data[index];
      }

  cout << "Dataset [ " << dataset_name << " ] sum [ " << dataset_vol_sum 
       << " ]" << endl;

  return 0;
  }
