// C++ program file expcorrect.cpp

/*
  By Nathan C. Hearn
     May 17, 2007

  Expansion correction main program.
*/


#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <stdlib.h>
#include "velocity_field.hpp"
#include "mixing_layer.hpp"


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  const int min_args = 12;

  if (argc < min_args)
    {
    cerr << endl;
    cerr << "Usage: " << argv[0] << " file_prefix start_index end_index" 
	 << endl;
    cerr << " map_prefix map_start_index map_end_index" << endl;
    cerr << "min_x max_x bin_width" << endl;
    cerr << "bubble_poly_massfract spike_poly_massfract" << endl;
    cerr << endl;
    return 1;
    }

  int argptr = 1;

  const std::string file_prefix = argv[argptr++];

  const unsigned int start_index 
    = static_cast<unsigned int>(atoi(argv[argptr++]));

  const unsigned int end_index
    = static_cast<unsigned int>(atoi(argv[argptr++]));

  if (end_index <= start_index)
    {
    cerr << "Error: End index must be greater than start index" << endl;
    return -1;
    }

  const std::string map_prefix = argv[argptr++];

  const unsigned int map_start_index 
    = static_cast<unsigned int>(atoi(argv[argptr++]));

  const unsigned int map_end_index
    = static_cast<unsigned int>(atoi(argv[argptr++]));

  if (map_end_index <= map_start_index)
    {
    cerr << "Error: Map end index must be greater than start index" << endl;
    return -1;
    }

  const double min_x = atof(argv[argptr++]);

  const double max_x = atof(argv[argptr++]);

  if (!(min_x < max_x))
    {
    cerr << "Error: Improper min_x/max_x settings" << endl;
    return -1;
    }

  const double bin_spacing = atof(argv[argptr++]);

  if (!(bin_spacing > 0.0))
    {
    cerr << "Error: Bin spacing must be a positive value" << endl;
    return -1;
    }

  const double bubble_poly = atof(argv[argptr++]);

  if ((bubble_poly < 0.0) || (bubble_poly > 1.0))
    {
    cerr << "Error: Min polyimide mass fraction must be in range [0,1]" 
	 << endl;
    return -1;
    }

  const double spike_poly = atof(argv[argptr++]);

  if ((spike_poly < 0.0) || (spike_poly > 1.0))
    {
    cerr << "Error: Max polyimide mass fraction must be in range [0,1]" 
	 << endl;
    return -1;
    }

  // Output the settings

  cout << "# Expansion correction output" << endl;
  cout << "# Flash data file prefix [ " << file_prefix << " ] Min index [ "
       << start_index << " ] Max index [ " << end_index << " ]" << endl;
  cout << "# 1D map file prefix [ " << map_prefix << " ] Min index [ "
       << map_start_index << " ] Max index [ " << map_end_index << " ]" 
       << endl;
  cout << "# x Min [ " << min_x << " ] Max [ " << max_x << " ]" << endl;
  cout << "# Bin width [ " << bin_spacing << " ]" << endl;
  cout << "# Polyimide mass fraction Min (bubble) [" << bubble_poly
       << " ] Max (spike) [ " << spike_poly << " ]" << endl;
  cout << endl;

  // Read the map

  const SimVelocityFields vel_fields(map_prefix, map_start_index, 
				     map_end_index);

  // Build the list of filenames

  std::vector<std::string> filelist;

  for (unsigned int idx = start_index; idx <= end_index; idx++)
    {
    std::stringstream strstream;

    strstream << file_prefix;

    strstream.width(4);
    strstream.fill('0');

    strstream << idx;

    filelist.push_back(strstream.str());
    }

  // Determine the mixing layer bounds

  std::vector<MixingLayer> mix_layers;

  get_mixing_layers_minmax(filelist, bubble_poly, spike_poly, mix_layers);
  
  // Process the mixing layers

  const unsigned int num_timesteps = mix_layers.size();

  const unsigned int num_cycles = num_timesteps - 1;

  double rt_layer_min = mix_layers[0].get_min_pos();
  double rt_layer_max = mix_layers[0].get_max_pos();

  cout << "Index\tTime\traw_min\traw_max\traw_width\tdelta_min\tdelta_max\t"
       << "delta_min_exp\tdelta_max_exp\tdelta_min_rt\tdelta_max_rt\t"
       << "rt_min\trt_max\trt_width" << endl;

  cout << start_index << "\t" << mix_layers[0].get_time() << "\t" 
       << rt_layer_min << "\t" << rt_layer_max << "\t" 
       << (rt_layer_max - rt_layer_min) << "\t"
       << 0.0 << "\t" << 0.0 << "\t" << 0.0 << "\t" << 0.0 << "\t"
       << 0.0 << "\t" << 0.0 << "\t" << 0.0 << "\t" << 0.0 << "\t" << 0.0
       << endl;

  for (unsigned int cycle = 0; cycle < num_cycles; cycle++)
    {
    const unsigned int lo_index = cycle;
    const unsigned int hi_index = cycle + 1;

    const unsigned int hi_file_index = hi_index + start_index;

    const double lo_time = mix_layers[lo_index].get_time();
    const double hi_time = mix_layers[hi_index].get_time();

    const double delta_time = hi_time - lo_time;

    // Determine the positions and changes of the mixing layer boundaries

    const double lo_layer_min_pos = mix_layers[lo_index].get_min_pos();
    const double lo_layer_max_pos = mix_layers[lo_index].get_max_pos();

    const double hi_layer_min_pos = mix_layers[hi_index].get_min_pos();
    const double hi_layer_max_pos = mix_layers[hi_index].get_max_pos();

    const double delta_min_pos = hi_layer_min_pos - lo_layer_min_pos;
    const double delta_max_pos = hi_layer_max_pos - lo_layer_max_pos;

    // Determine the lagrangian (expansion) changes

    const double lo_min_vel 
      = vel_fields.get_interp_vel(lo_time, lo_layer_min_pos);

    const double lo_max_vel 
      = vel_fields.get_interp_vel(lo_time, lo_layer_max_pos);

    const double hi_min_vel 
      = vel_fields.get_interp_vel(hi_time, hi_layer_min_pos);

    const double hi_max_vel 
      = vel_fields.get_interp_vel(hi_time, hi_layer_max_pos);

    const double exp_delta_min = delta_time * 0.5 * (lo_min_vel + hi_min_vel);
    const double exp_delta_max = delta_time * 0.5 * (lo_max_vel + hi_max_vel);

    // Update the effective RT layer boundaries

    const double rt_delta_min = delta_min_pos - exp_delta_min;
    const double rt_delta_max = delta_max_pos - exp_delta_max;

    rt_layer_min += rt_delta_min;
    rt_layer_max += rt_delta_max;

    // Output the results

    const double raw_hi_width = hi_layer_max_pos - hi_layer_min_pos;
    const double rt_width = rt_layer_max - rt_layer_min;

    cout << hi_file_index << "\t" << hi_time << "\t" 
	 << hi_layer_min_pos << "\t" << hi_layer_max_pos
	 << "\t" << raw_hi_width << "\t" << delta_min_pos << "\t" 
	 << delta_max_pos << "\t" << exp_delta_min << "\t" << exp_delta_max
	 << "\t" << rt_delta_min << "\t" << rt_delta_max << "\t"
	 << rt_layer_min << "\t" << rt_layer_max << "\t" << rt_width << endl;
    }

  return 0;
  }
