#!/usr/bin/env python

# generate_exp_plots.py

# By Nathan C. Hearn
#    November 13, 2008
#
# Plotter for data from expcorrect

# Make sure the QuickFlash/utils directory is in PYTHONPATH!!!


try :
    import psyco
except ImportError :
    pass


default_plot_title = ""


class SimData :

    def __init__(self, sim_name=None) :

        self.__sim_name = None

        self.__time_list = list()

        self.__raw_layer_min_list = list()
        self.__raw_layer_max_list = list()

        self.__corrected_layer_min_list = list()
        self.__corrected_layer_max_list = list()

        self.reset(sim_name)


    def reset(self, sim_name=None) :

        self.__sim_name = sim_name

        self.__time_list = list()

        self.__raw_layer_min_list = list()
        self.__raw_layer_max_list = list()

        self.__corrected_layer_min_list = list()
        self.__corrected_layer_max_list = list()
  

    def add_data(self, time, raw_layer_min_pos, raw_layer_max_pos,
                 corrected_layer_min_pos, corrected_layer_max_pos) :

        self.__time_list.append(time)

        self.__raw_layer_min_list.append(raw_layer_min_pos)
        self.__raw_layer_max_list.append(raw_layer_max_pos)

        self.__corrected_layer_min_list.append(corrected_layer_min_pos)
        self.__corrected_layer_max_list.append(corrected_layer_max_pos)


    def get_sim_name(self) :
        return self.__sim_name


    def get_num_entries(self) :
        return len(self.__time_list)


    def get_time_list(self) :
        return self.__time_list


    def get_raw_layer_min_list(self) :
        return self.__raw_layer_min_list


    def get_raw_layer_max_list(self) :
        return self.__raw_layer_max_list


    def get_corrected_layer_min_list(self) :
        return self.__corrected_layer_min_list


    def get_corrected_layer_max_list(self) :
        return self.__corrected_layer_max_list


    def get_raw_layer_widths(self) :

        layer_widths = list()

        num_entries = len(self.__time_list)

        for index in range(num_entries) :
            
            width = self.__raw_layer_max_list[index] \
                    - self.__raw_layer_min_list[index]
            
            layer_widths.append(width)

        return layer_widths

    def get_corrected_layer_widths(self) :

        layer_widths = list()

        num_entries = len(self.__time_list)

        for index in range(num_entries) :
            
            width = self.__corrected_layer_max_list[index] \
                    - self.__corrected_layer_min_list[index]
            
            layer_widths.append(width)

        return layer_widths


def read_sim_data(file_name, sim_name) :

        from plot_gen import read_table

        table_data = read_table(file_name)

        time_data = table_data["Time"]

        raw_min_data = table_data["raw_min"]
        raw_max_data = table_data["raw_max"]

        corrected_min_data = table_data["rt_min"]
        corrected_max_data = table_data["rt_max"]

        num_times = len(time_data)

        sim_data = SimData(sim_name)

        for index in range(num_times) :

            time = time_data[index]

            raw_min = raw_min_data[index]
            raw_max = raw_max_data[index]

            corrected_min = corrected_min_data[index]
            corrected_max = corrected_max_data[index]

            sim_data.add_data(time, raw_min, raw_max, corrected_min,
                              corrected_max)

        return sim_data


def main() :

    from sys import argv, exit, stderr
    from os import path
    from optparse import OptionParser
    
    from plot_gen import Plot

    ## Set up the command-line parser

    usage = str("%prog [options] sim_data_1[,sim_name_1] "
                + "[ sim_data_2[,sim_name_2] ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_prefix_str = str("output-prefix")
    argparser.add_option("--" + output_prefix_str, type="string",
                         dest="output_prefix",
                         help="Prefix for output plot files")
    argparser.set_defaults(output_prefix=None)

    plot_title_str = str("plot-title")
    argparser.add_option("--" + plot_title_str, type="string",
                         dest="plot_title",
                         help="Plot title")
    argparser.set_defaults(plot_title=None)

    background_vel_str = str("background-vel")
    argparser.add_option("--" + background_vel_str, type="string",
                         dest="background_vel",
                         help=("Background velocity (cm / s) "
                               + "with optional comma-separated time offset "
                               + "(s)"))
    argparser.set_defaults(background_vel=None)
    

    ## Process the command-line input

    opts, args = argparser.parse_args()

    argc = len(args)

    min_args = int(1)

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n")
        exit(1)

    ## Get the required arguments

    argptr = int(0)

    sim_file_list = list()
    sim_name_list = list()

    while (argptr < argc) :

        entry = args[argptr]

        words = entry.split(",")
        argptr += 1

        num_words = len(words)

        if (num_words > 2) :
            stderr.write("\nError: Sim data entry [ " + entry
                         + " ] must have one or two comma-separated elements"
                         + "\n\n")
            exit(-1)

        sim_data_file = words[0]

        if (num_words > 1) :
            sim_name = words[1]
        else :
            sim_name = sim_data_file

        sim_file_list.append(sim_data_file)
        sim_name_list.append(sim_name)


    num_files = len(sim_file_list)


    ## Get the optional arguments

    output_prefix = opts.output_prefix

    if (output_prefix is None) :
        output_prefix = ""
    elif (len(path.split(output_prefix)[1]) > 0) :
        output_prefix += "_"

    plot_title = opts.plot_title

    if (plot_title is None) :
        plot_title = default_plot_title

    if (plot_title != "") :
        plot_title += "  -  "

    background_vel = float(0.0)
    background_vel_time_offset = float(0.0)

    if (opts.background_vel is not None) :
        background_vel_words = opts.background_vel.split(",")

        num_background_vel_words = len(background_vel_words)

        if ((num_background_vel_words < 1) or (num_background_vel_words > 2)) :
            stderr.write("\nError: Background velocity requires one or two "
                         + "comma-separated entries\n\n")
            exit(-1)

        background_vel = float(background_vel_words[0])

        if (num_background_vel_words == 2) :
            background_vel_time_offset = float(background_vel_words[1])


    ## Collect the data

    sim_data_list = list()

    for file_index in range(num_files) :

        filename = sim_file_list[file_index]
        sim_name = sim_name_list[file_index]

        sim_data = read_sim_data(filename, sim_name)

        sim_data_list.append(sim_data)


    ## Plot the data

    width_plot = Plot(plot_title + "Expansion-Corrected Mixing Layer Widths",
                      "Time (ns)", "Width (microns)")
    
    bounds_plot = Plot(plot_title
                       + "Expansion-Corrected Mixing Layer Boundaries",
                       "Time (ns)", "Position (microns)")

    for sim_data in sim_data_list :

        sim_name = sim_data.get_sim_name()

        num_entries = sim_data.get_num_entries()

        orig_time_list = sim_data.get_time_list()
        
        orig_corrected_min_list = sim_data.get_corrected_layer_min_list()
        orig_corrected_max_list = sim_data.get_corrected_layer_max_list()

        time_list = list()
        
        corrected_min_list = list()
        corrected_max_list = list()

        corrected_width_list = list()

        for entry_index in range(num_entries) :

            ## Determine the position offsets due to background velocity

            orig_time = orig_time_list[entry_index]

            pos_offset = (-1.0) * (orig_time - background_vel_time_offset) \
                         * background_vel

            ## Convert to nanoseconds and microns from seconds and centimeters

            new_time = orig_time * 1.0e+09

            new_corrected_min \
                = (orig_corrected_min_list[entry_index] + pos_offset) * 1.0e+04
            
            new_corrected_max \
                = (orig_corrected_max_list[entry_index] + pos_offset) * 1.0e+04

            time_list.append(new_time)

            corrected_min_list.append(new_corrected_min)
            corrected_max_list.append(new_corrected_max)

            corrected_width_list.append(new_corrected_max - new_corrected_min)

        width_plot.add_plot(time_list, corrected_width_list,
                            legend_text=sim_name)

        bounds_plot.add_plot(time_list, corrected_min_list,
                             legend_text=(sim_name + " min"),
                             marker_index=1)
            
        bounds_plot.add_plot(time_list, corrected_max_list,
                             legend_text=(sim_name + " max"),
                             marker_index=2, use_prev_color=True)


    ## Save the plots

    width_plot_filename = output_prefix + "layer_width.svg"
    bounds_plot_filename = output_prefix + "layer_bounds.svg"

    width_plot.show_plot(width_plot_filename)
    bounds_plot.show_plot(bounds_plot_filename)

if (__name__ == "__main__") :
    main()

