#!/usr/bin/env python

# show_map.py

# By Nathan C. Hearn
#    February 7, 2007


try :
    import psyco
except ImportError :
    pass


resolution = 0.15  # microns


from sys import argv, exit, stderr, stdout


orig_file_suffix = "_orig.eps"
align_file_suffix = "_align.eps"

use_align_data = False


# Trajectories

class Trajectories :

    def __init__(self, filename=None) :
        
        self.__times = list()
        self.__interfaces = list()
        self.__traj = list()

        self.__num_times = int(0)
        self.__num_parts = int(0)

        if (filename is not None) :
            self.read_data(filename)

    def read_data(self, filename) :
        
        self.__times = list()
        self.__interfaces = list()
        self.__traj = list()

        self.__num_times = int(0)
        self.__num_parts = int(0)

        f = file(filename, "r")

        lines = f.readlines()

        num_lines = len(lines)

        if (num_lines > 0) :

            words = lines[0].split()

            if (len(words) != 2) :
                raise ValueError("Improper format for map file")

            file_timesteps = int(words[0])

            num_parts = int(words[1])

            # Set up storage for particles

            for part_index in range(num_parts) :
                self.__traj.append(list())

            num_step_values = num_parts + 2  # time, interface, pos

            timestep_count = int(0)

            for step_index in range(num_lines - 1) :

                file_line_index = step_index + 1

                words = lines[file_line_index].split()

                if (len(words) == num_step_values) :

                    time = float(words[0]) * 1.0e+09  # ns
                    interface_pos = float(words[1])  # cm

                    self.__times.append(time)
                    self.__interfaces.append(interface_pos)

                    for part_index in range(num_parts) :
                        
                        part_pos = float(words[part_index + 2])  #cm

                        self.__traj[part_index].append(part_pos)

                    timestep_count += 1
                else :
                    stdout.write("Skipping map file line [ "
                                 + str(file_line_index) + " ]\n")

            if (file_timesteps != timestep_count) :
                raise ValueError("Timestep count mismatch")
            
            self.__num_times = timestep_count
            self.__num_parts = num_parts

    def get_num_times(self) :
        return self.__num_times

    def get_num_parts(self) :
        return self.__num_parts


    def get_time(self, index=None) :

        ret_time = self.__times

        if (index is not None) :
            ret_time = self.__times[index]

        return ret_time


    def get_interface_pos(self, index=None) :

        ret_pos = self.__interfaces

        if (index is not None) :
            ret_pos = self.__interfaces[index]

        return ret_pos


    def get_trajectory(self, index) :
        return self.__traj[index]   


    def get_init_pos(self, index=None) :

        ret_pos = float(0.0)

        if (self.__num_times < 1) :
            raise ValueError("No trjaectory data")
        elif (index is not None) :

            if ((index < 0) or (index >= self.__num_parts)) :
                raise ValueError("Index out of range")
            else :                
                ret_pos = self.__traj[index][0]
        else :
            ret_pos = list()

            for traj in self.__traj :
                ret_pos.append(traj[0])
                
        return ret_pos


def select_indexes(pos_list, min_val, max_val) :

    index_list = list()

    num_pos = len(pos_list)

    for index in range(num_pos) :
        
        pos = pos_list[index]

        if ((not (pos < min_val)) and (pos < max_val)) :
            index_list.append(index)

    return index_list


def pare_indexes(index_list, skip) :

    ret_list = list()

    if (skip < 2) :
        ret_list = index_list
    else :
        num_indexes = len(index_list)

        for idx in range(0, num_indexes, skip) :
            ret_list.append(index_list[idx])

    return ret_list


def main() :

    # Parse the command line

    argc = len(argv)

    if (argc < 7) :
        stderr.write("Usage: " + argv[0] + " map_file min_init_x max_init_x "
                     + "trajectory_skip min_select_x max_select_x "
                     + "[ output_filename_prefix ]\n")
        exit(1)

    argPtr = int(1)

    map_file = str(argv[argPtr])
    argPtr += 1

    min_init_x = float(argv[argPtr])
    argPtr += 1

    max_init_x = float(argv[argPtr])
    argPtr += 1

    if (not (min_init_x < max_init_x)) :
        raise ValueError("Incompatible initial x limits")

    traj_skip = int(1)

    traj_skip = int(argv[argPtr])
    argPtr += 1

    if (traj_skip < 1) :
        raise ValueError("Trajectory skip must be a positive integer")

    min_select_x = float(argv[argPtr])
    argPtr += 1

    max_select_x = float(argv[argPtr])
    argPtr += 1

    save_fig = False

    fig_filename_prefix = None

    if (argPtr < argc) :
        fig_filename_prefix = str(argv[argPtr])
        argPtr += 1

        save_fig = True

    # Set up matplotlib

    import matplotlib

    if (save_fig) :
        matplotlib.use("PS")
    else :
        matplotlib.use("tkAgg")

    # matplotlib.rc("text.usetex : True")

    import pylab

    # Build the map

    traj = Trajectories(map_file)

    # Get some basic information

    times = traj.get_time()
    interface_pos = traj.get_interface_pos()

    traj_len = len(interface_pos)

    # Determine trajectories to be plotted

    num_tot_traj = traj.get_num_parts()

    init_pos_array = traj.get_init_pos()

    traj_indexes = range(num_tot_traj)
    traj_indexes = select_indexes(init_pos_array, min_init_x, max_init_x)
    traj_indexes = pare_indexes(traj_indexes, traj_skip)

    stderr.write("Plotting " + str(len(traj_indexes)) + " trajectories\n")

    # Set up the plot

    traj_fig = pylab.figure(1)
    traj_fig.clf()
    traj_fig.hold(True)

    traj_plot = traj_fig.add_subplot(1,1,1)

    percent_shown = 100.0 / float(traj_skip)

    percent_str = "%(val)2.2f" % { "val" : percent_shown }

    traj_plot.set_title("1D Tracer Trajectories - " + str(resolution)
                        + " micron resolution (" + percent_str + "% shown)")

    if (use_align_data) :
        traj_plot.set_xlabel("Position (cm) - Relative to Interface")
    else :
        traj_plot.set_xlabel("Position (cm)")
        
    traj_plot.set_ylabel("Time (ns)")

    # Plot the trajectories

    for traj_idx in traj_indexes :

        trajectory = traj.get_trajectory(traj_idx)

        if (use_align_data) :
            for i in range(traj_len) :
                trajectory[i] -= interface_pos[i]

        init_pos = init_pos_array[traj_idx]

        plotstyle = "c"

        if ((not (init_pos < min_select_x))
            and (not (init_pos > max_select_x))) :
            plotstyle = "g"
            
        traj_plot.plot(trajectory, times, plotstyle, linewidth=1)

    # Plot the interface

    if (use_align_data) :
        zero_array = list()

        for i in range(traj_len) :
            zero_array.append(0.0)

        traj_plot.plot(zero_array, times, ".r", linewidth=0, markersize=3.0)
    else :
        traj_plot.plot(interface_pos, times, ".r", linewidth=0, markersize=3.0)

    traj_plot.grid()

    ### Temporary
    # traj_plot.set_xlim(-0.05, 0.175)
    # traj_plot.set_ylim(5.0, 32.0)
    ### End Temporary
    
    if (save_fig) :
        if (use_align_data) :
            pylab.savefig(fig_filename_prefix + align_file_suffix)
        else :
            pylab.savefig(fig_filename_prefix + orig_file_suffix)
    else :
        pylab.show()

    


if (__name__ == "__main__") :
    main()


