// C++ header file velocity_field.cpp

/*
  By Nathan C. Hearn
     May 24, 2006

  Velocity field derived from 1D Flash simulations.
*/


#include "velocity_field.hpp"
#include <string>
#include <vector>
#include <iostream>
#include <algorithm>
#include <sstream>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_except.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"


// Class VelocityField

void VelocityField::reset()
  {
  time = 0.0;

  min_pos = 0.0;
  max_pos = 0.0;

  point_array.clear();
  }


void VelocityField::reset(const std::string & flash_datafile, 
			  const unsigned int buffer_size)
  {
  reset();
  load_file(flash_datafile, buffer_size);
  }


void VelocityField::reset(const VelocityField & source)
  {
  if (&source != this)
    {
    time = source.time;

    min_pos = source.min_pos;
    max_pos = source.max_pos;

    point_array = source.point_array;
    }
  }


double VelocityField::get_interp_vel(const double pos) const
  {
  // Find the neighboring cells

  unsigned int prev_cell = 0;
  unsigned int next_cell = 0;

  double delta_pos_prev = 0.0;
  double delta_next_prev = 0.0;

  get_nearest_points(pos, prev_cell, next_cell, delta_pos_prev, 
		     delta_next_prev);

  double high_weight = delta_pos_prev / delta_next_prev;

  if (high_weight > 1.0)
    high_weight = 1.0;
  else if (high_weight < 0.0)
    high_weight = 0.0;

  const double low_weight = 1.0 - high_weight;

  const double low_vel = point_array[prev_cell].get_vel();
  const double high_vel = point_array[next_cell].get_vel();

  const double vel = (low_vel * low_weight) + (high_vel * high_weight);

  return vel;
  }


void VelocityField::load_file(const std::string & flash_datafile, 
			      const unsigned int buffer_size)
  {
  using std::cerr;
  using std::endl;

  // Load the file

  QuickFlash::File::DataFile df(flash_datafile, buffer_size, true);

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (dims != 1)
    {
    cerr << "File [ " << flash_datafile << " ] is not one-dimensional" << endl;
    throw QuickFlash::Except("Incompatible file dimensions", __FILE__, __LINE__);
    }

  time = meshinfo.get_sim_time();

  const QuickFlash::File::Dataset & vel_set = df.get_dataset("velx");

  // Run through the blocks

  const unsigned int num_blocks = meshinfo.get_num_blocks();

  std::vector<double> pos_vect;
  QuickFlash::Block::BlockData<double> vel_data;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    if (meshinfo.is_leaf(block_index))
      {
      const QuickFlash::Block::BlockInfo & blockinfo 
	= meshinfo.get_block_info(block_index);

      vel_set.get_block_data(block_index, vel_data);

      const unsigned int num_cells = blockinfo.get_num_cells();

      for (unsigned int index = 0; index < num_cells; index++)
	{
	blockinfo.get_cell_center(index, pos_vect);

	const double pos = pos_vect[0];
	const double vel = vel_data[index];

	const VelocityPoint point(pos, vel);

	point_array.push_back(point);
	}
      }

  // Sort the array

  std::sort(point_array.begin(), point_array.end());

  // Get position bounds

  const unsigned int num_points = point_array.size();

  if (num_points < 2)
    {
    cerr << "Insufficient number of leaf cells" << endl << endl;
    throw QuickFlash::Except("Insufficient number of leaf cells", __FILE__,
			   __LINE__);
    }

  min_pos = point_array[0].get_pos();
  max_pos = point_array[num_points - 1].get_pos();
  }


void VelocityField::get_nearest_points(const double pos, 
				       unsigned int & prev_cell_index,
				       unsigned int & next_cell_index, 
				       double & delta_pos_prev,
				       double & delta_next_prev) const
  {
  const unsigned int num_points = point_array.size();

  if (num_points < 2)
    throw QuickFlash::Except("Insufficient points in list", __FILE__, __LINE__);

  unsigned int low = 0;
  unsigned int high = num_points - 1;

  double low_pos = point_array[low].get_pos();
  double high_pos = point_array[high].get_pos();

  if ((pos < low_pos) || (!(pos < high_pos)))
    throw QuickFlash::Except("Position out of bounds", __FILE__, __LINE__);

  while (low < (high - 1))
    {
    const unsigned int temp = (low + high) / 2;

    const double temp_pos = point_array[temp].get_pos();

    if (temp_pos > pos)
      high = temp;
    else
      low = temp;
    }

  // ERROR CHECKING???
  
  low_pos = point_array[low].get_pos();
  high_pos = point_array[high].get_pos();

  prev_cell_index = low;
  next_cell_index = high;

  delta_next_prev = high_pos - low_pos;
  delta_pos_prev = pos - low_pos;
  }


// Class SimVelocityFields

void SimVelocityFields::reset(const std::string & filename_base, 
			      const unsigned int start_index, 
			      const unsigned int end_index, 
			      const unsigned int num_digits,
			      const unsigned int buffer_size)
  {
  reset();

  // Determine the number of files to load

  const unsigned int num_fields = (end_index + 1) - start_index;

  if (num_fields < 2)
    throw QuickFlash::Except("Insufficient number of files", __FILE__,
			   __LINE__);

  vel_fields.resize(num_fields);

  for (unsigned int idx = 0; idx < num_fields; idx++)
    {
    const unsigned int file_index = idx + start_index;

    // Build the filename

    std::stringstream sstream;

    sstream << filename_base;

    sstream.width(num_digits);
    sstream.fill('0');

    sstream << file_index;

    // Load the file

    vel_fields[idx].reset(sstream.str(), buffer_size);
    }
  }


double SimVelocityFields::get_interp_vel(const double time, const double pos)
  const
  {
  // Find the neighboring times

  unsigned int prev_time = 0;
  unsigned int next_time = 0;

  double delta_time_prev = 0.0;
  double delta_next_prev = 0.0;

  find_nearest_times(time, prev_time, next_time, delta_time_prev, 
		     delta_next_prev);

  double high_weight = delta_time_prev / delta_next_prev;

  if (high_weight > 1.0)
    high_weight = 1.0;
  else if (high_weight < 0.0)
    high_weight = 0.0;

  const double low_weight = 1.0 - high_weight;

  const double low_vel = vel_fields[prev_time].get_interp_vel(pos);
  const double high_vel = vel_fields[next_time].get_interp_vel(pos);

  const double vel = (low_vel * low_weight) + (high_vel * high_weight);

  return vel;
  }


double SimVelocityFields::get_interp_acc(const double time, const double pos)
  const
  {
  // Find the neighboring times

  unsigned int prev_time = 0;
  unsigned int next_time = 0;

  double delta_time_prev = 0.0;
  double delta_next_prev = 0.0;

  find_nearest_times(time, prev_time, next_time, delta_time_prev, 
		     delta_next_prev);

  // For right now, just look at the change in local average velocity

  const double low_vel = vel_fields[prev_time].get_interp_vel(pos);
  const double high_vel = vel_fields[next_time].get_interp_vel(pos);

  const double delta_vel = high_vel - low_vel;

  return delta_vel / delta_next_prev;
  }


void SimVelocityFields::find_nearest_times(const double time, 
					   unsigned int & low_index,
					   unsigned int & high_index, 
					   double & delta_time_low,
					   double & delta_high_low) const
  {
  const unsigned int num_times = vel_fields.size();

  if (num_times < 2)
    throw QuickFlash::Except("Insufficient states in list", __FILE__, __LINE__);

  unsigned int low = 0;
  unsigned int high = num_times - 1;

  double low_time = vel_fields[low].get_time();
  double high_time = vel_fields[high].get_time();

  if ((time < low_time) || (!(time < high_time)))
    throw QuickFlash::Except("Time out of bounds", __FILE__, __LINE__);

  while (low < (high - 1))
    {
    const unsigned int temp = (low + high) / 2;

    const double temp_time = vel_fields[temp].get_time();

    if (temp_time > time)
      high = temp;
    else
      low = temp;
    }

  // ERROR CHECKING???
  
  low_time = vel_fields[low].get_time();
  high_time = vel_fields[high].get_time();

  low_index = low;
  high_index = high;

  delta_high_low = high_time - low_time;
  delta_time_low = time - low_time;
  }
