// C++ program file fileinfo.cpp

/*
  By Nathan C. Hearn
     November 19, 2007

  Simple file information display.
*/


#include <iostream>
#include <string>
#include <ostream>
#include <vector>
#include <list>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_siminfo.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_file_flashdefs.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_utils_file.hpp"


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, 
					option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 1;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name << " [ --time-only ] [ --tabular ]"
	 << endl << "\t"
	 << "[ --quiet] filename1 [ filename2 ...]" << endl << endl;
    return 1;
    }


  // Read the arguments

  unsigned int argptr = 0;

  std::list<std::string> file_name_list;

  while (argptr < num_args)
    file_name_list.push_back(argument_list[argptr++]);


  // Read the options

  bool time_only = false;

  bool output_table = false;

  bool quiet_operation = false;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "time-only")
      {
      time_only = true;
      }
    else if (keyword == "tabular")
      {
      output_table = true;
      }
    else if (keyword == "quiet")
      {
      quiet_operation = true;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }


  // Run through the files

  bool variables_set = false;

  std::vector<std::string> var_list;
  unsigned int num_vars = 0;

  std::list<std::string>::const_iterator filename_iter 
    = file_name_list.begin();

  while (filename_iter != file_name_list.end())
    {
    const std::string & filename = *filename_iter;

    if (!quiet_operation)
      cerr << "Processing file [ " << filename << " ]" << endl;

    QuickFlash::File::DataFile df(filename, 0, true);

    const QuickFlash::File::SimInfo & siminfo = df.get_sim_info();

    const double sim_time = siminfo.get_sim_time();

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    if (output_table && (!variables_set))
      {
      var_list = siminfo.get_mesh_variable_names();

      variables_set = true;

      // Output the header for the table

      cout << "Filename" 
	   << "\t" << "Sim_Time" 
	   << "\t" << "Mesh_Dims" 
	   << "\t" << "Num_Blocks" 
	   << "\t" << "Leaf_Blocks" 
	   << "\t" << "Min_Leaf_Refine_Level" 
	   << "\t" << "Max_Leaf_Refine_Level" 
	   << "\t" << "Min_Bounds" 
	   << "\t" << "Max_Bounds" 
	   << "\t" << "Base_Block_Dims" 
	   << "\t" << "Block_Dims";

      num_vars = var_list.size();

      for (unsigned int var_index = 0; var_index < num_vars; var_index++)
	{
	const std::string & var_name = var_list[var_index];

	cout << "\t" << var_name << "_Min" << "\t" << var_name << "_Max";
	}

      cout << endl;
      }

    if (time_only)
      cout << filename << "\t" << sim_time << endl;
    else
      {
      const unsigned int mesh_dims = meshinfo.get_dims();

      const unsigned int num_blocks = meshinfo.get_num_blocks();

      const unsigned int leaf_blocks = meshinfo.get_num_leaf_blocks();

      const unsigned int min_leaf_refine_level 
	= meshinfo.get_min_leaf_refine_level();

      const unsigned int max_leaf_refine_level 
	= meshinfo.get_max_leaf_refine_level();

      const std::vector<double> & min_bounds = meshinfo.get_volume_minbounds();
      const std::vector<double> & max_bounds = meshinfo.get_volume_maxbounds();

      const std::vector<unsigned int> & base_block_dims 
	= meshinfo.get_base_block_dims();

      const std::vector<unsigned int> & block_dims = meshinfo.get_block_dims();

      const unsigned int block_cells = meshinfo.get_num_block_cells();

      if (output_table)
	{
	cout << filename << "\t" << sim_time << "\t" << mesh_dims 
	     << "\t" << num_blocks << "\t" << leaf_blocks 
	     << "\t" << min_leaf_refine_level << "\t" << max_leaf_refine_level
	     << "\t" << min_bounds << "\t" << max_bounds 
	     << "\t" << base_block_dims << "\t" << block_dims;
	}
      else
	{
	cout << "File [ " << filename << " ]" << endl;

	cout << "  Simulation time [ " << sim_time << " ]" << endl;

	cout << "  Spatial dimensions [ " << mesh_dims << " ]" 
	     << endl;

	cout << "  Blocks [ " << num_blocks << " ]" << endl;

	cout << "  Leaf blocks [ " << leaf_blocks << " ]" << endl;

	cout << "  Min bounds [ " << min_bounds << " ]" << endl;

	cout << "  Max bounds [ " << max_bounds << " ]" << endl;

	cout << "  Base blocks [ " << base_block_dims << " ]" << endl;

	cout << "  Block dimensions [ " << block_dims << " ]" << endl;
	cout << "    Block cells [ " << block_cells << " ]" << endl;

	cout << "  Min leaf refine level [ " << min_leaf_refine_level 
	     << " ]" << endl;

	cout << "  Max leaf refine level [ " << max_leaf_refine_level 
	     << " ]" << endl;
	}

      // Get the list of datasets

      if (!output_table)
	var_list = siminfo.get_mesh_variable_names();

      const unsigned int num_datasets = var_list.size();

      for (unsigned int dataset_index = 0; dataset_index < num_datasets;
	   dataset_index++)
	{
	const std::string & dset_name = var_list[dataset_index];

	if (siminfo.mesh_variable_present(dset_name))
	  {
	  const QuickFlash::File::Dataset & dset = df.get_dataset(dset_name);

	  const double min_value = dset.get_min_value();
	  const double max_value = dset.get_max_value();

	  if (output_table)
	    cout << "\t" << min_value << "\t" << max_value;
	  else
	    {
	    cout << "  Dataset [ " << dset_name << " ]" << endl;

	    cout << "    Min value [ " << min_value << " ] Max value [ " 
		 << max_value << " ]" << endl;
	    }
	  }
	else if (output_table)
	  cout << "\t" << "N/A" << "\t" << "N/A";
	}

      if (output_table)
	cout << endl;
      else
	cout << endl << endl;
      }

    ++filename_iter;
    }

  return 0;
  }
