// C++ program file find_extent.cpp

/*
  By Nathan C. Hearn
     May 8, 2008

  Find the coordinate extents of the subset of space that has a data value
  below or above a given threshold.
*/


#include <iostream>
#include <string>
#include <vector>
#include <cstdlib>
#include <list>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_siminfo.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_tree_block.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_text.hpp"


const unsigned int read_buffer_size = 1000;


int find_extents(const QuickFlash::File::DataFile & df, 
		 const std::string & dataset_name, 
		 const double threshold_value, const bool find_region_above,
		 std::vector<double> & min_coords, 
		 std::vector<double> & max_coords)
  {
  int ret_val = -1;  // Error state

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  // Open the dataset

  const QuickFlash::File::Dataset & dset = df.get_dataset(dataset_name);

  QuickFlash::Block::BlockData<double> block_data;

  std::vector<double> cell_pos;

  bool found_interior = false;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    const std::vector<double> & block_min_coords = block_info.get_min_coords();
    const std::vector<double> & block_max_coords = block_info.get_max_coords();

    bool skip_branch = false;

    if (found_interior)
      {
      // See if we can skip this branch

      bool block_enclosed = true;

      for (unsigned int i = 0; i < dims; i++)
	{
	const double region_min = min_coords[i];
	const double region_max = max_coords[i];

	const double block_min = block_min_coords[i];
	const double block_max = block_max_coords[i];

	if ((block_min < region_min) || (block_max > region_max))
	  block_enclosed = false;
	}

      skip_branch = block_enclosed;
      }

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      dset.get_block_data(block_index, block_data);

      const unsigned int num_cells = block_info.get_num_cells();

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	const double cell_value = block_data[cell_index];

	bool use_cell = true;

	if (!find_region_above)
	  {
	  if (cell_value > threshold_value)
	    use_cell = false;
	  }
	else
	  {
	  if (cell_value < threshold_value)
	    use_cell = false;
	  }

	if (use_cell)
	  {
	  block_info.get_cell_center(cell_index, cell_pos);

	  if (found_interior)
	    {
	    for (unsigned int i = 0; i < dims; i++)
	      {
	      const double region_min = min_coords[i];
	      const double region_max = max_coords[i];

	      const double axis_cell_pos = cell_pos[i];

	      if (axis_cell_pos < region_min)
		min_coords[i] = axis_cell_pos;
	      else if (axis_cell_pos > region_max)
		max_coords[i] = axis_cell_pos;
	      }
	    }
	  else
	    {
	    min_coords = cell_pos;
	    max_coords = cell_pos;

	    found_interior = true;
	    }
	  }
	}
      }
    
    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  if (found_interior)
    ret_val = 0;  // OK

  return ret_val;
  }


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 3;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name << " dataset_name threshold_value"
	 << endl << "\t" << "[ --find-region-above ]"
	 << endl << "\t" << "[ --add-comment-marker=MARKER ]"
	 << endl << "\t" << "flash_filename1 [ flash_filename2 ... ]"
	 << endl << endl;
    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  const std::string dataset_name = argument_list[argptr++];

  const double threshold_value = std::atof(argument_list[argptr++].c_str());

  std::list<std::string> flash_filenames;

  while (argptr < num_args)
    flash_filenames.push_back(argument_list[argptr++]);

  // Parse the options

  bool find_region_above_threshold = false;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "find-region-above")
      {
      find_region_above_threshold = true;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }

  cout << "Dataset " << dataset_name << endl;
  cout << "ThresholdValue " << threshold_value << endl;
  
  cout << "FindRegionAboveThreshold ";

  if (find_region_above_threshold)
    cout << "true";
  else
    cout << "false";

  cout << endl;

  cout << endl;

  // Run through the files

  std::list<std::string>::const_iterator flash_filename_iter 
    = flash_filenames.begin();

  std::vector<double> min_coords;
  std::vector<double> max_coords;

  cout << "File" << "\t" << "Min_Coords" << "\t" << "Max_Coords" << endl;

  while (flash_filename_iter != flash_filenames.end())
    {
    const std::string & filename = *flash_filename_iter;

    QuickFlash::File::DataFile df(filename, read_buffer_size, true);

    const int state = find_extents(df, dataset_name, threshold_value,
				   find_region_above_threshold, min_coords,
				   max_coords);

    cout << filename;

    if (state == 0)
      cout << "\t" << min_coords << "\t" << max_coords;
    else
      cout << "\t" << "N/A" << "\t" << "N/A";

    cout << endl;

    ++ flash_filename_iter;
    }

  cout << endl;

  return 0;
  }
