// C++ program file find_flame_range.cpp

/*
  By Nathan C. Hearn
     March 7, 2008

  Program for determining approximate boundaries for the flame surface.

  Minimum and maximum values for the flame value are given by the user, as well
  as a number of axis sample points (called N for now)

  In 3D, planes normal to the x-direction are sampled along a regular NxN grid.
  If any sample points have flame values above the minimum flame value, that
  plane is considered behind the leading edge of the flame.  Similarly, if
  any sample points have flame values below the minimum flame value, that
  plane is considered ahead of the trailing edge of the flame.  A similar
  process occurs in 2D, with lines of N samples.

  The search algorithm is a bisection method, with the user specifying the
  number of iterations.
*/


#include <cstdlib>
#include <iostream>
#include <string>
#include "quickflash_file_datafile.hpp"
#include "flame_front.hpp"



int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  const int req_args = 6;

  if (argc < req_args)
    {
    cerr << endl << "Usage: " << argv[0] << " filename min_flame_value "
	 << "max_flame_value"
	 << endl << "\t"
	 << "axis_samples search_iterations" << endl << endl;
    return 1;
    }

  int argptr = 1;

  const std::string filename = argv[argptr++];

  const double min_flame_value = std::atof(argv[argptr++]);
  const double max_flame_value = std::atof(argv[argptr++]);

  if (!(min_flame_value < max_flame_value))
    {
    cerr << endl << "Error: Minimum flame value must be less than maximum"
	 << endl << endl;
    return -1;
    }

  const int int_axis_samples = std::atoi(argv[argptr++]);

  if (int_axis_samples < 1)
    {
    cerr << endl << "Error: Axis sample count must be positive" << endl 
	 << endl;
    return -1;
    }

  const unsigned int axis_samples 
    = static_cast<unsigned int>(int_axis_samples);

  const int int_search_iterations = std::atoi(argv[argptr++]);

  if (int_search_iterations < 1)
    {
    cerr << endl << "Error: Search iterations must be positive" << endl 
	 << endl;
    return -1;
    }

  const unsigned int search_iterations 
    = static_cast<unsigned int>(int_search_iterations);


  // Load the file

  QuickFlash::File::DataFile df(filename);


  // Perform the search

  double flame_min_x = 0.0;
  double flame_max_x = 0.0;

  double volume_min_x = 0.0;
  double volume_max_x = 0.0;

  find_flame_range(df, min_flame_value, max_flame_value, search_iterations, 
		   axis_samples, flame_min_x, flame_max_x, 
		   volume_min_x, volume_max_x);


  // Report the values

  cout << endl;

  cout << "Flame x-range " << flame_min_x << " " << flame_max_x << endl;
  cout << "Volume x-range " << volume_min_x << " " << volume_max_x << endl;

  cout << endl;

  return 0;
  }
