// C++ program file flame_front.cpp

/*
  By Nathan C. Hearn
     November 20, 2007

  Naive method for approximating the flame front position.
*/


#include "flame_front.hpp"
#include <vector>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_except.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_counters.hpp"
#include "file_defs.hpp"


const unsigned int buffer_size = 200;
const unsigned int cache_size = 1000;



double find_flame_front(const QuickFlash::File::DataFile & df, 
			const double flame_level,
			const unsigned int max_search_level, 
			const unsigned int axis_sample_points,
			double & volume_min_x, double & volume_max_x)
  {
  if (max_search_level < 1)
    throw QuickFlash::Except("At least one bisection level required", __FILE__,
			   __LINE__);

  if (axis_sample_points < 1)
    throw QuickFlash::Except("At least one sample point required", __FILE__, 
			   __LINE__);

  // Get the simulation information

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (dims < 1)
    throw QuickFlash::Except("At least one spatial dimension required", __FILE__,
			   __LINE__);

  // Open the dataset

  const QuickFlash::File::Dataset & dataset = df.get_dataset(Flame_Set_Name,
							   buffer_size);

  dataset.set_cache_size(cache_size);

  // Get the boundary information

  const std::vector<double> & min_bounds = meshinfo.get_volume_minbounds();
  const std::vector<double> & max_bounds = meshinfo.get_volume_maxbounds();

  const double min_x = min_bounds[0];
  const double max_x = max_bounds[0];

  volume_min_x = min_x;
  volume_max_x = max_x;

  // Build the sample counter

  QuickFlash::VectorCounter counter;

  std::vector<double> sample_start;
  std::vector<double> sample_delta;

  unsigned int num_sample_points = 1;

  if (dims > 1)
    {
    const std::vector<unsigned int> counter_dims((dims - 1), 
						 axis_sample_points); 

    counter.reset(counter_dims);
    
    sample_start.resize(dims - 1);
    sample_delta.resize(dims - 1);

    const double oneOver_axis_sample_points 
      = 1.0 / static_cast<double>(axis_sample_points);

    for (unsigned int i = 1; i < dims; i++)
      {
      const double axis_min = min_bounds[i];
      const double axis_max = max_bounds[i];

      const double axis_size = axis_max - axis_min;

      const double axis_delta = axis_size * oneOver_axis_sample_points;

      const double axis_start = axis_min + (0.5 * axis_delta);

      sample_start[i - 1] = axis_start;
      sample_delta[i - 1] = axis_delta;

      num_sample_points *= axis_sample_points;
      }
    }
  else
    {
    // Build a dummy counter for the 1D case

    const std::vector<unsigned int> counter_dims(1, 1);

    counter.reset(counter_dims);
    }
  
  const double oneOver_num_sample_points 
    = 1.0 / static_cast<double>(num_sample_points);

  double min_search_bound = min_x;
  double max_search_bound = max_x;

  bool block_loaded = false;

  unsigned int current_block_index = 0;

  QuickFlash::Block::BlockData<double> block_data;

  std::vector<double> pos(dims);

  // Find the flame front

  for (unsigned int search_level = 0; search_level < max_search_level; 
       search_level++)
    {
    const double test_position = 0.5 * (min_search_bound + max_search_bound);

    double accum_mean_flame_value = 0.0;

    pos[0] = test_position;

    counter.reset_counter();

    while (counter.in_bounds())
      {
      if (dims > 1)
	{
	for (unsigned int i = 1; i < dims; i++)
	  {
	  const unsigned int reduced_i = i - 1;

	  const unsigned int axis_pos_index = counter[reduced_i];

	  const double axis_pos = sample_start[reduced_i] 
	    + (static_cast<double>(axis_pos_index) * sample_delta[reduced_i]);

	  pos[i] = axis_pos;
	  }
	}

      unsigned int block_index = 0;
      unsigned int cell_index = 0;

      meshinfo.get_cell_index(pos, block_index, cell_index);

      if ((!block_loaded) || (current_block_index != block_index))
	{
	dataset.get_block_data(block_index, block_data);

	block_loaded = true;
	current_block_index = block_index;
	}

      const double cell_data = block_data[cell_index];

      accum_mean_flame_value += cell_data;

      counter.increment();
      }

    const double mean_flame_value 
      = accum_mean_flame_value * oneOver_num_sample_points;
    
    if (mean_flame_value < flame_level)
      max_search_bound = test_position;
    else
      min_search_bound = test_position;
    }

  return max_search_bound;
  }


void find_flame_range(const QuickFlash::File::DataFile & df, 
		      const double min_flame_level, 
		      const double max_flame_level, 
		      const unsigned int max_search_level, 
		      const unsigned int axis_sample_points,
		      double & flame_min_x, double & flame_max_x,
		      double & volume_min_x, double & volume_max_x)
  {
  if (max_search_level < 1)
    throw QuickFlash::Except("At least one bisection level required", __FILE__,
			   __LINE__);

  if (axis_sample_points < 1)
    throw QuickFlash::Except("At least one sample point required", __FILE__, 
			   __LINE__);

  // Get the simulation information

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (dims < 1)
    throw QuickFlash::Except("At least one spatial dimension required", __FILE__,
			   __LINE__);

  // Open the dataset

  const QuickFlash::File::Dataset & dataset = df.get_dataset(Flame_Set_Name,
							   buffer_size);

  dataset.set_cache_size(cache_size);

  // Get the boundary information

  const std::vector<double> & min_bounds = meshinfo.get_volume_minbounds();
  const std::vector<double> & max_bounds = meshinfo.get_volume_maxbounds();

  const double min_x = min_bounds[0];
  const double max_x = max_bounds[0];

  volume_min_x = min_x;
  volume_max_x = max_x;

  // Build the sample counter

  QuickFlash::VectorCounter counter;

  std::vector<double> sample_start;
  std::vector<double> sample_delta;

  unsigned int num_sample_points = 1;

  if (dims > 1)
    {
    const std::vector<unsigned int> counter_dims((dims - 1), 
						 axis_sample_points); 

    counter.reset(counter_dims);
    
    sample_start.resize(dims - 1);
    sample_delta.resize(dims - 1);

    const double oneOver_axis_sample_points 
      = 1.0 / static_cast<double>(axis_sample_points);

    for (unsigned int i = 1; i < dims; i++)
      {
      const double axis_min = min_bounds[i];
      const double axis_max = max_bounds[i];

      const double axis_size = axis_max - axis_min;

      const double axis_delta = axis_size * oneOver_axis_sample_points;

      const double axis_start = axis_min + (0.5 * axis_delta);

      sample_start[i - 1] = axis_start;
      sample_delta[i - 1] = axis_delta;

      num_sample_points *= axis_sample_points;
      }
    }
  else
    {
    // Build a dummy counter for the 1D case

    const std::vector<unsigned int> counter_dims(1, 1);

    counter.reset(counter_dims);
    }
  
  double front_min_search_bound = min_x;
  double front_max_search_bound = max_x;

  double back_min_search_bound = min_x;
  double back_max_search_bound = max_x;

  bool block_loaded = false;

  unsigned int current_block_index = 0;

  QuickFlash::Block::BlockData<double> block_data;

  std::vector<double> pos(dims);

  // Find the flame front (leading edge)

  for (unsigned int search_level = 0; search_level < max_search_level; 
       search_level++)
    {
    const double test_position 
      = 0.5 * (front_min_search_bound + front_max_search_bound);

    pos[0] = test_position;

    counter.reset_counter();

    bool above_min_flame_level = false;
    bool below_max_flame_level = false;

    bool end_early = false;

    while (counter.in_bounds() && (!end_early))
      {
      if (dims > 1)
	{
	for (unsigned int i = 1; i < dims; i++)
	  {
	  const unsigned int reduced_i = i - 1;

	  const unsigned int axis_pos_index = counter[reduced_i];

	  const double axis_pos = sample_start[reduced_i] 
	    + (static_cast<double>(axis_pos_index) * sample_delta[reduced_i]);

	  pos[i] = axis_pos;
	  }
	}

      unsigned int block_index = 0;
      unsigned int cell_index = 0;

      meshinfo.get_cell_index(pos, block_index, cell_index);

      if ((!block_loaded) || (current_block_index != block_index))
	{
	dataset.get_block_data(block_index, block_data);

	block_loaded = true;
	current_block_index = block_index;
	}

      const double cell_data = block_data[cell_index];

      if (cell_data > min_flame_level)
	above_min_flame_level = true;

      if (cell_data < max_flame_level)
	below_max_flame_level = true;

      if (above_min_flame_level && below_max_flame_level)
	end_early = true;

      counter.increment();
      }

    if (above_min_flame_level)
      front_min_search_bound = test_position;
    else
      front_max_search_bound = test_position;

    if (!below_max_flame_level)
      back_min_search_bound = test_position;
    else if (test_position < back_max_search_bound)
      back_max_search_bound = test_position;
    }


  // Now, find the trailing (back) edge of the flame

  if (front_max_search_bound < back_max_search_bound)
    back_max_search_bound = front_max_search_bound;

  for (unsigned int search_level = 0; search_level < max_search_level; 
       search_level++)
    {
    const double test_position 
      = 0.5 * (back_min_search_bound + back_max_search_bound);

    pos[0] = test_position;

    counter.reset_counter();

    bool below_max_flame_level = false;

    while (counter.in_bounds() && (!below_max_flame_level))
      {
      if (dims > 1)
	{
	for (unsigned int i = 1; i < dims; i++)
	  {
	  const unsigned int reduced_i = i - 1;

	  const unsigned int axis_pos_index = counter[reduced_i];

	  const double axis_pos = sample_start[reduced_i] 
	    + (static_cast<double>(axis_pos_index) * sample_delta[reduced_i]);

	  pos[i] = axis_pos;
	  }
	}

      unsigned int block_index = 0;
      unsigned int cell_index = 0;

      meshinfo.get_cell_index(pos, block_index, cell_index);

      if ((!block_loaded) || (current_block_index != block_index))
	{
	dataset.get_block_data(block_index, block_data);

	block_loaded = true;
	current_block_index = block_index;
	}

      const double cell_data = block_data[cell_index];

      if (cell_data < max_flame_level)
	below_max_flame_level = true;

      counter.increment();
      }

    if (below_max_flame_level)
      back_max_search_bound = test_position;
    else
      back_min_search_bound = test_position;
    }

  flame_min_x = back_min_search_bound;
  flame_max_x = front_max_search_bound;
  }
