// C++ program file flame_slicer.cpp

/*
  By Nathan C. Hearn
     November 20, 2007

  Slice and diagnostics generator for AFL simulations.
*/


#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <ostream>
#include <fstream>
#include <list>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_file_flashdefs.hpp"
#include "quickflash_file_siminfo.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_except.hpp"
#include "quickflash_slice_slicegrid.hpp"
#include "quickflash_slice_slicer.hpp"
#include "quickflash_color_colormap.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_image_imagegrid.hpp"
#include "quickflash_image_imagemagick.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_slice_output.hpp"
#include "flame_front.hpp"


const unsigned int default_cross_resolution = 256;

const char option_separator = ',';

const double deg_to_rad = (M_PI) / 180.0;

const double xy_roll_angle = 0.0 * deg_to_rad;
const double xy_polar_angle = 0.0 * deg_to_rad;
const double xy_azim_angle = 0.0 * deg_to_rad;

const double xz_roll_angle = 90.0 * deg_to_rad;
const double xz_polar_angle = 90.0 * deg_to_rad;
const double xz_azim_angle = -90.0 * deg_to_rad;

const double yz_roll_angle = 90.0 * deg_to_rad;
const double yz_polar_angle = 90.0 * deg_to_rad;
const double yz_azim_angle = 0.0 * deg_to_rad;


const unsigned int read_buffer_size = 50;
const unsigned int read_cache_size = 3000;

const double flame_value = 0.01;
const unsigned int max_search_level = 12;
const unsigned int axis_samples = 8;


const double image_height = 1.5e+06;
const double image_width = 10.0 * image_height;

const double image_width_buffer = 2.0e+06;

const char image_suffix[] = "png";


const double xsect_width = 1.5e+06;
const double xsect_height = 1.5e+06;

const unsigned int num_xsects = 9;

const double xsect_delta_x = image_width / static_cast<double>(num_xsects - 1);

const char xsect_name[] = "xsect";

const double xsect_top_bottom_delta = 0.01e+06;
const double xsect_flame_pos_delta = -1.0e+05;

const char xsect_index_type[] = "index";

const char xsect_top_type[] = "top";
const char xsect_bottom_type[] = "bottom";

const char xsect_flame_type[] = "flame";

const char xy_plane_label[] = "xy";
const char xz_plane_label[] = "xz";


void parse_dataset_name_option(const std::string & dataset_name_argument,
			       std::vector<std::string> & dataset_name_list)
  {
  dataset_name_list.clear();

  const std::string::size_type length = dataset_name_argument.size();

  std::string::size_type index = 0;

  std::string work_string;

  while (index < length)
    {
    std::string::size_type next_index 
      = dataset_name_argument.find(option_separator, index);

    if (next_index == std::string::npos)
      next_index = length;

    const std::string::size_type copy_length = next_index - index;

    work_string.assign(dataset_name_argument, index, copy_length);

    dataset_name_list.push_back(work_string);

    index = next_index + 1;
    }
  }


void filter_variable_names(const std::vector<std::string> & selected_variables,
			   const std::vector<std::string> & datafile_variables,
			   std::vector<std::string> & filtered_variables)
  {
  const unsigned int num_selected = selected_variables.size();

  const unsigned int num_datafile = datafile_variables.size();

  filtered_variables.clear();

  for (unsigned int datafile_index = 0; datafile_index < num_datafile;
       datafile_index++)
    {
    const std::string & datafile_name = datafile_variables[datafile_index];

    bool found_it = false;

    unsigned int selected_index = 0;

    while ((!found_it) && (selected_index < num_selected))
      {
      if (selected_variables[selected_index] == datafile_name)
	found_it = true;
      else
	selected_index++;
      }

    if (found_it)
      filtered_variables.push_back(datafile_name);
    }
  }


#ifdef USE_MAGICK
int write_image(const std::string & filename, 
		const QuickFlash::Slice::SliceGrid & data)
  {
  int ret_val = -1;  // Error state

  // Build the colormap

  double min_val = 0.0;
  double max_val = 0.0;

  data.get_minmax_values(min_val, max_val);

  if (!(min_val < max_val))
    max_val = (2.0 * std::fabs(min_val)) + 1.0;  // Make something up ...

  QuickFlash::Color::Colormap colormap(min_val, max_val);

  QuickFlash::Image::ImageGrid image(data.get_grid_data(), colormap);

  ret_val = QuickFlash::Image::ImageMagick::write_image(filename, image);

  return ret_val;
  }
#endif  // USE_MAGICK


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 1;

  if (num_args < min_args)
    {
    cerr << endl << "Usage: " << argv[0] << " flash_filename "
#ifdef USE_MAGICK
	 << endl << "\t"
	 << "[ --make-image ] [ --image-only ]"
#endif  // USE_MAGICK
	 << endl << "\t"
	 << "[ --cross-resolution=pixels ]"
	 << endl << "\t"
	 << "[ --datasets=dataset1,dataset2,... ]"
	 << endl << "\t"
	 << "  [ --output-prefix=filename_prefix ]" << endl << endl;
    return 1;
    }


  // Read the arguments

  unsigned int argptr = 0;

  const std::string filename = argument_list[argptr++];

  
  // Process the options

#ifdef USE_MAGICK
  bool output_image = false;
#endif  // USE_MAGICK

  bool output_data = true;

  unsigned int cross_resolution = default_cross_resolution;

  std::string output_prefix = filename;  // !!!

  std::vector<std::string> dataset_name_list;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "make-image")
      {
#ifdef USE_MAGICK
      output_image = true;
#else
      cerr << endl << "Error: Image output support not present" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "image-only")
      {
#ifdef USE_MAGICK
      output_image = true;
      output_data = false;
#else
      cerr << endl << "Error: Image output support not present" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "cross-resolution")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Cross-resolution pixel count required" << endl
	     << endl;
	return -1;
	}

      int int_cross_res = 0;

      current_option.get_value(int_cross_res);

      if (int_cross_res < 1)
	{
	cerr << endl << "Error: Cross-resolution must be a positive integer"
	     << endl << endl;
	return -1;
	}

      cross_resolution = static_cast<unsigned int>(int_cross_res);
      }
    else if (keyword == "datasets")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Datasets option requires list of dataset names"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(dataset_name_list);
      }
    else if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output prefix option requires a string"
	     << endl << endl;
	return -1;
	}

      output_prefix = current_option.get_value();
      }
    else
      {
      cerr << "Error: Keyword [ " << keyword << " ] not recognized" << endl
	   << endl;
      return -1;
      }

    ++option_iter;
    }


  // Set up some additional parameters

  const std::string output_file_prefix = output_prefix + "_";

  const std::string metadata_filename = output_file_prefix + "metadata.out";

  const unsigned int image_height_pixels = cross_resolution;
  const unsigned int image_width_pixels = 10 * image_height_pixels;

  const unsigned int xsect_width_pixels = cross_resolution;
  const unsigned int xsect_height_pixels = cross_resolution;


  // Open the file

  QuickFlash::File::DataFile df(filename, read_buffer_size, true);

  const QuickFlash::File::SimInfo & siminfo = df.get_sim_info();  // Var names

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (dims < 2)
    {
    cerr << endl << "Error: At least two dimensions required" << endl << endl;
    return -2;
    }

  // Open the metadata output file

  std::ofstream mdata(metadata_filename.c_str());

  // Find the approximate flame front and image bounds

  double volume_min_x = 0.0;
  double volume_max_x = 0.0;

  const double flame_pos = find_flame_front(df, flame_value, max_search_level,
					    axis_samples, volume_min_x,
					    volume_max_x);

  cerr << endl << "Estimated flame position [ " << flame_pos << " ]" << endl;

  double image_x_max = flame_pos + image_width_buffer;
  double image_x_min = image_x_max - image_width;

  if (image_x_max > volume_max_x)
    {
    image_x_max = volume_max_x;
    image_x_min = image_x_max - image_width;
    }
  else if (image_x_min < volume_min_x)
    {
    image_x_min = volume_min_x;
    image_x_max = image_x_min + image_width;
    }

  const double image_y_min = -0.5 * image_height;
  const double image_y_max = 0.5 * image_height;

  std::vector<double> grid_min_coords(2);
  std::vector<double> grid_max_coords(2);

  grid_min_coords[0] = image_x_min;
  grid_min_coords[1] = image_y_min;

  grid_max_coords[0] = image_x_max;
  grid_max_coords[1] = image_y_max;

  std::vector<double> grid_center(dims, 0.0);

  grid_center[0] = 0.5 * (image_x_min + image_x_max);
  grid_center[1] = 0.5 * (image_y_min + image_y_max);

  std::vector<double> grid_width(2);

  grid_width[0] = image_width;
  grid_width[1] = image_height;

  std::vector<unsigned int> grid_pixels(2);

  grid_pixels[0] = image_width_pixels;
  grid_pixels[1] = image_height_pixels;
  
  std::vector<double> xsect_grid_center(dims, 0.0);  // x-coord is variable

  std::vector<double> xsect_grid_min_coords(2);
  std::vector<double> xsect_grid_max_coords(2);

  xsect_grid_min_coords[0] = -(0.5) * xsect_width;
  xsect_grid_min_coords[1] = -(0.5) * xsect_height;

  xsect_grid_max_coords[0] = 0.5 * xsect_width;
  xsect_grid_max_coords[1] = 0.5 * xsect_height;

  std::vector<double> xsect_grid_width(2);

  xsect_grid_width[0] = xsect_width;
  xsect_grid_width[1] = xsect_height;

  std::vector<unsigned int> xsect_grid_pixels(2);

  xsect_grid_pixels[0] = xsect_width_pixels;
  xsect_grid_pixels[1] = xsect_height_pixels;

  // Find the top, bottom, and flame-front cross-section positions

  const double xsect_top_x_pos = volume_max_x - xsect_top_bottom_delta;
  const double xsect_bottom_x_pos = volume_min_x + xsect_top_bottom_delta;

  const double xsect_flame_x_pos = flame_pos + xsect_flame_pos_delta;

  // Build the comment list

  std::list<std::string> comment_list;

  std::stringstream strstream;

  strstream << "Flash file [ " << filename << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "Time [ " << meshinfo.get_sim_time() << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "Flame pos [ " << flame_pos << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "Grid center [ " << grid_center << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "Grid width [ " << grid_width << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "Grid min coords [ " << grid_min_coords << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "Grid max coords [ " << grid_max_coords << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "Grid pixels [ " << grid_pixels << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSect grid center [ " << xsect_grid_center << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSect grid width [ " << xsect_grid_width << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSect grid min coords [ " << xsect_grid_min_coords << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSect grid max coords [ " << xsect_grid_max_coords << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSect grid pixels [ " << xsect_grid_pixels << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSect top x-pos [ " << xsect_top_x_pos << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSect bottom x-pos [ " << xsect_bottom_x_pos << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSect flame-front x-pos [ " << xsect_flame_x_pos << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  // Report comments to stdout

  std::list<std::string>::const_iterator str_iter = comment_list.begin();
  std::list<std::string>::const_iterator end_str_iter = comment_list.end();

  while (str_iter != end_str_iter)
    {
    mdata << *str_iter << endl;
    ++str_iter;
    }

  // Construct the slice grid

  QuickFlash::Slice::SliceGrid grid_xy(grid_center, grid_width, grid_pixels, 
				     xy_azim_angle, xy_polar_angle, 
				     xy_roll_angle);
  QuickFlash::Slice::SliceGrid grid_xz(grid_center, grid_width, grid_pixels, 
				     xz_azim_angle, xz_polar_angle, 
				     xz_roll_angle);

  QuickFlash::Slice::SliceGrid xsect_grid;

  std::vector<unsigned int> subsample_dims(dims, 1);

  // Get the variable names

  const std::vector<std::string> & datafile_var_names 
    = siminfo.get_mesh_variable_names();

  std::vector<std::string> var_names;

  if (dataset_name_list.size() > 0)
    filter_variable_names(dataset_name_list, datafile_var_names, var_names);
  else
    var_names = datafile_var_names;

  const unsigned int num_vars = var_names.size();

  for (unsigned int var_index = 0; var_index < num_vars; var_index++)
    {
    const std::string & dataset_name = var_names[var_index];

    const QuickFlash::File::Dataset & dataset = df.get_dataset(dataset_name);

    const double dataset_min_value = dataset.get_min_value();
    const double dataset_max_value = dataset.get_max_value();

    std::list<std::string> current_comments = comment_list;

    strstream << "Dataset [ " + dataset_name + " ] Min_Value [ " 
	      << dataset_min_value
	      << " ] Max_Value [ " << dataset_max_value << " ]";

    const std::string dataset_comment = strstream.str();

    strstream.str("");

    current_comments.push_back(dataset_comment);

    mdata << dataset_comment << endl;
    cerr << dataset_comment << endl;

    const std::string current_output_file_prefix 
      = output_file_prefix + dataset_name;

    // Fill the grid

    slice_simple(meshinfo, dataset, subsample_dims, grid_xy);

    cerr << "  Finished filling xy grid" << endl;

    slice_simple(meshinfo, dataset, subsample_dims, grid_xz);

    cerr << "  Finished filling xz grid" << endl;

    double grid_xy_min_value = 0.0;
    double grid_xy_max_value = 0.0;

    grid_xy.get_minmax_values(grid_xy_min_value, grid_xy_max_value);

    double grid_xz_min_value = 0.0;
    double grid_xz_max_value = 0.0;

    grid_xz.get_minmax_values(grid_xz_min_value, grid_xz_max_value);

    if (output_data)
      {
      const std::string xy_output_filename 
	= current_output_file_prefix + "_" + xy_plane_label + ".csv";

      mdata << "  Data [ " << xy_plane_label 
	    << " ] File [ " << xy_output_filename 
	    << " ] Min_Coords [ " << grid_min_coords 
	    << " ] Max_Coords [ " << grid_max_coords
	    << " ] Min_Value [ " << grid_xy_min_value
	    << " ] Max_Value [ " << grid_xy_max_value << " ]" << endl;

      QuickFlash::Slice::write_grid(xy_output_filename, grid_xy, 
				  current_comments, true);

      const std::string xz_output_filename 
	= current_output_file_prefix + "_" + xz_plane_label + ".csv";

      mdata << "  Data [ " << xz_plane_label 
	    << " ] File [ " << xz_output_filename 
	    << " ] Min_Coords [ " << grid_min_coords 
	    << " ] Max_Coords [ " << grid_max_coords
	    << " ] Min_Value [ " << grid_xz_min_value
	    << " ] Max_Value [ " << grid_xz_max_value << " ]" << endl;

      QuickFlash::Slice::write_grid(xz_output_filename, grid_xz, 
				  current_comments, true);
      }

#ifdef USE_MAGICK
    if (output_image)
      {
      const std::string xy_image_filename 
	= current_output_file_prefix + "_" + xy_plane_label + "." 
	  + image_suffix;

      mdata << "  Image [ " << xy_plane_label 
	    << " ] File [ " << xy_image_filename 
	    << " ] Min_Coords [ " << grid_min_coords 
	    << " ] Max_Coords [ " << grid_max_coords
	    << " ] Min_Value [ " << grid_xy_min_value
	    << " ] Max_Value [ " << grid_xy_max_value << " ]" << endl;

      int error_state = write_image(xy_image_filename, grid_xy);

      if (error_state < 0)
	{
	cerr << endl << "Error: Unable to write xy image " 
	     << xy_image_filename << endl << endl;
	return -2;
	}

      const std::string xz_image_filename 
	= current_output_file_prefix + "_" + xz_plane_label + "." 
	  + image_suffix;

      mdata << "  Image [ " << xz_plane_label 
	    << " ] File [ " << xz_image_filename 
	    << " ] Min_Coords [ " << grid_min_coords 
	    << " ] Max_Coords [ " << grid_max_coords
	    << " ] Min_Value [ " << grid_xz_min_value
	    << " ] Max_Value [ " << grid_xz_max_value << " ]" << endl;

      error_state = write_image(xz_image_filename, grid_xz);

      if (error_state < 0)
	{
	cerr << endl << "Error: Unable to write xz image " 
	     << xz_image_filename << endl << endl;
	return -2;
	}
      }
#endif  // USE_MAGICK

    const std::string xsect_prefix = current_output_file_prefix + "_xsect_";

#if 0
    // Get cross-sections along sliced region

    for (unsigned int xsect_index = 0; xsect_index < num_xsects; xsect_index++)
      {
      std::list<std::string> xsect_comments = current_comments;

      // Build the filename

      strstream << xsect_index;

      const std::string xsect_index_str = strstream.str();

      strstream.str("");

      const std::string xsect_output_filename_nosuffix
	= xsect_prefix + "s" + xsect_index_str + ".";

      // Find the x-position and populate the grid

      const double xsect_x_pos = image_x_min
	+ (xsect_delta_x * static_cast<double>(xsect_index));

      xsect_grid_center[0] = xsect_x_pos;

      xsect_grid.reset(xsect_grid_center, xsect_grid_width, xsect_grid_pixels,
		       yz_azim_angle, yz_polar_angle, yz_roll_angle);

      slice_simple(meshinfo, dataset, subsample_dims, xsect_grid);

      cerr << "  Finished filling cross-section " << xsect_index << endl;

      double xsect_grid_min_value = 0.0;
      double xsect_grid_max_value = 0.0;

      xsect_grid.get_minmax_values(xsect_grid_min_value, xsect_grid_max_value);

      if (output_data)
	{
	const std::string output_filename 
	  = xsect_output_filename_nosuffix + "csv";

	strstream << "XSect [ Data ] Index [ " << xsect_index
		  << " ] Type [ " << xsect_index_type
		  << " ] Pos [ " << xsect_x_pos 
		  << " ] File [ " << output_filename 
		  << " ] Min_Coords [ " << xsect_grid_min_coords 
		  << " ] Max_Coords [ " << xsect_grid_max_coords
		  << " ] Min_Value [ " << xsect_grid_min_value
		  << " ] Max_Value [ " << xsect_grid_max_value 
		  << " ]";

	const std::string file_comment = strstream.str();

	strstream.str("");

	xsect_comments.push_back(file_comment);

	mdata << "  " << file_comment << endl;

	QuickFlash::Slice::write_grid(output_filename, xsect_grid, 
				    xsect_comments, true);
	}

#ifdef USE_MAGICK
      if (output_image)
	{
	const std::string image_filename 
	  = xsect_output_filename_nosuffix + image_suffix;

	strstream << "XSect [ Image ] Index [ " << xsect_index 
		  << " ] Type [ " << xsect_index_type
		  << " ] Pos [ " << xsect_x_pos 
		  << " ] File [ " << image_filename 
		  << " ] Min_Coords [ " << xsect_grid_min_coords 
		  << " ] Max_Coords [ " << xsect_grid_max_coords
		  << " ] Min_Value [ " << xsect_grid_min_value
		  << " ] Max_Value [ " << xsect_grid_max_value 
		  << " ]";

	const std::string file_comment = strstream.str();

	strstream.str("");

	mdata << "  " << file_comment << endl;

	const int error_state = write_image(image_filename, xsect_grid);

	if (error_state < 0)
	  {
	  cerr << endl << "Error: Unable to write image " << image_filename 
	       << endl << endl;
	  return -2;
	  }
	}
#endif  // USE_MAGICK
      }
#endif  // 0

    // Get the top cross-section

    std::list<std::string> xsect_comments_top = current_comments;

    // Build the filename

    const std::string xsect_top_output_filename_nosuffix
      = xsect_prefix + xsect_top_type + ".";

    // Set the x-position and populate the grid

    xsect_grid_center[0] = xsect_top_x_pos;

    xsect_grid.reset(xsect_grid_center, xsect_grid_width, xsect_grid_pixels,
		     yz_azim_angle, yz_polar_angle, yz_roll_angle);

    slice_simple(meshinfo, dataset, subsample_dims, xsect_grid);

    cerr << "  Finished filling top cross-section" << endl;

    double xsect_top_grid_min_value = 0.0;
    double xsect_top_grid_max_value = 0.0;

    xsect_grid.get_minmax_values(xsect_top_grid_min_value, 
				 xsect_top_grid_max_value);

    if (output_data)
      {
      const std::string output_filename 
	= xsect_top_output_filename_nosuffix + "csv";

      strstream << "XSect [ Data ] Type [ " << xsect_top_type
		  << " ] Pos [ " << xsect_top_x_pos 
		  << " ] File [ " << output_filename 
		  << " ] Min_Coords [ " << xsect_grid_min_coords 
		  << " ] Max_Coords [ " << xsect_grid_max_coords
		  << " ] Min_Value [ " << xsect_top_grid_min_value
		  << " ] Max_Value [ " << xsect_top_grid_max_value 
		  << " ]";

      const std::string file_comment = strstream.str();

      strstream.str("");

      xsect_comments_top.push_back(file_comment);

      mdata << "  " << file_comment << endl;

      QuickFlash::Slice::write_grid(output_filename, xsect_grid, 
				  xsect_comments_top, true);
      }

#ifdef USE_MAGICK
    if (output_image)
      {
      const std::string image_filename 
	= xsect_top_output_filename_nosuffix + image_suffix;

      strstream << "XSect [ Image ] Type [ " << xsect_top_type 
		<< " ] Pos [ " << xsect_top_x_pos 
		<< " ] File [ " << image_filename 
		<< " ] Min_Coords [ " << xsect_grid_min_coords 
		<< " ] Max_Coords [ " << xsect_grid_max_coords
		<< " ] Min_Value [ " << xsect_top_grid_min_value
		<< " ] Max_Value [ " << xsect_top_grid_max_value 
		<< " ]";

      const std::string file_comment = strstream.str();

      strstream.str("");

      mdata << "  " << file_comment << endl;

      const int error_state = write_image(image_filename, xsect_grid);

      if (error_state < 0)
	{
	cerr << endl << "Error: Unable to write image " << image_filename 
	     << endl << endl;
	return -2;
	}
      }
#endif  // USE_MAGICK


    // Get the bottom cross-section

    std::list<std::string> xsect_comments_bottom = current_comments;

    // Build the filename

    const std::string xsect_bottom_output_filename_nosuffix
      = xsect_prefix + xsect_bottom_type + ".";

    // Set the x-position and populate the grid

    xsect_grid_center[0] = xsect_bottom_x_pos;

    xsect_grid.reset(xsect_grid_center, xsect_grid_width, xsect_grid_pixels,
		     yz_azim_angle, yz_polar_angle, yz_roll_angle);

    slice_simple(meshinfo, dataset, subsample_dims, xsect_grid);

    cerr << "  Finished filling bottom cross-section" << endl;

    double xsect_bottom_grid_min_value = 0.0;
    double xsect_bottom_grid_max_value = 0.0;

    xsect_grid.get_minmax_values(xsect_bottom_grid_min_value, 
				 xsect_bottom_grid_max_value);

    if (output_data)
      {
      const std::string output_filename 
	= xsect_bottom_output_filename_nosuffix + "csv";

      strstream << "XSect [ Data ] Type [ " << xsect_bottom_type
		  << " ] Pos [ " << xsect_bottom_x_pos 
		  << " ] File [ " << output_filename 
		  << " ] Min_Coords [ " << xsect_grid_min_coords 
		  << " ] Max_Coords [ " << xsect_grid_max_coords
		  << " ] Min_Value [ " << xsect_bottom_grid_min_value
		  << " ] Max_Value [ " << xsect_bottom_grid_max_value 
		  << " ]";

      const std::string file_comment = strstream.str();

      strstream.str("");

      xsect_comments_bottom.push_back(file_comment);

      mdata << "  " << file_comment << endl;

      QuickFlash::Slice::write_grid(output_filename, xsect_grid, 
				  xsect_comments_bottom, true);
      }

#ifdef USE_MAGICK
    if (output_image)
      {
      const std::string image_filename 
	= xsect_bottom_output_filename_nosuffix + image_suffix;

      strstream << "XSect [ Image ] Type [ " << xsect_bottom_type 
		<< " ] Pos [ " << xsect_bottom_x_pos 
		<< " ] File [ " << image_filename 
		<< " ] Min_Coords [ " << xsect_grid_min_coords 
		<< " ] Max_Coords [ " << xsect_grid_max_coords
		<< " ] Min_Value [ " << xsect_bottom_grid_min_value
		<< " ] Max_Value [ " << xsect_bottom_grid_max_value 
		<< " ]";

      const std::string file_comment = strstream.str();

      strstream.str("");

      mdata << "  " << file_comment << endl;

      const int error_state = write_image(image_filename, xsect_grid);

      if (error_state < 0)
	{
	cerr << endl << "Error: Unable to write image " << image_filename 
	     << endl << endl;
	return -2;
	}
      }
#endif  // USE_MAGICK


    // Get the flame cross-section

    std::list<std::string> xsect_comments_flame = current_comments;

    // Build the filename

    const std::string xsect_flame_output_filename_nosuffix
      = xsect_prefix + xsect_flame_type + ".";

    // Set the x-position and populate the grid

    xsect_grid_center[0] = xsect_flame_x_pos;

    xsect_grid.reset(xsect_grid_center, xsect_grid_width, xsect_grid_pixels,
		     yz_azim_angle, yz_polar_angle, yz_roll_angle);

    slice_simple(meshinfo, dataset, subsample_dims, xsect_grid);

    cerr << "  Finished filling flame-front cross-section" << endl;

    double xsect_flame_grid_min_value = 0.0;
    double xsect_flame_grid_max_value = 0.0;

    xsect_grid.get_minmax_values(xsect_flame_grid_min_value, 
				 xsect_flame_grid_max_value);

    if (output_data)
      {
      const std::string output_filename 
	= xsect_flame_output_filename_nosuffix + "csv";

      strstream << "XSect [ Data ] Type [ " << xsect_flame_type
		  << " ] Pos [ " << xsect_flame_x_pos 
		  << " ] File [ " << output_filename 
		  << " ] Min_Coords [ " << xsect_grid_min_coords 
		  << " ] Max_Coords [ " << xsect_grid_max_coords
		  << " ] Min_Value [ " << xsect_flame_grid_min_value
		  << " ] Max_Value [ " << xsect_flame_grid_max_value 
		  << " ]";

      const std::string file_comment = strstream.str();

      strstream.str("");

      xsect_comments_flame.push_back(file_comment);

      mdata << "  " << file_comment << endl;

      QuickFlash::Slice::write_grid(output_filename, xsect_grid, 
				  xsect_comments_flame, true);
      }

#ifdef USE_MAGICK
    if (output_image)
      {
      const std::string image_filename 
	= xsect_flame_output_filename_nosuffix + image_suffix;

      strstream << "XSect [ Image ] Type [ " << xsect_flame_type 
		<< " ] Pos [ " << xsect_flame_x_pos 
		<< " ] File [ " << image_filename 
		<< " ] Min_Coords [ " << xsect_grid_min_coords 
		<< " ] Max_Coords [ " << xsect_grid_max_coords
		<< " ] Min_Value [ " << xsect_flame_grid_min_value
		<< " ] Max_Value [ " << xsect_flame_grid_max_value 
		<< " ]";

      const std::string file_comment = strstream.str();

      strstream.str("");

      mdata << "  " << file_comment << endl;

      const int error_state = write_image(image_filename, xsect_grid);

      if (error_state < 0)
	{
	cerr << endl << "Error: Unable to write image " << image_filename 
	     << endl << endl;
	return -2;
	}
      }
#endif  // USE_MAGICK


    // Free some memory

    dataset.clear_cache();
    }

  cout << endl;

  return 0;
  }
