// C++ program file collect_flame_data.cpp

/*
  By Nathan C. Hearn
     December 6, 2008

  Determine mass and mass fraction distributions for flame.  Select and save
  planes for later structure function analysis.
*/


// Include the MPI headers first to make MPI/xlC on BG/P happy ... (grrrr....)
#ifdef USE_MPI
#include <mpi.h>
#endif  // USE_MPI


#include <string>
#include <iostream>
#include <vector>
#include <fstream>
#include <cstdlib>
#include <cmath>
#include <list>
#include <sstream>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_slice_slicegrid.hpp"
#include "quickflash_slice_slicer.hpp"
#include "quickflash_slice_output.hpp"
#include "flame_dist.hpp"
#include "flame_front.hpp"
#include "decomp_utils.hpp"


const unsigned int default_flame_search_level = 7;
const unsigned int default_flame_search_samples = 10;

const char default_image_type[] = "png";

const unsigned int default_buffer_size = 50;
const unsigned int default_cache_size = 1000;

const char flam_name[] = "flam";
const char dens_name[] = "dens";

const char velx_name[] = "velx";
const char vely_name[] = "vely";
const char velz_name[] = "velz";

const double deg_to_rad = (M_PI) / 180.0;

const double xy_roll_angle = 0.0 * deg_to_rad;
const double xy_polar_angle = 0.0 * deg_to_rad;
const double xy_azim_angle = 0.0 * deg_to_rad;

const double xz_roll_angle = 90.0 * deg_to_rad;
const double xz_polar_angle = 90.0 * deg_to_rad;
const double xz_azim_angle = -90.0 * deg_to_rad;

const double yz_roll_angle = 90.0 * deg_to_rad;
const double yz_polar_angle = 90.0 * deg_to_rad;
const double yz_azim_angle = 0.0 * deg_to_rad;

const double slice_expansion_factor = 1.1;  // 10% increase

enum SliceOriginType { MeshOrigin, FlameTip, FlameMassCenter, 
		       FlameMassFractCenter };

const char origin_meshorigin_label[] = "mesh";
const char origin_flametip_label[] = "tip";
const char origin_flamemasscenter_label[] = "masscenter";
const char origin_flamemassfractcenter_label[] = "massfractcenter";

const SliceOriginType default_slice_origin_type = MeshOrigin;


int read_origin_type_label(const std::string & origin_type_label,
			   SliceOriginType & slice_origin_type)
  {
  int ret_val = 0;  // OK

  if (origin_type_label == origin_meshorigin_label)
    slice_origin_type = MeshOrigin;
  else if (origin_type_label == origin_flametip_label)
    slice_origin_type = FlameTip;
  else if (origin_type_label == origin_flamemasscenter_label)
    slice_origin_type = FlameMassCenter;
  else if (origin_type_label == origin_flamemassfractcenter_label)
    slice_origin_type = FlameMassFractCenter;
  else
    ret_val = -1;  // Error state

  return ret_val;
  }


int get_origin_type_label(const SliceOriginType slice_origin_type,
			  std::string & origin_type_label)
  {
  int ret_val = 0;  // OK

  switch (slice_origin_type)
    {
    case MeshOrigin :

      origin_type_label = origin_meshorigin_label;
      break;

    case FlameTip :

      origin_type_label = origin_flametip_label;
      break;

    case FlameMassCenter :

      origin_type_label = origin_flamemasscenter_label;
      break;

    case FlameMassFractCenter :

      origin_type_label = origin_flamemassfractcenter_label;
      break;

    default :

      ret_val = -1;  // Error state
    }

  return ret_val;
  }


void read_file_list(const std::string & file_list_filename, 
		    std::vector<std::string> & file_list)
  {
  file_list.clear();

  std::ifstream f(file_list_filename.c_str());

  std::string name_string;

  while (!(f.eof()))
    {
    f >> name_string;

    if (name_string.size() > 0)
      file_list.push_back(name_string);
    }
  }


void compute_flame_distribution(const std::string & output_filename,
				const std::string & comment_marker,
				const QuickFlash::File::DataFile & df,
				const double flame_envelope_min_x,
				const double flame_envelope_max_x,
				const double mass_distrib_bin_width,
				const double min_flame_level,
				const double max_flame_level,
				double & flame_center_mass_fract,
				double & flame_center_mass,
				double & flame_min_pos, double & flame_max_pos)
  {
  using std::endl;

  const double flame_envelope_width 
    = flame_envelope_max_x - flame_envelope_min_x;

  const unsigned int distrib_bin_count 
    = static_cast<unsigned int>(flame_envelope_width / mass_distrib_bin_width)
      + 1;

  FlameDistribution flame_dist(df, 0, flame_envelope_min_x, distrib_bin_count,
			       mass_distrib_bin_width, min_flame_level, 
			       max_flame_level);

  flame_center_mass_fract = flame_dist.get_center_pos_mass_fract();
  flame_center_mass = flame_dist.get_center_pos_mass();

  flame_min_pos = flame_dist.get_flame_min_pos();
  flame_max_pos = flame_dist.get_flame_max_pos();

  const std::vector<double> & distrib_bin_centers 
    = flame_dist.get_bin_centers();

  const std::vector<double> & distrib_bin_volumes 
    = flame_dist.get_bin_volumes();

  const std::vector<double> & distrib_mass_bins_total 
    = flame_dist.get_mass_bins_total();

  const std::vector<double> & distrib_dens_bins_total_mean
    = flame_dist.get_dens_bins_total_mean();

  const std::vector<double> & distrib_dens_bins_total_dispersion
    = flame_dist.get_dens_bins_total_dispersion();

  const std::vector<double> & distrib_mass_bins_burned 
    = flame_dist.get_mass_bins_burned();

  const std::vector<double> & distrib_dens_bins_burned_mean
    = flame_dist.get_dens_bins_burned_mean();

  const std::vector<double> & distrib_dens_bins_burned_dispersion
    = flame_dist.get_dens_bins_burned_dispersion();

  const std::vector<double> & distrib_mass_bins_unburned 
    = flame_dist.get_mass_bins_unburned();

  const std::vector<double> & distrib_dens_bins_unburned_mean
    = flame_dist.get_dens_bins_unburned_mean();

  const std::vector<double> & distrib_dens_bins_unburned_dispersion
    = flame_dist.get_dens_bins_unburned_dispersion();
  
  const std::vector<double> & distrib_burned_mass_fract_bins 
    = flame_dist.get_mass_fract_bins_burned();

  // Write out the data

  const std::string & flash_filename = df.get_filename();

  std::string flash_path;
  std::string flash_name;

  QuickFlash::Utils::split_name_path_file(flash_filename, flash_path, 
					  flash_name);

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const double sim_time = meshinfo.get_sim_time();

  std::ofstream flamedist(output_filename.c_str(), std::ios::trunc);

  flamedist << comment_marker << "DataFile " << flash_name << endl;

  flamedist << comment_marker << "SimulationTime " << sim_time << endl;
	
  flamedist << comment_marker << "MinFlameLevel " << min_flame_level 
	    << endl;
  flamedist << comment_marker << "MaxFlameLevel " << max_flame_level 
	    << endl;

  flamedist << comment_marker << "FlameEnvelopeMinPos " 
	    << flame_envelope_min_x << endl;
  flamedist << comment_marker << "FlameEnvelopeMaxPos " 
	    << flame_envelope_max_x << endl;

  flamedist << comment_marker << "MassDistribBinWidth " 
	    << mass_distrib_bin_width << endl;

  flamedist << comment_marker << "MassDistribArrayMin " 
	    << flame_dist.get_array_min_coord() << endl;

  flamedist << comment_marker << "MassDistribArrayMax " 
	    << flame_dist.get_array_max_coord() << endl;

  flamedist << comment_marker << "NumBins " << distrib_bin_count << endl;
	
  flamedist << comment_marker << "DistribArrayTotalMass " 
	    << flame_dist.get_total_mass() << endl;

  flamedist << comment_marker << "DistribArrayBurnedMass " 
	    << flame_dist.get_total_burned_mass() << endl;

  flamedist << comment_marker << "DistribArrayUnburnedMass " 
	    << flame_dist.get_total_unburned_mass() << endl;

  flamedist << comment_marker << "Data" << endl;

  flamedist << "Bin_Center" 
	    << "\t" << "Bin_Volume" 
	    << "\t" << "Bin_Mass" 
	    << "\t" << "Bin_Mean_Density" 
	    << "\t" << "Bin_Density_Dispersion" 
	    << "\t" << "Bin_Burned_Mass" 
	    << "\t" << "Bin_Burned_Mean_Density"
	    << "\t" << "Bin_Burned_Density_Dispersion" 
	    << "\t" << "Bin_Unburned_Mass" 
	    << "\t" << "Bin_Unburned_Mean_Density" 
	    << "\t" << "Bin_Unburned_Density_Dispersion" 
	    << "\t" << "Bin_Burned_Mass_Fract" << endl;

  for (unsigned int bin_index = 0; bin_index < distrib_bin_count;
       bin_index++)
    flamedist << distrib_bin_centers[bin_index] 
	      << "\t" << distrib_bin_volumes[bin_index] 
	      << "\t" << distrib_mass_bins_total[bin_index] 
	      << "\t" << distrib_dens_bins_total_mean[bin_index] 
	      << "\t" << distrib_dens_bins_total_dispersion[bin_index] 
	      << "\t" << distrib_mass_bins_burned[bin_index] 
	      << "\t" << distrib_dens_bins_burned_mean[bin_index] 
	      << "\t" << distrib_dens_bins_burned_dispersion[bin_index] 
	      << "\t" << distrib_mass_bins_unburned[bin_index] 
	      << "\t" << distrib_dens_bins_unburned_mean[bin_index] 
	      << "\t" << distrib_dens_bins_unburned_dispersion[bin_index] 
	      << "\t" << distrib_burned_mass_fract_bins[bin_index] << endl;
  }


void get_slice_pos_list(const double slice_origin, 
			const double rel_slice_min_x,
			const double rel_slice_max_x, 
			const double slice_spacing,
			std::vector<double> & slice_pos_list)
  {
  // NOTE: This function includes all slice positions, even ones outside volume

  slice_pos_list.clear();

  if (!(rel_slice_min_x < rel_slice_max_x))
    throw QuickFlash::Except("Min relative slice pos must be less than max",
			     __FILE__, __LINE__);

  if (!(slice_spacing > 0.0))
    throw QuickFlash::Except("Slice spacing must be positive", __FILE__, 
			     __LINE__);

  double current_rel_slice_pos = rel_slice_min_x;

  while ((!(current_rel_slice_pos > rel_slice_max_x)))
    {
    const double abs_slice_pos = current_rel_slice_pos + slice_origin;

    slice_pos_list.push_back(abs_slice_pos);

    current_rel_slice_pos += slice_spacing;
    }
  }


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

#ifdef USE_MPI
  const MPI_Comm mpi_comm = MPI_COMM_WORLD;

  MPI_Init(&argc, &argv);

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  const bool is_root = (mpi_rank == root_rank) ? true : false;
#else
  const bool is_root = true;
#endif  // USE_MPI

  // Read the command line arguments and options

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, 
					option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 8;  // Does not include Flash data file

  if (num_args < min_args)
    {
    const std::string qualified_exec_name = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(qualified_exec_name, 
					    exec_path, exec_name);

    if (is_root)
      cerr << endl << "Usage: " << exec_name 
	   << " slice_spacing slice_min_x slice_max_x"
	   << endl << "\t" << "min_flame_level max_flame_level"
	   << endl << "\t" << "mass_distrib_bin_width y-samples z-samples"
	   << endl << "\t"
	   << "[ --slice-x-origin=<mesh|tip|masscenter|massfractcenter> ]"
	   << endl << "\t" << "[ --skip-flame-search=FLAME_X_POS ]"
	   << endl << "\t" << "[ --output-prefix=FILENAME_PREFIX ]"
	   << endl << "\t" << "[ --add-comment-marks ]"
	   << endl << "\t" << "[ --buffer-size=NUM_BLOCKS ]"
	   << endl << "\t" << "[ --cache-size=NUM_BLOCKS ]"
	   << endl << "\t" << "[ --file-list=FILENAME ]"
	   << endl << "\t" << "[ flash_datafile1 flash_datafile2 ... ]"
	   << endl << endl;

    return 1;
    }

  // Read the options

  SliceOriginType slice_origin_type = default_slice_origin_type;

  bool skip_flame_search = false;
  double user_flame_pos = 0.0;

  bool use_output_prefix = false;
  std::string base_output_prefix;
  std::string base_output_path;
  std::string unqualified_output_prefix;  // No directory path component

  bool add_comment_markers = false;

  unsigned int buffer_size = default_buffer_size;
  unsigned int cache_size = default_cache_size;

  bool use_file_list = false;
  std::string file_list_filename;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "slice-x-origin")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl 
	       << "Error: Slice x-origin option requires an origin type"
	       << endl << endl;

	return -1;
	}

      const std::string & slice_origin_type_string 
	= current_option.get_value();

      if (read_origin_type_label(slice_origin_type_string, 
				 slice_origin_type) < 0)
	{
	if (is_root)
	  cerr << endl << "Error: Slice origin type [ " 
	       << slice_origin_type_string << " ] not recognized" << endl;

	return -1;
	}
      }
    else if (keyword == "skip-flame-search")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl 
	       << "Error: Skip flame search option requires flame x coord" 
	       << endl << endl;

	return -1;
	}

      skip_flame_search = true;

      current_option.get_value(user_flame_pos);
      }
    else if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: Output-prefix option requires an argument"
	       << endl << endl;

	return -1;
	}

      use_output_prefix = true;

      base_output_prefix = current_option.get_value();

      // Remove the directory path

      QuickFlash::Utils::split_name_path_file(base_output_prefix, 
					      base_output_path, 
					      unqualified_output_prefix);

      // Add an underscore, if not already present

      const unsigned int unqual_output_prefix_len 
	= unqualified_output_prefix.size();

      if (unqual_output_prefix_len > 0)
	if (unqualified_output_prefix[unqual_output_prefix_len - 1] != '_')
	  {
	  base_output_prefix += "_";
	  unqualified_output_prefix += "_";
	  }
      }
    else if (keyword == "add-comment-marks")
      {
      add_comment_markers = true;
      }
    else if (keyword == "buffer-size")
      {
      if (!current_option.value_present())
        {
	if (is_root)
	  cerr << endl 
	       << "Error: Buffer size option requires a non-negative integer"
	       << endl << endl;

	return -1;
	}

      current_option.get_value(buffer_size);
      }
    else if (keyword == "cache-size")
      {
      if (!current_option.value_present())
        {
	if (is_root)
	  cerr << endl 
	       << "Error: Cache size option requires a non-negative integer"
	       << endl << endl;

	return -1;
	}

      current_option.get_value(cache_size);
      }
    else if (keyword == "file-list")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: File list option requires a filename"
	       << endl << endl;

	return -1;
	}

      file_list_filename = current_option.get_value();

      use_file_list = true;
      }
    else
      {
      if (is_root)
	cerr << endl << "Error: Unrecognized option [ " << keyword << " ]"
	     << endl << endl;

      return -1;
      }

    ++option_iter;
    }


  // Read the arguments

  unsigned int argptr = 0;

  const double slice_spacing = std::atof(argument_list[argptr++].c_str());

  if (!(slice_spacing > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Slice spacing must be positive" << endl << endl;

    return -1;
    }

  const double rel_slice_min_x = std::atof(argument_list[argptr++].c_str());

  const double rel_slice_max_x = std::atof(argument_list[argptr++].c_str());

  if (!(rel_slice_min_x < rel_slice_max_x))
    {
    if (is_root)
      cerr << endl << "Error: Slice min x must be less than max x" << endl 
	   << endl;

    return -1;
    }

  const double min_flame_level = std::atof(argument_list[argptr++].c_str());

  if (!((min_flame_level > 0.0) && (min_flame_level < 1.0)))
    {
    if (is_root)
      cerr << endl << "Error: Min flame level must be in non-inclusive range "
	   << "(0.0, 1.0)" << endl << endl;

    return -1;
    }

  const double max_flame_level = std::atof(argument_list[argptr++].c_str());

  if (!((max_flame_level > min_flame_level) && (max_flame_level < 1.0)))
    {
    if (is_root)
      cerr << endl 
	   << "Error: Max flame level must lie between min level and 1.0"
	   << endl << endl;

    return -1;
    }

  const double mass_distrib_bin_width 
    = std::atof(argument_list[argptr++].c_str());

  if (!(mass_distrib_bin_width > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Mass distribution bin width must be positive"
	   << endl << endl;

    return -1;
    }

  const int y_samples_input = std::atoi(argument_list[argptr++].c_str());

  if (y_samples_input <= 1)
    {
    if (is_root)
      cerr << endl << "Error: At least two y-samples required." << endl 
	   << endl;

    return -1;
    }

  const unsigned int y_samples = static_cast<unsigned int>(y_samples_input);

  const int z_samples_input = std::atoi(argument_list[argptr++].c_str());

  if (z_samples_input <= 1)
    {
    if (is_root)
      cerr << endl << "Error: At least two z-samples required." << endl 
	   << endl;

    return -1;
    }

  const unsigned int z_samples = static_cast<unsigned int>(z_samples_input);

  std::vector<std::string> flash_datafile_list;

  if (use_file_list)
    read_file_list(file_list_filename, flash_datafile_list);

  while (argptr < num_args)
    flash_datafile_list.push_back(argument_list[argptr++]);


  const unsigned int num_files = flash_datafile_list.size();

  if (num_files < 1)
    {
    if (is_root)
      cerr << endl << "Error: No data files specified" << endl << endl;

    return -1;
    }


  // Set up the comment markers

  std::string comment_marker = "";

  if (add_comment_markers)
    comment_marker = "# ";


  // Run through the files

  bool proc_active = true;

  unsigned int start_file_index = 0;
  unsigned int last_file_index = 0;

  if (num_files < 1)
    proc_active = false;
  else
    {
    start_file_index = 0;
    last_file_index = num_files - 1;
    }

#ifdef USE_MPI
  if (mpi_size > 1)
    {
    std::vector<unsigned int> start_part_indexes;
    std::vector<unsigned int> part_counts;

    simple_linear_decomp(num_files, mpi_size, start_part_indexes, part_counts);

    const unsigned int proc_start_index = start_part_indexes[mpi_rank];
    const unsigned int proc_count = part_counts[mpi_rank];

    if (proc_count < 1)
      proc_active = false;
    else
      {
      start_file_index = proc_start_index;
      last_file_index = (start_file_index + part_counts[mpi_rank]) - 1;
      }
    }
#endif  // USE_MPI

  if (proc_active)
    {
    for (unsigned int local_file_index = start_file_index; 
	 local_file_index <= last_file_index; local_file_index++)
      {
      const std::string & flash_datafile 
	= flash_datafile_list[local_file_index];

      // Build the full output filename prefix

      std::string base_flash_datafile_path;
      std::string unqualified_flash_datafile_name;

      QuickFlash::Utils::split_name_path_file(flash_datafile, 
					      base_flash_datafile_path, 
					      unqualified_flash_datafile_name);

      const std::string full_output_prefix 
	= base_output_prefix + unqualified_flash_datafile_name + "_";

      const std::string unqualified_full_output_prefix 
	= unqualified_output_prefix + unqualified_flash_datafile_name + "_";

      // Open the metadata file

      std::string metadata_filename = full_output_prefix + "metadata.out";

      std::ofstream metadata(metadata_filename.c_str(), std::ios::trunc);

      metadata << "DataFile " << unqualified_flash_datafile_name << endl;

      // Open the file and obtain some information

      const QuickFlash::File::DataFile df(flash_datafile, buffer_size, true);
  
      const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

      const double time = meshinfo.get_sim_time();

      metadata << "SimTime " << time << endl;

      const unsigned int dims = meshinfo.get_dims();

      if (dims != 3)
	{
	if (is_root)
	  cerr << endl << "Error: Only 3D data supported" << endl << endl;

	return -2;
	}

      const std::vector<double> & min_bounds = meshinfo.get_volume_minbounds();
      const std::vector<double> & max_bounds = meshinfo.get_volume_maxbounds();

      metadata << "MinBounds " << min_bounds << endl;
      metadata << "MaxBounds " << max_bounds << endl;

      const double volume_min_x = min_bounds[0];
      const double volume_max_x = max_bounds[0];

      const double min_y = min_bounds[1];
      const double max_y = max_bounds[1];

      const double min_z = min_bounds[2];
      const double max_z = max_bounds[2];

      const double slice_center_y = 0.5 * (min_y + max_y);
      const double slice_center_z = 0.5 * (min_z + max_z);

      const double slice_width_y = max_y - min_y;
      const double slice_width_z = max_z - min_z;

      std::vector<double> slice_width(2);

      slice_width[0] = slice_width_y;
      slice_width[1] = slice_width_z;
  
      std::vector<unsigned int> slice_samples(2);

      slice_samples[0] = y_samples;
      slice_samples[1] = z_samples;

      // Prepare some datasets

      df.get_dataset(flam_name).set_cache_size(cache_size);
      df.get_dataset(dens_name).set_cache_size(cache_size);

      df.get_dataset(vely_name).set_cache_size(cache_size);
      df.get_dataset(velz_name).set_cache_size(cache_size);

      // Determine the range of slice x-values

      double flame_envelope_min_x = 0.0;
      double flame_envelope_max_x = 0.0;

      metadata << "MinFlameLevel " << min_flame_level << endl;
      metadata << "MaxFlameLevel " << max_flame_level << endl;

      if (skip_flame_search)
	{
	metadata << "UserFlamePos " << user_flame_pos << endl;

	flame_envelope_min_x = rel_slice_min_x;
	flame_envelope_max_x = rel_slice_max_x;

	if (slice_origin_type != MeshOrigin)
	  {
	  flame_envelope_min_x += user_flame_pos;
	  flame_envelope_max_x += user_flame_pos;
	  }
	}
      else
	{
	if (get_flame_bounds(df, min_flame_level, max_flame_level,
			     flame_envelope_min_x, flame_envelope_max_x) < 0)
	  {
	  if (is_root)
	    cerr << endl << "Unable to find flame boundaries" << endl << endl;

	  return -2;
	  }

	metadata << "FlameEnvelopeMinPos " << flame_envelope_min_x << endl;
	metadata << "FlameEnvelopeMaxPos " << flame_envelope_max_x << endl;
	}

      // Determine the distribution of mass

      metadata << "MassDistribBinSpacing " << mass_distrib_bin_width << endl;

      const std::string qualified_flamedist_name 
	= full_output_prefix + "flame_dist.csv";

      const std::string unqualified_flamedist_name 
	= unqualified_full_output_prefix + "flame_dist.csv";

      double flame_min_x = 0.0;
      double flame_max_x = 0.0;

      double flame_center_mass_fract = 0.0;
      double flame_center_mass = 0.0;

      compute_flame_distribution(qualified_flamedist_name, comment_marker, df,
				 flame_envelope_min_x, flame_envelope_max_x, 
				 mass_distrib_bin_width, min_flame_level, 
				 max_flame_level, flame_center_mass_fract, 
				 flame_center_mass, flame_min_x, 
				 flame_max_x);

      metadata << "MassDistribFile " << unqualified_flamedist_name << endl;

      metadata << "FlameCenterMassFract " << flame_center_mass_fract << endl;
      metadata << "FlameCenterMass " << flame_center_mass << endl;
      
      metadata << "DistribFlameMinPos " << flame_min_x << endl;
      metadata << "DistribFlameMaxPos " << flame_max_x << endl;

    
      // Determine reference position

      double slice_pos_origin = 0.0;

      switch (slice_origin_type)
	{
	case MeshOrigin :

	  slice_pos_origin = 0.0;
	  break;

	case FlameTip :

	  slice_pos_origin = flame_max_x;
	  break;

	case FlameMassCenter :

	  slice_pos_origin = flame_center_mass;
	  break;

	case FlameMassFractCenter :

	  slice_pos_origin = flame_center_mass_fract;
	  break;

	default :
	  
	  if (is_root)
	    cerr << "Error: Slice origin type not handled" << endl << endl;

	  return -2;
	}

      std::string slice_origin_type_label = "unknown";

      get_origin_type_label(slice_origin_type, slice_origin_type_label);

      metadata << "SliceOriginType " << slice_origin_type_label << endl;
      metadata << "SliceOrigin " << slice_pos_origin << endl;

      std::vector<double> slice_pos_list;

      get_slice_pos_list(slice_pos_origin, rel_slice_min_x, rel_slice_max_x,
			 slice_spacing, slice_pos_list);

      const unsigned int num_slices = slice_pos_list.size();

      metadata << "RelSliceMinPos " << rel_slice_min_x << endl;
      metadata << "RelSliceMaxPos " << rel_slice_max_x << endl;

      metadata << "NumSlices " << num_slices << endl;
      metadata << "SliceAbsolutePositions " << slice_pos_list << endl;
    

      // Collect the slice data

      std::vector<double> slice_center_coords(3, 0.0);

      slice_center_coords[1] = slice_center_y;
      slice_center_coords[2] = slice_center_z;

      QuickFlash::Slice::SliceGrid vel_y_slice;
      QuickFlash::Slice::SliceGrid vel_z_slice;

      const QuickFlash::File::Dataset & vely = df.get_dataset(vely_name);
      const QuickFlash::File::Dataset & velz = df.get_dataset(velz_name);

      for (unsigned int slice_index = 0; slice_index < num_slices; 
	   slice_index++)
	{
	const double current_slice_x = slice_pos_list[slice_index];

	if ((!(current_slice_x < volume_min_x)) 
	    && (!(current_slice_x > volume_max_x)))
	  {
	  slice_center_coords[0] = current_slice_x;

	  if (is_root)
	    cerr << "** Computing slice [ " << slice_index 
		 << " ] at x coordinate [ " << current_slice_x << " ]" << endl;
	  
	  vel_y_slice.reset(slice_center_coords, slice_width, slice_samples,
			    yz_azim_angle, yz_polar_angle, yz_roll_angle);
	  vel_z_slice.reset(slice_center_coords, slice_width, slice_samples,
			    yz_azim_angle, yz_polar_angle, yz_roll_angle);

	  QuickFlash::Slice::slice_simple(meshinfo, vely, vel_y_slice);
	  QuickFlash::Slice::slice_simple(meshinfo, velz, vel_z_slice);

	  std::list<std::string> slice_metadata_list;

	  std::stringstream comment_stream;

	  comment_stream.str("");
	
	  comment_stream << comment_marker << "DataFile " 
			 << unqualified_flash_datafile_name;

	  slice_metadata_list.push_back(comment_stream.str());

	  comment_stream.str("");

	  comment_stream << "SlicePosition " << current_slice_x;

	  slice_metadata_list.push_back(comment_stream.str());

	  // Build the output file names

	  std::stringstream index_stream;

	  index_stream << slice_index;

	  const std::string index_string = index_stream.str();

	  const std::string qualified_vely_name 
	    = full_output_prefix + "vely_slice_" + index_string + ".csv";

	  const std::string unqualified_vely_name 
	    = unqualified_full_output_prefix + "vely_slice_" + index_string 
  	      + ".csv";

	  const std::string qualified_velz_name 
	    = full_output_prefix + "velz_slice_" + index_string + ".csv";

	  const std::string unqualified_velz_name 
	    = unqualified_full_output_prefix + "velz_slice_" + index_string 
	      + ".csv";

	  // Save the slices

	  QuickFlash::Slice::write_grid(qualified_vely_name, vel_y_slice, 
					slice_metadata_list);

	  QuickFlash::Slice::write_grid(qualified_velz_name, vel_z_slice, 
					slice_metadata_list);

	  metadata << "SliceData " << slice_index << " " << current_slice_x 
		   << " " << unqualified_vely_name << " " 
		   << unqualified_velz_name << endl;
	  }
	else
	  metadata << "SliceData " << slice_index << " " << current_slice_x 
		   << " SKIPPED" << endl;
	}

      metadata << endl;

      cerr << "Completed " << flash_datafile << endl;
      }
    }

#ifdef USE_MPI
  MPI_Finalize();
#endif  // USE_MPI

  return 0;
  }
