// C++ program file flame_dist_analysis.cpp

/*
  By Nathan C. Hearn
     September 15, 2008

  Program for extracting various lateral averages from the flame brush.
*/



#include <iostream>
#include <fstream>
#include <cstdlib>
#include <list>
#include <vector>
#include <string>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_tree_block.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_utils_text.hpp"
#include "ranged_bin_arrays.hpp"


const char default_comment_marker[] = "# ";
const char distrib_data_start_marker[] = "DistribData";

const unsigned int default_flame_search_level = 7;
const unsigned int default_flame_search_samples = 10;

const unsigned int buffer_size = 100;



int get_flame_bounds(const QuickFlash::File::DataFile & df,
		     const double min_flame_value,
		     const double max_flame_value,
		     double & min_flame_x, double & max_flame_x)
  {
  int ret_val = -1;  // Error state

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const QuickFlash::File::Dataset & flame_set = df.get_dataset("flam");

  bool found_interior = false;

  double min_x = 0.0;
  double max_x = 0.0;

  QuickFlash::Block::BlockData<double> flame_data;

  std::vector<double> cell_min_bounds;
  std::vector<double> cell_max_bounds;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info = node_iter.get_block_info();

    const double block_min_x = block_info.get_min_coords()[0];
    const double block_max_x = block_info.get_max_coords()[0];

    bool skip_branch = false;

    if (found_interior)
      {
      skip_branch = true;  // Unless part of block falls outside region

      if ((block_min_x < min_x) || (block_max_x > max_x))
	skip_branch = false;
      }

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      flame_set.get_block_data(block_index, flame_data);

      const unsigned int num_cells = block_info.get_num_cells();

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	const double cell_value = flame_data[cell_index];

	if ((cell_value > min_flame_value) && (cell_value < max_flame_value))
	  {
	  block_info.get_cell_bounds(cell_index, cell_min_bounds, 
				     cell_max_bounds);

	  const double cell_min_x = cell_min_bounds[0];
	  const double cell_max_x = cell_max_bounds[0];

	  if (!found_interior)
	    {
	    min_x = cell_min_x;
	    max_x = cell_max_x;

	    found_interior = true;
	    }
	  else
	    {
	    if (cell_min_x < min_x)
	      min_x = cell_min_x;

	    if (cell_max_x > max_x)
	      max_x = cell_max_x;
	    }
	  }
	}
      }
    
    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  if (found_interior)
    {
    min_flame_x = min_x;
    max_flame_x = max_x;

    ret_val = 0;  // OK
    }

  return ret_val;
  }


int get_flame_distrib(const QuickFlash::File::DataFile & df,
		      const double bin_array_min_x,
		      const unsigned int num_bins, const double bin_width,
		      std::vector<double> & bin_centers,
		      std::vector<double> & bin_total_mass,
		      std::vector<double> & bin_total_mean_dens,
		      std::vector<double> & bin_total_dens_dispersion,
		      std::vector<double> & bin_burned_mass_fraction,
		      std::vector<double> & bin_burned_dens_dispersion)
  {
  int ret_val = -1;  // Error level

  QuickFlash::Stats::BinArray total_mass_bin_array(bin_array_min_x, num_bins, 
						 bin_width);

  QuickFlash::Stats::BinArray burned_mass_bin_array(bin_array_min_x, num_bins, 
						  bin_width);

  const double array_min_x = total_mass_bin_array.get_min_boundary();
  const double array_max_x = total_mass_bin_array.get_max_boundary();

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");
  const QuickFlash::File::Dataset & flam_set = df.get_dataset("flam");

  QuickFlash::Block::BlockData<double> dens_data;
  QuickFlash::Block::BlockData<double> flam_data;

  std::vector<double> cell_center;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info = node_iter.get_block_info();

    const double block_min_x = block_info.get_min_coords()[0];
    const double block_max_x = block_info.get_max_coords()[0];

    bool skip_branch = false;

    if ((block_max_x < array_min_x) || (block_min_x > array_max_x))
      skip_branch = true;

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      dens_set.get_block_data(block_index, dens_data);
      flam_set.get_block_data(block_index, flam_data);

      const unsigned int num_cells = block_info.get_num_cells();

      const double cell_x_width = block_info.get_cell_width()[0];

      // Assume the mesh has cartesian geometry

      const double cell_vol = block_info.get_cell_volume();

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	block_info.get_cell_center(cell_index, cell_center);

	const double cell_center_x = cell_center[0];

	if (total_mass_bin_array.in_bounds(cell_center_x, cell_x_width))
	  {
	  const double cell_dens = dens_data[cell_index];
	  const double cell_burned_dens = cell_dens * flam_data[cell_index];

	  // Accumulate the density, using volume as a weighting factor

	  total_mass_bin_array.add_data_spread(cell_center_x, cell_dens, 
					       cell_x_width, cell_vol);

	  burned_mass_bin_array.add_data_spread(cell_center_x, 
						cell_burned_dens, 
						cell_x_width, cell_vol);
	  }
	}
      }
    
    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  // Compute the centers, densities, masses, etc.

  bin_centers.resize(num_bins, 0.0);

  bin_total_mass.resize(num_bins, 0.0);
  bin_total_mean_dens.resize(num_bins, 0.0);
  bin_total_dens_dispersion.resize(num_bins, 0.0);

  bin_burned_mass_fraction.resize(num_bins, 0.0);
  bin_burned_dens_dispersion.resize(num_bins, 0.0);

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    bin_centers[bin_index] = total_mass_bin_array.get_bin_center(bin_index);

    const QuickFlash::Stats::Bin & total_bin = total_mass_bin_array[bin_index];
    const QuickFlash::Stats::Bin & burned_bin = burned_mass_bin_array[bin_index];

    const double total_mass = total_bin.get_weight_value_sum();

    const double total_mean_dens 
      = (total_mass > 0.0) ? total_bin.get_mean() : 0.0;

    const double total_dens_disp 
      = (total_mass > 0.0) ? total_bin.get_pop_stddev() : 0.0;

    const double burned_mass = burned_bin.get_weight_value_sum();

    const double burned_mass_fract 
      = (total_mass > 0.0) ? (burned_mass / total_mass) : 0.0;

    const double burned_dens_disp 
      = (burned_mass > 0.0) ? burned_bin.get_pop_stddev() : 0.0;

    // Store the values

    bin_total_mass[bin_index] = total_mass;
    bin_total_mean_dens[bin_index] = total_mean_dens;
    bin_total_dens_dispersion[bin_index] = total_dens_disp;

    bin_burned_mass_fraction[bin_index] = burned_mass_fract;
    bin_burned_dens_dispersion[bin_index] = burned_dens_disp;
    }

  ret_val = 0;  // OK -- ANYTHING WE SHOULD CHECK FOR???  (TOTAL MASS, ETC.?)

  return ret_val;
  }


int get_flame_center_mass_fract(const std::vector<double> & bin_centers,
				const std::vector<double> & bin_mass_fractions,
				double & center_pos)
  {
  // <x> = sum_i(x_i phi_i (1 - phi_i)) / sum_i(phi_i (1 - phi_i))

  int ret_val = -1;  // Error state

  const unsigned int num_bins = bin_centers.size();

  if ((num_bins > 0) && (bin_mass_fractions.size() == num_bins))
    {
    double accum_weight_pos = 0.0;
    double accum_weight = 0.0;

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      const double bin_pos = bin_centers[bin_index];
      const double mass_fract = bin_mass_fractions[bin_index];

      const double weight = mass_fract * (1.0 - mass_fract);

      accum_weight_pos += bin_pos * weight;
      accum_weight += weight;
      }

    if (accum_weight > 0.0)
      {
      center_pos = accum_weight_pos / accum_weight;

      ret_val = 0;  // OK
      }
    }

  return ret_val;
  }


int get_flame_center_mass(const std::vector<double> & bin_centers,
			  const std::vector<double> & bin_densities,
			  const std::vector<double> & bin_mass_fractions,
			  double & center_pos)
  {
  // <x> = sum_i(x_i dens_fuel_i dens_ash_i) / sum_i(dens_fuel_i dens_ash_i)

  int ret_val = -1;  // Error state

  const unsigned int num_bins = bin_centers.size();

  if ((num_bins > 0) && (bin_densities.size() == num_bins)
      && (bin_mass_fractions.size() == num_bins))
    {
    double accum_weight_pos = 0.0;
    double accum_weight = 0.0;

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      const double bin_pos = bin_centers[bin_index];
      const double bin_dens = bin_densities[bin_index];
      const double mass_fract = bin_mass_fractions[bin_index];

      const double weight 
	= ((1.0 - mass_fract) * bin_dens) * (mass_fract * bin_dens);

      accum_weight_pos += bin_pos * weight;
      accum_weight += weight;
      }

    if (accum_weight > 0.0)
      {
      center_pos = accum_weight_pos / accum_weight;

      ret_val = 0;  // OK
      }
    }

  return ret_val;
  }


int get_distrib_weight_mass(const QuickFlash::File::DataFile & df, 
			    const std::string & dataset_name, 
			    const double bin_array_min_x,
			    const unsigned int num_bins, 
			    const double bin_width,
			    QuickFlash::Stats::BinArray & bin_array)
  {
  int ret_val = -1;  // Error level

  bin_array.reset(bin_array_min_x, num_bins, bin_width);

  const double array_min_x = bin_array.get_min_boundary();
  const double array_max_x = bin_array.get_max_boundary();

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");
  const QuickFlash::File::Dataset & user_set = df.get_dataset(dataset_name);

  QuickFlash::Block::BlockData<double> dens_data;
  QuickFlash::Block::BlockData<double> user_data;

  std::vector<double> cell_center;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info = node_iter.get_block_info();

    const double block_min_x = block_info.get_min_coords()[0];
    const double block_max_x = block_info.get_max_coords()[0];

    bool skip_branch = false;

    if ((block_max_x < array_min_x) || (block_min_x > array_max_x))
      skip_branch = true;

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      dens_set.get_block_data(block_index, dens_data);
      user_set.get_block_data(block_index, user_data);

      const unsigned int num_cells = block_info.get_num_cells();

      const double cell_x_width = block_info.get_cell_width()[0];

      // Assume the mesh has cartesian geometry

      const double cell_vol = block_info.get_cell_volume();

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	block_info.get_cell_center(cell_index, cell_center);

	const double cell_center_x = cell_center[0];

	if (bin_array.in_bounds(cell_center_x, cell_x_width))
	  {
	  const double cell_value = user_data[cell_index];

	  // Accumulate the cell value, using mass as a weighting factor

	  const double cell_dens = dens_data[cell_index];

	  const double cell_mass = cell_dens * cell_vol;

	  bin_array.add_data_spread(cell_center_x, cell_value, 
				    cell_x_width, cell_mass);
	  }
	}
      }
    
    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  ret_val = 0;  // OK  --  ANYTHING WE SHOULD CHECK FOR???  (DIMENSIONS???)

  return ret_val;
  }


int get_range_areas(const QuickFlash::File::DataFile & df, 
		    const std::string & dataset_name, 
		    const double bin_array_min_x,
		    const unsigned int num_bins, const double bin_width,
		    const double range_min_value,
		    const std::vector<double> & range_max_values,
		    std::vector<double> & bin_centers,
		    std::vector< std::vector<double> > & range_bin_areas,
		    std::vector< std::vector<double> > & range_bin_densities)
  {
  int ret_val = -1;  // Error level

  // ERROR CHECKING???

  RangedBinArrays dens_bin_array_list(range_min_value, range_max_values, 
				      bin_array_min_x, num_bins, bin_width);

  RangedBinArrays area_bin_array_list(range_min_value, range_max_values,
				      bin_array_min_x, num_bins, bin_width);

  const double array_min_x = dens_bin_array_list.get_bin_array_min_bounds();
  const double array_max_x = dens_bin_array_list.get_bin_array_max_bounds();

  const double oneOver_bin_width = 1.0 / bin_width;

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");
  const QuickFlash::File::Dataset & user_set = df.get_dataset(dataset_name);

  QuickFlash::Block::BlockData<double> dens_data;
  QuickFlash::Block::BlockData<double> user_data;

  std::vector<double> cell_center;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info = node_iter.get_block_info();

    const double block_min_x = block_info.get_min_coords()[0];
    const double block_max_x = block_info.get_max_coords()[0];

    bool skip_branch = false;

    if ((block_max_x < array_min_x) || (block_min_x > array_max_x))
      skip_branch = true;

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      dens_set.get_block_data(block_index, dens_data);
      user_set.get_block_data(block_index, user_data);

      const unsigned int num_cells = block_info.get_num_cells();

      const double cell_x_width = block_info.get_cell_width()[0];

      // const double oneOver_cell_x_width = 1.0 / cell_x_width;

      // Assume the mesh has cartesian geometry

      const double cell_vol = block_info.get_cell_volume();

      // Compute a weight to distribute the contribution of the cell to area

      const double area_weight = oneOver_bin_width;

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	block_info.get_cell_center(cell_index, cell_center);

	const double cell_center_x = cell_center[0];

	if (dens_bin_array_list.pos_in_bin_array(cell_center_x, cell_x_width))
	  {
	  const double cell_value = user_data[cell_index];

	  // Accumulate the density

	  const double cell_dens = dens_data[cell_index];

	  dens_bin_array_list.add_data_spread(cell_value, cell_center_x,
					      cell_dens, cell_x_width, 
					      cell_vol);

	  // Accumulate the area 
	  // ... we need to adjust the weight to account for spread

	  area_bin_array_list.add_data_spread(cell_value, cell_center_x,
					      cell_vol, cell_x_width, 
					      area_weight);
	  }
	}
      }
    
    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  const unsigned int num_bin_arrays = dens_bin_array_list.get_num_bin_arrays();

  range_bin_areas.resize(num_bin_arrays);
  range_bin_densities.resize(num_bin_arrays);

  for (unsigned int bin_array_index = 0; bin_array_index < num_bin_arrays;
       bin_array_index++)
    {
    const QuickFlash::Stats::BinArray & dens_bin_array 
      = dens_bin_array_list[bin_array_index];

    const QuickFlash::Stats::BinArray & area_bin_array 
      = area_bin_array_list[bin_array_index];

    std::vector<double> & bin_densities = range_bin_densities[bin_array_index];

    bin_densities.resize(num_bins);

    std::vector<double> & bin_areas = range_bin_areas[bin_array_index];

    bin_areas.resize(num_bins);

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      const QuickFlash::Stats::Bin & dens_bin = dens_bin_array[bin_index];
      const QuickFlash::Stats::Bin & area_bin = area_bin_array[bin_index];

      if (dens_bin.get_weight_sum() > 0.0)
	bin_densities[bin_index] = dens_bin.get_mean();
      else
	bin_densities[bin_index] = 0.0;

      if (area_bin.get_weight_sum() > 0.0)
	bin_areas[bin_index] = area_bin.get_weight_value_sum();
      else
	bin_areas[bin_index] = 0.0;
      }
    }

  if (num_bin_arrays > 0)
    bin_centers = dens_bin_array_list[0].get_bin_centers();
  else
    bin_centers.resize(0);

  ret_val = 0;  // OK  --  ANYTHING WE SHOULD CHECK FOR???  (DIMENSIONS???)

  return ret_val;
  }


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  // Set up the command-line parser

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 4;

  if (num_args < min_args)
    {
    const std::string qualified_exec_name = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(qualified_exec_name, 
					  exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name << " bin_width"
	 << endl << "\t" << "min_brush_flame_level max_brush_flame_level"
	 << endl << "\t" << "[ --output-prefix=FILENAME_PREFIX ]"
	 << endl << "\t" << "[ --min-flame-brush-length=DISTANCE ]"
	 << endl << "\t" << "[ --pre-flame-sample-distance=DISTANCE ]"
	 << endl << "\t" << "[ --add-comment-marks[=COMMENT_MARKER] ]"
	 << endl << "\t" 
	 << "[ --get-area-fractions=LEVEL1_MAX_FLAME[,LEVEL2_MAX_FLAME,...] ]"
	 << endl << "\t" << "[ --distrib-datasets=DATASET1,[DATASET2,...] ]"
	 << endl << "\t" << "[ --compute-rms-difference=REF_VALUE ]"
	 << endl << "\t" << "flash_datafile1 [ flash_datafile2 ... ]"
	 << endl << endl;

    return 1;
    }


  // Read the required arguments

  unsigned int argptr = 0;

  const double bin_width = std::atof(argument_list[argptr++].c_str());

  if (!(bin_width > 0.0))
    {
    cerr << endl << "Error: Bin widths must be positive" << endl << endl;
    return -1;
    }

  const double min_flame_level = std::atof(argument_list[argptr++].c_str());

  const double max_flame_level = std::atof(argument_list[argptr++].c_str());

  if (!(min_flame_level < max_flame_level))
    {
    cerr << endl << "Error: Min flame level must be less than max"
	 << endl << endl;
    return -1;
    }

  if (!(min_flame_level > 0.0))
    cerr << endl << "Warning: Min flame level is not in range (0,1)"
	 << endl << endl;

  if (!(max_flame_level < 1.0))
    cerr << endl << "Warning: Max flame level is not in range (0,1)"
	 << endl << endl;

  std::list<std::string> filename_list;

  while (argptr < num_args)
    filename_list.push_back(argument_list[argptr++]);


  // Read the optional arguments

  std::string user_output_prefix;

  bool use_min_flame_brush_len = false;
  double min_flame_brush_len = 0.0;

  bool use_pre_flame_dist = false;
  double pre_flame_dist = 0.0;

  bool use_comment_marks = false;
  std::string user_comment_mark = default_comment_marker;

  bool compute_area_fractions = false;
  std::vector<double> area_max_fraction_levels;
  std::vector<std::string> area_max_fraction_strings;
  unsigned int num_area_fraction_levels = 0;

  const double area_min_fraction_level = 0.0;  // Fixed for now

  std::vector<std::string> distrib_dataset_list;

  bool compute_rms_diff = false;
  double rms_diff_ref_value = 0.0;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Output prefix option requires the prefix string"
	     << endl << endl;

	return -1;
	}

      user_output_prefix = current_option.get_value();

      // Add underscore character if more than just a path

      std::string prefix_path;
      std::string prefix_name;

      QuickFlash::Utils::split_name_path_file(user_output_prefix, prefix_path,
					    prefix_name);

      const unsigned int name_size = prefix_name.size();

      if (name_size > 0)
	if (prefix_name[name_size - 1] != '_')
	  user_output_prefix += '_';
      }
    else if (keyword == "min-flame-brush-length")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Minimum flame brush length option requires value"
	     << endl << endl;
	return -1;
	}

      use_min_flame_brush_len = true;

      current_option.get_value(min_flame_brush_len);

      if (!(min_flame_brush_len > 0.0))
	{
	cerr << endl 
	     << "Error: Minimum flame brush length value must be positive"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "pre-flame-sample-distance")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Pre-flame sample distance option requires value"
	     << endl << endl;
	return -1;
	}

      use_pre_flame_dist = true;

      current_option.get_value(pre_flame_dist);

      if (!(pre_flame_dist > 0.0))
	{
	cerr << endl 
	     << "Error: Pre-flame sample distance value must be positive"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword =="add-comment-marks")
      {
      use_comment_marks = true;

      if (current_option.value_present())
	user_comment_mark = current_option.get_value() + " ";
      }
    else if (keyword == "get-area-fractions")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Area fractions argument requires list of"
	     << endl << "\t" << "maximum mass fraction values"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(area_max_fraction_levels);
      current_option.get_value(area_max_fraction_strings);

      num_area_fraction_levels = area_max_fraction_levels.size();

      for (unsigned int level_index = 0; 
	   level_index < num_area_fraction_levels; level_index++)
	{
	const double level = area_max_fraction_levels[level_index];

	if (level_index > 0)
	  {
	  if (!(level > area_max_fraction_levels[level_index - 1]))
	    {
	    cerr << endl << "Error: Area levels must be in increasing order"
		 << endl << endl;
	    return -1;
	    }
	  }
	else if (!(level > area_min_fraction_level))
	  {
	  cerr << endl 
	       << "Error: Area levels must be greater than minimum level [ "
	       << area_min_fraction_level << " ]" << endl;
	  return -1;
	  }

	if ((!(level > 0.0)) || (!(level < 1.0)))
	  cerr << endl << "Warning: Area level [ " << level 
	       << " ] not in range (0,1)" << endl << endl;

	if ((!(level > min_flame_level)) || (!(level < max_flame_level)))
	  cerr << endl << "Warning: Area level [ " << level 
	       << " ] not in range (" << min_flame_level << "," 
	       << max_flame_level << ")" << endl << endl;
	}

      compute_area_fractions = true;
      }
    else if (keyword == "distrib-datasets")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Dataset option requires list of datasets"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(distrib_dataset_list);
      }
    else if (keyword == "compute-rms-difference")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Root mean square difference requires reference value"
	     << endl << endl;
	return -1;
	}

      compute_rms_diff = true;

      current_option.get_value(rms_diff_ref_value);
      }
    else
      {
      cerr << endl << "Error: Unrecognized option [ " << keyword << " ]"
	   << endl << endl;

      return -1;
      }

    ++option_iter;
    }


  // Setup additional quantities

  const std::string comment_marker 
    = (use_comment_marks) ? user_comment_mark : "";

  const unsigned int num_distrib_datasets = distrib_dataset_list.size();

  // Run through the files

  std::list<std::string>::const_iterator filename_iter = filename_list.begin();
  
  while (filename_iter != filename_list.end())
    {
    const std::string & filename = *filename_iter;

    // Setup the output file prefixes

    std::string filename_path;
    std::string filename_name;

    QuickFlash::Utils::split_name_path_file(filename, filename_path, 
					  filename_name);

    const std::string output_prefix = user_output_prefix + filename_name + "_";

    // Open the output file

    const std::string distrib_filename = output_prefix + "distrib_data.out";

    std::ofstream distdata(distrib_filename.c_str(), std::ios::trunc);

    // Load the data file

    const QuickFlash::File::DataFile df(filename, buffer_size, true);

    distdata << comment_marker << "DataFile " << filename_name << endl;
  
    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double time = meshinfo.get_sim_time();

    distdata << comment_marker << "SimTime " << time << endl;

    const unsigned int dims = meshinfo.get_dims();

    if (dims != 3)
      {
      cerr << endl << "Error: Only 3D data supported" << endl << endl;

      return -2;
      }

    distdata << comment_marker << "Dimensions " << dims << endl;

    const std::vector<double> & min_bounds = meshinfo.get_volume_minbounds();
    const std::vector<double> & max_bounds = meshinfo.get_volume_maxbounds();

    const double vol_min_x = min_bounds[0];
    const double vol_max_x = max_bounds[0];

    distdata << comment_marker << "MinBounds " << min_bounds << endl;
    distdata << comment_marker << "MaxBounds " << max_bounds << endl;

    distdata << endl;

    // Determine the extent of the flame

    double flame_min_x = 0.0;
    double flame_max_x = 0.0;

    if (get_flame_bounds(df, min_flame_level, max_flame_level,
			 flame_min_x, flame_max_x) < 0)
      {
      cerr << endl << "Error: Unable to find flame boundaries in file [ "
	   << filename << " ]" << endl << endl;
      return -2;
      }

    const double flame_width = flame_max_x - flame_min_x;

    distdata << comment_marker << "MinFlameLevel " << min_flame_level << endl;
    distdata << comment_marker << "MaxFlameLevel " << max_flame_level << endl;

    distdata << comment_marker << "MinFlamePos " << flame_min_x << endl;
    distdata << comment_marker << "MaxFlamePos " << flame_max_x << endl;
    distdata << comment_marker << "FlameWidth " << flame_width << endl;

    distdata << endl;

    // Determine the bin array layout

    double sample_min_x = flame_min_x;

    if (use_min_flame_brush_len)
      {
      const double flame_min_brush_pos = flame_max_x - min_flame_brush_len;

      if (flame_min_brush_pos < sample_min_x)
	sample_min_x = flame_min_brush_pos;

      distdata << comment_marker << "MinFlameBrushLen " << min_flame_brush_len
	       << endl;
      }

    if (sample_min_x < vol_min_x)
      sample_min_x = vol_min_x;

    double sample_max_x = flame_max_x;

    if (use_pre_flame_dist)
      {
      sample_max_x = flame_max_x + pre_flame_dist;

      distdata << comment_marker << "PreFlameSampleDist " << pre_flame_dist
	       << endl;
      }

    if (sample_max_x > vol_max_x)
      sample_max_x = vol_max_x;

    const double bin_array_width = sample_max_x - sample_min_x;

    const unsigned int num_bins 
      = static_cast<unsigned int>(bin_array_width / bin_width) + 1;

    const double bin_array_min_x = sample_min_x;

    const double bin_array_max_x 
      = bin_array_min_x + (static_cast<double>(num_bins) * bin_width);

    distdata << comment_marker << "BinWidth " << bin_width << endl;
    distdata << comment_marker << "NumBins " << num_bins << endl;
    distdata << comment_marker << "BinArrayMinPos " << bin_array_min_x << endl;
    distdata << comment_marker << "BinArrayMaxPos " << bin_array_max_x << endl;

    // Get the distribution of the basic flame parameters

    std::vector<double> bin_centers;
    std::vector<double> bin_total_mass;
    std::vector<double> bin_total_mean_dens;
    std::vector<double> bin_total_dens_dispersion;
    std::vector<double> bin_burned_mass_fraction;
    std::vector<double> bin_burned_dens_dispersion;
    
    if (get_flame_distrib(df, bin_array_min_x, num_bins, bin_width,
			  bin_centers, bin_total_mass, bin_total_mean_dens,
			  bin_total_dens_dispersion, bin_burned_mass_fraction,
			  bin_burned_dens_dispersion) < 0)
      {
      cerr << endl << "Error: Unable to get flame distributions for file [ "
	   << filename << " ]" << endl << endl;
      return -2;
      }

    // Compute values to describe center position of flame brush

    double flame_center_mass_fract = 0.0;

    if (get_flame_center_mass_fract(bin_centers, bin_burned_mass_fraction,
				    flame_center_mass_fract) < 0)
      {
      cerr << endl << "Error: Unable to compute flame center for file [ "
	   << filename << " ]" << endl << endl;
      return -2;
      }

    distdata << comment_marker << "FlameCenterMassFract " 
	     << flame_center_mass_fract << endl;

    double flame_center_mass = 0.0;

    if (get_flame_center_mass(bin_centers, bin_total_mean_dens,
			      bin_burned_mass_fraction, flame_center_mass) < 0)
      {
      cerr << endl << "Error: Unable to compute flame center for file [ "
	   << filename << " ]" << endl << endl;
      return -2;
      }

    distdata << comment_marker << "FlameCenterMass " << flame_center_mass 
	     << endl;

    distdata << endl;

    // Process the remaining user distributions

    distdata << comment_marker << "DistribDatasets " << distrib_dataset_list 
	     << endl;

    if (compute_rms_diff)
      distdata << comment_marker << "RootMeanSqDiffRefValue " 
	       << rms_diff_ref_value << endl;

    distdata << endl;

    std::vector< std::vector<double> > 
      bin_mean_values_list(num_distrib_datasets);

    std::vector< std::vector<double> > 
      bin_dispersion_list(num_distrib_datasets);

    std::vector< std::vector<double> >
      bin_rms_diff_list(num_distrib_datasets);

    for (unsigned int dset_index = 0; dset_index < num_distrib_datasets;
	 dset_index++)
      {
      const std::string & dataset_name = distrib_dataset_list[dset_index];

      QuickFlash::Stats::BinArray bin_array;
      
      if (get_distrib_weight_mass(df, dataset_name, bin_array_min_x,
				  num_bins, bin_width, bin_array) < 0)
	{
	cerr << endl 
	     << "Error: Unable to get distributions for variable [ " 
	     << dataset_name << " ] in file [ " << filename << " ]" << endl 
	     << endl;
	return -2;
	}

      std::vector<double> & mean_values = bin_mean_values_list[dset_index];

      mean_values.resize(num_bins);

      std::vector<double> & dispersion = bin_dispersion_list[dset_index];

      dispersion.resize(num_bins);

      std::vector<double> & rms_diff = bin_rms_diff_list[dset_index];

      if (compute_rms_diff)
	rms_diff.resize(num_bins);

      for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
	{
	const QuickFlash::Stats::Bin & bin = bin_array[bin_index];

	const double weight = bin.get_weight_sum();

	if (weight > 0.0)
	  {
	  mean_values[bin_index] = bin.get_mean();
	  dispersion[bin_index] = bin.get_pop_stddev();

	  if (compute_rms_diff)
	    rms_diff[bin_index] 
	      = bin.get_pop_root_mean_square_difference(rms_diff_ref_value);
	  }
	else
	  {
	  // For now, just put zero in the index

	  mean_values[bin_index] = 0.0;
	  dispersion[bin_index] = 0.0;

	  if (compute_rms_diff)
	    rms_diff[bin_index] = 0.0;
	  }
	}
      }

    // Get the area fractions, if desired

    std::vector< std::vector<double> > range_bin_areas;
    std::vector< std::vector<double> > range_bin_densities;

    if (compute_area_fractions)
      {
      distdata << comment_marker << "AreaFlameMinValue " 
	       << area_min_fraction_level << endl;

      distdata << comment_marker << "AreaFlameMaxValues " 
	       << area_max_fraction_strings << endl;

      distdata << endl;

      if (get_range_areas(df, "flam", bin_array_min_x, num_bins, bin_width,
			  area_min_fraction_level, area_max_fraction_levels, 
			  bin_centers, range_bin_areas, 
			  range_bin_densities) < 0)
	{
	cerr << endl 
	     << "Error: Unable to get flame cross-section areas in file [ " 
	     << filename << " ]" << endl << endl;
	return -2;
	}
      }

    // Output the data

    distdata << comment_marker << distrib_data_start_marker << endl << endl;

    distdata << "Bin_Index" << "\t" << "Bin_Center" << "\t" << "Mass" 
	     << "\t" << "Mean_Density" << "\t" << "Density_Dispersion"
	     << "\t" << "Burned_Mass_Fraction" << "\t" 
	     << "Burned_Density_Dispersion";

    for (unsigned int dset_index = 0; dset_index < num_distrib_datasets;
	 dset_index++)
      {
      const std::string & dataset_name = distrib_dataset_list[dset_index];

      distdata << "\t" << dataset_name << "_Mean_Value" << "\t" 
	       << dataset_name << "_Dispersion";

      if (compute_rms_diff)
	distdata << "\t" << dataset_name << "_RMS_Difference";
      }

    if (compute_area_fractions)
      {
      for (unsigned int range_index = 0; 
	   range_index < num_area_fraction_levels; range_index++)
	{
	const std::string column_prefix 
	  = "Flame_Level_" + area_max_fraction_strings[range_index] + "_";

	distdata << "\t" << column_prefix << "Area" << "\t" 
		 << column_prefix << "Density";
	}
      }

    distdata << endl;

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      distdata << bin_index << "\t" << bin_centers[bin_index]
	       << "\t" << bin_total_mass[bin_index] 
	       << "\t" << bin_total_mean_dens[bin_index]
	       << "\t" << bin_total_dens_dispersion[bin_index]
	       << "\t" << bin_burned_mass_fraction[bin_index]
	       << "\t" << bin_burned_dens_dispersion[bin_index];

      for (unsigned int dset_index = 0; dset_index < num_distrib_datasets;
	   dset_index++)
	{
	distdata << "\t" << bin_mean_values_list[dset_index][bin_index]
		 << "\t" << bin_dispersion_list[dset_index][bin_index];

	if (compute_rms_diff)
	  distdata << "\t" << bin_rms_diff_list[dset_index][bin_index];
	}

      if (compute_area_fractions)
	{
	for (unsigned int range_index = 0; 
	     range_index < num_area_fraction_levels; range_index++)
	  distdata << "\t" << range_bin_areas[range_index][bin_index]
		   << "\t" << range_bin_densities[range_index][bin_index];
	}

      distdata << endl;
      }

    distdata << endl;

    ++filename_iter;
    }

  return 0;
  }
