// C++ program file process_flame_data.cpp

/*
  By Nathan C. Hearn
     December 10, 2008

  Program for processing output from collect_flame_data.
*/


#ifdef USE_MPI
#include <mpi.h>
#include "mpi_utils.hpp"
#endif  // USE_MPI

#include <iostream>
#include <fstream>


void read_file_list(const std::string & file_list_filename, 
		    std::vector<std::string> & file_list)
  {
  file_list.clear();

  std::ifstream f(file_list_filename.c_str());

  std::string name_string;

  while (!(f.eof()))
    {
    f >> name_string;

    if (name_string.size() > 0)
      file_list.push_back(name_string);
    }
  }


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

#ifdef USE_MPI
  const MPI_Comm mpi_comm = MPI_COMM_WORLD;

  MPI_Init(&argc, &argv);

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  const bool is_root = (mpi_rank == root_rank) ? true : false;
#else
  const bool is_root = true;
#endif  // USE_MPI

  // Read the command line arguments and options

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, 
					option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 2;  // Does not include Flash data file

  if (num_args < min_args)
    {
    const std::string qualified_exec_name = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(qualified_exec_name, 
					    exec_path, exec_name);

    if (is_root)
      cerr << endl << "Usage: " << exec_name 
	   << " structure_function_bin_size max_radius "
	   << endl << "\t" << "[ --function-orders=N1,N2,... ]"
	   << endl << "\t" << "[ --output-prefix=FILENAME_PREFIX ]"
	   << endl << "\t" << "[ --add-comment-marks ]"
	   << endl << "\t" << "[ --file-list=FILENAME ]"
	   << endl << "\t" << "[ metadata_file1 metadata_file2 ... ]"
	   << endl << endl;

    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  const double struct_fn_bin_size = std::atof(argument_list[argptr++].c_str());

  if (!(struct_fn_bin_size > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Structure function bin size must be positive"
	   << endl << endl;

    return -1;
    }

  const double max_radius = std::atof(argument_list[argptr++].c_str());

  if (!(max_radius > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Maximum radius must be positive"
	   << endl << endl;

    return -1;
    }

  std::vector<std::string> metadata_file_list;

  while (argptr < num_args)
    metadata_file_list.push_back(argument_list[argptr++]);


  // Read the options

  std::vector<unsigned int> struct_function_orders;

  bool use_output_prefix = false;

  std::string base_output_prefix;
  std::string base_output_path;
  std::string unqualified_output_prefix;  // No directory path component

  bool add_comment_markers = false;


  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "function-orders")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: List of function orders required for option"
	       << endl << endl;

	return -1;
	}

      current_option.get_value(order_list);

      order_list_set = true;
      }
    else if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: Output-prefix option requires an argument"
	       << endl << endl;

	return -1;
	}

      use_output_prefix = true;

      base_output_prefix = current_option.get_value();

      // Remove the directory path

      QuickFlash::Utils::split_name_path_file(base_output_prefix, 
					      base_output_path, 
					      unqualified_output_prefix);

      // Add an underscore, if not already present

      const unsigned int unqual_output_prefix_len 
	= unqualified_output_prefix.size();

      if (unqual_output_prefix_len > 0)
	if (unqualified_output_prefix[unqual_output_prefix_len - 1] != '_')
	  {
	  base_output_prefix += "_";
	  unqualified_output_prefix += "_";
	  }
      }
    else if (keyword == "add-comment-marks")
      {
      add_comment_markers = true;
      }
    else if (keyword == "file-list")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: File list option requires a filename"
	       << endl << endl;

	return -1;
	}

      const std::string file_list_filename = current_option.get_value();

      std::string external_metadata_file_list;

      read_file_list(file_list_filename, external_metadata_file_list);

      const unsigned int num_ext_metadata_files 
	= external_metadata_file_list.size();

      for (unsigned int file_index = 0; file_index < num_ext_metadata_files;
	   file_index++)
	metadata_file_list.push_back(external_metadata_file_list[file_index]);
      }
    else
      {
      if (is_root)
	cerr << endl << "Error: Unrecognized option [ " << keyword << " ]"
	     << endl << endl;

      return -1;
      }

    ++option_iter;
    }

  // Get number of files

  const unsigned int num_files = metadata_file_list.size();

  if (num_files < 1)
    {
    if (is_root)
      cerr << endl << "Error: No data files specified" << endl << endl;

    return -1;
    }


  // Set up the comment markers

  std::string comment_marker = "";

  if (add_comment_markers)
    comment_marker = "# ";


  // Run through the files

  bool proc_active = true;

  unsigned int start_file_index = 0;
  unsigned int last_file_index = 0;

  if (num_files < 1)
    proc_active = false;
  else
    {
    start_file_index = 0;
    last_file_index = num_files - 1;
    }

#ifdef USE_MPI
  if (mpi_size > 1)
    {
    std::vector<unsigned int> start_part_indexes;
    std::vector<unsigned int> part_counts;

    simple_linear_decomp(num_files, mpi_size, start_part_indexes, part_counts);

    const unsigned int proc_start_index = start_part_indexes[mpi_rank];
    const unsigned int proc_count = part_counts[mpi_rank];

    if (proc_count < 1)
      proc_active = false;
    else
      {
      start_file_index = proc_start_index;
      last_file_index = (start_file_index + part_counts[mpi_rank]) - 1;
      }
    }
#endif  // USE_MPI

  if (proc_active)
    {
    for (unsigned int local_file_index = start_file_index; 
	 local_file_index <= last_file_index; local_file_index++)
      {
      const std::string & metadata_filename 
	= metadata_file_list[local_file_index];

      // Determine the metadata path and file name

      std::string base_metadata_file_path;
      std::string unqualified_metadata_filename;

      QuickFlash::Utils::split_name_path_file(metadata_filename, 
					      base_metadata_file_path, 
					      unqualified_metadata_filename);

      // Read the metadata file

      }
    }

#ifdef USE_MPI
  MPI_Finalize();
#endif  // USE_MPI

  return 0;
  }
