// C++ program file ranged_bin_arrays.cpp

/*
  By Nathan C. Hearn
     September 15, 2008

  Class for sorting data into separate BinArray objects based on an input 
  parameter.
*/


#include "ranged_bin_arrays.hpp"
#include <vector>
#include "quickflash_except.hpp"
#include "quickflash_stats_bins.hpp"


// Class RangedBinArrays

RangedBinArrays::RangedBinArrays() :
  num_bin_arrays(0), array_min_bounds(0.0), array_max_bounds(0.0), 
  range_min_value(0.0), range_max_values(), bin_array_list()
  { }


RangedBinArrays::RangedBinArrays(const double min_range_value, 
				 const std::vector<double> & max_range_values,
				 const double bin_array_min_bounds, 
				 const unsigned int num_bins, 
				 const double bin_width) :
  num_bin_arrays(0), array_min_bounds(0.0), array_max_bounds(0.0), 
  range_min_value(0.0), range_max_values(), bin_array_list()
  { 
  reset(min_range_value, max_range_values, bin_array_min_bounds, num_bins, 
	bin_width);
  }


RangedBinArrays::RangedBinArrays(const RangedBinArrays & source) :
  num_bin_arrays(0), array_min_bounds(0.0), array_max_bounds(0.0),
  range_min_value(0.0), range_max_values(), bin_array_list()
  { reset(source); }


void RangedBinArrays::reset()
  {
  num_bin_arrays = 0;

  array_min_bounds = 0.0;
  array_max_bounds = 0.0;

  range_min_value = 0.0;
  range_max_values.clear();

  bin_array_list.clear();
  }


void RangedBinArrays::reset(const double min_range_value, 
			    const std::vector<double> & max_range_values,
			    const double bin_array_min_bounds, 
			    const unsigned int num_bins, 
			    const double bin_width)
  {
  reset();

  // Create the arrays

  num_bin_arrays = max_range_values.size();
  
  if (num_bin_arrays > 0)
    {
    // Set up the bin arrays

    range_min_value = min_range_value;
    range_max_values = max_range_values;

    bin_array_list.resize(num_bin_arrays);

    double prev_range_value = min_range_value;

    for (unsigned int array_index = 0; array_index < num_bin_arrays;
	 array_index++)
      {
      const double range_value = range_max_values[array_index];

      // Make sure values are in order

      if (!(range_value > prev_range_value))
	throw QuickFlash::Except("Range values out of order", __FILE__,
			       __LINE__);

      // Build the BinArray object

      bin_array_list[array_index].reset(bin_array_min_bounds, num_bins, 
					bin_width);	
      }

    const QuickFlash::Stats::BinArray & first_array = bin_array_list[0];

    array_min_bounds = first_array.get_min_boundary();
    array_max_bounds = first_array.get_max_boundary();
    }
  }


void RangedBinArrays::reset(const RangedBinArrays & source)
  {
  if (&source != this)
    {
    num_bin_arrays = source.num_bin_arrays;

    array_min_bounds = source.array_min_bounds;
    array_max_bounds = source.array_max_bounds;

    range_min_value = source.range_min_value;
    range_max_values = source.range_max_values;

    bin_array_list = source.bin_array_list;
    }
  }


bool RangedBinArrays::pos_in_bin_array(const double bin_array_pos, 
				       const double width) const
  {
  bool ret_val = false;

  if (num_bin_arrays > 0)
    ret_val = bin_array_list[0].in_bounds(bin_array_pos, width);

  return ret_val;
  }


void RangedBinArrays::add_data(const double range_value, 
			       const double data_coordinate, 
			       const double data_value, const double weight)
  {
  unsigned int array_index = 0;

  if (locate_bin_array(range_value, array_index))
    bin_array_list[array_index].add_data(data_coordinate, data_value, weight);
  }


void RangedBinArrays::add_data_spread(const double range_value, 
				      const double data_coordinate, 
				      const double data_value, 
				      const double coordinate_spread,
				      const double weight)
  {
  unsigned int array_index = 0;

  if (locate_bin_array(range_value, array_index))
    bin_array_list[array_index].add_data_spread(data_coordinate, data_value, 
						coordinate_spread, weight);
  }


bool RangedBinArrays::locate_bin_array(const double range_value,
				       unsigned int & bin_array_index) const
  {
  bool ret_val = false;  // True if range_value corresponds to BinArray object

  if (num_bin_arrays > 0)
    {
    const double max_value = range_max_values[num_bin_arrays - 1];

    if ((!(range_value < range_min_value)) && (range_value < max_value))
      {
      // Use bisection search to find bin array index

      unsigned int lo_index = 0;
      unsigned int hi_index = num_bin_arrays - 1;

      while (lo_index < (hi_index - 1))
	{
	const unsigned int test_index = (lo_index + hi_index) / 2;

	const double test_max_value = range_max_values[test_index];

	if (range_value < test_max_value)
	  hi_index = test_index;
	else
	  lo_index = test_index;
	}

      if (range_value < range_max_values[lo_index])
	bin_array_index = lo_index;
      else
	bin_array_index = hi_index;

      ret_val = true;  // Found it!!
      }
    }

  return ret_val;
  }
