// C++ program file struct_fn.cpp

/*
  By Nathan C. Hearn
     February 8, 2008

  Main driver for structure function calculation.
*/


// Include the MPI headers first to make MPI/xlC on BG/P happy ... (grrrr....)

#ifdef USE_MPI
#include <mpi.h>
#include "mpi_utils.hpp"
#endif  // USE_MPI

#include <cstdlib>
#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>
#include <algorithm>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_slice_slicegrid.hpp"
#include "quickflash_slice_slicer.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_slice_output.hpp"
#include "flame_front.hpp"
#include "structure.hpp"
#include "flame_dist.hpp"

#ifdef USE_MAGICK
#include "quickflash_image_imagemagick.hpp"
#endif  // USE_MAGICK


const unsigned int default_flame_search_level = 7;
const unsigned int default_flame_search_samples = 10;

const char default_image_type[] = "png";

const unsigned int buffer_size = 100;
const unsigned int cache_size = 2000;

const char flam_name[] = "flam";
const char dens_name[] = "dens";

const char velx_name[] = "velx";
const char vely_name[] = "vely";
const char velz_name[] = "velz";

const double deg_to_rad = (M_PI) / 180.0;

const double xy_roll_angle = 0.0 * deg_to_rad;
const double xy_polar_angle = 0.0 * deg_to_rad;
const double xy_azim_angle = 0.0 * deg_to_rad;

const double xz_roll_angle = 90.0 * deg_to_rad;
const double xz_polar_angle = 90.0 * deg_to_rad;
const double xz_azim_angle = -90.0 * deg_to_rad;

const double yz_roll_angle = 90.0 * deg_to_rad;
const double yz_polar_angle = 90.0 * deg_to_rad;
const double yz_azim_angle = 0.0 * deg_to_rad;

const double slice_expansion_factor = 1.1;  // 10% increase

#ifdef USE_MPI
const int continue_code = 1;
const int finished_code = 0;
#endif  // USE_MPI


#ifdef USE_MAGICK
int write_image(const std::string & filename, 
		const QuickFlash::Slice::SliceGrid & data)
  {
  int ret_val = -1;  // Error state

  // Build the colormap

  double min_val = 0.0;
  double max_val = 0.0;

  data.get_minmax_values(min_val, max_val);

  if (!(min_val < max_val))
    max_val = (2.0 * std::fabs(min_val)) + 1.0;  // Make something up ...

  QuickFlash::Color::Colormap colormap(min_val, max_val);

  QuickFlash::Image::ImageGrid image(data.get_grid_data(), colormap);

  ret_val = QuickFlash::Image::ImageMagick::write_image(filename, image);

  return ret_val;
  }
#endif  // USE_MAGICK


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

#ifdef USE_MPI
  const MPI_Comm mpi_comm = MPI_COMM_WORLD;

  MPI_Init(&argc, &argv);

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  const bool is_root = (mpi_rank == root_rank) ? true : false;
#else
  const bool is_root = true;
#endif  // USE_MPI

  // Read the command line arguments and options

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 8;

  if (num_args < min_args)
    {
    if (is_root)
      {
      const std::string qualified_exec_name = argv[0];

      std::string exec_path;
      std::string exec_name;

      QuickFlash::Utils::split_name_path_file(qualified_exec_name, 
					    exec_path, exec_name);

      cerr << endl << "Usage: " << exec_name 
	   << " flash_datafile slice_spacing "
	   << endl << "\t" << "min_flame_level max_flame_level"
	   << endl << "\t" << "y-samples z-samples"
	   << endl << "\t" << "structure_function_bin_size max_radius"
	   << endl << "\t" << "[ --function-orders=N1,N2,... ]"
	   << endl << "\t" << "[ --follow-flame=MIN_REL_X,MAX_REL_X ]"
	   << endl << "\t" << "[ --flame-search-level=search_level ]"
	   << endl << "\t" << "[ --flame-search-samples=sample_points ]"
	   << endl << "\t" << "[ --skip-flame-search=FLAME_X_POS ]"
	   << endl << "\t" << "[ --output-prefix=FILENAME_PREFIX ]"
	   << endl << "\t" << "[ --add-comment-marks ]"
	   << endl << "\t" << "[ --save-x-axis-slices ]"
	   << endl << "\t" << "[ --save-cross-sect-slices ]"
	   << endl << "\t" << "[ --compute-mass-distrib=BIN_SPACING ]"
#ifdef USE_MAGICK
	   << endl << "\t" << "[ --make-images[=IMAGE_TYPE] ]"
	   << endl << "\t" << "[ --images-only ]"
#endif  // USE_MAGICK
	   << endl << endl;
      }

    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  const std::string flash_datafile = argument_list[argptr++];

  const double slice_spacing = std::atof(argument_list[argptr++].c_str());

  if (!(slice_spacing > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Slice spacing must be positive" << endl << endl;

    return -1;
    }

  const double min_flame_level = std::atof(argument_list[argptr++].c_str());

  if (!((min_flame_level > 0.0) && (min_flame_level < 1.0)))
    {
    if (is_root)
      cerr << endl 
	   << "Error: Min flame level must be in non-inclusive range "
	   << "(0.0, 1.0)" << endl << endl;

    return -1;
    }

  const double max_flame_level = std::atof(argument_list[argptr++].c_str());

  if (!((max_flame_level > min_flame_level) && (max_flame_level < 1.0)))
    {
    if (is_root)
      cerr << endl 
	   << "Error: Max flame level must lie between min level and 1.0"
	   << endl << endl;

    return -1;
    }

  const int y_samples_input = std::atoi(argument_list[argptr++].c_str());

  if (y_samples_input <= 1)
    {
    if (is_root)
      cerr << endl << "Error: At least two y-samples required." << endl 
	   << endl;

    return -1;
    }

  const unsigned int y_samples = static_cast<unsigned int>(y_samples_input);

  const int z_samples_input = std::atoi(argument_list[argptr++].c_str());

  if (z_samples_input <= 1)
    {
    if (is_root)
      cerr << endl << "Error: At least two z-samples required." << endl 
	   << endl;

    return -1;
    }

  const unsigned int z_samples = static_cast<unsigned int>(z_samples_input);

  const double struct_fn_bin_size = std::atof(argument_list[argptr++].c_str());

  if (!(struct_fn_bin_size > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Structure function bin size must be positive"
	   << endl << endl;

    return -1;
    }

  const double struct_fn_max_radius 
    = std::atof(argument_list[argptr++].c_str());

  if (!(struct_fn_max_radius > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Structure function max radius must be positive"
	   << endl << endl;

    return -1;
    }


  // Read the options

  bool order_list_set = false;
  std::vector<unsigned int> order_list;

  bool follow_flame = false;

  double flame_rel_slice_min_x = 0.0;  // Relative to flame position
  double flame_rel_slice_max_x = 0.0;  // ...

  unsigned int flame_search_level = default_flame_search_level;
  unsigned int flame_search_samples = default_flame_search_samples;

  bool skip_flame_search = false;
  double user_flame_pos = 0.0;

  bool use_output_prefix = false;
  std::string base_output_prefix;
  std::string base_output_path;
  std::string unqualified_output_prefix;  // No directory path component

  bool add_comment_markers = false;

  bool save_x_slices = false;
  bool save_cross_sect_slices = false;

  bool compute_mass_distrib = false;
  double mass_distrib_bin_width = 0.0;

#ifdef USE_MAGICK
  bool save_slice_images = false;

  std::string slice_image_type = default_image_type;
#endif  // USE_MAGICK

  bool save_slice_data = true;  // Unless only images are requested

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "function-orders")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: List of function orders required for option"
	       << endl << endl;

	return -1;
	}

      current_option.get_value(order_list);

      order_list_set = true;
      }
    else if (keyword == "follow-flame")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: Follow-flame option requires two arguments"
	       << endl << endl;

	return -1;
	}

      std::vector<double> follow_flame_params;

      current_option.get_value(follow_flame_params);

      if (follow_flame_params.size() != 2)
	{
	if (is_root)
	  cerr << endl << "Error:  Follow-flame option requires two arguments"
	       << endl << endl;

	return -1;
	}

      follow_flame = true;

      flame_rel_slice_min_x = follow_flame_params[0];
      flame_rel_slice_max_x = follow_flame_params[1];

      if (!(flame_rel_slice_min_x < flame_rel_slice_max_x))
	{
	if (is_root)
	  cerr << endl << "Error: Follow-flame min_x must be less than max_x"
	       << endl << endl;

	return -1;
	}
      }
    else if (keyword == "flame-search-level")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl 
	       << "Error: Flame-search-level option requires one argument" 
	       << endl << endl;

	return -1;
	}

      current_option.get_value(flame_search_level);

      if (flame_search_level < 3)
	{
	if (is_root)
	  cerr << endl << "Error: Flame-search level must be greater than 2"
	       << endl << endl;

	return -1;
	}
      }
    else if (keyword == "flame-search-samples")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl 
	       << "Error: Flame-search samples option requires one argument" 
	       << endl << endl;

	return -1;
	}

      current_option.get_value(flame_search_samples);

      if (flame_search_level < 1)
	{
	if (is_root)
	  cerr << endl << "Error: Flame-search samples must be positive"
	       << endl << endl;

	return -1;
	}
      }
    else if (keyword == "skip-flame-search")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl 
	       << "Error: Skip flame search option requires flame x coord" 
	       << endl << endl;

	return -1;
	}

      skip_flame_search = true;

      current_option.get_value(user_flame_pos);
      }
    else if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: Output-prefix option requires an argument"
	       << endl << endl;

	return -1;
	}

      use_output_prefix = true;

      base_output_prefix = current_option.get_value();

      // Remove the directory path

      QuickFlash::Utils::split_name_path_file(base_output_prefix, 
					    base_output_path, 
					    unqualified_output_prefix);
      }
    else if (keyword == "add-comment-marks")
      {
      add_comment_markers = true;
      }
    else if (keyword == "save-x-axis-slices")
      {
      save_x_slices = true;
      }
    else if (keyword == "save-cross-sect-slices")
      {
      save_cross_sect_slices = true;
      }
    else if (keyword == "compute-mass-distrib")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl 
	       << "Error: Mass distribution option requires a bin width" 
	       << endl << endl;

	return -1;
	}

      compute_mass_distrib = true;

      current_option.get_value(mass_distrib_bin_width);

      if (!(mass_distrib_bin_width > 0.0))
	{
	if (is_root)
	  cerr << endl << "Error: Mass distribution bin width must be positive"
	       << endl << endl;

	return -1;
	}
      }
    else if (keyword == "make-images")
      {
#ifdef USE_MAGICK
      save_slice_images = true;

      if (current_option.value_present())
	slice_image_type = current_option.get_value();
#else
      cerr << endl << "Error: Image output support not present" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "images-only")
      {
#ifdef USE_MAGICK
      // NOTE: Without --make-images, uses default image type

      save_slice_images = true; 
      save_slice_data = false;
#else
      cerr << endl << "Error: Image output support not present" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else
      {
      if (is_root)
	cerr << endl << "Error: Unrecognized option [ " << keyword << " ]"
	     << endl << endl;

      return -1;
      }

    ++option_iter;
    }

  // Get some additional quantities

  std::string base_flash_datafile_path;
  std::string unqualified_flash_datafile_name;

  QuickFlash::Utils::split_name_path_file(flash_datafile, 
					base_flash_datafile_path, 
					unqualified_flash_datafile_name);

  // Perform some checks

  if (!order_list_set)
    {
    order_list.resize(1);
    order_list[0] = 2;
    }

  const unsigned int num_orders = order_list.size();


  // Set up the comment markers

  std::string comment_marker = "";

  if (add_comment_markers)
    comment_marker = "# ";


  // Perform the calculation

  if (!is_root)
    {
#ifdef USE_MPI
    bool finished = false;

    while (!finished)
      {
      int do_calculation = 0;
      
      MPI_Bcast(&do_calculation, 1, MPI_INT, root_rank, mpi_comm);

      if (do_calculation == finished_code)
	finished = true;
      else
	get_scalar_struct_fn_worker(mpi_comm);
      }
#endif  // USE_MPI
    }
  else
    {
    // Open the metadata file

    std::string metadata_filename;

    if (use_output_prefix)
      {
      metadata_filename += base_output_path;

      if (unqualified_output_prefix.size() > 0)
	metadata_filename += unqualified_output_prefix + "_";
      }

    metadata_filename += unqualified_flash_datafile_name + "_metadata.out";

    std::ofstream metadata(metadata_filename.c_str(), std::ios::trunc);

    metadata << "DataFile " << unqualified_flash_datafile_name << endl;

    // Open the file and obtain some information

    const QuickFlash::File::DataFile df(flash_datafile, buffer_size, true);
  
    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double time = meshinfo.get_sim_time();

    metadata << "SimTime " << time << endl;

    const unsigned int dims = meshinfo.get_dims();

    if (dims != 3)
      {
      cerr << endl << "Error: Only 3D data supported" << endl << endl;

      return -2;
      }

    const std::vector<double> & min_bounds = meshinfo.get_volume_minbounds();
    const std::vector<double> & max_bounds = meshinfo.get_volume_maxbounds();

    metadata << "MinBounds " << min_bounds << endl;
    metadata << "MaxBounds " << max_bounds << endl;

    const double min_y = min_bounds[1];
    const double max_y = max_bounds[1];

    const double min_z = min_bounds[2];
    const double max_z = max_bounds[2];

    const double slice_center_y = 0.5 * (min_y + max_y);
    const double slice_center_z = 0.5 * (min_z + max_z);

    const double slice_width_y = max_y - min_y;
    const double slice_width_z = max_z - min_z;

    std::vector<double> slice_width(2);

    slice_width[0] = slice_width_y;
    slice_width[1] = slice_width_z;
  
    std::vector<unsigned int> slice_samples(2);

    slice_samples[0] = y_samples;
    slice_samples[1] = z_samples;

    // Prepare some datasets

    const QuickFlash::File::Dataset & flam = df.get_dataset(flam_name);

    flam.set_cache_size(cache_size);

    const QuickFlash::File::Dataset & dens = df.get_dataset(dens_name);

    dens.set_cache_size(cache_size);
  
    // Determine the range of slice x-values

    double volume_min_x = 0.0;
    double volume_max_x = 0.0;
  
    double flame_front_x = 0.0;
    double flame_back_x = 0.0;

    if (skip_flame_search)
      {
      metadata << "UserFlamePos " << user_flame_pos << endl;

      flame_front_x = user_flame_pos;
      flame_back_x = user_flame_pos;

      volume_min_x = meshinfo.get_volume_minbounds()[0];
      volume_max_x = meshinfo.get_volume_maxbounds()[0];
      }
    else
      {
      metadata << "MinFlameLevel " << min_flame_level << endl;
      metadata << "MaxFlameLevel " << max_flame_level << endl;

      metadata << "FlameSearchLevel " << flame_search_level << endl;
      metadata << "FlameSearchSamples " << flame_search_samples << endl;

      double flame_envelope_min_x = 0.0;
      double flame_envelope_max_x = 0.0;

      find_flame_range(df, min_flame_level, max_flame_level, 
		       flame_search_level, flame_search_samples, 
		       flame_envelope_min_x, flame_envelope_max_x, 
		       volume_min_x, volume_max_x);

      metadata << "FlameEnvelopeMinPos " << flame_envelope_min_x << endl;
      metadata << "FlameEnvelopeMaxPos " << flame_envelope_max_x << endl;

      if (compute_mass_distrib)
	{
	// Determine the distribution of mass

	metadata << "MassDistribBinSpacing " << mass_distrib_bin_width << endl;

	FlameDistribution flame_dist(df, 0, mass_distrib_bin_width,
				     flame_envelope_min_x, 
				     flame_envelope_max_x);

	const unsigned int num_distrib_bins = flame_dist.get_num_bins();

	const std::vector<double> & bin_centers 
	  = flame_dist.get_bin_centers();

	const std::vector<double> & mass_bins_total 
	  = flame_dist.get_mass_bins_total();

	const std::vector<double> & mass_bins_burned 
	  = flame_dist.get_mass_bins_burned();

	const std::vector<double> & mass_bins_unburned 
	  = flame_dist.get_mass_bins_unburned();

	const std::vector<double> & flame_mass_fract_bins 
	  = flame_dist.get_flame_mass_fract_bins();

	// Build the output file names
      
	const std::string basename 
	  = unqualified_flash_datafile_name + "_flame_dist.csv" ;

	std::string qualified_datafile_name = "";  // With directory path
	std::string unqualified_datafile_name = "";  // Without directory path

	if (use_output_prefix)
	  {
	  qualified_datafile_name += base_output_path;

	  if (unqualified_output_prefix.size() > 0)
	    {
	    qualified_datafile_name += unqualified_output_prefix + "_";
	    unqualified_datafile_name += unqualified_output_prefix + "_";
	    }
	  }

	qualified_datafile_name += basename;
	unqualified_datafile_name += basename;

	// Write the file

	std::ofstream f(qualified_datafile_name.c_str(), std::ios::trunc);

	f << comment_marker << "DataFile " << unqualified_flash_datafile_name 
	  << endl;
	f << comment_marker << "SimulationTime " << time << endl;
	
	f << comment_marker << "MinFlameLevel " << min_flame_level << endl;
	f << comment_marker << "MaxFlameLevel " << max_flame_level << endl;

	f << comment_marker << "FlameEnvelopeMinPos " << flame_envelope_min_x 
	  << endl;
	f << comment_marker << "FlameEnvelopeMaxPos " << flame_envelope_max_x 
	  << endl;

	f << comment_marker << "MassDistribBinWidth " << mass_distrib_bin_width
	  << endl;

	f << comment_marker << "MassDistribArrayMin " 
	  << flame_dist.get_array_min_coord() << endl;

	f << comment_marker << "MassDistribArrayMax " 
	  << flame_dist.get_array_max_coord() << endl;

	f << comment_marker << "NumBins " << num_distrib_bins << endl;
	
	f << comment_marker << "DistribArrayTotalMass " 
	  << flame_dist.get_total_mass() << endl;

	f << comment_marker << "DistribArrayBurnedMass " 
	  << flame_dist.get_total_burned_mass() << endl;

	f << comment_marker << "DistribArrayUnburnedMass " 
	  << flame_dist.get_total_unburned_mass() << endl;

	f << comment_marker << "Data" << endl;

	f << "Bin_Center\tBin_Mass\tBin_Burned_Mass\tBin_Unburned_Mass\t"
	  << "Burned_Mass_Fract" << endl;

	for (unsigned int bin_index = 0; bin_index < num_distrib_bins;
	     bin_index++)
	  f << bin_centers[bin_index] << "\t" << mass_bins_total[bin_index] 
	    << "\t" << mass_bins_burned[bin_index] << "\t" 
	    << mass_bins_unburned[bin_index] << "\t" 
	    << flame_mass_fract_bins[bin_index] << endl;

	metadata << "MassDistribFile " << unqualified_datafile_name << endl;
	}

      flame_front_x = flame_envelope_max_x;
      flame_back_x = flame_envelope_min_x;
      }

    metadata << "FlameFrontEdgePos " << flame_front_x << endl;

    // Use leading edge of flame (flame_envelope_max_x) as position reference

    double slice_area_min_x = volume_min_x;
    double slice_area_max_x = volume_max_x;

    if (follow_flame)
      {
      metadata << "FollowFlameMinBound " << flame_rel_slice_min_x << endl;
      metadata << "FollowFlameMaxBound " << flame_rel_slice_max_x << endl;

      const double flame_slice_min_x 
	= flame_front_x + flame_rel_slice_min_x;

      const double flame_slice_max_x 
	= flame_front_x + flame_rel_slice_max_x;

      if (slice_area_min_x < flame_slice_min_x)
	slice_area_min_x = flame_slice_min_x;

      if (slice_area_max_x > flame_slice_max_x)
	slice_area_max_x = flame_slice_max_x;
      }

    metadata << "FlameSliceMinPos " << slice_area_min_x << endl;
    metadata << "FlameSliceMaxPos " << slice_area_max_x << endl;

    const double slice_min_x = slice_area_min_x;

    const double fl_num_slices 
      = (slice_area_max_x - slice_area_min_x) / slice_spacing;

    const unsigned int min_slices 
      = static_cast<unsigned int>(fl_num_slices + 0.25);

    const unsigned int num_slices = (min_slices > 0) ? min_slices : 1;

    metadata << "NumSlices " << num_slices << endl;

    metadata << "StructFnOrderList " << order_list << endl;

    // Set up the datasets

    //    const QuickFlash::File::Dataset & velx
    //      = df.get_dataset(velx_name);

    const QuickFlash::File::Dataset & vely = df.get_dataset(vely_name);

    vely.set_cache_size(cache_size);

    const QuickFlash::File::Dataset & velz = df.get_dataset(velz_name);

    velz.set_cache_size(cache_size);

    // Output the x-axis slices

    if (save_x_slices)
      {
      const double x_slice_center = 0.5 * (slice_area_min_x 
					   + slice_area_max_x);

      const double slice_area_width = slice_area_max_x - slice_area_min_x;

      const double x_slice_width = slice_area_width * slice_expansion_factor;
      const double x_slice_half_width = 0.5 * x_slice_width;

      const double x_slice_min_x = x_slice_center - x_slice_half_width;
      const double x_slice_max_x = x_slice_center + x_slice_half_width;

      // Use y-samples as basis for slice resolution

      const double x_slice_pixel_width 
	= (max_y - min_y) / static_cast<double>(y_samples);

      const double x_pixels_fl = x_slice_width / x_slice_pixel_width;

      const unsigned int x_pixels 
	= std::max(static_cast<unsigned int>(2), 
		   static_cast<unsigned int>(x_pixels_fl));

      std::vector<double> slice_center(3);

      slice_center[0] = x_slice_center;
      slice_center[1] = slice_center_y;
      slice_center[2] = slice_center_z;

      std::vector<double> slice_width_xy(2);

      slice_width_xy[0] = x_slice_width;
      slice_width_xy[1] = slice_width_y;

      std::vector<double> slice_width_xz(2);

      slice_width_xz[0] = x_slice_width;
      slice_width_xz[1] = slice_width_z;

      std::vector<unsigned int> pixels_xy(2);

      pixels_xy[0] = x_pixels;
      pixels_xy[1] = y_samples;

      std::vector<unsigned int> pixels_xz(2);

      pixels_xz[0] = x_pixels;
      pixels_xz[1] = z_samples;

      QuickFlash::Slice::SliceGrid vel_slice_xy(slice_center, slice_width_xy,
					      pixels_xy, xy_azim_angle,
					      xy_polar_angle, xy_roll_angle);

      QuickFlash::Slice::SliceGrid vel_slice_xz(slice_center, slice_width_xz,
					      pixels_xz, xz_azim_angle,
					      xz_polar_angle, xz_roll_angle);

      const std::string vely_slice_output_prefix 
	= unqualified_flash_datafile_name + "_vely_slice_";

      const std::string velz_slice_output_prefix
	= unqualified_flash_datafile_name + "_velz_slice_";

      std::string qualified_vely_slice_output_prefix = "";  // With path
      std::string unqualified_vely_slice_output_prefix = "";  // Without path
      
      std::string qualified_velz_slice_output_prefix = "";  // With path
      std::string unqualified_velz_slice_output_prefix = "";  // Without path

      if (use_output_prefix)
	{
	qualified_vely_slice_output_prefix += base_output_path;
	qualified_velz_slice_output_prefix += base_output_path;

	if (unqualified_output_prefix.size() > 0)
	  {
	  qualified_vely_slice_output_prefix
	    += unqualified_output_prefix + "_";

	  unqualified_vely_slice_output_prefix
	    += unqualified_output_prefix + "_";

	  qualified_velz_slice_output_prefix
	    += unqualified_output_prefix + "_";

	  unqualified_velz_slice_output_prefix
	    += unqualified_output_prefix + "_";
	  }
	}

      qualified_vely_slice_output_prefix += vely_slice_output_prefix;
      unqualified_vely_slice_output_prefix += vely_slice_output_prefix;

      qualified_velz_slice_output_prefix += velz_slice_output_prefix;
      unqualified_velz_slice_output_prefix += velz_slice_output_prefix;

      // Make the vely slices

      QuickFlash::Slice::slice_simple(meshinfo, vely, vel_slice_xy);
      QuickFlash::Slice::slice_simple(meshinfo, vely, vel_slice_xz);

      if (save_slice_data)
	{
	const std::string qualified_vely_xy_output_filename
	  = qualified_vely_slice_output_prefix + "xy.csv";

	const std::string unqualified_vely_xy_output_filename
	  = unqualified_vely_slice_output_prefix + "xy.csv";

	const std::string qualified_vely_xz_output_filename
	  = qualified_vely_slice_output_prefix + "xz.csv";

	const std::string unqualified_vely_xz_output_filename
	  = unqualified_vely_slice_output_prefix + "xz.csv";

	QuickFlash::Slice::write_grid(unqualified_vely_xy_output_filename,
				    vel_slice_xy, true);

	metadata << "SliceData xy Dataset vely Min " 
		 << x_slice_min_x << "," << min_y 
		 << " Max " << x_slice_max_x << "," << max_y << " " 
		 << qualified_vely_xy_output_filename << endl;

	QuickFlash::Slice::write_grid(unqualified_vely_xz_output_filename,
				    vel_slice_xz, true);

	metadata << "SliceData xz Dataset vely Min " 
		 << x_slice_min_x << "," << min_z
		 << " Max " << x_slice_max_x << "," << max_z << " " 
		 << qualified_vely_xz_output_filename << endl;
	}

#ifdef USE_MAGICK
      if (save_slice_images)
	{
	const std::string qualified_vely_xy_image_filename
	  = qualified_vely_slice_output_prefix + "xy." + slice_image_type;

	const std::string unqualified_vely_xy_image_filename
	  = unqualified_vely_slice_output_prefix + "xy." + slice_image_type;

	const std::string qualified_vely_xz_image_filename
	  = qualified_vely_slice_output_prefix + "xz." + slice_image_type;

	const std::string unqualified_vely_xz_image_filename
	  = unqualified_vely_slice_output_prefix + "xz." + slice_image_type;

	write_image(unqualified_vely_xy_image_filename, vel_slice_xy);

	metadata << "SliceImage xy Dataset vely Min " 
		 << x_slice_min_x << "," << min_y 
		 << " Max " << x_slice_max_x << "," << max_y << " " 
		 << qualified_vely_xy_image_filename << endl;

	write_image(unqualified_vely_xz_image_filename, vel_slice_xz);

	metadata << "SliceImage xz Dataset vely Min " 
		 << x_slice_min_x << "," << min_z 
		 << " Max " << x_slice_max_x << "," << max_z << " " 
		 << qualified_vely_xz_image_filename << endl;
	}
#endif  // USE_MAGICK

      // Make the velz slices

      QuickFlash::Slice::slice_simple(meshinfo, velz, vel_slice_xy);
      QuickFlash::Slice::slice_simple(meshinfo, velz, vel_slice_xz);

      if (save_slice_data)
	{
	const std::string qualified_velz_xy_output_filename
	  = qualified_velz_slice_output_prefix + "xy.csv";

	const std::string unqualified_velz_xy_output_filename
	  = unqualified_velz_slice_output_prefix + "xy.csv";

	const std::string qualified_velz_xz_output_filename
	  = qualified_velz_slice_output_prefix + "xz.csv";

	const std::string unqualified_velz_xz_output_filename
	  = unqualified_velz_slice_output_prefix + "xz.csv";

	QuickFlash::Slice::write_grid(unqualified_velz_xy_output_filename,
				    vel_slice_xy, true);

	metadata << "SliceData xy Dataset velz Min " 
		 << x_slice_min_x << "," << min_y 
		 << " Max " << x_slice_max_x << "," << max_y << " " 
		 << qualified_velz_xy_output_filename << endl;

	QuickFlash::Slice::write_grid(unqualified_velz_xz_output_filename,
				    vel_slice_xz, true);

	metadata << "SliceData xz Dataset velz Min " 
		 << x_slice_min_x << "," << min_z
		 << " Max " << x_slice_max_x << "," << max_z << " " 
		 << qualified_velz_xz_output_filename << endl;
	}

#ifdef USE_MAGICK
      if (save_slice_images)
	{
	const std::string qualified_velz_xy_image_filename
	  = qualified_velz_slice_output_prefix + "xy." + slice_image_type;

	const std::string unqualified_velz_xy_image_filename
	  = unqualified_velz_slice_output_prefix + "xy." + slice_image_type;

	const std::string qualified_velz_xz_image_filename
	  = qualified_velz_slice_output_prefix + "xz." + slice_image_type;

	const std::string unqualified_velz_xz_image_filename
	  = unqualified_velz_slice_output_prefix + "xz." + slice_image_type;

	write_image(unqualified_velz_xy_image_filename, vel_slice_xy);

	metadata << "SliceImage xy Dataset velz Min " 
		 << x_slice_min_x << "," << min_y 
		 << " Max " << x_slice_max_x << "," << max_y << " " 
		 << qualified_velz_xy_image_filename << endl;

	write_image(unqualified_velz_xz_image_filename, vel_slice_xz);

	metadata << "SliceImage xz Dataset velz Min " 
		 << x_slice_min_x << "," << min_z 
		 << " Max " << x_slice_max_x << "," << max_z << " " 
		 << qualified_velz_xz_image_filename << endl;
	}
#endif  // USE_MAGICK
      }

    // Do the samples

    std::vector<double> slice_center_coords(3, 0.0);

    slice_center_coords[1] = slice_center_y;
    slice_center_coords[2] = slice_center_z;

    QuickFlash::Slice::SliceGrid vel_y_slice;
    QuickFlash::Slice::SliceGrid vel_z_slice;

    GridVector vel_slice_data(2);  // Pointers to y, z velocity slice data

    for (unsigned int i = 0; i < 2; i++)
      vel_slice_data[i] = 0;

    std::vector<double> struct_fn_r_radii;
    std::vector<unsigned int> struct_fn_r_counts;
    std::vector< std::vector<double> > struct_fn_r_values(num_orders);
    std::vector< std::vector<double> > struct_fn_r_stddev(num_orders);

    std::vector<double> struct_fn_y_positions;
    std::vector<unsigned int> struct_fn_y_counts;
    std::vector< std::vector<double> > struct_fn_y_values(num_orders);
    std::vector< std::vector<double> > struct_fn_y_stddev(num_orders);

    std::vector<double> struct_fn_z_positions;
    std::vector<unsigned int> struct_fn_z_counts;
    std::vector< std::vector<double> > struct_fn_z_values(num_orders);
    std::vector< std::vector<double> > struct_fn_z_stddev(num_orders);

    double current_slice_x = slice_min_x;

    for (unsigned int slice_index = 0; slice_index < num_slices; slice_index++)
      {
      slice_center_coords[0] = current_slice_x;

      cerr << "** Computing slice [ " << slice_index << " ] at x coordinate [ "
	   << current_slice_x << " ]" << endl;

      vel_y_slice.reset(slice_center_coords, slice_width, slice_samples,
			yz_azim_angle, yz_polar_angle, yz_roll_angle);
      vel_z_slice.reset(slice_center_coords, slice_width, slice_samples,
			yz_azim_angle, yz_polar_angle, yz_roll_angle);

      QuickFlash::Slice::slice_simple(meshinfo, vely, vel_y_slice);
      QuickFlash::Slice::slice_simple(meshinfo, velz, vel_z_slice);

      vel_slice_data[0] = &(vel_y_slice.get_grid_data());
      vel_slice_data[1] = &(vel_z_slice.get_grid_data());

      const QuickFlash::Block::BlockInfo & slice_info 
	= vel_y_slice.get_grid_info();

#ifdef USE_MPI
      if (mpi_size > 1)
	{
	int do_calc = continue_code;

	MPI_Bcast(&do_calc, 1, MPI_INT, root_rank, mpi_comm);

	get_scalar_struct_fn_master(slice_info, vel_slice_data, 
				    struct_fn_bin_size, struct_fn_max_radius,
				    order_list, 
				    struct_fn_r_radii, struct_fn_r_counts, 
				    struct_fn_r_values, struct_fn_r_stddev, 
				    struct_fn_y_positions, struct_fn_y_counts, 
				    struct_fn_y_values, struct_fn_y_stddev, 
				    struct_fn_z_positions, struct_fn_z_counts, 
				    struct_fn_z_values, struct_fn_z_stddev, 
				    mpi_comm);
	}
      else
	get_scalar_struct_fn(slice_info, vel_slice_data, struct_fn_bin_size,
			     struct_fn_max_radius, order_list,
			     struct_fn_r_radii, struct_fn_r_counts, 
			     struct_fn_r_values, struct_fn_r_stddev, 
			     struct_fn_y_positions, struct_fn_y_counts, 
			     struct_fn_y_values, struct_fn_y_stddev, 
			     struct_fn_z_positions, struct_fn_z_counts, 
			     struct_fn_z_values, struct_fn_z_stddev);
#else
      get_scalar_struct_fn(slice_info, vel_slice_data, struct_fn_bin_size,
			   struct_fn_max_radius, order_list, 
			   struct_fn_r_radii, struct_fn_r_counts, 
			   struct_fn_r_values, struct_fn_r_stddev, 
			   struct_fn_y_positions, struct_fn_y_counts, 
			   struct_fn_y_values, struct_fn_y_stddev, 
			   struct_fn_z_positions, struct_fn_z_counts, 
			   struct_fn_z_values, struct_fn_z_stddev);
#endif  // USE_MPI

      // Build the output file names
      
      std::stringstream basename_stream;

      basename_stream << unqualified_flash_datafile_name << "_slice_" 
		      << slice_index << ".csv";

      const std::string basename = basename_stream.str();

      std::string qualified_datafile_name = "";  // With directory path
      std::string unqualified_datafile_name = "";  // Without directory path

      if (use_output_prefix)
	{
	qualified_datafile_name += base_output_path;

	if (unqualified_output_prefix.size() > 0)
	  {
	  qualified_datafile_name += unqualified_output_prefix + "_";
	  unqualified_datafile_name += unqualified_output_prefix + "_";
	  }
	}

      qualified_datafile_name += basename;
      unqualified_datafile_name += basename;

      // Create the file and write the header

      std::ofstream f(qualified_datafile_name.c_str(), std::ios::trunc);

      f << comment_marker << "DataFile " << unqualified_flash_datafile_name 
	<< endl;
      f << comment_marker << "SimulationTime " << time << endl;
      f << comment_marker << "StructureFunctionOrders " << order_list << endl;
      f << comment_marker << "SlicePos " << current_slice_x << endl;

      f << comment_marker << "SliceSpacing " << slice_spacing << endl;

      if (!skip_flame_search)
	{
	f << comment_marker << "MinFlameLevel " << min_flame_level << endl;
	f << comment_marker << "MaxFlameLevel " << max_flame_level << endl;

	f << comment_marker << "FlameEnvelopeMinPos " << flame_back_x 
	  << endl;
	f << comment_marker << "FlameEnvelopeMaxPos " << flame_front_x 
	  << endl;
	}
      else
	f << comment_marker << "UserFlamePos " << user_flame_pos << endl;

      f << comment_marker << "SliceSamples " << y_samples << " " << z_samples
	<< endl;

      f << comment_marker << "BinSize " << struct_fn_bin_size << endl;
      f << comment_marker << "MaxRadius " << struct_fn_max_radius << endl;

      f << comment_marker << "FollowFlame ";

      if (follow_flame)
	f << "true";
      else
	f << "false";
      
      f << endl;

      if (follow_flame)
	{
	f << comment_marker << "FlameRel_min_x " << flame_rel_slice_min_x 
	  << endl;
	f << comment_marker << "FlameRel_max_x " << flame_rel_slice_max_x 
	  << endl;
	}

      if (!skip_flame_search)
	{
	f << comment_marker << "FlameSearchLevel " << flame_search_level 
	  << endl;
	f << comment_marker << "FlameSearchSamples " << flame_search_samples
	  << endl;
	}

      // Write out the data here ...

      f << comment_marker << "StructureData" << endl;

      f << "Bin_Center";

      for (unsigned int order_list_index = 0; order_list_index < num_orders;
	   order_list_index++)
	f << "\t" << "Order_" << order_list[order_list_index] << "_r_count"
	  << "\t" << "Order_" << order_list[order_list_index] << "_r_value"
	  << "\t" << "Order_" << order_list[order_list_index] << "_r_stddev"
	  << "\t" << "Order_" << order_list[order_list_index] << "_y_count"
	  << "\t" << "Order_" << order_list[order_list_index] << "_y_value"
	  << "\t" << "Order_" << order_list[order_list_index] << "_y_stddev"
	  << "\t" << "Order_" << order_list[order_list_index] << "_z_count"
	  << "\t" << "Order_" << order_list[order_list_index] << "_z_value"
	  << "\t" << "Order_" << order_list[order_list_index] << "_z_stddev";

      f << endl;

      const unsigned int num_bins = struct_fn_r_radii.size();

      for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
	{
	f << struct_fn_r_radii[bin_index];

	for (unsigned int order_list_index = 0; order_list_index < num_orders;
	     order_list_index++)
	  f << "\t" << struct_fn_r_counts[bin_index]
	    << "\t" << struct_fn_r_values[order_list_index][bin_index]
	    << "\t" << struct_fn_r_stddev[order_list_index][bin_index]
	    << "\t" << struct_fn_y_counts[bin_index]
	    << "\t" << struct_fn_y_values[order_list_index][bin_index]
	    << "\t" << struct_fn_y_stddev[order_list_index][bin_index]
	    << "\t" << struct_fn_z_counts[bin_index]
	    << "\t" << struct_fn_z_values[order_list_index][bin_index]
	    << "\t" << struct_fn_z_stddev[order_list_index][bin_index];

	f << endl;
	}

      f << endl;

      metadata << "Slice " << unqualified_datafile_name << endl;

      current_slice_x += slice_spacing;
      }

#ifdef USE_MPI
    if ((mpi_size > 1) && (is_root))
      {
      int do_calc = finished_code;

      MPI_Bcast(&do_calc, 1, MPI_INT, root_rank, mpi_comm);
      }
#endif  // USE_MPI
    }

#ifdef USE_MPI
  MPI_Finalize();
#endif  // USE_MPI

  return 0;
  }


