// C++ program file struct_fn_xy.cpp

/*
  By Nathan C. Hearn
     February 8, 2008

  Main driver for structure function calculation.  
  (Only for cardinal directions.)
*/


// Include the MPI headers first to make MPI/xlC on BG/P happy ... (grrrr....)

#ifdef USE_MPI
#include <mpi.h>
#include "mpi_utils.hpp"
#endif  // USE_MPI

#include <cstdlib>
#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>
#include <algorithm>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_slice_slicegrid.hpp"
#include "quickflash_slice_slicer.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_slice_output.hpp"
#include "flame_front.hpp"
#include "structure_xy.hpp"
#include "flame_dist.hpp"

#ifdef USE_MAGICK
#include "quickflash_image_imagemagick.hpp"
#endif  // USE_MAGICK


const unsigned int default_flame_search_level = 7;
const unsigned int default_flame_search_samples = 10;

const char default_image_type[] = "png";

const unsigned int default_buffer_size = 50;
const unsigned int default_cache_size = 1000;

const char flam_name[] = "flam";
const char dens_name[] = "dens";

const char velx_name[] = "velx";
const char vely_name[] = "vely";
const char velz_name[] = "velz";

const double deg_to_rad = (M_PI) / 180.0;

const double xy_roll_angle = 0.0 * deg_to_rad;
const double xy_polar_angle = 0.0 * deg_to_rad;
const double xy_azim_angle = 0.0 * deg_to_rad;

const double xz_roll_angle = 90.0 * deg_to_rad;
const double xz_polar_angle = 90.0 * deg_to_rad;
const double xz_azim_angle = -90.0 * deg_to_rad;

const double yz_roll_angle = 90.0 * deg_to_rad;
const double yz_polar_angle = 90.0 * deg_to_rad;
const double yz_azim_angle = 0.0 * deg_to_rad;

const double slice_expansion_factor = 1.1;  // 10% increase

enum SliceOriginType { MeshOrigin, FlameTip, FlameMassCenter, 
		       FlameMassFractCenter };

const char origin_meshorigin_label[] = "mesh";
const char origin_flametip_label[] = "tip";
const char origin_flamemasscenter_label[] = "masscenter";
const char origin_flamemassfractcenter_label[] = "massfractcenter";

const SliceOriginType default_slice_origin_type = MeshOrigin;

#ifdef USE_MPI
const int continue_code = 1;
const int finished_code = 0;
#endif  // USE_MPI


int read_origin_type_label(const std::string & origin_type_label,
			   SliceOriginType & slice_origin_type)
  {
  int ret_val = 0;  // OK

  if (origin_type_label == origin_meshorigin_label)
    slice_origin_type = MeshOrigin;
  else if (origin_type_label == origin_flametip_label)
    slice_origin_type = FlameTip;
  else if (origin_type_label == origin_flamemasscenter_label)
    slice_origin_type = FlameMassCenter;
  else if (origin_type_label == origin_flamemassfractcenter_label)
    slice_origin_type = FlameMassFractCenter;
  else
    ret_val = -1;  // Error state

  return ret_val;
  }


int get_origin_type_label(const SliceOriginType slice_origin_type,
			  std::string & origin_type_label)
  {
  int ret_val = 0;  // OK

  switch (slice_origin_type)
    {
    case MeshOrigin :
      origin_type_label = origin_meshorigin_label;
      break;

    case FlameTip :
      origin_type_label = origin_flametip_label;
      break;

    case FlameMassCenter :
      origin_type_label = origin_flamemasscenter_label;
      break;

    case FlameMassFractCenter :
      origin_type_label = origin_flamemassfractcenter_label;
      break;

    default :
      ret_val = -1;  // Error state
    }

  return ret_val;
  }


void compute_flame_distribution(const std::string & output_filename,
				const std::string & comment_marker,
				const QuickFlash::File::DataFile & df,
				const double flame_envelope_min_x,
				const double flame_envelope_max_x,
				const double mass_distrib_bin_width,
				const double min_flame_level,
				const double max_flame_level,
				double & flame_center_mass_fract,
				double & flame_center_mass,
				double & flame_min_pos, double & flame_max_pos)
  {
  using std::endl;

  const double flame_envelope_width 
    = flame_envelope_max_x - flame_envelope_min_x;

  const unsigned int distrib_bin_count 
    = static_cast<unsigned int>(flame_envelope_width / mass_distrib_bin_width)
      + 1;

  FlameDistribution flame_dist(df, 0, flame_envelope_min_x, distrib_bin_count,
			       mass_distrib_bin_width, min_flame_level, 
			       max_flame_level);

  flame_center_mass_fract = flame_dist.get_center_pos_mass_fract();
  flame_center_mass = flame_dist.get_center_pos_mass();

  flame_min_pos = flame_dist.get_flame_min_pos();
  flame_max_pos = flame_dist.get_flame_max_pos();

  const std::vector<double> & distrib_bin_centers 
    = flame_dist.get_bin_centers();

  const std::vector<double> & distrib_bin_volumes 
    = flame_dist.get_bin_volumes();

  const std::vector<double> & distrib_mass_bins_total 
    = flame_dist.get_mass_bins_total();

  const std::vector<double> & distrib_dens_bins_total_mean
    = flame_dist.get_dens_bins_total_mean();

  const std::vector<double> & distrib_dens_bins_total_dispersion
    = flame_dist.get_dens_bins_total_dispersion();

  const std::vector<double> & distrib_mass_bins_burned 
    = flame_dist.get_mass_bins_burned();

  const std::vector<double> & distrib_dens_bins_burned_mean
    = flame_dist.get_dens_bins_burned_mean();

  const std::vector<double> & distrib_dens_bins_burned_dispersion
    = flame_dist.get_dens_bins_burned_dispersion();

  const std::vector<double> & distrib_mass_bins_unburned 
    = flame_dist.get_mass_bins_unburned();

  const std::vector<double> & distrib_dens_bins_unburned_mean
    = flame_dist.get_dens_bins_unburned_mean();

  const std::vector<double> & distrib_dens_bins_unburned_dispersion
    = flame_dist.get_dens_bins_unburned_dispersion();
  
  const std::vector<double> & distrib_burned_mass_fract_bins 
    = flame_dist.get_mass_fract_bins_burned();

  // Write out the data

  const std::string & flash_filename = df.get_filename();

  std::string flash_path;
  std::string flash_name;

  QuickFlash::Utils::split_name_path_file(flash_filename, flash_path, 
					  flash_name);

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const double sim_time = meshinfo.get_sim_time();

  std::ofstream flamedist(output_filename.c_str(), std::ios::trunc);

  flamedist << comment_marker << "DataFile " << flash_name << endl;

  flamedist << comment_marker << "SimulationTime " << sim_time << endl;
	
  flamedist << comment_marker << "MinFlameLevel " << min_flame_level 
	    << endl;
  flamedist << comment_marker << "MaxFlameLevel " << max_flame_level 
	    << endl;

  flamedist << comment_marker << "FlameEnvelopeMinPos " 
	    << flame_envelope_min_x << endl;
  flamedist << comment_marker << "FlameEnvelopeMaxPos " 
	    << flame_envelope_max_x << endl;

  flamedist << comment_marker << "MassDistribBinWidth " 
	    << mass_distrib_bin_width << endl;

  flamedist << comment_marker << "MassDistribArrayMin " 
	    << flame_dist.get_array_min_coord() << endl;

  flamedist << comment_marker << "MassDistribArrayMax " 
	    << flame_dist.get_array_max_coord() << endl;

  flamedist << comment_marker << "NumBins " << distrib_bin_count << endl;
	
  flamedist << comment_marker << "DistribArrayTotalMass " 
	    << flame_dist.get_total_mass() << endl;

  flamedist << comment_marker << "DistribArrayBurnedMass " 
	    << flame_dist.get_total_burned_mass() << endl;

  flamedist << comment_marker << "DistribArrayUnburnedMass " 
	    << flame_dist.get_total_unburned_mass() << endl;

  flamedist << comment_marker << "Data" << endl;

  flamedist << "Bin_Center" 
	    << "\t" << "Bin_Volume" 
	    << "\t" << "Bin_Mass" 
	    << "\t" << "Bin_Mean_Density" 
	    << "\t" << "Bin_Density_Dispersion" 
	    << "\t" << "Bin_Burned_Mass" 
	    << "\t" << "Bin_Burned_Mean_Density"
	    << "\t" << "Bin_Burned_Density_Dispersion" 
	    << "\t" << "Bin_Unburned_Mass" 
	    << "\t" << "Bin_Unburned_Mean_Density" 
	    << "\t" << "Bin_Unburned_Density_Dispersion" 
	    << "\t" << "Bin_Burned_Mass_Fract" << endl;

  for (unsigned int bin_index = 0; bin_index < distrib_bin_count;
       bin_index++)
    flamedist << distrib_bin_centers[bin_index] 
	      << "\t" << distrib_bin_volumes[bin_index] 
	      << "\t" << distrib_mass_bins_total[bin_index] 
	      << "\t" << distrib_dens_bins_total_mean[bin_index] 
	      << "\t" << distrib_dens_bins_total_dispersion[bin_index] 
	      << "\t" << distrib_mass_bins_burned[bin_index] 
	      << "\t" << distrib_dens_bins_burned_mean[bin_index] 
	      << "\t" << distrib_dens_bins_burned_dispersion[bin_index] 
	      << "\t" << distrib_mass_bins_unburned[bin_index] 
	      << "\t" << distrib_dens_bins_unburned_mean[bin_index] 
	      << "\t" << distrib_dens_bins_unburned_dispersion[bin_index] 
	      << "\t" << distrib_burned_mass_fract_bins[bin_index] << endl;
  }


void get_slice_pos_list(const double slice_origin, 
			const double rel_slice_min_x,
			const double rel_slice_max_x, 
			const double slice_spacing, 
			const double volume_min_bound, 
			const double volume_max_bound,
			std::vector<double> & slice_pos_list)
  {
  slice_pos_list.clear();

  if (!(rel_slice_min_x < rel_slice_max_x))
    throw QuickFlash::Except("Min relative slice pos must be less than max",
			     __FILE__, __LINE__);

  if (!(slice_spacing > 0.0))
    throw QuickFlash::Except("Slice spacing must be positive", __FILE__, 
			     __LINE__);

  if (!(volume_min_bound < volume_max_bound))
    throw QuickFlash::Except("Min volume bound must be less than max",
			     __FILE__, __LINE__);

  double current_slice_pos = rel_slice_min_x;

  while ((!(current_slice_pos > rel_slice_max_x))
	 && (current_slice_pos < volume_max_bound))
    {
    if ((!(current_slice_pos < rel_slice_min_x)) 
	&& (current_slice_pos > volume_min_bound))
      {
      const double abs_slice_pos = current_slice_pos + slice_origin;

      slice_pos_list.push_back(abs_slice_pos);
      }

    current_slice_pos += slice_spacing;
    }
  }


#ifdef USE_MAGICK
int write_image(const std::string & filename, 
		const QuickFlash::Slice::SliceGrid & data)
  {
  int ret_val = -1;  // Error state

  // Build the colormap

  double min_val = 0.0;
  double max_val = 0.0;

  data.get_minmax_values(min_val, max_val);

  if (!(min_val < max_val))
    max_val = (2.0 * std::fabs(min_val)) + 1.0;  // Make something up ...

  QuickFlash::Color::Colormap colormap(min_val, max_val);

  QuickFlash::Image::ImageGrid image(data.get_grid_data(), colormap);

  ret_val = QuickFlash::Image::ImageMagick::write_image(filename, image);

  return ret_val;
  }
#endif  // USE_MAGICK


void write_structure_file(const std::string & filename,
			  const std::list<std::string> & comment_list,
			  const std::vector<unsigned int> & order_list,
			  const std::vector<double> & struct_fn_positions,
			  const std::vector<unsigned int> & struct_fn_y_counts,
			  const std::vector< std::vector<double> > 
			    & struct_fn_y_values,
			  const std::vector< std::vector<double> > 
			    & struct_fn_y_stddev,
			  const std::vector<unsigned int> & struct_fn_z_counts,
			  const std::vector< std::vector<double> >
			    & struct_fn_z_values,
			  const std::vector< std::vector<double> > 
			    & struct_fn_z_stddev,
			  const std::string & comment_marker="")
  {
  using std::endl;

  const unsigned int num_orders = order_list.size();

  std::ofstream f(filename.c_str(), std::ios::trunc);

  std::list<std::string>::const_iterator comment_iter = comment_list.begin();

  while (comment_iter != comment_list.end())
    {
    f << comment_marker << *comment_iter << endl;
    ++comment_iter;
    }

  f << endl;

  f << comment_marker << "StructureData" << endl;

  f << "Bin_Center" << "\t" << "Bin_y_Counts";

  for (unsigned int order_list_index = 0; order_list_index < num_orders;
	   order_list_index++)
    f << "\t" << "Order_" << order_list[order_list_index] << "_y_value" 
      << "\t" << "Order_" << order_list[order_list_index] << "_y_stddev";

  f << "\t" << "Bin_z_Counts";

  for (unsigned int order_list_index = 0; order_list_index < num_orders;
       order_list_index++)
    f << "\t" << "Order_" << order_list[order_list_index] << "_z_value"
      << "\t" << "Order_" << order_list[order_list_index] << "_z_stddev";

  f << endl;

  const unsigned int num_bins = struct_fn_positions.size();

  // ERROR CHECKING???

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    f << struct_fn_positions[bin_index];

    f << "\t" << struct_fn_y_counts[bin_index];

    for (unsigned int order_list_index = 0; order_list_index < num_orders;
	 order_list_index++)
      f << "\t" << struct_fn_y_values[order_list_index][bin_index]
	<< "\t" << struct_fn_y_stddev[order_list_index][bin_index];

    f << "\t" << struct_fn_z_counts[bin_index];

    for (unsigned int order_list_index = 0; order_list_index < num_orders;
	 order_list_index++)
      f << "\t" << struct_fn_z_values[order_list_index][bin_index]
	<< "\t" << struct_fn_z_stddev[order_list_index][bin_index];

    f << endl;
    }

  f << endl;
  }


void write_q_file(const std::string & filename,
		  const std::list<std::string> & comment_list,
		  const char q_axis,
		  const std::vector<double> & q_bin_radii,
		  const std::vector<unsigned int> & q_bin_cell_count,
		  const std::vector<double> & q_bin_accum,
		  const std::string & comment_marker="")
  {
  using std::endl;

  std::string q_name = "q_";

  for (unsigned int n = 0; n < 2; n++)
    q_name += q_axis;

  std::ofstream f(filename.c_str(), std::ios::trunc);

  std::list<std::string>::const_iterator comment_iter = comment_list.begin();

  while (comment_iter != comment_list.end())
    {
    f << comment_marker << *comment_iter << endl;
    ++comment_iter;
    }

  f << endl;

  f << comment_marker << "Data_" << q_name << endl;

  f << "Bin_Center_Radius" << "\t" << "Bin_Cell_Count" << "\t" 
    << q_name << "_value" << endl;

  const unsigned int num_bins = q_bin_radii.size();

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    f << q_bin_radii[bin_index] << "\t" << q_bin_cell_count[bin_index]
      << "\t" << q_bin_accum[bin_index] << endl;

  f << endl;
  }


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

#ifdef USE_MPI
  const MPI_Comm mpi_comm = MPI_COMM_WORLD;

  MPI_Init(&argc, &argv);

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  const bool is_root = (mpi_rank == root_rank) ? true : false;
#else
  const bool is_root = true;
#endif  // USE_MPI

  // Read the command line arguments and options

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, 
					option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 11;

  if (num_args < min_args)
    {
    if (is_root)
      {
      const std::string qualified_exec_name = argv[0];

      std::string exec_path;
      std::string exec_name;

      QuickFlash::Utils::split_name_path_file(qualified_exec_name, 
					      exec_path, exec_name);

      cerr << endl << "Usage: " << exec_name 
	   << " flash_datafile slice_spacing slice_min_x slice_max_x"
	   << endl << "\t" << "min_flame_level max_flame_level"
	   << endl << "\t" << "mass_distrib_bin_width y-samples z-samples"
	   << endl << "\t" << "structure_function_bin_size max_radius"
	   << endl << "\t" << "[ --function-orders=N1,N2,... ]"
	   << endl << "\t" 
	   << "[ --slice-x-origin=<mesh|tip|masscenter|massfractcenter> ]"
	   << endl << "\t" << "[ --skip-flame-search=FLAME_X_POS ]"
	   << endl << "\t" << "[ --output-prefix=FILENAME_PREFIX ]"
	   << endl << "\t" << "[ --add-comment-marks ]"
	   << endl << "\t" << "[ --save-x-axis-slices ]"
	   << endl << "\t" << "[ --save-cross-sect-slices ]"
#ifdef USE_MAGICK
	   << endl << "\t" << "[ --make-images[=IMAGE_TYPE] ]"
	   << endl << "\t" << "[ --images-only ]"
#endif  // USE_MAGICK
	   << endl << "\t" << "[ --buffer-size=NUM_BLOCKS ]"
	   << endl << "\t" << "[ --cache-size=NUM_BLOCKS ]"
	   << endl << endl;
      }

    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  const std::string flash_datafile = argument_list[argptr++];

  const double slice_spacing = std::atof(argument_list[argptr++].c_str());

  if (!(slice_spacing > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Slice spacing must be positive" << endl << endl;

    return -1;
    }

  const double rel_slice_min_x = std::atof(argument_list[argptr++].c_str());

  const double rel_slice_max_x = std::atof(argument_list[argptr++].c_str());

  if (!(rel_slice_min_x < rel_slice_max_x))
    {
    if (is_root)
      cerr << endl << "Error: Slice min x must be less than max x" << endl 
	   << endl;

    return -1;
    }

  const double min_flame_level = std::atof(argument_list[argptr++].c_str());

  if (!((min_flame_level > 0.0) && (min_flame_level < 1.0)))
    {
    if (is_root)
      cerr << endl 
	   << "Error: Min flame level must be in non-inclusive range "
	   << "(0.0, 1.0)" << endl << endl;

    return -1;
    }

  const double max_flame_level = std::atof(argument_list[argptr++].c_str());

  if (!((max_flame_level > min_flame_level) && (max_flame_level < 1.0)))
    {
    if (is_root)
      cerr << endl 
	   << "Error: Max flame level must lie between min level and 1.0"
	   << endl << endl;

    return -1;
    }

  const double mass_distrib_bin_width 
    = std::atof(argument_list[argptr++].c_str());

  if (!(mass_distrib_bin_width > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Mass distribution bin width must be positive"
	   << endl << endl;

    return -1;
    }

  const int y_samples_input = std::atoi(argument_list[argptr++].c_str());

  if (y_samples_input <= 1)
    {
    if (is_root)
      cerr << endl << "Error: At least two y-samples required." << endl 
	   << endl;

    return -1;
    }

  const unsigned int y_samples = static_cast<unsigned int>(y_samples_input);

  const int z_samples_input = std::atoi(argument_list[argptr++].c_str());

  if (z_samples_input <= 1)
    {
    if (is_root)
      cerr << endl << "Error: At least two z-samples required." << endl 
	   << endl;

    return -1;
    }

  const unsigned int z_samples = static_cast<unsigned int>(z_samples_input);

  const double struct_fn_bin_size = std::atof(argument_list[argptr++].c_str());

  if (!(struct_fn_bin_size > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Structure function bin size must be positive"
	   << endl << endl;

    return -1;
    }

  const double struct_fn_max_radius 
    = std::atof(argument_list[argptr++].c_str());

  if (!(struct_fn_max_radius > 0.0))
    {
    if (is_root)
      cerr << endl << "Error: Structure function max radius must be positive"
	   << endl << endl;

    return -1;
    }


  // Read the options

  bool order_list_set = false;
  std::vector<unsigned int> order_list;

  SliceOriginType slice_origin_type = default_slice_origin_type;

  bool skip_flame_search = false;
  double user_flame_pos = 0.0;

  bool use_output_prefix = false;
  std::string base_output_prefix;
  std::string base_output_path;
  std::string unqualified_output_prefix;  // No directory path component

  bool add_comment_markers = false;

  bool save_x_slices = false;
  bool save_cross_sect_slices = false;

#ifdef USE_MAGICK
  bool save_slice_images = false;

  std::string slice_image_type = default_image_type;
#endif  // USE_MAGICK

  bool save_slice_data = true;  // Unless only images are requested

  unsigned int buffer_size = default_buffer_size;
  unsigned int cache_size = default_cache_size;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "function-orders")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: List of function orders required for option"
	       << endl << endl;

	return -1;
	}

      current_option.get_value(order_list);

      order_list_set = true;
      }
    else if (keyword == "slice-x-origin")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl 
	       << "Error: Slice x-origin option requires an origin type"
	       << endl << endl;

	return -1;
	}

      const std::string & slice_origin_type_string 
	= current_option.get_value();

      if (read_origin_type_label(slice_origin_type_string, 
				 slice_origin_type) < 0)
	{
	if (is_root)
	  cerr << endl << "Error: Slice origin type [ " 
	       << slice_origin_type_string << " ] not recognized" << endl;

	return -1;
	}
      }
    else if (keyword == "skip-flame-search")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl 
	       << "Error: Skip flame search option requires flame x coord" 
	       << endl << endl;

	return -1;
	}

      skip_flame_search = true;

      current_option.get_value(user_flame_pos);
      }
    else if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: Output-prefix option requires an argument"
	       << endl << endl;

	return -1;
	}

      use_output_prefix = true;

      base_output_prefix = current_option.get_value();

      // Remove the directory path

      QuickFlash::Utils::split_name_path_file(base_output_prefix, 
					    base_output_path, 
					    unqualified_output_prefix);
      }
    else if (keyword == "add-comment-marks")
      {
      add_comment_markers = true;
      }
    else if (keyword == "save-x-axis-slices")
      {
      save_x_slices = true;
      }
    else if (keyword == "save-cross-sect-slices")
      {
      save_cross_sect_slices = true;
      }
    else if (keyword == "make-images")
      {
#ifdef USE_MAGICK
      save_slice_images = true;

      if (current_option.value_present())
	slice_image_type = current_option.get_value();
#else
      if (is_root)
        cerr << endl << "Error: Image output support not present" << endl 
	     << endl;

      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "images-only")
      {
#ifdef USE_MAGICK
      // NOTE: Without --make-images, uses default image type

      save_slice_images = true; 
      save_slice_data = false;
#else
      if (is_root)
        cerr << endl << "Error: Image output support not present" << endl 
	     << endl;

      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "buffer-size")
      {
      if (!current_option.value_present())
        {
	if (is_root)
	  cerr << endl 
	       << "Error: Buffer size option requires a non-negative integer"
	       << endl << endl;

	return -1;
	}

      current_option.get_value(buffer_size);
      }
    else if (keyword == "cache-size")
      {
      if (!current_option.value_present())
        {
	if (is_root)
	  cerr << endl 
	       << "Error: Cache size option requires a non-negative integer"
	       << endl << endl;

	return -1;
	}

      current_option.get_value(cache_size);
      }
    else
      {
      if (is_root)
	cerr << endl << "Error: Unrecognized option [ " << keyword << " ]"
	     << endl << endl;

      return -1;
      }

    ++option_iter;
    }

  // Get some additional quantities

  std::string base_flash_datafile_path;
  std::string unqualified_flash_datafile_name;

  QuickFlash::Utils::split_name_path_file(flash_datafile, 
					  base_flash_datafile_path, 
					  unqualified_flash_datafile_name);

  // Perform some checks

  if (!order_list_set)
    {
    order_list.resize(1);
    order_list[0] = 2;
    }

  const unsigned int num_orders = order_list.size();


  // Set up the comment markers

  std::string comment_marker = "";

  if (add_comment_markers)
    comment_marker = "# ";


  // Perform the calculation

  if (!is_root)
    {
#ifdef USE_MPI
    bool finished = false;

    while (!finished)
      {
      int do_calculation = 0;
      
      MPI_Bcast(&do_calculation, 1, MPI_INT, root_rank, mpi_comm);

      if (do_calculation == finished_code)
	finished = true;
      else
	get_scalar_struct_fn_worker(mpi_comm);
      }
#endif  // USE_MPI
    }
  else
    {
    // Open the metadata file

    std::string metadata_filename;

    if (use_output_prefix)
      {
      metadata_filename += base_output_path;

      if (unqualified_output_prefix.size() > 0)
	metadata_filename += unqualified_output_prefix + "_";
      }

    metadata_filename += unqualified_flash_datafile_name + "_metadata.out";

    std::ofstream metadata(metadata_filename.c_str(), std::ios::trunc);

    metadata << "DataFile " << unqualified_flash_datafile_name << endl;

    // Open the file and obtain some information

    const QuickFlash::File::DataFile df(flash_datafile, buffer_size, true);
  
    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double time = meshinfo.get_sim_time();

    metadata << "SimTime " << time << endl;

    const unsigned int dims = meshinfo.get_dims();

    if (dims != 3)
      {
      cerr << endl << "Error: Only 3D data supported" << endl << endl;

      return -2;
      }

    const std::vector<double> & min_bounds = meshinfo.get_volume_minbounds();
    const std::vector<double> & max_bounds = meshinfo.get_volume_maxbounds();

    metadata << "MinBounds " << min_bounds << endl;
    metadata << "MaxBounds " << max_bounds << endl;

    const double volume_min_x = min_bounds[0];
    const double volume_max_x = max_bounds[0];

    const double min_y = min_bounds[1];
    const double max_y = max_bounds[1];

    const double min_z = min_bounds[2];
    const double max_z = max_bounds[2];

    const double slice_center_y = 0.5 * (min_y + max_y);
    const double slice_center_z = 0.5 * (min_z + max_z);

    const double slice_width_y = max_y - min_y;
    const double slice_width_z = max_z - min_z;

    std::vector<double> slice_width(2);

    slice_width[0] = slice_width_y;
    slice_width[1] = slice_width_z;
  
    std::vector<unsigned int> slice_samples(2);

    slice_samples[0] = y_samples;
    slice_samples[1] = z_samples;

    // Prepare some datasets

    const QuickFlash::File::Dataset & flam = df.get_dataset(flam_name);

    flam.set_cache_size(cache_size);

    const QuickFlash::File::Dataset & dens = df.get_dataset(dens_name);

    dens.set_cache_size(cache_size);
  
    // Determine the range of slice x-values

    double flame_envelope_min_x = 0.0;
    double flame_envelope_max_x = 0.0;

    metadata << "MinFlameLevel " << min_flame_level << endl;
    metadata << "MaxFlameLevel " << max_flame_level << endl;

   if (skip_flame_search)
      {
      metadata << "UserFlamePos " << user_flame_pos << endl;

      flame_envelope_min_x = rel_slice_min_x;
      flame_envelope_max_x = rel_slice_max_x;

      if (slice_origin_type != MeshOrigin)
	{
	flame_envelope_min_x += user_flame_pos;
	flame_envelope_max_x += user_flame_pos;
	}
      }
    else
      {
      if (get_flame_bounds(df, min_flame_level, max_flame_level,
			   flame_envelope_min_x, flame_envelope_max_x) < 0)
	{
	cerr << endl << "Unable to find flame boundaries" << endl << endl;
	return -2;
	}

      metadata << "FlameEnvelopeMinPos " << flame_envelope_min_x << endl;
      metadata << "FlameEnvelopeMaxPos " << flame_envelope_max_x << endl;
      }
      
    // Determine the distribution of mass

    metadata << "MassDistribBinSpacing " << mass_distrib_bin_width << endl;

    const std::string flamedist_basename 
      = unqualified_flash_datafile_name + "_flame_dist.csv" ;

    std::string qualified_flamedist_name = "";  // With directory path
    std::string unqualified_flamedist_name = "";  // Without directory path

    if (use_output_prefix)
      {
      qualified_flamedist_name += base_output_path;

      if (unqualified_output_prefix.size() > 0)
	{
	qualified_flamedist_name += unqualified_output_prefix + "_";
	unqualified_flamedist_name += unqualified_output_prefix + "_";
	}
      }

    qualified_flamedist_name += flamedist_basename;
    unqualified_flamedist_name += flamedist_basename;

    double flame_min_x = 0.0;
    double flame_max_x = 0.0;

    double flame_center_mass_fract = 0.0;
    double flame_center_mass = 0.0;

    compute_flame_distribution(qualified_flamedist_name, comment_marker, df,
			       flame_envelope_min_x, flame_envelope_max_x, 
			       mass_distrib_bin_width, min_flame_level, 
			       max_flame_level, flame_center_mass_fract, 
			       flame_center_mass, flame_min_x, 
			       flame_max_x);

    metadata << "MassDistribFile " << unqualified_flamedist_name << endl;

    metadata << "FlameCenterMassFract " << flame_center_mass_fract << endl;
    metadata << "FlameCenterMass " << flame_center_mass << endl;

    metadata << "DistribFlameMinPos " << flame_min_x << endl;
    metadata << "DistribFlameMaxPos " << flame_max_x << endl;

    // Determine reference position

    double slice_pos_origin = 0.0;

    switch (slice_origin_type)
      {
      case MeshOrigin :
	slice_pos_origin = 0.0;
	break;

      case FlameTip :
	slice_pos_origin = flame_max_x;
	break;

      case FlameMassCenter :
	slice_pos_origin = flame_center_mass;
	break;

      case FlameMassFractCenter :
	slice_pos_origin = flame_center_mass_fract;
	break;

      default :
	if (is_root)
	  cerr << "Error: Slice origin type not handled" << endl << endl;

	return -2;
      }

    std::string slice_origin_type_label = "unknown";

    get_origin_type_label(slice_origin_type, slice_origin_type_label);

    metadata << "SliceOriginType " << slice_origin_type_label << endl;
    metadata << "SliceOrigin " << slice_pos_origin << endl;

    // Compute the list of slices

    metadata << "SliceRelativeMinBound " << rel_slice_min_x << endl;
    metadata << "SliceRelativeMaxBound " << rel_slice_max_x << endl;

    std::vector<double> slice_pos_list;

    get_slice_pos_list(slice_pos_origin, rel_slice_min_x, rel_slice_max_x,
		       slice_spacing, volume_min_x, volume_max_x, 
		       slice_pos_list);

    const unsigned int num_slices = slice_pos_list.size();

    metadata << "NumSlices " << num_slices << endl;
    metadata << "SliceAbsolutePositions " << slice_pos_list << endl;
    
    // Collect the slice data

    metadata << "StructFnOrderList " << order_list << endl;

    // Set up the datasets

    //    const QuickFlash::File::Dataset & velx
    //      = df.get_dataset(velx_name);

    const QuickFlash::File::Dataset & vely = df.get_dataset(vely_name);

    vely.set_cache_size(cache_size);

    const QuickFlash::File::Dataset & velz = df.get_dataset(velz_name);

    velz.set_cache_size(cache_size);

    // Output the x-axis slices

    if (save_x_slices)
      {
      const double slice_area_min_x = slice_pos_origin + rel_slice_min_x;
      const double slice_area_max_x = slice_pos_origin + rel_slice_max_x;

      const double slice_center_x = 0.5 * (slice_area_min_x 
					   + slice_area_max_x);

      const double slice_area_width = slice_area_max_x - slice_area_min_x;

      const double full_slice_width 
	= slice_area_width * slice_expansion_factor;

      const double full_slice_half_width = 0.5 * full_slice_width;

      const double full_slice_min_x = slice_center_x - full_slice_half_width;
      const double full_slice_max_x = slice_center_x + full_slice_half_width;

      // Use y-samples as basis for slice resolution

      const double slice_pixel_width_y
	= (max_y - min_y) / static_cast<double>(y_samples);

      const double x_pixels_fl = full_slice_width / slice_pixel_width_y;

      const unsigned int x_pixels 
	= std::max(static_cast<unsigned int>(2), 
		   static_cast<unsigned int>(x_pixels_fl));

      std::vector<double> slice_center(3);

      slice_center[0] = slice_center_x;
      slice_center[1] = slice_center_y;
      slice_center[2] = slice_center_z;

      std::vector<double> slice_width_xy(2);

      slice_width_xy[0] = full_slice_width;
      slice_width_xy[1] = slice_width_y;

      std::vector<double> slice_width_xz(2);

      slice_width_xz[0] = full_slice_width;
      slice_width_xz[1] = slice_width_z;

      std::vector<unsigned int> pixels_xy(2);

      pixels_xy[0] = x_pixels;
      pixels_xy[1] = y_samples;

      std::vector<unsigned int> pixels_xz(2);

      pixels_xz[0] = x_pixels;
      pixels_xz[1] = z_samples;

      QuickFlash::Slice::SliceGrid vel_slice_xy(slice_center, slice_width_xy,
						pixels_xy, xy_azim_angle,
						xy_polar_angle, xy_roll_angle);

      QuickFlash::Slice::SliceGrid vel_slice_xz(slice_center, slice_width_xz,
						pixels_xz, xz_azim_angle,
						xz_polar_angle, xz_roll_angle);

      const std::string vely_slice_output_prefix 
	= unqualified_flash_datafile_name + "_vely_slice_";

      const std::string velz_slice_output_prefix
	= unqualified_flash_datafile_name + "_velz_slice_";

      std::string qualified_vely_slice_output_prefix = "";  // With path
      std::string unqualified_vely_slice_output_prefix = "";  // Without path
      
      std::string qualified_velz_slice_output_prefix = "";  // With path
      std::string unqualified_velz_slice_output_prefix = "";  // Without path

      if (use_output_prefix)
	{
	qualified_vely_slice_output_prefix += base_output_path;
	qualified_velz_slice_output_prefix += base_output_path;

	if (unqualified_output_prefix.size() > 0)
	  {
	  qualified_vely_slice_output_prefix
	    += unqualified_output_prefix + "_";

	  unqualified_vely_slice_output_prefix
	    += unqualified_output_prefix + "_";

	  qualified_velz_slice_output_prefix
	    += unqualified_output_prefix + "_";

	  unqualified_velz_slice_output_prefix
	    += unqualified_output_prefix + "_";
	  }
	}

      qualified_vely_slice_output_prefix += vely_slice_output_prefix;
      unqualified_vely_slice_output_prefix += vely_slice_output_prefix;

      qualified_velz_slice_output_prefix += velz_slice_output_prefix;
      unqualified_velz_slice_output_prefix += velz_slice_output_prefix;

      // Make the vely slices

      QuickFlash::Slice::slice_simple(meshinfo, vely, vel_slice_xy);
      QuickFlash::Slice::slice_simple(meshinfo, vely, vel_slice_xz);

      if (save_slice_data)
	{
	const std::string qualified_vely_xy_output_filename
	  = qualified_vely_slice_output_prefix + "xy.csv";

	const std::string unqualified_vely_xy_output_filename
	  = unqualified_vely_slice_output_prefix + "xy.csv";

	const std::string qualified_vely_xz_output_filename
	  = qualified_vely_slice_output_prefix + "xz.csv";

	const std::string unqualified_vely_xz_output_filename
	  = unqualified_vely_slice_output_prefix + "xz.csv";

	QuickFlash::Slice::write_grid(unqualified_vely_xy_output_filename,
				    vel_slice_xy, true);

	metadata << "SliceData xy Dataset vely Min " 
		 << full_slice_min_x << "," << min_y 
		 << " Max " << full_slice_max_x << "," << max_y << " " 
		 << qualified_vely_xy_output_filename << endl;

	QuickFlash::Slice::write_grid(unqualified_vely_xz_output_filename,
				    vel_slice_xz, true);

	metadata << "SliceData xz Dataset vely Min " 
		 << full_slice_min_x << "," << min_z
		 << " Max " << full_slice_max_x << "," << max_z << " " 
		 << qualified_vely_xz_output_filename << endl;
	}

#ifdef USE_MAGICK
      if (save_slice_images)
	{
	const std::string qualified_vely_xy_image_filename
	  = qualified_vely_slice_output_prefix + "xy." + slice_image_type;

	const std::string unqualified_vely_xy_image_filename
	  = unqualified_vely_slice_output_prefix + "xy." + slice_image_type;

	const std::string qualified_vely_xz_image_filename
	  = qualified_vely_slice_output_prefix + "xz." + slice_image_type;

	const std::string unqualified_vely_xz_image_filename
	  = unqualified_vely_slice_output_prefix + "xz." + slice_image_type;

	write_image(unqualified_vely_xy_image_filename, vel_slice_xy);

	metadata << "SliceImage xy Dataset vely Min " 
		 << full_slice_min_x << "," << min_y 
		 << " Max " << full_slice_max_x << "," << max_y << " " 
		 << qualified_vely_xy_image_filename << endl;

	write_image(unqualified_vely_xz_image_filename, vel_slice_xz);

	metadata << "SliceImage xz Dataset vely Min " 
		 << full_slice_min_x << "," << min_z 
		 << " Max " << full_slice_max_x << "," << max_z << " " 
		 << qualified_vely_xz_image_filename << endl;
	}
#endif  // USE_MAGICK

      // Make the velz slices

      QuickFlash::Slice::slice_simple(meshinfo, velz, vel_slice_xy);
      QuickFlash::Slice::slice_simple(meshinfo, velz, vel_slice_xz);

      if (save_slice_data)
	{
	const std::string qualified_velz_xy_output_filename
	  = qualified_velz_slice_output_prefix + "xy.csv";

	const std::string unqualified_velz_xy_output_filename
	  = unqualified_velz_slice_output_prefix + "xy.csv";

	const std::string qualified_velz_xz_output_filename
	  = qualified_velz_slice_output_prefix + "xz.csv";

	const std::string unqualified_velz_xz_output_filename
	  = unqualified_velz_slice_output_prefix + "xz.csv";

	QuickFlash::Slice::write_grid(unqualified_velz_xy_output_filename,
				    vel_slice_xy, true);

	metadata << "SliceData xy Dataset velz Min " 
		 << full_slice_min_x << "," << min_y 
		 << " Max " << full_slice_max_x << "," << max_y << " " 
		 << qualified_velz_xy_output_filename << endl;

	QuickFlash::Slice::write_grid(unqualified_velz_xz_output_filename,
				    vel_slice_xz, true);

	metadata << "SliceData xz Dataset velz Min " 
		 << full_slice_min_x << "," << min_z
		 << " Max " << full_slice_max_x << "," << max_z << " " 
		 << qualified_velz_xz_output_filename << endl;
	}

#ifdef USE_MAGICK
      if (save_slice_images)
	{
	const std::string qualified_velz_xy_image_filename
	  = qualified_velz_slice_output_prefix + "xy." + slice_image_type;

	const std::string unqualified_velz_xy_image_filename
	  = unqualified_velz_slice_output_prefix + "xy." + slice_image_type;

	const std::string qualified_velz_xz_image_filename
	  = qualified_velz_slice_output_prefix + "xz." + slice_image_type;

	const std::string unqualified_velz_xz_image_filename
	  = unqualified_velz_slice_output_prefix + "xz." + slice_image_type;

	write_image(unqualified_velz_xy_image_filename, vel_slice_xy);

	metadata << "SliceImage xy Dataset velz Min " 
		 << full_slice_min_x << "," << min_y 
		 << " Max " << full_slice_max_x << "," << max_y << " " 
		 << qualified_velz_xy_image_filename << endl;

	write_image(unqualified_velz_xz_image_filename, vel_slice_xz);

	metadata << "SliceImage xz Dataset velz Min " 
		 << full_slice_min_x << "," << min_z 
		 << " Max " << full_slice_max_x << "," << max_z << " " 
		 << qualified_velz_xz_image_filename << endl;
	}
#endif  // USE_MAGICK
      }

    // Do the samples

    std::vector<double> slice_center_coords(3, 0.0);

    slice_center_coords[1] = slice_center_y;
    slice_center_coords[2] = slice_center_z;

    QuickFlash::Slice::SliceGrid vel_y_slice;
    QuickFlash::Slice::SliceGrid vel_z_slice;

    GridVector vel_slice_data(2);  // Pointers to y, z velocity slice data

    for (unsigned int i = 0; i < 2; i++)
      vel_slice_data[i] = 0;

    std::vector<double> struct_fn_y_positions;
    std::vector<unsigned int> struct_fn_y_counts;
    std::vector< std::vector<double> > struct_fn_y_values(num_orders);
    std::vector< std::vector<double> > struct_fn_y_stddev(num_orders);

    std::vector<double> struct_fn_z_positions;
    std::vector<unsigned int> struct_fn_z_counts;
    std::vector< std::vector<double> > struct_fn_z_values(num_orders);
    std::vector< std::vector<double> > struct_fn_z_stddev(num_orders);

    std::vector<double> q_yy_bin_radii;
    std::vector<unsigned int> q_yy_bin_cell_count;
    std::vector<double> q_yy_bin_accum;

    std::vector<double> q_zz_bin_radii;
    std::vector<unsigned int> q_zz_bin_cell_count;
    std::vector<double> q_zz_bin_accum;

    for (unsigned int slice_index = 0; slice_index < num_slices; slice_index++)
      {
      const double current_slice_x = slice_pos_list[slice_index];

      slice_center_coords[0] = current_slice_x;

      cerr << "** Computing slice [ " << slice_index << " ] at x coordinate [ "
	   << current_slice_x << " ]" << endl;

      vel_y_slice.reset(slice_center_coords, slice_width, slice_samples,
			yz_azim_angle, yz_polar_angle, yz_roll_angle);
      vel_z_slice.reset(slice_center_coords, slice_width, slice_samples,
			yz_azim_angle, yz_polar_angle, yz_roll_angle);

      QuickFlash::Slice::slice_simple(meshinfo, vely, vel_y_slice);
      QuickFlash::Slice::slice_simple(meshinfo, velz, vel_z_slice);

      vel_slice_data[0] = &(vel_y_slice.get_grid_data());
      vel_slice_data[1] = &(vel_z_slice.get_grid_data());

      const QuickFlash::Block::BlockInfo & slice_info 
	= vel_y_slice.get_grid_info();

#ifdef USE_MPI
      if (mpi_size > 1)
	{
	int do_calc = continue_code;

	MPI_Bcast(&do_calc, 1, MPI_INT, root_rank, mpi_comm);

	get_scalar_struct_fn_master(slice_info, vel_slice_data, 
				    struct_fn_bin_size, struct_fn_max_radius,
				    order_list,
				    struct_fn_y_positions, struct_fn_y_counts, 
				    struct_fn_y_values, struct_fn_y_stddev, 
				    struct_fn_z_positions, struct_fn_z_counts, 
				    struct_fn_z_values, struct_fn_z_stddev, 
				    q_yy_bin_radii, q_yy_bin_cell_count,
				    q_yy_bin_accum, 
				    q_zz_bin_radii, q_zz_bin_cell_count,
				    q_zz_bin_accum,
				    mpi_comm);
	}
      else
	get_scalar_struct_fn(slice_info, vel_slice_data, struct_fn_bin_size,
			     struct_fn_max_radius, order_list,
			     struct_fn_y_positions, struct_fn_y_counts, 
			     struct_fn_y_values, struct_fn_y_stddev, 
			     struct_fn_z_positions, struct_fn_z_counts, 
			     struct_fn_z_values, struct_fn_z_stddev, 
			     q_yy_bin_radii, q_yy_bin_cell_count,
			     q_yy_bin_accum, 
			     q_zz_bin_radii, q_zz_bin_cell_count,
			     q_zz_bin_accum);
#else
      get_scalar_struct_fn(slice_info, vel_slice_data, struct_fn_bin_size,
			   struct_fn_max_radius, order_list,  
			   struct_fn_y_positions, struct_fn_y_counts, 
			   struct_fn_y_values, struct_fn_y_stddev, 
			   struct_fn_z_positions, struct_fn_z_counts, 
			   struct_fn_z_values, struct_fn_z_stddev, 
			   q_yy_bin_radii, q_yy_bin_cell_count,
			   q_yy_bin_accum, 
			   q_zz_bin_radii, q_zz_bin_cell_count,
			   q_zz_bin_accum);
#endif  // USE_MPI

      // Set up the list of comments for the structure and integral scale files

      std::list<std::string> comment_list;

      std::stringstream comment_stream;
      
      comment_stream.str("");

      comment_stream << comment_marker << "DataFile " 
		     << unqualified_flash_datafile_name;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");

      comment_stream << "SimulationTime " << time;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");

      comment_stream << "StructureFunctionOrders " << order_list;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");

      comment_stream << "SlicePos " << current_slice_x;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");

      comment_stream << "SliceSpacing " << slice_spacing;

      comment_list.push_back(comment_stream.str());

      if (!skip_flame_search)
	{
	comment_stream.str("");

	comment_stream << "FlameEnvelopeMinPos " << flame_envelope_min_x;

	comment_list.push_back(comment_stream.str());

	comment_stream.str("");

	comment_stream << "FlameEnvelopeMaxPos " << flame_envelope_max_x;

	comment_list.push_back(comment_stream.str());
	}
      else
	{
	comment_stream.str("");

	comment_stream << "UserFlamePos " << user_flame_pos;

	comment_list.push_back(comment_stream.str());
	}

      comment_stream.str("");
	
      comment_stream << "MinFlameLevel " << min_flame_level;

      comment_list.push_back(comment_stream.str());
      
      comment_stream.str("");
	
      comment_stream << "MaxFlameLevel " << max_flame_level;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "FlameMinPos " << flame_min_x;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "FlameMaxPos " << flame_max_x;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "FlameCenterMassFract " << flame_center_mass_fract;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "FlameCenterMass " << flame_center_mass;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "SliceOriginType " << slice_origin_type_label;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "SlicePosOrigin " << slice_pos_origin;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "SliceSpacing " << slice_spacing;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "SliceRelativeMinPos " << rel_slice_min_x;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "SliceRelativeMaxPos " << rel_slice_max_x;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "AbsoluteSlicePositions " << slice_pos_list;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "SliceGridDims " << y_samples << "," << z_samples;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "StructFnBinSize " << struct_fn_bin_size;

      comment_list.push_back(comment_stream.str());

      comment_stream.str("");
	
      comment_stream << "StructFnMaxRadius " << struct_fn_max_radius;

      comment_list.push_back(comment_stream.str());

      // Write the structure file

      std::stringstream structfile_basename_stream;

      structfile_basename_stream << unqualified_flash_datafile_name 
				 << "_slice_" << slice_index << "_struct.csv";

      const std::string structfile_basename = structfile_basename_stream.str();

      std::string qualified_structfile_name = "";  // With directory path
      std::string unqualified_structfile_name = "";  // Without directory path

      if (use_output_prefix)
	{
	qualified_structfile_name += base_output_path;

	if (unqualified_output_prefix.size() > 0)
	  {
	  qualified_structfile_name += unqualified_output_prefix + "_";
	  unqualified_structfile_name += unqualified_output_prefix + "_";
	  }
	}

      qualified_structfile_name += structfile_basename;
      unqualified_structfile_name += structfile_basename;

      write_structure_file(qualified_structfile_name, comment_list,
			   order_list, struct_fn_y_positions,
			   struct_fn_y_counts, struct_fn_y_values,
			   struct_fn_y_stddev, struct_fn_z_counts,
			   struct_fn_z_values, struct_fn_z_stddev,
			   comment_marker);

      // Write the q_yy file

      std::stringstream q_yy_basename_stream;

      q_yy_basename_stream << unqualified_flash_datafile_name
			   << "_slice_" << slice_index << "_q_yy.csv";

      const std::string q_yy_file_basename = q_yy_basename_stream.str();

      std::string qualified_q_yy_file_name = "";  // With directory path
      std::string unqualified_q_yy_file_name = "";  // Without directory path

      if (use_output_prefix)
	{
	qualified_q_yy_file_name += base_output_path;

	if (unqualified_output_prefix.size() > 0)
	  {
	  qualified_q_yy_file_name += unqualified_output_prefix + "_";
	  unqualified_q_yy_file_name += unqualified_output_prefix + "_";
	  }
	}

      qualified_q_yy_file_name += q_yy_file_basename;
      unqualified_q_yy_file_name += q_yy_file_basename;

      write_q_file(qualified_q_yy_file_name, comment_list, 'y', q_yy_bin_radii,
		   q_yy_bin_cell_count, q_yy_bin_accum, comment_marker);

      // Write the q_zz file

      std::stringstream q_zz_basename_stream;

      q_zz_basename_stream << unqualified_flash_datafile_name
			   << "_slice_" << slice_index << "_q_zz.csv";

      const std::string q_zz_file_basename = q_zz_basename_stream.str();

      std::string qualified_q_zz_file_name = "";  // With directory path
      std::string unqualified_q_zz_file_name = "";  // Without directory path

      if (use_output_prefix)
	{
	qualified_q_zz_file_name += base_output_path;

	if (unqualified_output_prefix.size() > 0)
	  {
	  qualified_q_zz_file_name += unqualified_output_prefix + "_";
	  unqualified_q_zz_file_name += unqualified_output_prefix + "_";
	  }
	}

      qualified_q_zz_file_name += q_zz_file_basename;
      unqualified_q_zz_file_name += q_zz_file_basename;

      write_q_file(qualified_q_zz_file_name, comment_list, 'z', q_zz_bin_radii,
		   q_zz_bin_cell_count, q_zz_bin_accum, comment_marker);

      // Save a reference to the files generated

      metadata << "Slice " << unqualified_structfile_name << " " 
	       << unqualified_q_yy_file_name << " " 
	       << unqualified_q_zz_file_name << endl;
      }

#ifdef USE_MPI
    if ((mpi_size > 1) && (is_root))
      {
      int do_calc = finished_code;

      MPI_Bcast(&do_calc, 1, MPI_INT, root_rank, mpi_comm);
      }
#endif  // USE_MPI
    }

#ifdef USE_MPI
  MPI_Finalize();
#endif  // USE_MPI

  return 0;
  }


