// C++ program file structure_xy.cpp

/*
  By Nathan C. Hearn
     February 1, 2008

  Functions for computing structure functions.  (Only for cardinal directions.)

  For right now, it is assumed that periodic boundary conditions exist.
*/


// Include the MPI headers first to make MPI/xlC on BG/P happy ... (grrrr....)

#ifdef USE_MPI
#include <mpi.h>
#include "decomp_utils.hpp"
#include "mpi_utils.hpp"
#endif  // USE_MPI

#include "structure_xy.hpp"
#include <cmath>
#include <vector>
#include <algorithm>
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_except.hpp"
#include "quickflash_geometry.hpp"
#include "quickflash_stats_bins.hpp"


#include <iostream>
using std::cerr;
using std::endl;


void get_scalar_struct_fn(const QuickFlash::Block::BlockInfo & grid_info,
			  const GridVector & vel_data, 
			  const double fn_bin_width, 
			  const double fn_max_radius, 
			  const std::vector<unsigned int> & fn_order_list,
			  std::vector<double> & y_bin_positions,
			  std::vector<unsigned int> & y_bin_cell_count,
			  std::vector< std::vector<double> > 
			    & fn_values_y_list,
			  std::vector< std::vector<double> >
			    & fn_stddev_y_list,
			  std::vector<double> & z_bin_positions,
			  std::vector<unsigned int> & z_bin_cell_count,
			  std::vector< std::vector<double> > 
			    & fn_values_z_list,
			  std::vector< std::vector<double> >
			    & fn_stddev_z_list,
			  std::vector<double> & q_yy_bin_radii,
			  std::vector<unsigned int> & q_yy_bin_cell_count,
			  std::vector<double> & q_yy_bin_accum,
			  std::vector<double> & q_zz_bin_radii,
			  std::vector<unsigned int> & q_zz_bin_cell_count,
			  std::vector<double> & q_zz_bin_accum)
  {
  const unsigned int num_orders = fn_order_list.size();

  if (num_orders < 1)
    throw QuickFlash::Except("No function orders specified", __FILE__, __LINE__);

  const unsigned int num_cells = grid_info.get_num_cells();

  std::vector<QuickFlash::Stats::BinArray> function_bins_y_list;
  std::vector<QuickFlash::Stats::BinArray> function_bins_z_list;

  QuickFlash::Stats::BinArray q_yy_bins;
  QuickFlash::Stats::BinArray q_zz_bins;

  get_scalar_struct_fn(grid_info, vel_data, fn_bin_width, fn_max_radius, 
		       fn_order_list, 0, num_cells,
		       function_bins_y_list, function_bins_z_list,
		       q_yy_bins, q_zz_bins);

  const QuickFlash::Stats::BinArray & prototype_y_bins = function_bins_y_list[0];

  y_bin_positions = prototype_y_bins.get_bin_centers();

  const unsigned int num_y_bins = y_bin_positions.size();

  y_bin_cell_count.resize(num_y_bins);

  for (unsigned int bin_index = 0; bin_index < num_y_bins; bin_index++)
    y_bin_cell_count[bin_index] 
      = prototype_y_bins[bin_index].get_num_entries();

  fn_values_y_list.resize(num_orders);
  fn_stddev_y_list.resize(num_orders);

  for (unsigned int order_index = 0; order_index < num_orders; order_index++)
    {
    const QuickFlash::Stats::BinArray & function_bins_y
      = function_bins_y_list[order_index];

    std::vector<double> & fn_values = fn_values_y_list[order_index];
    std::vector<double> & stddev_values = fn_stddev_y_list[order_index];

    fn_values.resize(num_y_bins);
    stddev_values.resize(num_y_bins);

    for (unsigned int bin_index = 0; bin_index < num_y_bins; bin_index++)
      {
      fn_values[bin_index] = function_bins_y[bin_index].get_mean();
      stddev_values[bin_index] = function_bins_y[bin_index].get_pop_stddev();
      }
    }

  const QuickFlash::Stats::BinArray & prototype_z_bins = function_bins_z_list[0];

  z_bin_positions = prototype_z_bins.get_bin_centers();

  const unsigned int num_z_bins = z_bin_positions.size();

  z_bin_cell_count.resize(num_z_bins);

  for (unsigned int bin_index = 0; bin_index < num_z_bins; bin_index++)
    z_bin_cell_count[bin_index] 
      = prototype_z_bins[bin_index].get_num_entries();

  fn_values_z_list.resize(num_orders);
  fn_stddev_z_list.resize(num_orders);

  for (unsigned int order_index = 0; order_index < num_orders; order_index++)
    {
    const QuickFlash::Stats::BinArray & function_bins_z
      = function_bins_z_list[order_index];

    std::vector<double> & fn_values = fn_values_z_list[order_index];
    std::vector<double> & stddev_values = fn_stddev_z_list[order_index];

    fn_values.resize(num_z_bins);
    stddev_values.resize(num_z_bins);

    for (unsigned int bin_index = 0; bin_index < num_z_bins; bin_index++)
      {
      fn_values[bin_index] = function_bins_z[bin_index].get_mean();
      stddev_values[bin_index] = function_bins_z[bin_index].get_pop_stddev();
      }
    }

  // Process the q_yy bins

  q_yy_bin_radii = q_yy_bins.get_bin_centers();

  const unsigned int num_q_yy_bins = q_yy_bin_radii.size();

  q_yy_bin_cell_count.resize(num_q_yy_bins);
  q_yy_bin_accum.resize(num_q_yy_bins);

  for (unsigned int bin_index = 0; bin_index < num_q_yy_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & bin = q_yy_bins[bin_index];

    q_yy_bin_cell_count[bin_index] = bin.get_num_entries();
    q_yy_bin_accum[bin_index] = bin.get_weight_value_sum();
    }

  // Process the q_zz bins

  q_zz_bin_radii = q_zz_bins.get_bin_centers();

  const unsigned int num_q_zz_bins = q_zz_bin_radii.size();

  q_zz_bin_cell_count.resize(num_q_zz_bins);
  q_zz_bin_accum.resize(num_q_zz_bins);

  for (unsigned int bin_index = 0; bin_index < num_q_zz_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & bin = q_zz_bins[bin_index];

    q_zz_bin_cell_count[bin_index] = bin.get_num_entries();
    q_zz_bin_accum[bin_index] = bin.get_weight_value_sum();
    }
  }


void get_scalar_struct_fn(const QuickFlash::Block::BlockInfo & grid_info,
			  const GridVector & vel_data, 
			  const double fn_bin_width, 
			  const double fn_max_radius,
			  const std::vector<unsigned int> & fn_order_list,
			  const unsigned int start_proc_cell_index,
			  const unsigned int num_proc_cells,
			  std::vector<QuickFlash::Stats::BinArray>
			    & function_bins_y_list,
			  std::vector<QuickFlash::Stats::BinArray>
			    & function_bins_z_list,
			  QuickFlash::Stats::BinArray & q_yy_bins,
			  QuickFlash::Stats::BinArray & q_zz_bins)
  {
  if (num_proc_cells > 0)
    {
    const unsigned int num_orders = fn_order_list.size();

    if (num_orders < 1)
      throw QuickFlash::Except("At least one function order required", __FILE__,
			     __LINE__);

    // Check the input and collect parameters

    const unsigned int dims = grid_info.get_space_dims();

    if (dims < 1)
      throw QuickFlash::Except("At least one dimension required", __FILE__,
			     __LINE__);

    if (dims > 2)
      throw QuickFlash::Except("Only one or two dimensions supported", __FILE__,
			     __LINE__);

    if (vel_data.size() != dims)
      throw QuickFlash::Except("Need one velocity component per grid dimension",
			   __FILE__, __LINE__);

    const unsigned int num_cells = grid_info.get_num_cells();

    if (num_cells < 1)
      throw QuickFlash::Except("No cells present", __FILE__, __LINE__);

    const unsigned int end_proc_cell_index 
      = (start_proc_cell_index + num_proc_cells) - 1;

    if (end_proc_cell_index >= num_cells)
      throw QuickFlash::Except("Requested grid indexes out of range", __FILE__,
			     __LINE__);

    const std::vector<unsigned int> & grid_dims = grid_info.get_block_dims();

    const unsigned int num_y_cells = grid_dims[0];

    const unsigned int num_z_cells = (dims > 1) ? grid_dims[1] : 0;

    for (unsigned int set_index = 0; set_index < dims; set_index++)
      {
      if (vel_data[set_index] == 0)
	throw QuickFlash::Except("Non-existent velocity data", __FILE__, 
			       __LINE__);

      if (vel_data[set_index]->get_block_dims() != grid_dims)
	throw QuickFlash::Except("Incompatible grid dimensions", __FILE__, 
			       __LINE__);
      }

    if (!(fn_bin_width > 0.0))
      throw QuickFlash::Except("Bin width must be positive", __FILE__, 
			       __LINE__);

    if (!(fn_max_radius > 0.0))
      throw QuickFlash::Except("Maximum radius must be positive", __FILE__, 
			     __LINE__);

    std::vector<double> grid_width;

    grid_info.get_block_width(grid_width);

    const double grid_y_width = grid_width[0];
    const double grid_z_width = (dims > 1) ? grid_width[1] : 0.0;

    const std::vector<double> & cell_width = grid_info.get_cell_width();

    const double cell_y_width = cell_width[0];
    const double cell_z_width = (dims > 1) ? cell_width[1] : 0.0;

    const double grid_cell_vol = grid_info.get_cell_volume();  // Cell area

    // Determine the number of bins -- use same radius, width for all

    const double fl_min_num_bins = ((fn_max_radius / fn_bin_width) + 0.25);

    const unsigned int num_bins 
      = static_cast<unsigned int>(fl_min_num_bins) + 1;
  
    const double function_max_radius 
      = static_cast<double>(num_bins) * fn_bin_width;

    function_bins_y_list.resize(num_orders);

    if (dims > 1)
      function_bins_z_list.resize(num_orders);
    else
      function_bins_z_list.clear();

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      {
      function_bins_y_list[order_index].reset(0.0, function_max_radius, 
					      num_bins);

      if (dims > 1)
	function_bins_z_list[order_index].reset(0.0, function_max_radius, 
						num_bins);
      }

    // Set up q_yy and q_zz arrays -- center the first bin at zero radius

    const unsigned int num_q_yy_bins = std::max(static_cast<unsigned int>(1), 
						num_y_cells / 2);

    const unsigned int num_q_zz_bins 
      = (dims > 1) ? std::max(static_cast<unsigned int>(1), 
			      num_z_cells / 2) : 0;

    const double q_yy_min_radius = -0.5 * cell_y_width;
    const double q_zz_min_radius = -0.5 * cell_z_width;

    q_yy_bins.reset(q_yy_min_radius, num_q_yy_bins, cell_y_width);

    if (dims > 1)
      q_zz_bins.reset(q_zz_min_radius, num_q_zz_bins, cell_z_width);
    else
      q_zz_bins.reset();

    const double q_yy_max_radius = q_yy_bins.get_max_boundary();

    const double q_zz_max_radius 
      = (dims > 1) ? q_zz_bins.get_max_boundary() : 0.0;

    // Reserve some memory

    std::vector<unsigned int> source_index(dims);

    std::vector<double> position_coords(dims);
    std::vector<double> source_coords(dims);

    std::vector<double> position_vel(dims);
    std::vector<double> source_vel(dims);

    std::vector<double> vel_diff(dims);
    
    // Run through the requested positions

    std::vector<unsigned int> position_index_vector(dims);

    for (unsigned int position_index = start_proc_cell_index; 
	 position_index <= end_proc_cell_index; position_index++)
      {
      if ((position_index % 1000) == 0)
	cerr << "    Pixel " << position_index << endl;

      grid_info.get_cell_center(position_index, position_coords);

      grid_info.get_vector_index(position_index, position_index_vector);

      const unsigned int position_y_index = position_index_vector[0];

      const unsigned int position_z_index 
	= (dims > 1) ? position_index_vector[1] : 0;

      for (unsigned int i = 0; i < dims; i++)
	position_vel[i] = (*vel_data[i])[position_index];

      const double position_y_vel = position_vel[0];
      const double position_z_vel = (dims > 1) ? position_vel[1] : 0.0;

      const double position_y_pos = position_coords[0];
      const double position_z_pos = (dims > 1) ? position_coords[1] : 0.0;

      // NEED TO GENERALIZE FOR ARBITRARY DIMENSIONS????

      // Get the zero-radius q_yy and q_zz data

      q_yy_bins.add_data(0.0, (position_y_vel * position_y_vel), 
			 grid_cell_vol);

      if (dims > 1)
	q_zz_bins.add_data(0.0, (position_z_vel * position_z_vel), 
			   grid_cell_vol);

      // Collect the y-data

      source_index = position_index_vector;  // Reset the index vector

      for (unsigned int y_index = 0; y_index < num_y_cells; y_index++)
	{
	const bool is_myself = (y_index == position_y_index) ? true : false;

	// We are only concerned about displacements in positive-y direction

	source_index[0] = y_index;

	grid_info.get_cell_center(source_index, source_coords);

	const double source_y_pos = source_coords[0];

	const double source_y_vel = (*vel_data[0])[source_index];

	const double vel_y_diff = source_y_vel - position_y_vel;

	const double abs_vel_y_diff = std::fabs(vel_y_diff);

	// Assume that displacements further than sim width are unnecessary

	const double base_y_disp = source_y_pos - position_y_pos;

	// Impose periodic boundary conditions

	const double periodic_y_disp 
	  = (base_y_disp > 0.0) ? base_y_disp : (base_y_disp + grid_y_width);

	// Collect the data

	if (periodic_y_disp < function_max_radius)
	  {
	  for (unsigned int order_index = 0; order_index < num_orders; 
	       order_index++)
	    {
	    const unsigned int fn_order = fn_order_list[order_index];

	    double fn_element = 1.0;

	    for (unsigned int mult_index = 0; mult_index < fn_order; 
		 mult_index++)
	      fn_element *= abs_vel_y_diff;

	    // For right now, just add point data
	      
	    QuickFlash::Stats::BinArray & bin_array 
	      = function_bins_y_list[order_index];

	    bin_array.add_data_spread(periodic_y_disp, fn_element,
				      cell_y_width);
	    }
	  }

	if (!is_myself)
	  if (periodic_y_disp < q_yy_max_radius)
	    {
	    // We have already added the zero-displacement term above 

	    const double vel_corr_elem = position_y_vel * source_y_vel;

	    q_yy_bins.add_data(periodic_y_disp, vel_corr_elem, grid_cell_vol);
	    }
	}

      // Collect the z-data

      if (dims > 1)
	{
	source_index = position_index_vector;  // Reset the index vector

	for (unsigned int z_index = 0; z_index < num_z_cells; z_index++)
	  {
	  const bool is_myself = (z_index == position_z_index) ? true : false;

	  // We are only concerned about displacements in positive-z direction

	  source_index[1] = z_index;

	  grid_info.get_cell_center(source_index, source_coords);

	  const double source_z_pos = source_coords[1];

	  const double source_z_vel = (*vel_data[1])[source_index];

	  const double vel_z_diff = source_z_vel - position_z_vel;

	  const double abs_vel_z_diff = std::fabs(vel_z_diff);

	  // Assume that displacements further than sim width are unnecessary

	  const double base_z_disp = source_z_pos - position_z_pos;

	  // Impose periodic boundary conditions

	  const double periodic_z_disp 
	    = (base_z_disp > 0.0) ? base_z_disp : (base_z_disp + grid_z_width);

	  // Collect the data

	  if (periodic_z_disp < function_max_radius)
	    {
	    for (unsigned int order_index = 0; order_index < num_orders; 
		 order_index++)
	      {
	      const unsigned int fn_order = fn_order_list[order_index];

	      double fn_element = 1.0;

	      for (unsigned int mult_index = 0; mult_index < fn_order; 
		   mult_index++)
	      fn_element *= abs_vel_z_diff;
	      
	      // For right now, just add point data
	      
	      QuickFlash::Stats::BinArray & bin_array 
		= function_bins_z_list[order_index];

	      bin_array.add_data_spread(periodic_z_disp, fn_element,
					cell_z_width);
	      }
	    }

	  if (!is_myself)
	    if (periodic_z_disp < q_zz_max_radius)
	      {
	      // We have already added the zero-displacement term above 

	      const double vel_corr_elem = position_z_vel * source_z_vel;

	      q_zz_bins.add_data(periodic_z_disp, vel_corr_elem, 
				 grid_cell_vol);
	      }
	  }
	}
      }
    }
  }


#ifdef USE_MPI

const int mpi_tag_start_proc_cell_index = 100;
const int mpi_tag_fn_order = 200;
const int mpi_tag_num_proc_cells = 300;
const int mpi_tag_bin_width = 400;
const int mpi_tag_max_radius = 500;
const int mpi_tag_grid_dims = 600;
const int mpi_tag_min_bounds = 700;
const int mpi_tag_max_bounds = 800;
const int mpi_tag_vel_data = 900;
const int mpi_tag_value_weights = 1000;
const int mpi_tag_weights = 1100;


void collect_stats_master(const QuickFlash::Stats::BinArray & bin_array,
			  std::vector<unsigned int> & bin_data_count,
			  std::vector<double> & bin_means,
			  std::vector<double> & bin_stddev,
			  const MPI_Comm mpi_comm)
  {
  // Get MPI-related values

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank != root_rank)
    throw QuickFlash::Except("Server function can only be called by local root",
			   __FILE__, __LINE__);

  // Set up data structures

  const unsigned int num_bins = bin_array.get_num_bins();

  std::vector<unsigned int> accum_bin_counts(num_bins);
  std::vector<double> accum_value_weights(num_bins);
  std::vector<double> accum_sqValue_weights(num_bins);
  std::vector<double> accum_weights(num_bins);

  // Get local data

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & bin = bin_array[bin_index];

    accum_bin_counts[bin_index] = bin.get_num_entries();

    accum_value_weights[bin_index] = bin.get_weight_value_sum();
    accum_sqValue_weights[bin_index] = bin.get_weight_sqValue_sum();

    accum_weights[bin_index] = bin.get_weight_sum();
    }

  // Collect data from other processes

  if (mpi_size > 1)
    {
    std::vector<unsigned int> global_bin_counts(num_bins);
    std::vector<double> global_value_weights(num_bins);
    std::vector<double> global_sqValue_weights(num_bins);
    std::vector<double> global_weights(num_bins);
   
    sum_vect_master(accum_bin_counts, global_bin_counts, mpi_comm);

    sum_vect_master(accum_value_weights, global_value_weights, mpi_comm);
    sum_vect_master(accum_sqValue_weights, global_sqValue_weights, mpi_comm);

    sum_vect_master(accum_weights, global_weights, mpi_comm);

    // Move global data to accum structures using swap

    accum_bin_counts.swap(global_bin_counts);
    accum_value_weights.swap(global_value_weights);
    accum_sqValue_weights.swap(global_sqValue_weights);
    accum_weights.swap(global_weights);
    }

  // Process the data

  bin_data_count = accum_bin_counts;

  bin_means.resize(num_bins);
  bin_stddev.resize(num_bins);

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const double weight = accum_weights[bin_index];

    double mean_val = 0.0;
    double stddev_val = 0.0;

    if (weight > 0.0)
      {
      const double oneOver_weight = 1.0 / weight;

      const double value_weight = accum_value_weights[bin_index];
      const double sqValue_weight = accum_value_weights[bin_index];

      mean_val = value_weight * oneOver_weight;

      // Use sigma^2 = [ SUM(w_i * x_i^2) / SUM(w_i) ] - (N * MEAN(x_i)^2)

      const double sq_stddev 
	= (sqValue_weight * oneOver_weight) - (mean_val * mean_val);

      if (sq_stddev > 0.0)
	stddev_val = std::sqrt(sq_stddev);
      }

    bin_means[bin_index] = mean_val;
    bin_stddev[bin_index] = stddev_val;
    }
  }


void collect_stats_worker(const QuickFlash::Stats::BinArray & bin_array,
			  const MPI_Comm mpi_comm)
  {
  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank == root_rank)
    throw QuickFlash::Except("Worker function must not be called by local root",
			   __FILE__, __LINE__);

  // Get data

  const unsigned int num_bins = bin_array.get_num_bins();

  std::vector<unsigned int> local_bin_counts(num_bins);
  std::vector<double> local_value_weights(num_bins);
  std::vector<double> local_sqValue_weights(num_bins);
  std::vector<double> local_weights(num_bins);

  // Get global data

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & bin = bin_array[bin_index];

    local_bin_counts[bin_index] = bin.get_num_entries();

    local_value_weights[bin_index] = bin.get_weight_value_sum();
    local_sqValue_weights[bin_index] = bin.get_weight_sqValue_sum();

    local_weights[bin_index] = bin.get_weight_sum();
    }

  // Send the data to root

  sum_vect_worker(root_rank, local_bin_counts, mpi_comm);

  sum_vect_worker(root_rank, local_value_weights, mpi_comm);
  sum_vect_worker(root_rank, local_sqValue_weights, mpi_comm);

  sum_vect_worker(root_rank, local_weights, mpi_comm);
  }


void collect_stats_worker_empty(const MPI_Comm mpi_comm)
  {
  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank == root_rank)
    throw QuickFlash::Except("Worker function must not be called by local root",
			   __FILE__, __LINE__);

  // I have no data, so just send a null vector

  const std::vector<unsigned int> null_int_vect(0);
  const std::vector<double> null_fp_vect(0);

  sum_vect_worker(root_rank, null_int_vect, mpi_comm);  // Bin counts

  sum_vect_worker(root_rank, null_fp_vect, mpi_comm);  // Weight--value sums
  sum_vect_worker(root_rank, null_fp_vect, mpi_comm);  // Weight--sqValue sums

  sum_vect_worker(root_rank, null_fp_vect, mpi_comm);  // Weight sums
  }


void collect_accum_master(const QuickFlash::Stats::BinArray & bin_array,
			  std::vector<unsigned int> & bin_data_count,
			  std::vector<double> & bin_accum_values,
			  const MPI_Comm mpi_comm)
  {
  // Get MPI-related values

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank != root_rank)
    throw QuickFlash::Except("Server function can only be called by local root",
			   __FILE__, __LINE__);

  // Set up data structures

  const unsigned int num_bins = bin_array.get_num_bins();

  std::vector<unsigned int> accum_bin_counts(num_bins);
  std::vector<double> accum_weight_value_sums(num_bins);

  // Get local data

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & bin = bin_array[bin_index];

    accum_bin_counts[bin_index] = bin.get_num_entries();

    accum_weight_value_sums[bin_index] = bin.get_weight_value_sum();
    }

  // Collect data from other processes

  if (mpi_size > 1)
    {
    std::vector<unsigned int> global_bin_counts(num_bins);
    std::vector<double> global_weight_value_sums(num_bins);
   
    sum_vect_master(accum_bin_counts, global_bin_counts, mpi_comm);

    sum_vect_master(accum_weight_value_sums, global_weight_value_sums, 
		    mpi_comm);

    // Move global data to accum structures using swap

    accum_bin_counts.swap(global_bin_counts);
    accum_weight_value_sums.swap(global_weight_value_sums);
    }

  // Process the data

  bin_data_count = accum_bin_counts;
  bin_accum_values = accum_weight_value_sums;
  }


void collect_accum_worker(const QuickFlash::Stats::BinArray & bin_array,
			  const MPI_Comm mpi_comm)
  {
  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank == root_rank)
    throw QuickFlash::Except("Worker function must not be called by local root",
			   __FILE__, __LINE__);

  // Get data

  const unsigned int num_bins = bin_array.get_num_bins();

  std::vector<unsigned int> local_bin_counts(num_bins);
  std::vector<double> local_weight_value_sums(num_bins);

  // Get global data

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & bin = bin_array[bin_index];

    local_bin_counts[bin_index] = bin.get_num_entries();
    local_weight_value_sums[bin_index] = bin.get_weight_value_sum();
    }

  // Send the data to root

  sum_vect_worker(root_rank, local_bin_counts, mpi_comm);
  sum_vect_worker(root_rank, local_weight_value_sums, mpi_comm);
  }


void collect_accum_worker_empty(const MPI_Comm mpi_comm)
  {
  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank == root_rank)
    throw QuickFlash::Except("Worker function must not be called by local root",
			   __FILE__, __LINE__);

  // I have no data, so just send a null vector

  const std::vector<unsigned int> null_int_vect(0);
  const std::vector<double> null_fp_vect(0);

  sum_vect_worker(root_rank, null_int_vect, mpi_comm);  // Bin counts
  sum_vect_worker(root_rank, null_fp_vect, mpi_comm);  // Weight--value sums
  }


void get_scalar_struct_fn_master(const QuickFlash::Block::BlockInfo & grid_info,
				 const GridVector & vel_data, 
				 const double fn_bin_width, 
				 const double fn_max_radius, 
				 const std::vector<unsigned int> 
				   & fn_order_list,
				 std::vector<double> & y_bin_positions,
				 std::vector<unsigned int> & y_bins_cell_count,
				 std::vector< std::vector<double> > 
				   & fn_values_y_list,
				 std::vector< std::vector<double> >
				   & fn_stddev_y_list,
				 std::vector<double> & z_bin_positions,
				 std::vector<unsigned int> & z_bins_cell_count,
				 std::vector< std::vector<double> > 
				   & fn_values_z_list,
				 std::vector< std::vector<double> >
				   & fn_stddev_z_list,
				 std::vector<double> & q_yy_bin_radii,
				 std::vector<unsigned int> 
				   & q_yy_bin_cell_count,
				 std::vector<double> & q_yy_bin_accum,
				 std::vector<double> & q_zz_bin_radii,
				 std::vector<unsigned int> 
				   & q_zz_bin_cell_count,
				 std::vector<double> & q_zz_bin_accum,
				 const MPI_Comm mpi_comm)
  {
  const unsigned int dims = grid_info.get_space_dims();

  const std::vector<unsigned int> & grid_dims = grid_info.get_block_dims();

  for (unsigned int i = 0; i < dims; i++)
    {
    if (vel_data[i] == 0)
      throw QuickFlash::Except("Non-existant velocity data", __FILE__, __LINE__);

    if (vel_data[i]->get_block_dims() != grid_dims)
      throw QuickFlash::Except("Incompatible velocity data dimensions", __FILE__,
			     __LINE__);
    }

  const unsigned int num_orders = fn_order_list.size();

  if (num_orders < 1)
    throw QuickFlash::Except("At least one order must be specified", __FILE__,
			   __LINE__);

  // Get MPI-related values

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank != root_rank)
    throw QuickFlash::Except("Server function can only be called by local root",
			   __FILE__, __LINE__);

  const unsigned int mpi_size_unsigned = static_cast<unsigned int>(mpi_size);

  // Get some basic data about the space

  const std::vector<double> & min_bounds = grid_info.get_min_coords();
  const std::vector<double> & max_bounds = grid_info.get_max_coords();

  const unsigned int num_cells = grid_info.get_num_cells();

  // Determine the distribution of cells to be processed

  std::vector<unsigned int> start_proc_cell_indexes;
  std::vector<unsigned int> num_proc_cells;

  simple_linear_decomp(num_cells, mpi_size, start_proc_cell_indexes, 
		       num_proc_cells);

  const unsigned int my_start_cell_index = start_proc_cell_indexes[0];
  const unsigned int my_num_cells = num_proc_cells[0];

  // Send the data and layout to everyone else

  for (unsigned int proc_id = 1; proc_id < mpi_size_unsigned; proc_id++)
    {
    const unsigned int start_cell_index_proc
      = start_proc_cell_indexes[proc_id];

    const unsigned int num_cells_proc
      = num_proc_cells[proc_id];

    int start_cell_index_proc_mpi = static_cast<int>(start_cell_index_proc);
    int num_cells_proc_mpi = static_cast<int>(num_cells_proc);

    MPI_Send(&start_cell_index_proc_mpi, 1, MPI_UNSIGNED, proc_id, 
	     mpi_tag_start_proc_cell_index, mpi_comm);

    MPI_Send(&num_cells_proc_mpi, 1, MPI_UNSIGNED, proc_id, 
	     mpi_tag_num_proc_cells, mpi_comm);
    }

  // Send the global data

  double bin_width_mpi = fn_bin_width;
  double max_radius_mpi = fn_max_radius;

  MPI_Bcast(&bin_width_mpi, 1, MPI_DOUBLE, mpi_rank, mpi_comm);

  MPI_Bcast(&max_radius_mpi, 1, MPI_DOUBLE, mpi_rank, mpi_comm);

  bcast_vect_server(fn_order_list, mpi_comm);

  bcast_vect_server(grid_dims, mpi_comm);

  bcast_vect_server(min_bounds, mpi_comm);
  bcast_vect_server(max_bounds, mpi_comm);

  for (unsigned int i = 0; i < dims; i++)
    bcast_vect_server(vel_data[i]->get_data(), mpi_comm);

  // Do my work

  std::vector<QuickFlash::Stats::BinArray> function_bins_y_list;
  std::vector<QuickFlash::Stats::BinArray> function_bins_z_list;

  QuickFlash::Stats::BinArray q_yy_bins;
  QuickFlash::Stats::BinArray q_zz_bins;

  get_scalar_struct_fn(grid_info, vel_data, fn_bin_width, fn_max_radius, 
		       fn_order_list, my_start_cell_index, my_num_cells, 
		       function_bins_y_list, function_bins_z_list,
		       q_yy_bins, q_zz_bins);

  // Collect the statistics

  y_bin_positions = function_bins_y_list[0].get_bin_centers();
  
  fn_values_y_list.resize(num_orders);
  fn_stddev_y_list.resize(num_orders);

  for (unsigned int order_index = 0; order_index < num_orders; order_index++)
    collect_stats_master(function_bins_y_list[order_index], y_bins_cell_count,
			 fn_values_y_list[order_index], 
			 fn_stddev_y_list[order_index], mpi_comm);

  z_bin_positions = function_bins_z_list[0].get_bin_centers();
  
  fn_values_z_list.resize(num_orders);
  fn_stddev_z_list.resize(num_orders);

  for (unsigned int order_index = 0; order_index < num_orders; order_index++)
    collect_stats_master(function_bins_z_list[order_index], z_bins_cell_count,
			 fn_values_z_list[order_index], 
			 fn_stddev_z_list[order_index], mpi_comm);

  q_yy_bin_radii = q_yy_bins.get_bin_centers();

  collect_accum_master(q_yy_bins, q_yy_bin_cell_count, q_yy_bin_accum, 
		       mpi_comm);

  q_zz_bin_radii = q_zz_bins.get_bin_centers();

  collect_accum_master(q_zz_bins, q_zz_bin_cell_count, q_zz_bin_accum, 
		       mpi_comm);
  }
  

void get_scalar_struct_fn_worker(const MPI_Comm mpi_comm)
  {
  MPI_Status status;

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank == root_rank)
    throw QuickFlash::Except("Worker function must not be called by local root",
			   __FILE__, __LINE__);

  unsigned int start_proc_cell_index = 0;
  unsigned int num_proc_cells = 0;

  MPI_Recv(&start_proc_cell_index, 1, MPI_UNSIGNED, root_rank, 
	   mpi_tag_start_proc_cell_index, mpi_comm, &status);
  
  MPI_Recv(&num_proc_cells, 1, MPI_UNSIGNED, root_rank, 
	   mpi_tag_num_proc_cells, mpi_comm, &status);


  // Get information about the binning

  double bin_width = 0.0;
  double max_radius = 0.0;

  MPI_Bcast(&bin_width, 1, MPI_DOUBLE, root_rank, mpi_comm);

  MPI_Bcast(&max_radius, 1, MPI_DOUBLE, root_rank, mpi_comm);


  // Get the function orders

  std::vector<unsigned int> fn_order_list;

  bcast_vect_client(root_rank, fn_order_list, mpi_comm);


  // Get the grid information

  std::vector<unsigned int> grid_dims;

  bcast_vect_client(root_rank, grid_dims, mpi_comm);

  std::vector<double> min_bounds;
  std::vector<double> max_bounds;

  bcast_vect_client(root_rank, min_bounds, mpi_comm);
  bcast_vect_client(root_rank, max_bounds, mpi_comm);

  QuickFlash::Block::BlockInfo gridinfo(min_bounds, max_bounds, grid_dims,
				      QuickFlash::Geometry::Cartesian);

  const unsigned int dims = gridinfo.get_space_dims();

  const unsigned int num_cells = gridinfo.get_num_cells();

  // Set up the velocity data

  std::vector< QuickFlash::Block::BlockData<double> > vel_data(dims);

  std::vector<double> axis_vel_data_array;

  for (unsigned int i = 0; i < dims; i++)
    {
    QuickFlash::Block::BlockData<double> & axis_vel_data = vel_data[i];

    axis_vel_data.reset(grid_dims);

    bcast_vect_client(root_rank, axis_vel_data_array, mpi_comm);

    if (axis_vel_data_array.size() != num_cells)
      throw QuickFlash::Except("Incompatible velocity data array", __FILE__,
			     __LINE__);

    axis_vel_data.set_data(axis_vel_data_array);
    }


  // Get the local statistics

  if (num_cells > 0)
    {
    GridVector vel_data_ptrs(dims);

    for (unsigned int i = 0; i < dims; i++)
      vel_data_ptrs[i] = &vel_data[i];

    std::vector<QuickFlash::Stats::BinArray> function_bins_y_list;
    std::vector<QuickFlash::Stats::BinArray> function_bins_z_list;

    QuickFlash::Stats::BinArray q_yy_bins;
    QuickFlash::Stats::BinArray q_zz_bins;
  
    get_scalar_struct_fn(gridinfo, vel_data_ptrs, bin_width, max_radius,
			 fn_order_list, start_proc_cell_index, num_proc_cells, 
			 function_bins_y_list, function_bins_z_list,
			 q_yy_bins, q_zz_bins);

    // Get the weights and return to root; assume root has correct bin centers

    const unsigned int num_orders = fn_order_list.size();

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      collect_stats_worker(function_bins_y_list[order_index], mpi_comm);

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      collect_stats_worker(function_bins_z_list[order_index], mpi_comm);

    collect_accum_worker(q_yy_bins, mpi_comm);
    collect_accum_worker(q_zz_bins, mpi_comm);
    }
  else
    {
    // I have no data, so just send zeros

    const unsigned int num_orders = fn_order_list.size();

    // y-list

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      collect_stats_worker_empty(mpi_comm);

    // z-list

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      collect_stats_worker_empty(mpi_comm);

    collect_accum_worker_empty(mpi_comm);  // Q_yy
    collect_accum_worker_empty(mpi_comm);  // Q_zz
    }
  }


#endif  // USE_MPI
