// C++ program file flash_resample.cpp

/*
  By Nathan C. Hearn
     May 21, 2007

  Resampling routines for Flash.
*/


#include "flash_resample.hpp"
#include "quickflash.hpp"
#include <vector>
#include <string>
#include <iostream>


// Some constants

const unsigned int buffer_size = 100;
const unsigned int cache_size = 100;  // Must be greater than zero!!!!


// Helpful typedefs

typedef QuickFlash::Cache::IndexCache< QuickFlash::Block::BlockData<double> > 
  BlockCache;


// Declare persistant objects

QuickFlash::File::DataFile df;

std::vector<const QuickFlash::File::Dataset *> dataset_list;

std::vector<BlockCache *> datacache_list;

const QuickFlash::File::MeshInfo * mesh_info = 0;

unsigned int num_blocks = 0;

unsigned int dims = 0;

std::vector<double> min_bounds;
std::vector<double> max_bounds;


// Function definitions

int FTOC(quickflash_init_resample)(const char file_name[])
  {
  using std::cerr;
  using std::endl;

  int ret_val = -1;  // Error state

  if (FTOC(quickflash_clean_up)() == 0)
    {
    const std::string filename = file_name;

    bool ok = true;

    try
      {
      df.reset(filename, buffer_size);
      }
    catch (QuickFlash::Except ex)
      {
      cerr << "Exception caught in quickflash_init_resample" << endl;
      ok = false;
      }

    if (ok)
      {
      const QuickFlash::File::MeshInfo & local_meshinfo = df.get_mesh_info();

      num_blocks = local_meshinfo.get_num_blocks();

      dims = local_meshinfo.get_dims();

      min_bounds = local_meshinfo.get_volume_minbounds();
      max_bounds = local_meshinfo.get_volume_maxbounds();

      mesh_info = &local_meshinfo;

      ret_val = 0;  // OK
      }
    }

  return ret_val;
  }


int FTOC(quickflash_add_dataset)(const char dataset_name[])
  {
  using std::cerr;
  using std::endl;

  int ret_val = -1;  // Error state

  const std::string setname = dataset_name;

  unsigned int set_index = 0;

  bool ok = true;

  try
    {
    const QuickFlash::File::Dataset & set = df.get_dataset(setname);

    set_index = dataset_list.size();

    dataset_list.push_back(&set);

    BlockCache * set_cache = new BlockCache(num_blocks, cache_size);

    datacache_list.push_back(set_cache);
    }
  catch (QuickFlash::Except ex)
    {
    cerr << "Exception caught in quickflash_add_dataset" << endl;
    ok = false;
    }

  if (ok)
    ret_val = static_cast<int>(set_index);

  return ret_val;
  }


bool FTOC(quickflash_in_bounds)(const double & x, const double & y,
			      const double & z)
  {
  bool in_bounds = false;

  if (mesh_info != 0)
    {
    // See if any of the positions are out of bounds

    in_bounds = true;

    if (dims > 0)
      if ((x < min_bounds[0]) || (x > max_bounds[0]))
	in_bounds = false;

    if (dims > 1)
      if ((y < min_bounds[1]) || (y > max_bounds[1]))
	in_bounds = false;

    if (dims > 2)
      if ((z < min_bounds[2]) || (z > max_bounds[2]))
	in_bounds = false;
    }

  return in_bounds;
  }


int FTOC(quickflash_dims)()
  {
  int ret_val = -1;

  // Only return dimensions if a valid mesh_info object exists

  if (mesh_info != 0)
    ret_val = static_cast<int>(dims);

  return ret_val;
  }


int FTOC(quickflash_min_bounds)(double & x, double & y, double & z)
  {
  int ret_val = -1;

  // Only return boundary if a valid mesh_info object exists

  if (mesh_info != 0)
    {
    if (dims > 0)
      x = min_bounds[0];
    else
      x = 0.0;

    if (dims > 1)
      y = min_bounds[1];
    else
      y = 0.0;

    if (dims > 2)
      z = min_bounds[2];
    else
      z = 0.0;

    ret_val = 0;  // OK
    }

  return ret_val;
  }


int FTOC(quickflash_max_bounds)(double & x, double & y, double & z)
  {
  int ret_val = -1;

  // Only return boundary if a valid mesh_info object exists

  if (mesh_info != 0)
    {
    if (dims > 0)
      x = max_bounds[0];
    else
      x = 0.0;

    if (dims > 1)
      y = max_bounds[1];
    else
      y = 0.0;

    if (dims > 2)
      z = max_bounds[2];
    else
      z = 0.0;

    ret_val = 0;  // OK
    }

  return ret_val;
  }


int FTOC(quickflash_get_block_cell)(const double & x, const double & y, 
				  const double & z, int & block_index,
				  int & cell_index)
  {
  using std::cerr;
  using std::endl;

  int ret_val = -1;  // Error state

  if (mesh_info != 0)
    {
    const QuickFlash::File::MeshInfo & local_meshinfo = *mesh_info;

    std::vector<double> pos(dims, 0.0);

    if (dims > 0)
      pos[0] = x;

    if (dims > 1)
      pos[1] = y;

    if (dims > 2)
      pos[2] = z;

    // See if position is in-bounds

    bool in_bounds = true;

    for (unsigned int i = 0; i < dims; i++)
      if ((pos[i] < min_bounds[i]) || (pos[i] > max_bounds[i]))
	in_bounds = false;

    if (!in_bounds)
      {
      ret_val = 1;  // Out of bounds

      // CAN WE STILL GET VALID BLOCK INDEX AND CELL INDEX FOR NEAREST CELL???

      block_index = -1;
      cell_index = -1;
      }
    else
      {
      unsigned int block = 0;
      unsigned int cell = 0;

      bool ok = true;

      try
	{
	local_meshinfo.get_cell_index(pos, block, cell);
	}
      catch (QuickFlash::Except ex)
	{
	cerr << "Exception caught in quickflash_get_block_cell" << endl;

	ok = false;
	}

      if (ok)
	{
	block_index = static_cast<int>(block);
	cell_index = static_cast<int>(cell);

	ret_val = 0;  // In bounds
	}
      }
    }

  return ret_val;
  }


int FTOC(quickflash_get_value)(const int dataset_index, 
			       const int & block_index,
			       const int & cell_index, double & value)
  {
  int ret_val = -1;  // Error state  -- EXCEPTION HANDLING?????

  if (dataset_index >= 0)
    {
    const unsigned int setindex = static_cast<unsigned int>(dataset_index);

    if (setindex < dataset_list.size())
      {
      if ((block_index > 0) && (cell_index > 0))
	{
	const unsigned int block = static_cast<unsigned int>(block_index);
	const unsigned int cell = static_cast<unsigned int>(cell_index);

	if (block < num_blocks)
	  {
	  // Make sure the block is in the cache

	  if (!(datacache_list[setindex]->data_present(block)))
	    {
	    // Get the block and place it in the cache

	    QuickFlash::Block::BlockData<double> blockdata;

	    dataset_list[setindex]->get_block_data(block, blockdata);

	    datacache_list[setindex]->add_data(block, blockdata);
	    }

	  // Copy the data

	  const QuickFlash::Block::BlockData<double> & blockdata
	    = (*(datacache_list[setindex]))[block];

	  const unsigned int num_cells = blockdata.get_num_cells();

	  if (cell < num_cells)
	    {
	    value = blockdata[cell];

	    ret_val = 0;  // OK
	    }
	  }
	}
      }
    }

  return ret_val;
  }
      

int FTOC(quickflash_clean_up)()
  {
  int ret_val = -1;  // Error state -- EXCEPTION HANDLING???

  // See if anything exists

  if (mesh_info != 0)
    ret_val = 0;  // OK ... for now

  // Clean out everything!!!

  const unsigned int num_sets = dataset_list.size();

  for (unsigned int index = 0; index < num_sets; index++)
    {
    dataset_list[index] = 0;  // Set pointer to null; I don't own the object

    delete datacache_list[index];
    datacache_list[index] = 0;
    }

  dataset_list.clear();
  datacache_list.clear();

  mesh_info = 0;  // I don't own this

  num_blocks = 0;

  dims = 0;

  min_bounds.clear();
  max_bounds.clear();

  df.reset();

  return ret_val;
  }
