// C++ header file mixing_layer.hpp

/*
  By Nathan C. Hearn
     May 16, 2007

  Classes for storing mixing layer information for each timestep.
*/


#ifndef MIXING_LAYER_HPP
#define MIXING_LAYER_HPP


#include <vector>
#include "quickflash_file_datafile.hpp"
#include "quickflash_block_blockdata.hpp"
#include "stat_bins.hpp"


class MixingLayer
  {
  public :

    typedef QuickFlash::Block::BlockData<double> DoubleBlock;
    typedef QuickFlash::Block::BlockData<unsigned int> UIntBlock;


  public:

    MixingLayer();

    MixingLayer(const QuickFlash::File::DataFile & datafile, 
		const std::vector<double> & min_sample_pos,
		const std::vector<double> & max_sample_pos,
		const std::vector<double> & sample_spacing,
		const unsigned int buffer_size=0);

    MixingLayer(const MixingLayer & source);

    ~MixingLayer() { }

    MixingLayer & operator=(const MixingLayer & source)
      {
      reset(source);
      return *this;
      }

    bool operator<(const double ref_time)
      { return (sim_time < ref_time); }

    bool operator<(const MixingLayer & source)
      { return (sim_time < source.sim_time); }

    double get_sim_time() const { return sim_time; }

    const std::vector<double> & get_min_sample_coord() const 
      { return min_coord; }

    const std::vector<double> & get_max_sample_coord() const 
      { return max_coord; }

    const std::vector<double> & get_sample_spacing() const 
      { return bin_spacing; }

    // double get_int_pos() const { return int_pos; }

    void reset();
      
    void reset(const QuickFlash::File::DataFile & datafile, 
	       const std::vector<double> & min_sample_coord, 
	       const std::vector<double> & max_sample_coord, 
	       const std::vector<double> & sample_spacing,
	       const unsigned int buffer_size=0);

    void reset(const MixingLayer & source);

    const std::vector<double> & get_bin_array_centers() const
      { return dens_array.get_bin_centers(); }

    void get_bin_masses(std::vector<double> & bin_masses) const
      { dens_array.get_sum_weight_values(bin_masses); }

    void get_bin_poly_masses(std::vector<double> & bin_masses) const
      { poly_array.get_sum_weight_values(bin_masses); }

    void get_vol_avg_dens(std::vector<double> & bin_dens) const
      { dens_array.get_means(bin_dens); }

    void get_vol_avg_poly_dens(std::vector<double> & bin_dens) const
      { poly_dens_array.get_means(bin_dens); }

    void get_vol_avg_poly(std::vector<double> & bin_poly) const
      { poly_array.get_means(bin_poly); }
    
    void get_bin_entries(std::vector<unsigned int> & bin_entries) const
      { dens_array.get_bin_entries(bin_entries); }

    void get_vol_avg_dens(DoubleBlock & bin_dens) const
      { dens_grid.get_means(bin_dens); }

    void get_vol_avg_poly_dens(DoubleBlock & bin_dens) 
      const
      { poly_dens_grid.get_means(bin_dens); }

    void get_bin_entries(UIntBlock & bin_entries) const
      { dens_grid.get_bin_entries(bin_entries); }

    void get_lateral_spectral_power_dens(const unsigned int bin_index,
					 std::vector<double> & power) const;

    void get_lateral_spectral_power_dens(DoubleBlock & power) const;

    void get_lateral_spectral_power_poly_dens(const unsigned int bin_index,
					      std::vector<double> & power) 
      const;

    void get_lateral_spectral_power_poly_dens(DoubleBlock & power) 
      const;

    void get_secondary_power_ratio_dens(std::vector<double> & bin_power_ratio,
				    std::vector<unsigned int> & secondary_freq)
      const;

    void get_secondary_power_ratio_poly_dens(std::vector<double> 
					     & bin_power_ratio,
				    std::vector<unsigned int> & secondary_freq)
      const;


    void get_abs_atwood_number(DoubleBlock & atwood_numbers) const;

    void get_lateral_peak_atwood_number(std::vector<double> & atwood_numbers)
      const;

  private :

    void load_data(const QuickFlash::File::DataFile & datafile, 
		   const unsigned int buffer_size=0);


  private :

    static const char dens_name[];
    static const char foam_name[];

    static const unsigned int default_buffer_size;

    unsigned int dims;

    std::vector<unsigned int> grid_dims;

    double sim_time;

    std::vector<double> min_coord;
    std::vector<double> max_coord;

    std::vector<double> bin_spacing;

    // double int_pos;

    Stats::BinArray dens_array;
    Stats::BinArray poly_dens_array;
    Stats::BinArray poly_array;

    Stats::BinGrid dens_grid;
    Stats::BinGrid poly_dens_grid;
  };


#endif  // MIXING_LAYER_HPP
