// C++ header file stat_bins.hpp

/*
  By Nathan C. Hearn
     September 20, 2006

  Statistical binning classes.
*/


#ifndef STAT_BINS_HPP
#define STAT_BINS_HPP


#include <math.h>
#include <vector>
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"


namespace Stats
{

class Bin
  {
  public:
    Bin() :
      num_entries(0), sum_weight(0.0), sum_weight_values(0.0),
      sum_weight_sqValues(0.0), min_value(0.0), max_value(0.0)
      { }
    Bin(const Bin & source) :
      num_entries(0), sum_weight(0.0), sum_weight_values(0.0),
      sum_weight_sqValues(0.0), min_value(0.0), max_value(0.0)
      { reset(source); }
    ~Bin() { }

    Bin & operator=(const Bin & source)
      {
      reset(source);
      return *this;
      }

    Bin & operator+=(const Bin & source)
      {
      add_bin(source);
      return *this;
      }

    void reset();

    void reset(const Bin & source);

    void add_bin(const Bin & source);

    void add_entry(const double value, const double weight=1.0);      

    unsigned int get_num_entries() const { return num_entries; }

    bool entry_present() const
      { 
      bool present = false;
      
      if (num_entries > 0)
	present = true;

      return present;
      }

    double get_sum_weights() const { return sum_weight; }

    double get_sum_weight_values() const { return sum_weight_values; }

    double get_mean() const 
      { 
      double mean_val = 0.0;

      if (sum_weight > 0.0)
	mean_val = sum_weight_values / sum_weight;

      return mean_val;
      }

    double get_pop_sqStddev() const;
    double get_pop_stddev() const { return sqrt(get_pop_sqStddev()); }

    double get_min_value() const { return min_value; }
    double get_max_value() const { return max_value; }

  private:
    unsigned int num_entries;

    double sum_weight;

    double sum_weight_values;
    double sum_weight_sqValues;

    double min_value;
    double max_value;
  };


class BinArray
  {
  public:
    BinArray() :
      num_bins(0), bin_min_bounds(), bin_max_bounds(), bin_center_coords(),
      bins()
      { }
    BinArray(const unsigned int num_bins, const double min_boundary, 
	     const double max_boundary) :
      num_bins(0), bin_min_bounds(), bin_max_bounds(), bin_center_coords(),
      bins()
      { reset(num_bins, min_boundary, max_boundary); }
    BinArray(const std::vector<double> bin_min_boundaries, 
	     const std::vector<double> bin_max_boundaries) :
      num_bins(0), bin_min_bounds(), bin_max_bounds(), bin_center_coords(),
      bins()
      { reset(bin_min_boundaries, bin_max_boundaries); }
    BinArray(const BinArray & source) :
      num_bins(0), bin_min_bounds(), bin_max_bounds(), bin_center_coords(),
      bins()
      { reset(source); }
    ~BinArray() { }

    BinArray & operator=(const BinArray & source)
      {
      reset(source);
      return *this;
      }

    const Bin & operator[](const unsigned int bin_index) const
      { return bins[bin_index]; }

    void reset();

    void reset(const unsigned int num_bins, const double min_boundary, 
	       const double max_boundary);

    void reset(const std::vector<double> bin_min_boundaries, 
	       const std::vector<double> bin_max_boundaries);

    void reset(const BinArray & source);

    void add_data(const double coordinate, const double value,
		  const double weight=1.0);
    void add_data_spread(const double coordinate, const double value,
			 const double spread, const double weight=1.0);

    unsigned int get_num_bins() const { return num_bins; }

    bool in_bin(const unsigned int bin_index, const double coordinate, 
		const double spread=0.0) const;

    double get_overlap_fract(const unsigned int bin_index, 
			     const double coordinate,
			     const double spread=0.0) const;

    const Bin & get_bin(const unsigned int bin_index) const 
      { return bins[bin_index]; }

    double get_bin_center(const unsigned int bin_index) const
      { return bin_center_coords[bin_index]; }
    double get_bin_width(const unsigned int bin_index) const
      { return bin_widths[bin_index]; }

    const std::vector<double> & get_bin_centers() const
      { return bin_center_coords; }

    void get_means(std::vector<double> & bin_data) const;
    void get_sum_weights(std::vector<double> & bin_data) const;
    void get_sum_weight_values(std::vector<double> & bin_data) const;

    void get_bin_entries(std::vector<unsigned int> & bin_data) const;

    double get_bin_min_bounds(const unsigned int bin_index) const
      { return bin_min_bounds[bin_index]; }
    double get_bin_max_bounds(const unsigned int bin_index) const
      { return bin_max_bounds[bin_index]; }
				

  private:
    unsigned int num_bins;

    std::vector<double> bin_min_bounds;
    std::vector<double> bin_max_bounds;

    std::vector<double> bin_center_coords;
    std::vector<double> bin_widths;

    std::vector<Bin> bins;
  };


class BinGrid
  {
  public :

    BinGrid();

    BinGrid(const std::vector<unsigned int> & dims, 
	    const std::vector<double> & min_coords,
	    const std::vector<double> & max_coords);

    BinGrid(const BinGrid & source);

    ~BinGrid() { }

    BinGrid & operator=(const BinGrid & source)
      {
      reset(source);
      return *this;
      }

    const Bin & operator[](const unsigned int bin_index) const
      { return grid_data[bin_index]; }

    const Bin & operator[](const std::vector<unsigned int> & bin_index) const
      { return grid_data[bin_index]; }

    void reset();

    void reset(const std::vector<unsigned int> & dims, 
	       const std::vector<double> & min_coords,
	       const std::vector<double> & max_coords);

    void reset(const BinGrid & source);

    unsigned int get_space_dims() const { return space_dims; }

    unsigned int get_num_bins() const { return num_bins; }

    const std::vector<unsigned int> & get_grid_dims() const 
      { return grid_info.get_block_dims(); }

    const QuickFlash::Block::BlockInfo & get_grid_info() const 
      { return grid_info; }

    void add_data(const std::vector<double> & coords, const double value,
		  const double weight=1.0);

    void add_data_spread(const std::vector<double> & coords,
			 const double value,
			 const std::vector<double> &  spread, 
			 const double weight=1.0);

    void get_means(QuickFlash::Block::BlockData<double> & bin_data) const;

    void get_sum_weights(QuickFlash::Block::BlockData<double> & bin_data) const;

    void get_sum_weight_values(QuickFlash::Block::BlockData<double> & bin_data) 
      const;

    void get_bin_entries(QuickFlash::Block::BlockData<unsigned int> & bin_data) 
      const;

  private :

    double get_overlap_fract(const std::vector<unsigned int> & bin_index, 
			     const std::vector<double> & coords,
			     const std::vector<double> & spread) const;


  private :

    unsigned int space_dims;

    unsigned int num_bins;

    QuickFlash::Block::BlockInfo grid_info;
    QuickFlash::Block::BlockData<Bin> grid_data;
  };


}  // End namespace Stats


#endif  // STAT_BINS_HPP
