// C++ program file line_domain.cpp

/*
  By Nathan C. Hearn
     January 26, 2007

  Class for defining rectangular lineout region parallel to x-axis.
*/


#include "line_domain.hpp"
#include <vector>
#include "quickflash_except.hpp"


void LineDomain::reset(const double min_x, const double max_x,
		       const std::vector<double> & center_coords,
		       const double domain_radius)
  {
  dims = center_coords.size();

  if (dims < 2)
    throw QuickFlash::Except("Insufficient number of dimensions", __FILE__, 
			__LINE__);

  if (!(min_x < max_x))
    throw QuickFlash::Except("Incompatible min/max x-values", __FILE__, __LINE__);

  start_coords = center_coords;
  start_coords[0] = min_x;

  end_coords = center_coords;
  end_coords[0] = max_x;

  radius = domain_radius;

  if (!(radius > 0.0))
    throw QuickFlash::Except("Line width must be a positive value", __FILE__,
			__LINE__);
  }


bool LineDomain::in_domain(const std::vector<double> & point) const
  {
  bool in_bounds = false;

  if (point.size() != dims)
    throw QuickFlash::Except("Point has incompatible dimensions", __FILE__,
			__LINE__);

  // For right now, the x-axis is special, and cylinder is really a rectangle

  const double min_x = start_coords[0];
  const double max_x = end_coords[0];

  const double point_x = point[0];

  if ((!(point_x < min_x)) && (point_x < max_x))
    {
    in_bounds = true;  // OK so far ...

    unsigned int i = 1;

    while (in_bounds && (i < dims))
      {
      const double min_coord = start_coords[i] - radius;
      const double max_coord = start_coords[i] + radius;

      const double coord = point[i];

      if ((coord < min_coord) || (coord > max_coord))
	in_bounds = false;
      else
	i++;
      }
    }

  return in_bounds;
  }

bool LineDomain::in_domain(const std::vector<double> & point, 
			   const std::vector<double> & width) const
  {
  if ((point.size() != dims) || (width.size() != dims))
    throw QuickFlash::Except("Incompatible point/width vectors", __FILE__, 
			__LINE__);

  std::vector<double> min_coords = point;
  std::vector<double> max_coords = point;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double half_width = width[i] * 0.5;

    min_coords[i] -= half_width;
    max_coords[i] += half_width;
    }

  return in_domain_box(min_coords, max_coords);
  }


bool LineDomain::in_domain_box(const std::vector<double> & min_coords,
			       const std::vector<double> & max_coords) const
  {
  bool in_bounds = false;

  // For right now, the x-axis is special, and cylinder is really a rectangle

  const double min_x = start_coords[0];
  const double max_x = end_coords[0];

  const double min_coords_x = min_coords[0];
  const double max_coords_x = max_coords[0];

  if ((min_coords_x < max_x) && (max_coords_x > min_x))
    {
    in_bounds = true;  // OK so far...

    unsigned int i = 1;

    while (in_bounds && (i < dims))
      {
      const double min_bound = start_coords[i] - radius;
      const double max_bound = start_coords[i] + radius;

      const double min_coords_axis = min_coords[i];
      const double max_coords_axis = max_coords[i];

      if ((min_coords_axis > max_bound) || (max_coords_axis < min_bound))
	in_bounds = false;
      else
	i++;
      }
    }

  return in_bounds;
  }
