// C++ program file lineout.cpp

/*
  By Nathan C. Hearn
     September 21, 2006

  Program for gathering information for flash files along strip in x-direction.
*/


#include <iostream>
#include <vector>
#include <string>
#include <list>
#include "quickflash_except.hpp"
#include "line_domain.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_stats_bins.hpp"


const std::string poly_name = "poly";
const std::string dens_name = "dens";

const unsigned int buffer_size = 200;


void build_filename(const std::string filename_base, 
		    const unsigned int num_digits, const unsigned int index,
                    std::string & filename)
  {
  std::stringstream name_stream;

  name_stream << filename_base;

  name_stream.width(num_digits);
  name_stream.fill('0');

  name_stream << index;

  filename = name_stream.str();
  }



int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  if (argc < 2)
    {
    cerr << "Usage: " << argv[0] << " min_x max_x num_bins y y_width "
	 << "filename_base num_index_digits start_index end_index" 
	 << endl;
    return 1;
    }

  int argPtr = 1;

  const double min_x = atof(argv[argPtr++]);
  const double max_x = atof(argv[argPtr++]);

  if (!(min_x < max_x))
    throw QuickFlash::Except("Incompatible min/max x values", __FILE__, 
			     __LINE__);

  const unsigned int num_bins
    = static_cast<unsigned int>(atoi(argv[argPtr++]));

  if (num_bins < 1)
    throw QuickFlash::Except("Must have at least one bin", __FILE__, __LINE__);

  const double y = atof(argv[argPtr++]);

  const double y_width = atof(argv[argPtr++]);

  const double radius = y_width * 0.5;

  if (!(radius > 0.0))
    throw QuickFlash::Except("Width must be positive", __FILE__, __LINE__);

  const std::string filename_base = argv[argPtr++];

  const unsigned int index_digits 
    = static_cast<unsigned int>(atoi(argv[argPtr++]));

  const unsigned int start_index 
    = static_cast<unsigned int>(atoi(argv[argPtr++]));

  const unsigned int end_index 
    = static_cast<unsigned int>(atoi(argv[argPtr++]));

  if (end_index < start_index)
    throw QuickFlash::Except("Incompatible start/end indexes", __FILE__, 
			     __LINE__);

  // Set up the domain

  std::vector<double> start_center(2);

  start_center[0] = min_x;
  start_center[1] = y;

  const LineDomain domain(min_x, max_x, start_center, radius);

  // Build the bins

  QuickFlash::Stats::BinArray poly_bins(min_x, max_x, num_bins);
  QuickFlash::Stats::BinArray dens_bins(min_x, max_x, num_bins);

  // Run through the files

  for (unsigned int index = start_index; index <= end_index; index++)
    {
    poly_bins.reset_bins();
    dens_bins.reset_bins();

    std::string filename;

    build_filename(filename_base, index_digits, index, filename);

    cerr << "Opening file [ " << filename << " ]" << endl;

    const QuickFlash::File::DataFile df(filename, buffer_size);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const unsigned int dims = meshinfo.get_dims();

    if (dims != 2)
      throw QuickFlash::Except("Incompatible dimensions", __FILE__, __LINE__);

    const double time = meshinfo.get_sim_time();

    // Output the header

    cout << "# File " << filename << endl;
    cout << "# Time " << time << " s" << endl;

    cout << "bin_index" << "\t" << "bin_center" << "\t" << "num_entries" 
	 << "\t" << "poly_mass" << "\t" << "mean_density" << endl;

    // Load the data sets

    const QuickFlash::File::Dataset & poly = df.get_dataset(poly_name);
    const QuickFlash::File::Dataset & dens = df.get_dataset(dens_name);

    // Run through the blocks

    const unsigned int num_blocks = meshinfo.get_num_blocks();

    std::vector<double> block_center;
    std::vector<double> block_width;

    QuickFlash::Block::BlockData<double> poly_block;
    QuickFlash::Block::BlockData<double> dens_block;

    for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
      {
      if (meshinfo.is_leaf(block_index))
	{
	meshinfo.get_block_center(block_index, block_center);
	meshinfo.get_block_width(block_index, block_width);

	if (domain.in_domain(block_center, block_width))
	  {
	  // Load the block information

	  const QuickFlash::Block::BlockInfo & block_info
	    = meshinfo.get_block_info(block_index);

	  // Load the block

	  poly.get_block_data(block_index, poly_block);
	  dens.get_block_data(block_index, dens_block);

	  const std::vector<double> & cell_width = block_info.get_cell_width();

	  const double cell_side_length_x = cell_width[0];

	  const double cell_volume = block_info.get_cell_volume();

	  const unsigned int num_cells = block_info.get_num_cells();

	  std::vector<double> cell_center;

	  for (unsigned int cell_index = 0; cell_index < num_cells; 
	       cell_index++)
	    {
	    // Get the center coordinates

	    block_info.get_cell_center(cell_index, cell_center);

	    if (domain.in_domain(cell_center, cell_width))
	      {
	      const double cell_x_center = cell_center[0];

	      const double poly_value = poly_block[cell_index];
	      const double dens_value = dens_block[cell_index];

	      const double cell_mass = cell_volume * dens_value;
	      // const double poly_mass = poly_value * cell_mass;

	      poly_bins.add_data_spread(cell_x_center, poly_value, 
					cell_side_length_x, cell_mass);
	      dens_bins.add_data_spread(cell_x_center, dens_value, 
					cell_side_length_x, cell_volume);
	      }
	    }
	  }
	}
      }

    // Report on the findings

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      const QuickFlash::Stats::Bin & poly = poly_bins.get_bin(bin_index);
      const QuickFlash::Stats::Bin & dens = dens_bins.get_bin(bin_index);

      const double bin_center = poly_bins.get_bin_center(bin_index);

      const unsigned int num_entries = poly.get_num_entries();

      const double poly_mass = poly.get_weight_value_sum();

      const double dens_mean = dens.get_mean();

      cout << bin_index << "\t" << bin_center << "\t" << num_entries << "\t"
	   << poly_mass << "\t" << dens_mean << "\t" << endl;
      }
    }

  return 0;
  }
