// C++ program file file accum_grid.cpp

/*
  By Nathan C. Hearn
     July 6, 2007

  Uniform Grid for accumulating values.
*/


#include "accum_grid.hpp"
#include <vector>
#include "quickflash_geometry.hpp"
#include "quickflash_except.hpp"
#include "quickflash_counters.hpp"


// Helper functions

double volume_overlap(const std::vector<double> & primary_min_coords,
		      const std::vector<double> & primary_max_coords,
		      const std::vector<double> & secondary_min_coords,
		      const std::vector<double> & secondary_max_coords)
  {
  double overlap_volume = 0.0;  // Default value

  const unsigned int dims = primary_min_coords.size();

  if ((primary_max_coords.size() != dims) 
      || (secondary_min_coords.size() != dims)
      || (secondary_max_coords.size() != dims))
    throw QuickFlash::Except("Incompatible boundary vectors", __FILE__,
			   __LINE__);

  // Get overlap

  bool overlapping = true;

  double volume = 1.0;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double primary_min = primary_min_coords[i];
    const double primary_max = primary_max_coords[i];

    const double secondary_min = secondary_min_coords[i];
    const double secondary_max = secondary_max_coords[i];

    const double overlap_min = (primary_min > secondary_min) 
      ? primary_min : secondary_min;

    const double overlap_max = (primary_max < secondary_max) 
        ? primary_max : secondary_max;

    if (!(overlap_min < overlap_max))
      overlapping = false;

    // Get the volume and center position of the overlap

    volume *= overlap_max - overlap_min;
    }

  if (overlapping)
    overlap_volume = volume;

  return overlap_volume;
  }


// Class AccumGrid

void AccumGrid::reset()
  {
  dims = 0;

  info.reset();
  data.reset();

  half_window_width.clear();
  bin_window_scale = 0.0;
  }


void AccumGrid::reset(const std::vector<double> & min_coords,
		      const std::vector<double> & max_coords,
		      const std::vector<unsigned int> & grid_dims,
		      const std::vector<double> & bin_window_size)
  {
  dims = grid_dims.size();

  if ((min_coords.size() != dims) || (max_coords.size() != dims))
    throw QuickFlash::Except("Incompatible coordiante vector lengths", 
			   __FILE__, __LINE__);

  const unsigned int window_vector_size = bin_window_size.size();

  if ((window_vector_size != dims) && (window_vector_size != 0))
    throw QuickFlash::Except("Incompatible window size vector length", __FILE__,
			   __LINE__);

  info.reset(min_coords, max_coords, grid_dims, QuickFlash::Geometry::Cartesian);
  data.reset(grid_dims);

  // Determine the window size

  std::vector<double> window_width(dims);

  if (window_vector_size == 0)
    window_width = info.get_cell_width();
  else
    {
    window_width = bin_window_size;

    for (unsigned int i = 0; i < dims; i++)
      if (!(window_width[i] > 0.0))
	throw QuickFlash::Except("Window widths must be positive", __FILE__,
			       __LINE__);
    }

  half_window_width.resize(dims);

  for (unsigned int i = 0; i < dims; i++)
    half_window_width[i] = 0.5 * window_width[i];

  double window_volume = 1.0;

  for (unsigned int i = 0; i < dims; i++)
    window_volume *= window_width[i];

  const double bin_volume = info.get_cell_volume();

  bin_window_scale = bin_volume / window_volume;

  clear_grid();
  }


void AccumGrid::reset(const AccumGrid & source)
  {
  if (&source != this)
    {
    info = source.info;
    data = source.data;
    }
  }


void AccumGrid::clear_grid()
  {
  const unsigned int num_cells = info.get_num_cells();

  for (unsigned int index = 0; index < num_cells; index++)
    data[index] = 0.0;
  }


void AccumGrid::distribute_value(const std::vector<double> & center_pos,
				 const double value)
  {
  if (center_pos.size() != dims)
    throw QuickFlash::Except("Incompatible coordinate vector", __FILE__, 
			   __LINE__);

  // Add value to the cell containing center_pos (if it exists)

  // THIS IGNORES THE WINDOWING!!!!!

  if (info.in_block(center_pos))
    {
    const unsigned int cell_index = info.get_nearest_cell(center_pos);

    data[cell_index] += value;
    }
  }


void AccumGrid::distribute_value(const std::vector<double> & center_pos,
				 const std::vector<double> & width,
				 const double value)
  {
  if ((center_pos.size() != dims) || (width.size() != dims))
    throw QuickFlash::Except("Incompatible coordinate or width vector", 
			   __FILE__, __LINE__);

  for (unsigned int i = 0; i < dims; i++)
    if (!(width[i] > 0.0))
      throw QuickFlash::Except("Width elements must be positive", __FILE__,
			     __LINE__);

  // Add value to the cells overlapping this region 
  // ... scaled by the fractional overlap area

  std::vector<double> min_coords = center_pos;
  std::vector<double> max_coords = center_pos;

  std::vector<double> min_envelope = center_pos;
  std::vector<double> max_envelope = center_pos;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double half_width = 0.5 * width[i];

    min_coords[i] -= half_width;
    max_coords[i] += half_width;

    // Over-estimate the applicable window-width overlap region

    const double envelope_cell_width = half_window_width[i] + half_width;

    min_envelope[i] -= envelope_cell_width;
    max_envelope[i] += envelope_cell_width;
    }

  double volume = 1.0;

  for (unsigned int i = 0; i < dims; i++)
    volume *= width[i];

  const double value_over_volume = value / volume;

  // Determine the range of applicable coordinates

  std::vector<unsigned int> min_index(dims);
  std::vector<unsigned int> max_index(dims);

  info.get_nearest_cell(min_envelope, min_index);
  info.get_nearest_cell(max_envelope, max_index);

  // Set up the counter object

  std::vector<unsigned int> index_range(dims);

  for (unsigned int i = 0; i < dims; i++)
    index_range[i] = (max_index[i] - min_index[i]) + 1;

  QuickFlash::VectorCounter counter(index_range);

  // Add the scaled value to the cells

  std::vector<unsigned int> bin_index(dims);

  std::vector<double> bin_center_coords(dims);
  std::vector<double> bin_min_coords(dims);
  std::vector<double> bin_max_coords(dims);

  while (counter.in_bounds())
    {
    bin_index = min_index;

    for (unsigned int i = 0; i < dims; i++)
      bin_index[i] += counter[i];

    info.get_cell_center(bin_index, bin_center_coords);

    for (unsigned int i = 0; i < dims; i++)
      {
      const double center_coord = bin_center_coords[i];
      const double half_width = half_window_width[i];

      bin_min_coords[i] = center_coord - half_width;
      bin_max_coords[i] = center_coord + half_width;
      }

    const double overlap = volume_overlap(min_coords, max_coords,
					  bin_min_coords, bin_max_coords);

    const double scale = overlap * bin_window_scale;

    data[bin_index] += scale * value_over_volume;

    counter.increment();
    }
  }
