#!/usr/bin/env python

# plot_histograms.py

# By Nathan C. Hearn
#    July 2, 2008
#
# Produces histograms for output of mass_bins


try :
    import psyco
except ImportError :
    pass


plot_suffix = "png"
matplotlib_mode = "Agg"

font_size = "x-small"


def main() :

    from sys import argv, exit, stderr, stdout
    from optparse import OptionParser

    from read_bins import read_bins

    import matplotlib
    matplotlib.use(matplotlib_mode)

    import pylab

    # Set up the command line parser

    usage = str("%prog [options] mass_bin_file")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_prefix_str = str("output-prefix")
    argparser.add_option("--" + output_prefix_str, type="string",
                         dest="output_prefix", help="Output file prefix")
    argparser.set_defaults(output_prefix=None)

    start_entry_str = str("start-entry")
    argparser.add_option("--" + start_entry_str, type="int",
                         dest="start_entry", help="Initial entry index")
    argparser.set_defaults(start_entry=None)

    end_entry_str = str("end-entry")
    argparser.add_option("--" + end_entry_str, type="int",
                         dest="end_entry", help="Initial entry index")
    argparser.set_defaults(end_entry=None)

    target_mass_above_str = str("target-mass-above")
    argparser.add_option("--" + target_mass_above_str, type="float",
                         dest="target_mass_above",
                         help="Estimate value enclosing given mass "
                              + "(at or above value)")
    argparser.set_defaults(target_mass_above=None)


    # Process the command line input

    opts, args = argparser.parse_args()

    argc = len(args)

    req_args = int(1)

    if (argc != req_args) :
        argparser.print_help(file=stderr)
        exit(1)

    argptr = int(0)

    mass_bin_file = args[argptr]
    argptr += 1

    output_prefix = ""

    if (opts.output_prefix is not None) :
        output_prefix = opts.output_prefix
        
    start_entry = opts.start_entry
    end_entry = opts.end_entry

    target_mass_above = opts.target_mass_above

    find_target_mass_above = False

    if (target_mass_above is not None) :
        find_target_mass_above = True


    ## ERROR CHECKING???

    ## Get the basic information

    mass_bins, compute_radii = read_bins(mass_bin_file)

    num_entries = mass_bins.get_num_entries()

    dataset = mass_bins.get_dataset_name()

    min_value = mass_bins.get_min_value()
    max_value = mass_bins.get_max_value()

    num_bins = mass_bins.get_num_bins()
    bin_values = mass_bins.get_bin_values()

    global_min_mass, global_max_mass = mass_bins.get_global_min_max_mass(True)

    if ((global_min_mass is None) or (global_max_mass is None)) :
        raise RuntimeError("No mass found")

    zero_mass_value = 0.5 * global_min_mass  ## For log scaling

    global_min_radius = None
    global_max_radius = None

    if (compute_radii) :
        global_min_radius, global_max_radius \
            = mass_bins.get_global_min_max_radii()
        
    zero_radius_value = 0.5 * global_min_radius  ## For log scaling


    ## Produce the plots

    if (start_entry is None) :
        start_entry = int(0)

    if (end_entry is None) :
        end_entry = num_entries - 1

    for entry_index in range(start_entry, (end_entry + 1)) :

        data_file = mass_bins.get_filename(entry_index)

        sim_time = mass_bins.get_sim_time(entry_index)

        total_mass = mass_bins.get_total_mass(entry_index)

        mass_below = mass_bins.get_mass_below(entry_index)
        mass_above = mass_bins.get_mass_above(entry_index)

        bin_masses = mass_bins.get_bin_masses(entry_index)

        sum_bin_masses = mass_bins.get_sum_bin_masses(entry_index)

        sum_mass = sum_bin_masses + mass_below + mass_above

        stderr.write("Processing data for file " + data_file + "\n")
        stderr.write("  Mass Reported [ " + str(total_mass) + " ] Actual [ "
                     + str(sum_mass) + " ]\n")

        bin_mass_enclosed_above \
            = mass_bins.get_bin_masses_enclosed_above(entry_index)

        bin_mass_enclosed_below \
            = mass_bins.get_bin_masses_enclosed_below(entry_index)

        bin_min_radii = mass_bins.get_bin_min_radii(entry_index)
        bin_max_radii = mass_bins.get_bin_max_radii(entry_index)

        ## Print some file data to stdout

        stdout.write("File " + data_file)

        if (find_target_mass_above) :

            value_mass_above \
                = mass_bins.get_value_from_mass_above(entry_index,
                                                      target_mass_above)

            stdout.write(" MassValueAbove")

            if (value_mass_above is not None) :
                stdout.write(" " + str(value_mass_above))
            else :
                stdout.write(" NA")

        stdout.write("\n")

        ## Build the mass histogram plot

        plot_title = dataset + " Bins - " + data_file + " - Time " \
                     + str(sim_time) + " s - Mass " + str(total_mass)

        x_title = dataset + " Bins"
        y_title = "Bin Mass"

        x_columns = list()
        x_labels = list()

        x_columns.append(int(0))
        x_labels.append("< %(num)#5.2e" % { "num" : min_value })

        for bin_index in range(num_bins) :
            x_columns.append(bin_index + 1)
            x_labels.append("%(num)#5.2e" % { "num" : bin_values[bin_index] })

        x_columns.append(int(num_bins + 1))
        x_labels.append("> %(num)#5.2e" % { "num" : max_value })

        fig = pylab.figure(figsize=(8, 6), dpi=128)

        fig.subplots_adjust(bottom=0.18)

        fig.clf()
        fig.hold(True)

        hist = fig.add_subplot(1,1,1)

        hist.semilogy()

        hist.set_title(plot_title, fontsize=font_size)

        hist.set_xlabel(x_title)
        hist.set_ylabel(y_title)

        bar_indexes = range(num_bins + 2)
        
        bar_data = list()

        bar_data.append(mass_below)
        bar_data.extend(bin_masses)
        bar_data.append(mass_above)

        for index in bar_indexes :
            if (bar_data[index] < global_min_mass) :
                bar_data[index] = zero_mass_value

        hist.bar(bar_indexes, bar_data)

        hist.set_xticks(x_columns)
        hist.set_xticklabels(x_labels, fontsize=font_size, rotation="vertical",
                             horizontalalignment="center")
        # hist.set_yticklabels(fontsize=font_size)

        hist.set_ylim(zero_mass_value, global_max_mass)

        hist_filename = output_prefix + data_file + "_" + dataset + "_" \
                        + "mass_hist" + "." + plot_suffix

        pylab.savefig(hist_filename, orientation="portrait")

        ## Build the mass enclosed above plot

        plot_title = dataset + " Bins - " + data_file + " - Time " \
                     + str(sim_time) + " s - Mass " + str(total_mass)

        x_title = dataset + " Bins"
        y_title = "Mass Enclosed At and Above Bin Value"

        x_columns = list()
        x_labels = list()

        x_columns.append(int(0))
        x_labels.append("< %(num)#5.2e" % { "num" : min_value })

        for bin_index in range(num_bins) :
            x_columns.append(bin_index + 1)
            x_labels.append("%(num)#5.2e" % { "num" : bin_values[bin_index] })

        x_columns.append(int(num_bins + 1))
        x_labels.append("> %(num)#5.2e" % { "num" : max_value })
                           

        fig = pylab.figure(figsize=(8, 6), dpi=128)

        fig.subplots_adjust(bottom=0.18)

        fig.clf()
        fig.hold(True)

        plot_above = fig.add_subplot(1,1,1)

        plot_above.set_title(plot_title, fontsize=font_size)

        plot_above.set_xlabel(x_title)
        plot_above.set_ylabel(y_title)

        plot_indexes = range(num_bins + 2)
        
        above_data = list()

        above_data.append(total_mass)
        above_data.extend(bin_mass_enclosed_above)
        above_data.append(mass_above)

        for index in plot_indexes :
            if (above_data[index] < global_min_mass) :
                above_data[index] = zero_mass_value

        plot_above.plot(plot_indexes, above_data)

        plot_above.set_xticks(x_columns)
        plot_above.set_xticklabels(x_labels, fontsize=font_size,
                                   rotation="vertical",
                                   horizontalalignment="center")
        # plot_above.set_yticklabels(fontsize=font_size)

        plot_above.set_ylim(zero_mass_value, total_mass)

        plot_above_filename = output_prefix + data_file + "_" + dataset + "_" \
                              + "mass_enclosed_above" + "." + plot_suffix

        pylab.savefig(plot_above_filename, orientation="portrait")

        ## Build the mass enclosed below plot

        plot_title = dataset + " Bins - " + data_file + " - Time " \
                     + str(sim_time) + " s - Mass " + str(total_mass)

        x_title = dataset + " Bins"
        y_title = "Mass Enclosed At and Below Bin Value"

        x_columns = list()
        x_labels = list()

        x_columns.append(int(0))
        x_labels.append("< %(num)#5.2e" % { "num" : min_value })

        for bin_index in range(num_bins) :
            x_columns.append(bin_index + 1)
            x_labels.append("%(num)#5.2e" % { "num" : bin_values[bin_index] })

        x_columns.append(int(num_bins + 1))
        x_labels.append("> %(num)#5.2e" % { "num" : max_value })
                           

        fig = pylab.figure(figsize=(8, 6), dpi=128)

        fig.subplots_adjust(bottom=0.18)

        fig.clf()
        fig.hold(True)

        plot_below = fig.add_subplot(1,1,1)

        plot_below.set_title(plot_title, fontsize=font_size)

        plot_below.set_xlabel(x_title)
        plot_below.set_ylabel(y_title)

        plot_indexes = range(num_bins + 2)
        
        below_data = list()

        below_data.append(mass_below)
        below_data.extend(bin_mass_enclosed_below)
        below_data.append(total_mass)

        for index in plot_indexes :
            if (below_data[index] < global_min_mass) :
                below_data[index] = zero_mass_value

        plot_below.plot(plot_indexes, below_data)

        plot_below.set_xticks(x_columns)
        plot_below.set_xticklabels(x_labels, fontsize=font_size,
                                   rotation="vertical",
                                   horizontalalignment="center")
        # plot_below.set_yticklabels(fontsize=font_size)

        plot_below.set_ylim(zero_mass_value, total_mass)

        plot_below_filename = output_prefix + data_file + "_" + dataset + "_" \
                              + "mass_enclosed_below" + "." + plot_suffix

        pylab.savefig(plot_below_filename, orientation="portrait")

        if (compute_radii) :
            
            ## Build the density-radius plots

            plot_title = dataset + " Bins - " + data_file + " - Time " \
                         + str(sim_time) + " s - Mass " + str(total_mass)

            x_title = dataset + " Bins"
            y_title = "Bin Radius Range"

            x_columns = list()
            x_labels = list()

            for bin_index in range(num_bins) :
                x_columns.append(bin_index)
                x_labels.append("%(num)#5.2e"
                                % { "num" : bin_values[bin_index] })

            fig = pylab.figure(figsize=(8, 6), dpi=128)

            fig.subplots_adjust(bottom=0.18)

            fig.clf()
            fig.hold(True)

            rad_hist = fig.add_subplot(1,1,1)

            rad_hist.semilogy()

            rad_hist.set_title(plot_title, fontsize=font_size)

            rad_hist.set_xlabel(x_title)
            rad_hist.set_ylabel(y_title)

            rad_hist_indexes = range(num_bins)

            min_radius_data = list(bin_min_radii)

            for index in rad_hist_indexes :
                if (min_radius_data[index] is None) :
                    min_radius_data[index] = zero_radius_value
                elif (min_radius_data[index] < global_min_radius) :
                    min_radius_data[index] = zero_radius_value

            max_radius_data = list(bin_max_radii)

            for index in rad_hist_indexes :
                if (max_radius_data[index] is None) :
                    max_radius_data[index] = zero_radius_value
                elif (max_radius_data[index] < global_min_radius) :
                    max_radius_data[index] = zero_radius_value

            rad_hist_heights = list()

            for index in rad_hist_indexes :
                rad_hist_heights.append(max_radius_data[index]
                                        - min_radius_data[index])

            rad_hist.bar(rad_hist_indexes, rad_hist_heights,
                         bottom=min_radius_data)

            rad_hist.set_xticks(x_columns)
            rad_hist.set_xticklabels(x_labels, fontsize=font_size,
                                     rotation="vertical",
                                     horizontalalignment="center")
            # hist.set_yticklabels(fontsize=font_size)

            rad_hist.set_ylim(global_min_radius, global_max_radius)

            rad_hist_filename = output_prefix + data_file + "_" + dataset \
                                + "_" + "radius_hist" + "." + plot_suffix

            pylab.savefig(rad_hist_filename, orientation="portrait")

            ## Build the enclosed mass versus radius plots

            plot_title = "Inferred Mass vs Radii - " + data_file + " - Time " \
                         + str(sim_time) + " s - Mass " + str(total_mass)

            x_title = "Radius"
            y_title = "Mass Enclosed Below Radius"

            above_masses = list()
            above_radii = list()

            for bin_index in range(num_bins) :

                bin_max_rad = bin_max_radii[bin_index]

                if (bin_max_rad is not None) :
                    above_masses.append(bin_mass_enclosed_above[bin_index])
                    above_radii.append(bin_max_rad)

            fig = pylab.figure(figsize=(8, 6), dpi=128)

            fig.clf()
            fig.hold(True)

            above_mass_rad_plot = fig.add_subplot(1,1,1)

            above_mass_rad_plot.set_title(plot_title, fontsize=font_size)

            above_mass_rad_plot.set_xlabel(x_title)
            above_mass_rad_plot.set_ylabel(y_title)

            above_mass_rad_plot.loglog()

            above_mass_rad_plot.plot(above_radii, above_masses)

            above_mass_rad_plot.set_xlim(global_min_radius, global_max_radius)
            above_mass_rad_plot.set_ylim(zero_mass_value, total_mass)

            above_mass_rad_plot.grid()

            above_mass_rad_plot_filename = output_prefix + data_file + "_" \
                                           + dataset + "_" \
                                           + "mass_radius_above" + "." \
                                           + plot_suffix

            pylab.savefig(above_mass_rad_plot_filename, orientation="portrait")
        

if (__name__ == "__main__") :
    main()

        
