// C++ program file list_blocks.cpp

/*
  By Nathan C. Hearn
     November 9, 2008

  Program for extracting block information.
*/


#include <iostream>
#include <vector>
#include <list>
#include <string>
#include <fstream>
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_tree_block.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blocktype.hpp"


const char default_comment_marker[] = "#";


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, 
					option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 1;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name << " [ --output-prefix=PREFIX ]"
	 << endl << "\t"
	 << "[ --min-refine-level=MIN_LEVEL ] [ --max-refine-level=MAX_LEVEL ]"
	 << endl << "\t"
	 << "[ --finest-blocks ] [ --add-comment-marker[=MARKER] ]"
	 << endl << "\t"
	 << "flash_filename1 [ flash_filename2 ... ]"
	 << endl << endl;

    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  std::list<std::string> flash_file_list;

  while (argptr < num_args)
    flash_file_list.push_back(argument_list[argptr++]);

  // Parse the options

  std::string base_output_prefix = "";

  bool limit_min_refine_level = false;
  unsigned int min_refine_level = 0;

  bool limit_max_refine_level = false;
  unsigned int max_refine_level = 0;

  bool finest_blocks_only = false;

  bool use_comment_marks = false;
  std::string user_comment_marker = default_comment_marker;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output prefix is required with option"
	     << endl << endl;
	return -1;
	}

      base_output_prefix = current_option.get_value();

      // Add an underscore if prefix contains more than just a path

      std::string base_prefix_path;
      std::string base_prefix_name;

      QuickFlash::Utils::split_name_path_file(base_output_prefix, 
					      base_prefix_path, 
					      base_prefix_name);

      const unsigned int base_prefix_name_len = base_prefix_name.size();

      if (base_prefix_name_len > 0)
	if (base_prefix_name[base_prefix_name_len - 1] != '_')
	  base_output_prefix += "_";
      }
    else if (keyword == "min-refine-level")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Min refine level requires a positive level"
	     << endl << endl;

	return -1;
	}
      
      current_option.get_value(min_refine_level);

      if (min_refine_level < 1)
	{
	cerr << endl << "Error: Refinement levels must be positive integers"
	     << endl << endl;

	return -1;
	}

      limit_min_refine_level = true;
      }
    else if (keyword == "max-refine-level")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Max refine level requires a positive level"
	     << endl << endl;

	return -1;
	}
      
      current_option.get_value(max_refine_level);

      if (max_refine_level < 1)
	{
	cerr << endl << "Error: Refinement levels must be positive integers"
	     << endl << endl;

	return -1;
	}

      limit_max_refine_level = true;
      }
    else if (keyword == "finest-blocks")
      {
      finest_blocks_only = true;
      }
    else if (keyword == "add-comment-marker")
      {
      use_comment_marks = true;

      if (current_option.value_present())
	user_comment_marker = current_option.get_value();
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }


  // Perform additional checks

  if (limit_min_refine_level && limit_max_refine_level)
    if (min_refine_level > max_refine_level)
      {
      cerr << endl << "Error: Min refinement level can not exceed max level"
	   << endl << endl;
      return -1;
      }

  // Set up some useful quantities

  const std::string comment_marker 
    = use_comment_marks ? (user_comment_marker + " ") : "";

  // Run through the files

  std::list<std::string>::const_iterator file_iter = flash_file_list.begin();

  while (file_iter != flash_file_list.end())
    {
    const std::string & flash_filename = *file_iter;

    // Open the data file

    const QuickFlash::File::DataFile df(flash_filename, 0, true);

    const QuickFlash::File::MeshInfo & mesh_info = df.get_mesh_info();

    const unsigned int dims = mesh_info.get_dims();

    const double sim_time = mesh_info.get_sim_time();

    // Open the output file

    const std::string output_prefix = base_output_prefix + flash_filename;

    const std::string output_filename = output_prefix + "_blocks.csv";

    std::ofstream f(output_filename.c_str(), std::ios::trunc);

    f << comment_marker << "DataFile " << flash_filename << endl;
    f << comment_marker << "MeshDims " << dims << endl;
    f << comment_marker << "SimTime " << sim_time << endl;

    f << endl;

    f << comment_marker << "NumBlocks " << mesh_info.get_num_blocks() << endl;

    f << comment_marker << "NumLeafBlocks " << mesh_info.get_num_leaf_blocks()
      << endl;

    f << comment_marker << "BaseBlocks " << mesh_info.get_base_block_dims()
      << endl;

    f << comment_marker << "VolumeMinBounds " 
      << mesh_info.get_volume_minbounds() << endl;
      
    f << comment_marker << "VolumeMaxBounds " 
      << mesh_info.get_volume_maxbounds() << endl;
      
    f << comment_marker << "VolumeCenter " << mesh_info.get_volume_center()
      << endl;

    f << comment_marker << "VolumeWidth " << mesh_info.get_volume_width()
      << endl;

    f << comment_marker << "MinLeafRefineLevel " 
      << mesh_info.get_min_leaf_refine_level() << endl;

    f << comment_marker << "MaxLeafRefineLevel " 
      << mesh_info.get_max_leaf_refine_level() << endl;

    f << endl;

    if (limit_min_refine_level)
      f << comment_marker << "MinListRefineLevel " << min_refine_level << endl;

    if (limit_max_refine_level)
      f << comment_marker << "MaxListRefineLevel " << max_refine_level << endl;

    f << comment_marker << "FinestBlocksOnly ";

    if (finest_blocks_only)
      f << "true";
    else
      f << "false";

    f << endl;

    f << endl;

    f << comment_marker << "DataStart" << endl;

    f << endl;

    f << "Block_Index" << "\t" << "Block_Type" << "\t" << "Refine_Level" 
      << "\t" << "Block_Min_Coords" << "\t" << "Block_Max_Coords" 
      << "\t" << "Block_Center" << "\t" << "Block_Width" << endl;

    // Run through the blocks using a tree iterator

    QuickFlash::Tree::ConstTreeIter node_iter = mesh_info.get_tree_iter();

    while (node_iter.valid_node())
      {
      const QuickFlash::Block::BlockInfo & block_info 
	= node_iter.get_block_info();

      const unsigned int refine_level = node_iter.get_refinement_level();

      const bool is_leaf = node_iter.is_leaf();

      // Determine if node or branch are to be included

      bool skip_branch = false;
      bool skip_node = false;

      if (limit_max_refine_level && (refine_level > max_refine_level))
	{
	skip_branch = true;
	skip_node = true;
	}
      else if (limit_min_refine_level && (refine_level < min_refine_level))
	skip_node = true;
      else if (finest_blocks_only)
	{
	if (!is_leaf)
	  {
	  if (limit_max_refine_level)
	    {
	    if (refine_level < max_refine_level)
	      skip_node = true;
	    }
	  else
	    skip_node = true;
	  }
	}

      if (!skip_node)
	{
	const unsigned int block_index = node_iter.get_block_index();

	const QuickFlash::Block::BlockType block_type 
	  = mesh_info.get_block_type(block_index);

	std::string block_type_str;

	QuickFlash::Block::get_block_type_label(block_type, block_type_str);

	const std::vector<double> & min_coords = block_info.get_min_coords();
	const std::vector<double> & max_coords = block_info.get_max_coords();

	std::vector<double> block_center;
	std::vector<double> block_width;

	block_info.get_block_center(block_center);
	block_info.get_block_width(block_width);

	f << block_index << "\t" << block_type_str << "\t" << refine_level
	  << "\t" << min_coords << "\t" << max_coords << "\t" << block_center
	  << "\t" << block_width << endl;
	}

      if (skip_branch)
	node_iter.next_branch();
      else
	node_iter.next();    
      }

    f << endl;

    ++file_iter;
    }

  return 0;
  }
