#!/usr/bin/env python

# generate_layer_plots.py

# By Nathan C. Hearn
#    August 26, 2008
#
# Plotter for data from layer_distrib

# Make sure the QuickFlash/utils directory is in PYTHONPATH!!!


try :
    import psyco
except ImportError :
    pass


default_plot_title = ""


class SimData :

    def __init__(self, name, prefix) :

        self.__name = name
        self.__prefix = prefix

        self.__time_list = list()

        self.__layer_min_list = list()
        self.__layer_max_list = list()

        self.__envelope_min_list = list()
        self.__envelope_max_list = list()


    def get_metadata_filename(self, index) :

        index_str = "%(num)04d" % { "num" : index }

        filename = self.__prefix + index_str + "_metadata.out"

        return filename


    def add_data(self, time, layer_min_pos, layer_max_pos,
                 envelope_min_pos, envelope_max_pos) :

        self.__time_list.append(time)

        self.__layer_min_list.append(layer_min_pos)
        self.__layer_max_list.append(layer_max_pos)

        self.__envelope_min_list.append(envelope_min_pos)
        self.__envelope_max_list.append(envelope_max_pos)
        

    def get_name(self) :
        return self.__name


    def get_num_entries(self) :
        return len(self.__time_list)


    def get_time_list(self) :
        return self.__time_list


    def get_layer_min_list(self) :
        return self.__layer_min_list


    def get_layer_max_list(self) :
        return self.__layer_max_list


    def get_envelope_min_list(self) :
        return self.__envelope_min_list


    def get_envelope_max_list(self) :
        return self.__envelope_max_list


    def get_layer_widths(self) :

        layer_widths = list()

        num_entries = len(self.__time_list)

        for index in range(num_entries) :
            width = self.__layer_max_list[index] - self.__layer_min_list[index]
            layer_widths.append(width)

        return layer_widths


def main() :

    from sys import argv, exit, stderr
    from os import path
    from optparse import OptionParser
    
    from plot_gen import Plot

    from metadata_reader import parse_distrib_metadata

    ## Set up the command-line parser

    usage = str("%prog [options] start_index end_index "
                + "metadata_prefix1,data_name_1[,background_vel_1] "
                + "[ metadata_prefix_2,data_name_2[,background_vel_2] ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_prefix_str = str("output-prefix")
    argparser.add_option("--" + output_prefix_str, type="string",
                         dest="output_prefix",
                         help="Prefix for output plot files")
    argparser.set_defaults(output_prefix=None)

    plot_title_str = str("plot-title")
    argparser.add_option("--" + plot_title_str, type="string",
                         dest="plot_title",
                         help="Plot title")
    argparser.set_defaults(plot_title=None)

    background_vel_str = str("background-vel")
    argparser.add_option("--" + background_vel_str, type="string",
                         dest="background_vel",
                         help=("Background velocity (cm / s) "
                               + "with optional comma-separated time offset "
                               + "(s)"))
    argparser.set_defaults(background_vel=None)

    envelope_file_prefix_str = str("envelope-file-prefix")
    argparser.add_option("--" + envelope_file_prefix_str, type="string",
                         dest="envelope_file_prefix",
                         help="File prefix for maximal envelope plot")
    argparser.set_defaults(envelope_file_prefix=None)
    

    ## Process the command-line input

    opts, args = argparser.parse_args()

    argc = len(args)

    min_args = int(3)

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n")
        exit(1)

    ## Get the required arguments

    argptr = int(0)

    start_index = int(args[argptr])
    argptr += 1

    if (start_index < 0) :
        stderr.write("\nError: Start index [ " + str(start_index)
                     + " ] must be non-negative\n\n")
        exit(-1)

    end_index = int(args[argptr])
    argptr += 1

    if (end_index < start_index) :
        stderr.write("\nError: End index [ " + str(end_index)
                     + " ] can not be less than start index\n\n")
        exit(-1)

    sim_set_list = list()
    background_vel_list = list()





    

    while (argptr < argc) :

        entry = args[argptr]

        words = entry.split(",")
        argptr += 1

        num_words = len(words)

        if (num_words < 2) :
            stderr.write("\nError: Metadata entry [ " + entry
                         + " ] must have two comma-separated elements\n\n")
            exit(-1)

        if (num_words > 3) :
            stderr.write("\nError: Metadata entry requires two or three "
                         + "comma-separated elements\n\n")
            exit(-1)

        metadata_prefix = words[0]
        metadata_name = words[1]

        sim_set_container = SimData(metadata_name, metadata_prefix)

        sim_set_list.append(sim_set_container)

    ## Get the optional arguments

    output_prefix = opts.output_prefix

    if (output_prefix is None) :
        output_prefix = ""
    elif (len(path.split(output_prefix)[1]) > 0) :
        output_prefix += "_"

    plot_title = opts.plot_title

    if (plot_title is None) :
        plot_title = default_plot_title

    if (plot_title != "") :
        plot_title += "  -  "

    background_vel = float(0.0)
    background_vel_time_offset = float(0.0)

    if (opts.background_vel is not None) :
        background_vel_words = opts.background_vel.split(",")

        num_background_vel_words = len(background_vel_words)

        if ((num_background_vel_words < 1) or (num_background_vel_words > 2)) :
            stderr.write("\nError: Background velocity requires one or two "
                         + "comma-separated entries\n\n")
            exit(-1)

        background_vel = float(background_vel_words[0])

        if (num_background_vel_words == 2) :
            background_vel_time_offset = float(background_vel_words[1])

    envelope_set = None

    if (opts.envelope_file_prefix is not None) :
        envelope_set = SimData("Envelope", opts.envelope_file_prefix)


    ## Get derived data

    num_files = (end_index + 1) - start_index

    ## Collect the data

    for sim_set in sim_set_list :

        for index in range(num_files) :

            file_index = index + start_index

            filename = sim_set.get_metadata_filename(file_index)

            file_data, slice_info = parse_distrib_metadata(filename)

            if ((not file_data.get_value("FoundLayerMin"))
                or (not file_data.get_value("FoundLayerMax"))) :
                
                stderr.write("\nWarning: Layer boundaries not found in file "
                             + filename + "\n\n")

            else :

                sim_time = file_data.get_value("SimTime")

                layer_min_pos = file_data.get_value("LayerMinPos")
                layer_max_pos = file_data.get_value("LayerMaxPos")

                envelope_min_pos = file_data.get_value("MinEnvelopePos")
                envelope_max_pos = file_data.get_value("MaxEnvelopePos")

                sim_set.add_data(sim_time, layer_min_pos, layer_max_pos,
                                 envelope_min_pos, envelope_max_pos)

    if (envelope_set is not None) :

        for index in range(num_files) :

            file_index = index + start_index

            filename = envelope_set.get_metadata_filename(file_index)

            file_data, slice_info = parse_distrib_metadata(filename)

            if ((not file_data.get_value("FoundLayerMin"))
                or (not file_data.get_value("FoundLayerMax"))) :
                
                stderr.write("\nWarning: Layer boundaries not found in file "
                             + filename + "\n\n")

            else :

                sim_time = file_data.get_value("SimTime")

                layer_min_pos = file_data.get_value("LayerMinPos")
                layer_max_pos = file_data.get_value("LayerMaxPos")

                envelope_min_pos = file_data.get_value("MinEnvelopePos")
                envelope_max_pos = file_data.get_value("MaxEnvelopePos")

                envelope_set.add_data(sim_time, layer_min_pos, layer_max_pos,
                                      envelope_min_pos, envelope_max_pos)

    ## Plot the data

    width_plot = Plot(plot_title + "Mixing Layer Widths", "Time (ns)",
                      "Width (microns)")
    
    bounds_plot = Plot(plot_title + "Mixing Layer Boundaries", "Time (ns)",
                       "Position (microns)")

    for sim_set in sim_set_list :

        sim_name = sim_set.get_name()

        num_entries = sim_set.get_num_entries()

        orig_time_list = sim_set.get_time_list()
        
        orig_layer_min_list = sim_set.get_layer_min_list()
        orig_layer_max_list = sim_set.get_layer_max_list()

        time_list = list()
        
        layer_min_list = list()
        layer_max_list = list()

        layer_width_list = list()

        for entry_index in range(num_entries) :

            ## Determine the position offsets due to background velocity

            orig_time = orig_time_list[entry_index]

            pos_offset = (-1.0) * (orig_time - background_vel_time_offset) \
                         * background_vel

            ## Convert to nanoseconds and microns from seconds and centimeters

            new_time = orig_time * 1.0e+09

            new_layer_min = (orig_layer_min_list[entry_index] + pos_offset) \
                            * 1.0e+04
            
            new_layer_max = (orig_layer_max_list[entry_index] + pos_offset) \
                            * 1.0e+04

            time_list.append(new_time)

            layer_min_list.append(new_layer_min)
            layer_max_list.append(new_layer_max)

            layer_width_list.append(new_layer_max - new_layer_min)

        width_plot.add_plot(time_list, layer_width_list, legend_text=sim_name)

        bounds_plot.add_plot(time_list, layer_min_list,
                             legend_text=(sim_name + " min"),
                             marker_index=1)
            
        bounds_plot.add_plot(time_list, layer_max_list,
                             legend_text=(sim_name + " max"),
                             marker_index=2, use_prev_color=True)

    if (envelope_set is not None) :

        sim_name = "Envelope"

        num_entries = envelope_set.get_num_entries()

        orig_time_list = envelope_set.get_time_list()
        
        orig_envelope_min_list = envelope_set.get_envelope_min_list()
        orig_envelope_max_list = envelope_set.get_envelope_max_list()

        time_list = list()
        
        envelope_min_list = list()
        envelope_max_list = list()

        for entry_index in range(num_entries) :

            ## Determine the position offsets due to background velocity

            orig_time = orig_time_list[entry_index]

            pos_offset = (-1.0) * (orig_time - background_vel_time_offset) \
                         * background_vel

            ## Convert to nanoseconds and microns from seconds and centimeters

            new_time = orig_time * 1.0e+09

            new_envelope_min = (orig_envelope_min_list[entry_index]
                                + pos_offset) * 1.0e+04
            
            new_envelope_max = (orig_envelope_max_list[entry_index]
                                + pos_offset) * 1.0e+04

            time_list.append(new_time)

            envelope_min_list.append(new_envelope_min)
            envelope_max_list.append(new_envelope_max)

        bounds_plot.add_plot(time_list, envelope_min_list,
                             legend_text=(sim_name + " min"))
            
        bounds_plot.add_plot(time_list, envelope_max_list,
                             legend_text=(sim_name + " max"),
                             use_prev_color=True)

    ## Save the plots

    width_plot_filename = output_prefix + "layer_width.svg"
    bounds_plot_filename = output_prefix + "layer_bounds.svg"

    width_plot.show_plot(width_plot_filename)
    bounds_plot.show_plot(bounds_plot_filename)

if (__name__ == "__main__") :
    main()

