#!/usr/bin/env python

# get_layer_center.py

# By Nathan C. Hearn
#    August 26, 2008
#
# Determines layer center from layer_distrib output using different criteria

# Make sure the QuickFlash/utils directory is in PYTHONPATH!!!


try :
    import psyco
except ImportError :
    pass

class TimestepData :

    def __init__(self, metadata_file=None) :

        self.__datafile_name = None

        self.__sim_time = None

        self.__envelope_min_pos = None
        self.__envelope_max_pos = None
        self.__envelope_center = None
        self.__envelope_width = None

        self.__layer_min_pos = None
        self.__layer_max_pos = None
        self.__layer_center = None
        self.__layer_width = None

        self.__mass_center = None
        self.__mass_fraction_center = None

        if (metadata_file is not None) :
            self.reset(metadata_file)

    def reset(self, metadata_file=None) :

        from metadata_reader import parse_distrib_metadata

        self.__datafile_name = None

        self.__sim_time = None

        self.__envelope_min_pos = None
        self.__envelope_max_pos = None
        self.__envelope_center = None
        self.__envelope_width = None

        self.__layer_min_pos = None
        self.__layer_max_pos = None
        self.__layer_center = None
        self.__layer_width = None

        self.__mass_center = None
        self.__mass_fraction_center = None

        if (metadata_file is not None) :

            file_data, slice_info = parse_distrib_metadata(metadata_file)

            self.__datafile_name = file_data.get_group_name()
            self.__sim_time = file_data.get_value("SimTime")
            
            self.__envelope_min_pos = file_data.get_value("MinEnvelopePos")
            self.__envelope_max_pos = file_data.get_value("MaxEnvelopePos")

            self.__envelope_center = 0.5 * (self.__envelope_min_pos
                                            + self.__envelope_max_pos)

            self.__envelope_width = self.__envelope_max_pos \
                                    - self.__envelope_min_pos

            self.__layer_min_pos = file_data.get_value("LayerMinPos")
            self.__layer_max_pos = file_data.get_value("LayerMaxPos")

            self.__layer_center \
                = 0.5 * (self.__layer_min_pos + self.__layer_max_pos)

            self.__layer_width = self.__layer_max_pos - self.__layer_min_pos

            distribution_file = file_data.get_value("LayerDistribFile")

            self.__read_distrib_file(distribution_file)


    def get_datafile_name(self) :
        return self.__datafile_name

    def get_sim_time(self) :
        return self.__sim_time

    def get_envelope_min_pos(self) :
        return self.__envelope_min_pos

    def get_envelope_max_pos(self) :
        return self.__envelope_max_pos

    def get_envelope_center(self) :
        return self.__envelope_center

    def get_envelope_width(self) :
        return self.__envelope_width

    def get_layer_min_pos(self) :
        return self.__layer_min_pos

    def get_layer_max_pos(self) :
        return self.__layer_max_pos

    def get_layer_center(self) :
        return self.__layer_center

    def get_layer_width(self) :
        return self.__layer_width

    def get_mass_center(self) :
        return self.__mass_center

    def get_mass_fract_center(self) :
        return self.__mass_fract_center
    

    def __read_distrib_file(self, distribution_file) :

        from plot_gen import read_table

        table_data = read_table(distribution_file, table_start_line="BinData")

        bin_center = table_data["Bin_Center"]
        bin_mass = table_data["Bin_Mass"]
        bin_foam_mass = table_data["Bin_Foam_Mass"]

        num_bins = len(bin_center)

        if (num_bins < 1) :
            raise RuntimeError("No distribution data present")

        accum_pos_mass_weight = float(0.0)
        accum_mass_weight = float(0.0)

        accum_pos_mass_fract_weight = float(0.0)
        accum_mass_fract_weight = float(0.0)

        for bin_index in range(num_bins) :

            pos = bin_center[bin_index]

            mass = bin_mass[bin_index]

            if (mass > 0.0) :

                foam_mass = bin_foam_mass[bin_index]
            
                poly_mass = mass - foam_mass

                poly_mass_fract = poly_mass / mass

                mass_weight = foam_mass * poly_mass

                if (mass_weight < 0.0) :
                    mass_weight = 0.0
                
                mass_fract_weight = poly_mass_fract * (1.0 - poly_mass_fract)

                if (mass_fract_weight < 0.0) :
                    mass_fract_weight = 0.0

                accum_pos_mass_weight += pos * mass_weight
                accum_mass_weight += mass_weight

                accum_pos_mass_fract_weight += pos * mass_fract_weight
                accum_mass_fract_weight += mass_fract_weight

        if (accum_mass_weight > 0.0) :
            self.__mass_center = accum_pos_mass_weight / accum_mass_weight
        else :
            self.__mass_center = None

        if (accum_mass_fract_weight > 0.0) :
            self.__mass_fract_center = accum_pos_mass_fract_weight \
                                       / accum_mass_fract_weight
        else :
            self.__mass_fract_center = None


def main() :

    from sys import exit, stderr, stdout
    from os import path
    from optparse import OptionParser
    
    from metadata_reader import parse_distrib_metadata

    ## Set up the command-line parser

    usage = str("%prog [options] metadata_file_1 [ metadata_file_2 ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    background_vel_str = str("background-vel")
    argparser.add_option("--" + background_vel_str, type="string",
                         dest="background_vel",
                         help=("Background velocity (with optional "
                               + "comma-separated time offset)"))
    argparser.set_defaults(background_vel=None)

    ## Process the command-line input

    opts, args = argparser.parse_args()

    argc = len(args)

    min_args = int(1)

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n")
        exit(1)

    ## Get the required arguments

    argptr = int(0)

    metadata_file_list = list()

    while (argptr < argc) :
        metadata_file_list.append(str(args[argptr]))
        argptr += 1

    ## Get the optional arguments

    background_vel = float(0.0)
    background_vel_time_offset = float(0.0)

    if (opts.background_vel is not None) :

        background_vel_words = opts.background_vel.split(",")

        num_background_vel_words = len(background_vel_words)

        if ((num_background_vel_words < 1) or (num_background_vel_words > 2)) :
            stderr.write("\nError: Background velocity requires one or two "
                         + "comma-separated entries\n\n")
            exit(-1)


        background_vel = float(background_vel_words[0])

        if (num_background_vel_words == 2) :
            background_vel_time_offset = float(background_vel_words[1])

    ## Collect and output the data

    stdout.write("Data_File" + "\t" + "Sim_Time" + "\t" + "Layer_Min_Pos"
                 + "\t" + "Layer_Max_Pos" + "\t" + "Layer_Center" + "\t"
                 + "Layer_Mass_Center" + "\t" + "Layer_Mass_Fraction_Center\n")

    for metadata_file in metadata_file_list :

        timestep_data = TimestepData(metadata_file)

        datafile = timestep_data.get_datafile_name()
        sim_time = timestep_data.get_sim_time()

        background_offset = background_vel * (sim_time
                                              - background_vel_time_offset)

        layer_min_pos = timestep_data.get_layer_min_pos() - background_offset
        layer_max_pos = timestep_data.get_layer_max_pos() - background_offset

        layer_center = timestep_data.get_layer_center() - background_offset

        mass_center = timestep_data.get_mass_center() - background_offset
        
        mass_fract_center = timestep_data.get_mass_fract_center() \
                            - background_offset

        stdout.write(datafile + "\t" + str(sim_time) + "\t"
                     + str(layer_min_pos) + "\t" + str(layer_max_pos) + "\t"
                     + str(layer_center) + "\t" + str(mass_center) + "\t"
                     + str(mass_fract_center) + "\n")

    stdout.write("\n")

if (__name__ == "__main__") :
    main()

