// C++ program file layer_distribution.cpp

/*
  By Nathan C. Hearn
     April 3, 2008

  Container for layer distribution data.
*/


#include "layer_distribution.hpp"
#include <vector>
#include <algorithm>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_except.hpp"
#include "quickflash_stats_bins.hpp"
#include "quickflash_tree_block.hpp"
#include "file_defs.hpp"


// Class LayerDistribution

LayerDistribution::LayerDistribution() :
  total_mass(0.0), poly_mass(0.0), chbr_mass(0.0), foam_mass(0.0),
  bin_array_min_coord(0.0), bin_array_max_coord(0.0), num_bins(0), 
  bin_centers(), 
  total_mass_bins(), poly_mass_bins(), chbr_mass_bins(), foam_mass_bins(),
  total_dens_variance_bins(), poly_dens_variance_bins(), 
  chbr_dens_variance_bins(), foam_dens_variance_bins()
  { }


LayerDistribution::LayerDistribution(const QuickFlash::File::DataFile & df, 
				     const unsigned int search_axis,
				     const double sample_bin_width,
				     const double min_coord, 
				     const double max_coord,
				     const bool limit_sample_radius,
				     const double max_sample_radius) :
  total_mass(0.0), poly_mass(0.0), chbr_mass(0.0), foam_mass(0.0),
  bin_array_min_coord(0.0), bin_array_max_coord(0.0), num_bins(0),
  bin_centers(), 
  total_mass_bins(), poly_mass_bins(), chbr_mass_bins(), foam_mass_bins(),
  total_dens_variance_bins(), poly_dens_variance_bins(), 
  chbr_dens_variance_bins(), foam_dens_variance_bins()
  { 
  reset(df, search_axis, sample_bin_width, min_coord, max_coord, 
	limit_sample_radius, max_sample_radius);
  }


LayerDistribution::LayerDistribution(const LayerDistribution & source) :
  total_mass(0.0), poly_mass(0.0), chbr_mass(0.0), foam_mass(0.0),
  bin_array_min_coord(0.0), bin_array_max_coord(0.0), num_bins(0),
  bin_centers(), 
  total_mass_bins(), poly_mass_bins(), chbr_mass_bins(), foam_mass_bins(),
  total_dens_variance_bins(), poly_dens_variance_bins(), 
  chbr_dens_variance_bins(), foam_dens_variance_bins()
  { reset(source); }


void LayerDistribution::reset()
  {
  total_mass = 0.0;
  poly_mass = 0.0;
  chbr_mass = 0.0;
  foam_mass = 0.0;

  bin_array_min_coord = 0.0;
  bin_array_max_coord = 0.0;

  num_bins = 0;

  bin_centers.clear();

  total_mass_bins.clear();
  poly_mass_bins.clear();
  chbr_mass_bins.clear();
  foam_mass_bins.clear();

  total_dens_variance_bins.clear();
  poly_dens_variance_bins.clear();
  chbr_dens_variance_bins.clear();
  foam_dens_variance_bins.clear();
  }


void LayerDistribution::reset(const QuickFlash::File::DataFile & df, 
			      const unsigned int search_axis,
			      const double sample_bin_width,
			      const double min_coord, const double max_coord,
			      const bool limit_sample_radius,
			      const double max_sample_radius)
  {
  reset();

  if (!(sample_bin_width > 0.0))
    throw QuickFlash::Except("Bin width must be positive", __FILE__, __LINE__);

  if (!(min_coord < max_coord))
    throw QuickFlash::Except("Minimum coordinate must be less than maximum",
			   __FILE__, __LINE__);

  if (limit_sample_radius)
    if (!(max_sample_radius > 0.0))
      throw QuickFlash::Except("Maximum sample radius must be positive", 
			     __FILE__, __LINE__);

  const double sq_max_radius = max_sample_radius * max_sample_radius;


  // Get the mesh info

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (search_axis >= dims)
    throw QuickFlash::Except("Search axis index out of range", __FILE__, 
			   __LINE__);

  const std::vector<double> & sim_min_bounds = meshinfo.get_volume_minbounds();
  const std::vector<double> & sim_max_bounds = meshinfo.get_volume_maxbounds();

  std::vector<double> sim_center(dims);

  for (unsigned int i = 0; i < dims; i++)
    sim_center[i] = 0.5 * (sim_min_bounds[i] + sim_max_bounds[i]);


  // Get the datasets

  const QuickFlash::File::Dataset & dens_set = df.get_dataset(dens_dataset_name);

  const QuickFlash::File::Dataset & poly_set = df.get_dataset(poly_dataset_name);
  // const QuickFlash::File::Dataset & chbr_set = df.get_dataset(chbr_dataset_name);
  const QuickFlash::File::Dataset & foam_set = df.get_dataset(foam_dataset_name);


  // Set up the bins -- use center specified by min_coord, max_coord

  const double min_array_width = max_coord - min_coord;

  num_bins = static_cast<unsigned int>(min_array_width / sample_bin_width) + 1;

  const double array_width = static_cast<float>(num_bins) * sample_bin_width;

  const double bin_array_center = 0.5 * (min_coord + max_coord);

  const double half_array_width = 0.5 * array_width;

  bin_array_min_coord = bin_array_center - half_array_width;
  bin_array_max_coord = bin_array_center + half_array_width;

  QuickFlash::Stats::BinArray total_dens_array(bin_array_min_coord,
					     num_bins, sample_bin_width);

  QuickFlash::Stats::BinArray poly_dens_array(bin_array_min_coord,
					    num_bins, sample_bin_width);

  QuickFlash::Stats::BinArray chbr_dens_array(bin_array_min_coord,
					    num_bins, sample_bin_width);

  QuickFlash::Stats::BinArray foam_dens_array(bin_array_min_coord,
					    num_bins, sample_bin_width);

  // Run through the data

  QuickFlash::Block::BlockData<double> dens_data;

  QuickFlash::Block::BlockData<double> poly_data;
  // QuickFlash::Block::BlockData<double> chbr_data;  // INFER FOR NOW!!
  QuickFlash::Block::BlockData<double> foam_data;

  std::vector<double> cell_pos;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    const double block_min_pos = block_info.get_min_coords()[search_axis];
    const double block_max_pos = block_info.get_max_coords()[search_axis];

    bool skip_branch = false;

    if ((bin_array_min_coord > block_max_pos)
	|| (bin_array_max_coord < block_min_pos))
      skip_branch = true;

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      const double cell_volume = block_info.get_cell_volume();  // CARTESIAN!!
      const double cell_axis_width = block_info.get_cell_width()[search_axis];

      const unsigned int num_cells = block_info.get_num_cells();

      bool block_loaded = false;

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	block_info.get_cell_center(cell_index, cell_pos);

	bool use_cell = true;

	if (limit_sample_radius)
	  {
	  double sq_cell_pos_radius = 0.0;

	  for (unsigned int i = 0; i < dims; i++)
	    if (i != search_axis)
	      {
	      const double axis_comp = cell_pos[i] - sim_center[i];

	      sq_cell_pos_radius += axis_comp * axis_comp;
	      }

	  if (sq_cell_pos_radius > sq_max_radius)
	    use_cell = false;
	  }

	if (use_cell)
	  {
	  if (!block_loaded)
	    {
	    dens_set.get_block_data(block_index, dens_data);

	    poly_set.get_block_data(block_index, poly_data);
	    // chbr_set.get_block_data(block_index, chbr_data);
	    foam_set.get_block_data(block_index, foam_data);

	    block_loaded = true;
	    }

	  const double search_axis_pos = cell_pos[search_axis];

	  const double cell_dens = dens_data[cell_index];

	  const double poly_dens = poly_data[cell_index] * cell_dens;
	  // const double chbr_dens = chbr_data[cell_index] * cell_dens;
	  const double foam_dens = foam_data[cell_index] * cell_dens;

	  // FOR NOW, ASSUME CHBR IS [ TOTAL - (POLY + FOAM) ]

	  const double chbr_dens 
	    = std::max(cell_dens - (poly_dens + foam_dens), 0.0);

	  // WE SHOULD CLIP THE CELLS CROSSING MIN_COORD OR MAX_COORD!!!

	  total_dens_array.add_data_spread(search_axis_pos, cell_dens,
					   cell_axis_width, cell_volume);
					   
	  poly_dens_array.add_data_spread(search_axis_pos, poly_dens,
					  cell_axis_width, cell_volume);
	  chbr_dens_array.add_data_spread(search_axis_pos, chbr_dens,
					  cell_axis_width, cell_volume);
	  foam_dens_array.add_data_spread(search_axis_pos, foam_dens,
					  cell_axis_width, cell_volume);
	  }
	}
      }

    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  // Build the member arrays

  bin_centers = total_dens_array.get_bin_centers();

  total_mass_bins.resize(num_bins, 0.0);
  poly_mass_bins.resize(num_bins, 0.0);
  chbr_mass_bins.resize(num_bins, 0.0);
  foam_mass_bins.resize(num_bins, 0.0);

  total_dens_variance_bins.resize(num_bins, 0.0);
  poly_dens_variance_bins.resize(num_bins, 0.0);
  chbr_dens_variance_bins.resize(num_bins, 0.0);
  foam_dens_variance_bins.resize(num_bins, 0.0);

  double array_sum_total_mass = 0.0;
  double array_sum_poly_mass = 0.0;
  double array_sum_chbr_mass = 0.0;
  double array_sum_foam_mass = 0.0;


  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & total_bin = total_dens_array[bin_index];

    const QuickFlash::Stats::Bin & poly_bin = poly_dens_array[bin_index];
    const QuickFlash::Stats::Bin & chbr_bin = chbr_dens_array[bin_index];
    const QuickFlash::Stats::Bin & foam_bin = foam_dens_array[bin_index];

    // The total mass is the mean value times the weight (volume) of the bin

    if (total_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = total_bin.get_weight_value_sum();

      array_sum_total_mass += bin_mass;

      total_mass_bins[bin_index] = bin_mass;
      total_dens_variance_bins[bin_index] = total_bin.get_pop_variance();
      }

    if (poly_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = poly_bin.get_weight_value_sum();

      array_sum_poly_mass += bin_mass;

      poly_mass_bins[bin_index] = bin_mass;
      poly_dens_variance_bins[bin_index] = poly_bin.get_pop_variance();
      }

    if (chbr_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = chbr_bin.get_weight_value_sum();

      array_sum_chbr_mass += bin_mass;

      chbr_mass_bins[bin_index] = bin_mass;
      chbr_dens_variance_bins[bin_index] = chbr_bin.get_pop_variance();
      }

    if (foam_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = foam_bin.get_weight_value_sum();

      array_sum_foam_mass += bin_mass;

      foam_mass_bins[bin_index] = bin_mass;
      foam_dens_variance_bins[bin_index] = foam_bin.get_pop_variance();
      }
    }

  total_mass = array_sum_total_mass;
  poly_mass = array_sum_poly_mass;
  chbr_mass = array_sum_chbr_mass;
  foam_mass = array_sum_foam_mass;
  }
  

void LayerDistribution::reset(const LayerDistribution & source)
  {
  if (&source != this)
    {
    total_mass = source.total_mass;
    poly_mass = source.poly_mass;
    chbr_mass = source.chbr_mass;
    foam_mass = source.foam_mass;

    bin_array_min_coord = source.bin_array_min_coord;
    bin_array_max_coord = source.bin_array_max_coord;

    num_bins = source.num_bins;

    bin_centers = source.bin_centers;

    total_mass_bins = source.total_mass_bins;
    poly_mass_bins = source.poly_mass_bins;
    chbr_mass_bins = source.chbr_mass_bins;
    foam_mass_bins = source.foam_mass_bins;

    total_dens_variance_bins = source.total_dens_variance_bins;
    poly_dens_variance_bins = source.poly_dens_variance_bins;
    chbr_dens_variance_bins = source.chbr_dens_variance_bins;
    foam_dens_variance_bins = source.foam_dens_variance_bins;
    }
  }





