// C++ program file layer_slices.cpp

/*
  By Nathan C. Hearn
     April 5, 2008

  All-inclusive 2D/3D data slicing tools for Omega laser experiment 
  simulations.
*/


#include "layer_slices.hpp"
#include <vector>
#include <algorithm>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_except.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_stats_bins.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_tree_block.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_utils.hpp"
#include "quickflash_geometry.hpp"
#include "file_defs.hpp"


// Class LayerSlice1D

LayerSlice1D::LayerSlice1D() :   
  total_mass(0.0), poly_mass(0.0), chbr_mass(0.0), foam_mass(0.0),
  bin_array_min_coord(0.0), bin_array_max_coord(0.0), num_bins(0), 
  bin_centers(), 
  total_mass_bins(), poly_mass_bins(), chbr_mass_bins(), foam_mass_bins(),
  total_dens_variance_bins(), poly_dens_variance_bins(), 
  chbr_dens_variance_bins(), foam_dens_variance_bins()
  { }


LayerSlice1D::LayerSlice1D(const QuickFlash::File::DataFile & df,
			   const unsigned int search_axis,
			   const unsigned int slice_axis,
			   const double search_axis_center_coord,
			   const double slice_width,
			   const double accum_bin_width,
			   const bool limit_sample_radius,
			   const double max_sample_radius) :
  total_mass(0.0), poly_mass(0.0), chbr_mass(0.0), foam_mass(0.0),
  bin_array_min_coord(0.0), bin_array_max_coord(0.0), num_bins(0), 
  bin_centers(), 
  total_mass_bins(), poly_mass_bins(), chbr_mass_bins(), foam_mass_bins(),
  total_dens_variance_bins(), poly_dens_variance_bins(), 
  chbr_dens_variance_bins(), foam_dens_variance_bins()
  { 
  reset(df, search_axis, slice_axis, search_axis_center_coord, slice_width,
	accum_bin_width, limit_sample_radius, max_sample_radius);
  }


LayerSlice1D::LayerSlice1D(const LayerSlice1D & source) :
  total_mass(0.0), poly_mass(0.0), chbr_mass(0.0), foam_mass(0.0),
  bin_array_min_coord(0.0), bin_array_max_coord(0.0), num_bins(0), 
  bin_centers(), 
  total_mass_bins(), poly_mass_bins(), chbr_mass_bins(), foam_mass_bins(),
  total_dens_variance_bins(), poly_dens_variance_bins(), 
  chbr_dens_variance_bins(), foam_dens_variance_bins()
  { reset(source); }


void LayerSlice1D::reset()
  {
  total_mass = 0.0;
  poly_mass = 0.0;
  chbr_mass = 0.0;
  foam_mass = 0.0;

  bin_array_min_coord = 0.0;
  bin_array_max_coord = 0.0;

  num_bins = 0;

  bin_centers.clear();

  total_mass_bins.clear();
  poly_mass_bins.clear();
  chbr_mass_bins.clear();
  foam_mass_bins.clear();

  total_dens_variance_bins.clear();
  poly_dens_variance_bins.clear();
  chbr_dens_variance_bins.clear();
  foam_dens_variance_bins.clear();
  }


void LayerSlice1D::reset(const QuickFlash::File::DataFile & df,
			 const unsigned int search_axis,
			 const unsigned int slice_axis,
			 const double search_axis_center_coord,
			 const double slice_width,
			 const double accum_bin_width,
			 const bool limit_sample_radius,
			 const double max_sample_radius)
  {
  reset();

  if (search_axis == slice_axis)
    throw QuickFlash::Except("Slice axis and search axis must be different");

  if (!(slice_width > 0.0))
    throw QuickFlash::Except("Slice width must be positive", __FILE__, __LINE__);

  if (!(accum_bin_width > 0.0))
    throw QuickFlash::Except("Bin width must be positive", __FILE__, __LINE__);

  if (limit_sample_radius)
    if (!(max_sample_radius > 0.0))
      throw QuickFlash::Except("Maximum sample radius must be positive", 
			     __FILE__, __LINE__);

  const double sq_max_radius = max_sample_radius * max_sample_radius;


  // Get the mesh info

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (search_axis >= dims)
    throw QuickFlash::Except("Search axis index out of range", __FILE__, 
			   __LINE__);

  if (slice_axis >= dims)
    throw QuickFlash::Except("Slice axis index out of range", __FILE__, 
			   __LINE__);

  const std::vector<double> & sim_min_bounds = meshinfo.get_volume_minbounds();
  const std::vector<double> & sim_max_bounds = meshinfo.get_volume_maxbounds();

  std::vector<double> sim_center(dims);

  for (unsigned int i = 0; i < dims; i++)
    sim_center[i] = 0.5 * (sim_min_bounds[i] + sim_max_bounds[i]);

  const double slice_axis_sim_center = sim_center[slice_axis];

  const double half_slice_width = 0.5 * slice_width;

  const double min_search_axis_coord 
    = search_axis_center_coord - half_slice_width;

  const double max_search_axis_coord 
    = search_axis_center_coord + half_slice_width;


  // Get the datasets

  const QuickFlash::File::Dataset & dens_set = df.get_dataset(dens_dataset_name);

  const QuickFlash::File::Dataset & poly_set = df.get_dataset(poly_dataset_name);
  // const QuickFlash::File::Dataset & chbr_set = df.get_dataset(chbr_dataset_name);
  const QuickFlash::File::Dataset & foam_set = df.get_dataset(foam_dataset_name);


  // Set up the bins

  const double min_slice_coord 
    = limit_sample_radius ? (slice_axis_sim_center - max_sample_radius) 
      : sim_min_bounds[slice_axis];

  const double max_slice_coord 
    = limit_sample_radius ? (slice_axis_sim_center + max_sample_radius) 
      : sim_max_bounds[slice_axis];

  const double min_slice_length = max_slice_coord - min_slice_coord;

  num_bins = static_cast<unsigned int>(min_slice_length / accum_bin_width) + 1;

  const double slice_length = static_cast<float>(num_bins) * accum_bin_width;

  const double bin_array_center = 0.5 * (min_slice_coord + max_slice_coord);

  const double half_slice_length = 0.5 * slice_length;

  bin_array_min_coord = bin_array_center - half_slice_length;
  bin_array_max_coord = bin_array_center + half_slice_length;

  QuickFlash::Stats::BinArray total_dens_array(bin_array_min_coord,
					     num_bins, accum_bin_width);

  QuickFlash::Stats::BinArray poly_dens_array(bin_array_min_coord,
					    num_bins, accum_bin_width);

  QuickFlash::Stats::BinArray chbr_dens_array(bin_array_min_coord,
					    num_bins, accum_bin_width);

  QuickFlash::Stats::BinArray foam_dens_array(bin_array_min_coord,
					    num_bins, accum_bin_width);


  // Run through the data

  QuickFlash::Block::BlockData<double> dens_data;

  QuickFlash::Block::BlockData<double> poly_data;
  // QuickFlash::Block::BlockData<double> chbr_data;  // INFER FOR NOW!!
  QuickFlash::Block::BlockData<double> foam_data;

  std::vector<double> cell_pos;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    const double block_search_axis_min_pos 
      = block_info.get_min_coords()[search_axis];

    const double block_search_axis_max_pos 
      = block_info.get_max_coords()[search_axis];

    const double block_slice_axis_min_pos 
      = block_info.get_min_coords()[slice_axis];

    const double block_slice_axis_max_pos 
      = block_info.get_max_coords()[slice_axis];

    bool skip_branch = false;

    if ((min_search_axis_coord > block_search_axis_max_pos)
	|| (max_search_axis_coord < block_search_axis_min_pos))
      skip_branch = true;
    else if ((bin_array_min_coord > block_slice_axis_max_pos)
	|| (bin_array_max_coord < block_slice_axis_min_pos))
      skip_branch = true;

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      const double cell_volume = block_info.get_cell_volume();  // CARTESIAN!!

      const double cell_search_axis_width 
	= block_info.get_cell_width()[search_axis];

      const double oneOver_cell_search_axis_width 
	= 1.0 / cell_search_axis_width;

      const double cell_search_axis_half_width = 0.5 * cell_search_axis_width;

      const double cell_slice_axis_width 
	= block_info.get_cell_width()[slice_axis];

      const unsigned int num_cells = block_info.get_num_cells();

      bool block_loaded = false;

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	block_info.get_cell_center(cell_index, cell_pos);

	bool use_cell = true;

	if (limit_sample_radius)
	  {
	  double sq_cell_pos_radius = 0.0;

	  for (unsigned int i = 0; i < dims; i++)
	    if (i != search_axis)
	      {
	      const double axis_comp = cell_pos[i] - sim_center[i];

	      sq_cell_pos_radius += axis_comp * axis_comp;
	      }

	  if (sq_cell_pos_radius > sq_max_radius)
	    use_cell = false;
	  }

	if (use_cell)
	  {
	  const double search_axis_pos = cell_pos[search_axis];
	  const double slice_axis_pos = cell_pos[slice_axis];

	  // Determine the fraction of overlap along the search axis

	  const double cell_search_axis_min 
	    = search_axis_pos - cell_search_axis_half_width;

	  const double cell_search_axis_max
	    = search_axis_pos + cell_search_axis_half_width;

	  const double search_axis_overlap_min 
	    = std::max(cell_search_axis_min, min_search_axis_coord);

	  const double search_axis_overlap_max
	    = std::min(cell_search_axis_max, max_search_axis_coord);

	  if (search_axis_overlap_min < search_axis_overlap_max)
	    {
	    const double overlap_fraction 
	      = (search_axis_overlap_max - search_axis_overlap_min) 
	        * oneOver_cell_search_axis_width;

	    const double overlap_cell_volume = overlap_fraction * cell_volume;

	    if (!block_loaded)
	      {
	      dens_set.get_block_data(block_index, dens_data);

	      poly_set.get_block_data(block_index, poly_data);
	      // chbr_set.get_block_data(block_index, chbr_data);
	      foam_set.get_block_data(block_index, foam_data);

	      block_loaded = true;
	      }

	    const double cell_dens = dens_data[cell_index];

	    const double poly_dens = poly_data[cell_index] * cell_dens;
	    // const double chbr_dens = chbr_data[cell_index] * cell_dens;
	    const double foam_dens = foam_data[cell_index] * cell_dens;

	    // FOR NOW, ASSUME CHBR IS [ TOTAL - (POLY + FOAM) ]

	    const double chbr_dens 
	      = std::max(cell_dens - (poly_dens + foam_dens), 0.0);

	    // WE SHOULD CLIP THE CELLS CROSSING MIN/MAX_SLICE_COORD!!!

	    total_dens_array.add_data_spread(slice_axis_pos, cell_dens,
					     cell_slice_axis_width, 
					     overlap_cell_volume);
					   
	    poly_dens_array.add_data_spread(slice_axis_pos, poly_dens,
					    cell_slice_axis_width, 
					    overlap_cell_volume);
					   
	    chbr_dens_array.add_data_spread(slice_axis_pos, chbr_dens,
					    cell_slice_axis_width, 
					    overlap_cell_volume);
					   
	    foam_dens_array.add_data_spread(slice_axis_pos, foam_dens,
					    cell_slice_axis_width, 
					    overlap_cell_volume);
	    }
	  }
	}
      }

    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }


  // Build the member arrays
  
  bin_centers = total_dens_array.get_bin_centers();

  total_mass_bins.resize(num_bins, 0.0);
  poly_mass_bins.resize(num_bins, 0.0);
  chbr_mass_bins.resize(num_bins, 0.0);
  foam_mass_bins.resize(num_bins, 0.0);

  total_dens_variance_bins.resize(num_bins, 0.0);
  poly_dens_variance_bins.resize(num_bins, 0.0);
  chbr_dens_variance_bins.resize(num_bins, 0.0);
  foam_dens_variance_bins.resize(num_bins, 0.0);

  double array_sum_total_mass = 0.0;
  double array_sum_poly_mass = 0.0;
  double array_sum_chbr_mass = 0.0;
  double array_sum_foam_mass = 0.0;

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & total_bin = total_dens_array[bin_index];

    const QuickFlash::Stats::Bin & poly_bin = poly_dens_array[bin_index];
    const QuickFlash::Stats::Bin & chbr_bin = chbr_dens_array[bin_index];
    const QuickFlash::Stats::Bin & foam_bin = foam_dens_array[bin_index];

    // The total mass is the mean value times the weight (volume) of the bin

    if (total_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = total_bin.get_weight_value_sum();

      array_sum_total_mass += bin_mass;

      total_mass_bins[bin_index] = bin_mass;
      total_dens_variance_bins[bin_index] = total_bin.get_pop_variance();
      }

    if (poly_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = poly_bin.get_weight_value_sum();

      array_sum_poly_mass += bin_mass;

      poly_mass_bins[bin_index] = bin_mass;
      poly_dens_variance_bins[bin_index] = poly_bin.get_pop_variance();
      }

    if (chbr_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = chbr_bin.get_weight_value_sum();

      array_sum_chbr_mass += bin_mass;

      chbr_mass_bins[bin_index] = bin_mass;
      chbr_dens_variance_bins[bin_index] = chbr_bin.get_pop_variance();
      }

    if (foam_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = foam_bin.get_weight_value_sum();

      array_sum_foam_mass += bin_mass;

      foam_mass_bins[bin_index] = bin_mass;
      foam_dens_variance_bins[bin_index] = foam_bin.get_pop_variance();
      }
    }

  total_mass = array_sum_total_mass;
  poly_mass = array_sum_poly_mass;
  chbr_mass = array_sum_chbr_mass;
  foam_mass = array_sum_foam_mass;
  }


void LayerSlice1D::reset(const LayerSlice1D & source)
  {
  if (&source != this)
    {
    total_mass = source.total_mass;
    poly_mass = source.poly_mass;
    chbr_mass = source.chbr_mass;
    foam_mass = source.foam_mass;

    bin_array_min_coord = source.bin_array_min_coord;
    bin_array_max_coord = source.bin_array_max_coord;

    num_bins = source.num_bins;

    bin_centers = source.bin_centers;

    total_mass_bins = source.total_mass_bins;
    poly_mass_bins = source.poly_mass_bins;
    chbr_mass_bins = source.chbr_mass_bins;
    foam_mass_bins = source.foam_mass_bins;

    total_dens_variance_bins = source.total_dens_variance_bins;
    poly_dens_variance_bins = source.poly_dens_variance_bins;
    chbr_dens_variance_bins = source.chbr_dens_variance_bins;
    foam_dens_variance_bins = source.foam_dens_variance_bins;
    }
  }


// Class LayerSlice2D

LayerSlice2D::LayerSlice2D() :
  total_mass(0.0), poly_mass(0.0), chbr_mass(0.0), foam_mass(0.0),
  grid_info(), slice_to_file_index_map(),
  total_mass_bins(), poly_mass_bins(), chbr_mass_bins(), foam_mass_bins()
  { }


LayerSlice2D::LayerSlice2D(const QuickFlash::File::DataFile & df, 
			   const unsigned int search_axis,
			   const double search_axis_center_coord,
			   const double slice_width,
			   const std::vector<double> & accum_bin_width,
			   const bool limit_sample_radius,
			   const double max_sample_radius) :
  total_mass(0.0), poly_mass(0.0), chbr_mass(0.0), foam_mass(0.0),
  grid_info(), slice_to_file_index_map(),
  total_mass_bins(), poly_mass_bins(), chbr_mass_bins(), foam_mass_bins()
  { 
  reset(df, search_axis, search_axis_center_coord, slice_width, 
	accum_bin_width, limit_sample_radius, max_sample_radius);
  }


LayerSlice2D::LayerSlice2D(const LayerSlice2D & source) :
  total_mass(0.0), poly_mass(0.0), chbr_mass(0.0), foam_mass(0.0),
  grid_info(), slice_to_file_index_map(),
  total_mass_bins(), poly_mass_bins(), chbr_mass_bins(), foam_mass_bins()
  { reset(source); }


void LayerSlice2D::reset()
  {
  total_mass = 0.0;
  poly_mass = 0.0;
  chbr_mass = 0.0;
  foam_mass = 0.0;

  grid_info.reset();

  slice_to_file_index_map.clear();

  total_mass_bins.reset();
  poly_mass_bins.reset();
  chbr_mass_bins.reset();
  foam_mass_bins.reset();
  }


void LayerSlice2D::reset(const QuickFlash::File::DataFile & df, 
			 const unsigned int search_axis,
			 const double search_axis_center_coord,
			 const double slice_width,
			 const std::vector<double> & accum_bin_width,
			 const bool limit_sample_radius,
			 const double max_sample_radius)
  {
  const unsigned int slice_dims = 2;
  const unsigned int file_dims = 3;  // Required

  if (search_axis >= file_dims)
    throw QuickFlash::Except("Search axis index out of range", __FILE__, 
			   __LINE__);

  if (!(slice_width > 0.0))
    throw QuickFlash::Except("Slice width must be positive", __FILE__, __LINE__);

  if (accum_bin_width.size() != slice_dims)
    throw QuickFlash::Except("Bin width vector must have two elements", __FILE__,
			   __LINE__);

  for (unsigned int i = 0; i < slice_dims; i++)
    if (!(accum_bin_width[i] > 0.0))
      throw QuickFlash::Except("Bin width elements must be positive", __FILE__,
			     __LINE__);

  if (limit_sample_radius)
    if (!(max_sample_radius > 0.0))
      throw QuickFlash::Except("Maximum sample radius must be positive", 
			     __FILE__, __LINE__);

  const double sq_max_radius = max_sample_radius * max_sample_radius;

  
  // Get the 

  slice_to_file_index_map.resize(slice_dims);

  switch (search_axis)
    {
    case 0 :
      slice_to_file_index_map[0] = 1;  // File y-axis -> slice x-axis
      slice_to_file_index_map[1] = 2;  // File z-axis -> slice y-axis
      break;

    case 1 :
      slice_to_file_index_map[0] = 0;
      slice_to_file_index_map[1] = 2;
      break;

    case 2 :
      slice_to_file_index_map[0] = 0;
      slice_to_file_index_map[1] = 2;
      break;
    }


  // Get the mesh info

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  if (meshinfo.get_dims() != file_dims)
    throw QuickFlash::Except("Only 3D datasets supported for 2D slices", 
			   __FILE__, __LINE__);

  const std::vector<double> & sim_min_bounds = meshinfo.get_volume_minbounds();
  const std::vector<double> & sim_max_bounds = meshinfo.get_volume_maxbounds();

  std::vector<double> sim_center(file_dims);

  for (unsigned int i = 0; i < file_dims; i++)
    sim_center[i] = 0.5 * (sim_min_bounds[i] + sim_max_bounds[i]);

  const double half_slice_width = 0.5 * slice_width;

  const double min_search_axis_coord 
    = search_axis_center_coord - half_slice_width;

  const double max_search_axis_coord 
    = search_axis_center_coord + half_slice_width;


  // Get the datasets

  const QuickFlash::File::Dataset & dens_set = df.get_dataset(dens_dataset_name);

  const QuickFlash::File::Dataset & poly_set = df.get_dataset(poly_dataset_name);
  // const QuickFlash::File::Dataset & chbr_set = df.get_dataset(chbr_dataset_name);
  const QuickFlash::File::Dataset & foam_set = df.get_dataset(foam_dataset_name);


  // Set up the bins

  std::vector<double> slice_min_coords(slice_dims);
  std::vector<double> slice_max_coords(slice_dims);

  std::vector<unsigned int> slice_grid_dims(slice_dims);

  for (unsigned int i = 0; i < slice_dims; i++)
    {
    const unsigned int file_axis = slice_to_file_index_map[i];

    const double file_axis_sim_center = sim_center[file_axis];

    const double file_axis_slice_min_coord
      = limit_sample_radius ? (file_axis_sim_center - max_sample_radius) 
        : sim_min_bounds[file_axis];

    const double file_axis_slice_max_coord
      = limit_sample_radius ? (file_axis_sim_center + max_sample_radius) 
        : sim_max_bounds[file_axis];

    slice_min_coords[i] = file_axis_slice_min_coord;
    slice_max_coords[i] = file_axis_slice_max_coord;

    const double axis_min_slice_length 
      = file_axis_slice_max_coord - file_axis_slice_min_coord;

    const double axis_bin_width = accum_bin_width[i];

    const unsigned int axis_bins 
      = static_cast<unsigned int>(axis_min_slice_length / axis_bin_width) + 1;

    const double axis_slice_width 
      = static_cast<double>(axis_bins) * axis_bin_width;

    const double axis_bin_grid_center 
      = 0.5 * (file_axis_slice_min_coord + file_axis_slice_max_coord);

    const double axis_half_slice_length = 0.5 * axis_slice_width;

    slice_min_coords[i] = axis_bin_grid_center - axis_half_slice_length;
    slice_max_coords[i] = axis_bin_grid_center + axis_half_slice_length;

    slice_grid_dims[i] = axis_bins;
    }

  QuickFlash::Stats::BinGrid total_dens_grid(slice_min_coords, accum_bin_width,
					   slice_grid_dims);

  QuickFlash::Stats::BinGrid poly_dens_grid(slice_min_coords, accum_bin_width,
					  slice_grid_dims);

  QuickFlash::Stats::BinGrid chbr_dens_grid(slice_min_coords, accum_bin_width,
					  slice_grid_dims);

  QuickFlash::Stats::BinGrid foam_dens_grid(slice_min_coords, accum_bin_width,
					  slice_grid_dims);


  // Run through the data

  QuickFlash::Block::BlockData<double> dens_data;

  QuickFlash::Block::BlockData<double> poly_data;
  // QuickFlash::Block::BlockData<double> chbr_data;  // INFER FOR NOW!!
  QuickFlash::Block::BlockData<double> foam_data;

  std::vector<double> file_cell_pos(file_dims);
  std::vector<double> slice_cell_pos(slice_dims);

  std::vector<double> slice_block_min_bounds(slice_dims);
  std::vector<double> slice_block_max_bounds(slice_dims);

  std::vector<double> slice_cell_width(slice_dims);

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    const std::vector<double> & file_current_min_bounds 
      = block_info.get_min_coords();

    const std::vector<double> & file_current_max_bounds 
      = block_info.get_max_coords();

    const double block_search_axis_min_pos 
      = file_current_min_bounds[search_axis];

    const double block_search_axis_max_pos 
      = file_current_max_bounds[search_axis];

    for (unsigned int i = 0; i < slice_dims; i++)
      {
      const unsigned int file_axis = slice_to_file_index_map[i];

      slice_block_min_bounds[i] = file_current_min_bounds[file_axis];
      slice_block_max_bounds[i] = file_current_max_bounds[file_axis];
      }

    bool skip_branch = false;

    if ((min_search_axis_coord > block_search_axis_max_pos)
	|| (max_search_axis_coord < block_search_axis_min_pos))
      skip_branch = true;
    else
      {
      const double area_overlap
        = QuickFlash::Block::Utils::get_volume_overlap(slice_min_coords,
						     slice_max_coords,
						     slice_block_min_bounds,
						     slice_block_max_bounds);

      if (!(area_overlap > 0.0))
	skip_branch = true;
      }
						     
    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      const double cell_volume = block_info.get_cell_volume();  // CARTESIAN!!

      const std::vector<double> & file_cell_width 
	= block_info.get_cell_width();

      const double cell_search_axis_width = file_cell_width[search_axis];

      const double oneOver_cell_search_axis_width 
	= 1.0 / cell_search_axis_width;

      const double cell_search_axis_half_width = 0.5 * cell_search_axis_width;

      for (unsigned int i = 0; i < slice_dims; i++)
	{
	const unsigned int file_axis = slice_to_file_index_map[i];

	slice_cell_width[i] = file_cell_width[file_axis];
	}

      const unsigned int num_cells = block_info.get_num_cells();

      bool block_loaded = false;

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	block_info.get_cell_center(cell_index, file_cell_pos);

	bool use_cell = true;

	if (limit_sample_radius)
	  {
	  double sq_cell_pos_radius = 0.0;

	  for (unsigned int i = 0; i < file_dims; i++)
	    if (i != search_axis)
	      {
	      const double axis_comp = file_cell_pos[i] - sim_center[i];

	      sq_cell_pos_radius += axis_comp * axis_comp;
	      }

	  if (sq_cell_pos_radius > sq_max_radius)
	    use_cell = false;
	  }

	if (use_cell)
	  {
	  const double search_axis_pos = file_cell_pos[search_axis];

	  // Determine the fraction of overlap along the search axis

	  const double cell_search_axis_min 
	    = search_axis_pos - cell_search_axis_half_width;

	  const double cell_search_axis_max
	    = search_axis_pos + cell_search_axis_half_width;

	  const double search_axis_overlap_min 
	    = std::max(cell_search_axis_min, min_search_axis_coord);

	  const double search_axis_overlap_max
	    = std::min(cell_search_axis_max, max_search_axis_coord);

	  if (search_axis_overlap_min < search_axis_overlap_max)
	    {
	    const double overlap_fraction 
	      = (search_axis_overlap_max - search_axis_overlap_min) 
	        * oneOver_cell_search_axis_width;

	    const double overlap_cell_volume = overlap_fraction * cell_volume;

	    for (unsigned int i = 0; i < slice_dims; i++)
	      {
	      const unsigned int file_axis = slice_to_file_index_map[i];

	      slice_cell_pos[i] = file_cell_pos[file_axis];
	      }

	    if (!block_loaded)
	      {
	      dens_set.get_block_data(block_index, dens_data);

	      poly_set.get_block_data(block_index, poly_data);
	      // chbr_set.get_block_data(block_index, chbr_data);
	      foam_set.get_block_data(block_index, foam_data);

	      block_loaded = true;
	      }

	    const double cell_dens = dens_data[cell_index];

	    const double poly_dens = poly_data[cell_index] * cell_dens;
	    // const double chbr_dens = chbr_data[cell_index] * cell_dens;
	    const double foam_dens = foam_data[cell_index] * cell_dens;

	    // FOR NOW, ASSUME CHBR IS [ TOTAL - (POLY + FOAM) ]

	    const double chbr_dens 
	      = std::max(cell_dens - (poly_dens + foam_dens), 0.0);

	    // WE SHOULD CLIP THE CELLS CROSSING SLICE_MIN/MAX_COORDS!!!

	    total_dens_grid.add_data_spread(slice_cell_pos, cell_dens,
					    slice_cell_width, 
					    overlap_cell_volume);
					    
	    poly_dens_grid.add_data_spread(slice_cell_pos, poly_dens,
					   slice_cell_width, 
					   overlap_cell_volume);
					    
	    chbr_dens_grid.add_data_spread(slice_cell_pos, chbr_dens,
					   slice_cell_width, 
					   overlap_cell_volume);
					    
	    foam_dens_grid.add_data_spread(slice_cell_pos, foam_dens,
					   slice_cell_width, 
					   overlap_cell_volume);
	    }
	  }
	}
      }

    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  // Build the member arrays

  grid_info.reset(slice_min_coords, slice_max_coords, slice_grid_dims,
		  QuickFlash::Geometry::Cartesian);

  total_mass_bins.reset(slice_grid_dims);
  poly_mass_bins.reset(slice_grid_dims);
  chbr_mass_bins.reset(slice_grid_dims);
  foam_mass_bins.reset(slice_grid_dims);

  const unsigned int num_bins = grid_info.get_num_cells();

  double grid_sum_total_mass = 0.0;
  double grid_sum_poly_mass = 0.0;
  double grid_sum_chbr_mass = 0.0;
  double grid_sum_foam_mass = 0.0;

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & total_bin = total_dens_grid[bin_index];

    const QuickFlash::Stats::Bin & poly_bin = poly_dens_grid[bin_index];
    const QuickFlash::Stats::Bin & chbr_bin = chbr_dens_grid[bin_index];
    const QuickFlash::Stats::Bin & foam_bin = foam_dens_grid[bin_index];

    // The total mass is the mean value times the weight (volume) of the bin

    if (total_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = total_bin.get_weight_value_sum();

      grid_sum_total_mass += bin_mass;

      total_mass_bins[bin_index] = bin_mass;
      }
    else
      total_mass_bins[bin_index] = 0.0;

    if (poly_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = poly_bin.get_weight_value_sum();

      grid_sum_poly_mass += bin_mass;

      poly_mass_bins[bin_index] = bin_mass;
      }
    else
      poly_mass_bins[bin_index] = 0.0;

    if (chbr_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = chbr_bin.get_weight_value_sum();

      grid_sum_chbr_mass += bin_mass;

      chbr_mass_bins[bin_index] = bin_mass;
      }
    else
      chbr_mass_bins[bin_index] = 0.0;

    if (foam_bin.get_weight_sum() > 0.0)
      {
      const double bin_mass = foam_bin.get_weight_value_sum();

      grid_sum_foam_mass += bin_mass;

      foam_mass_bins[bin_index] = bin_mass;
      }
    else
      foam_mass_bins[bin_index] = 0.0;
    }

  total_mass = grid_sum_total_mass;
  poly_mass = grid_sum_poly_mass;
  chbr_mass = grid_sum_chbr_mass;
  foam_mass = grid_sum_foam_mass;
  }

      
void LayerSlice2D::reset(const LayerSlice2D & source)
  {
  if (&source != this)
    {
    total_mass = source.total_mass;
    poly_mass = source.poly_mass;
    chbr_mass = source.chbr_mass;
    foam_mass = source.foam_mass;

    grid_info = source.grid_info;

    slice_to_file_index_map = source.slice_to_file_index_map;

    total_mass_bins = source.total_mass_bins;
    poly_mass_bins = source.poly_mass_bins;
    chbr_mass_bins = source.chbr_mass_bins;
    foam_mass_bins = source.foam_mass_bins;
    }
  }
