// C++ program file mass_histogram.cpp

/*
  By Nathan C. Hearn
     April 11, 2008

  Mass histogram for ranges of data values.
*/


#include <iostream>
#include <cstdlib>
#include <algorithm>
#include <vector>
#include <list>
#include <fstream>
#include <string>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_stats_bins.hpp"
#include "quickflash_tree_block.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "file_defs.hpp"
#include "layer_detection.hpp"

const unsigned int buffer_size = 500;

const char default_comment_marker[] = "# ";


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 3;

  if (num_args < min_args)
    {
    cerr << endl << "Usage: " << argv[0] << " dataset num_bins"
	 << endl << "\t"
	 << "[ --max-sample-radius=RADIUS ] [ --exp-bin-widths ]"
	 << endl << "\t"
	 << "[ --min-value=VALUE ] [ --max-value=VALUE ]"
	 << endl << "\t"
	 << "[ --min-x-coord=POSITION ] [ --max-x-coord=POSITION ]"
	 << endl << "\t"
	 << "[ --use-envelope=DATASET,MIN_VALUE,MAX_VALUE ]"
	 << endl << "\t"
	 << "[ --output-prefix=PREFIX ] [ --add-comment-marker[=MARKER] ]"
	 << endl << "\t"
	 << "filename1 [ filename2 ... ]"
	 << endl << endl;
    return 1;
    }


  // Read the required arguments

  unsigned int argptr = 0;

  const std::string dataset_name = argument_list[argptr++];

  const unsigned int num_bins = std::atoi(argument_list[argptr++].c_str());

  if (num_bins < 1)
    {
    cerr << endl << "Error: Number of bins must be a positive integer" 
	 << endl << endl;
    return -1;
    }

  std::list<std::string> datafile_list;

  while (argptr < num_args)
    datafile_list.push_back(argument_list[argptr++]);


  // Parse the options

  bool limit_sample_radius = false;
  double max_sample_radius = 0.0;

  bool use_dataset_min = true;
  double user_min_value = 0.0;

  bool use_dataset_max = true;
  double user_max_value = 0.0;

  bool use_user_min_x = false;
  double user_min_x = 0.0;

  bool use_user_max_x = false;
  double user_max_x = 0.0;

  bool use_envelope = false;
  std::string envelope_dataset;
  double envelope_min_value = 0.0;
  double envelope_max_value = 0.0;

  bool use_exp_bin_widths = false;

  bool use_user_output_prefix = false;
  std::string user_output_prefix = "";

  bool add_comment_marker = false;
  std::string comment_marker = "";

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter 
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "max-sample-radius")
      {
      if (!current_option.value_present())
        {
        cerr << endl << "Error: Max sample radius requires a floating point "
             "argument" << endl << endl;
	return -1;
	}

      current_option.get_value(max_sample_radius);

      if (!(max_sample_radius > 0.0))
	{
	cerr << endl << "Error: Max sample radius must be positive"
	     << endl << endl;
	return -1;
	}

      limit_sample_radius = true;
      }
    else if (keyword == "exp-bin-widths")
      {
      use_exp_bin_widths = true;
      }
    else if (keyword == "min-value")
      {
      if (!current_option.value_present())
        {
        cerr << endl << "Error: Min data value requires an argument"
             << endl << endl;
	return -1;
	}

      current_option.get_value(user_min_value);

      use_dataset_min = false;
      }
    else if (keyword == "max-value")
      {
      if (!current_option.value_present())
        {
        cerr << endl << "Error: Max data value requires an argument"
             << endl << endl;
	return -1;
	}

      current_option.get_value(user_max_value);

      use_dataset_min = false;
      }
    else if (keyword == "min-x-coord")
      {
      if (!current_option.value_present())
        {
        cerr << endl << "Error: Min x coordinate requires an argument"
             << endl << endl;
	return -1;
	}

      current_option.get_value(user_min_x);

      use_user_min_x = true;
      }
    else if (keyword == "max-x-coord")
      {
      if (!current_option.value_present())
        {
        cerr << endl << "Error: Max x coordinate requires and argument"
             << endl << endl;
	return -1;
	}

      current_option.get_value(user_max_x);

      use_user_max_x = true;
      }
    else if (keyword == "use-envelope")
      {
      if (!current_option.value_present())
        {
        cerr << endl << "Error: Envelope option requires arguments"
             << endl << endl;
	return -1;
	}

      std::vector<std::string> envelope_opts;

      current_option.get_value(envelope_opts);

      if (envelope_opts.size() != 3)
	{
	cerr << endl << "Error: Envelope option requires arguments" << endl 
	     << endl;
	return -1;
	}

      envelope_dataset = envelope_opts[0];

      envelope_min_value = std::atof(envelope_opts[1].c_str());
      envelope_max_value = std::atof(envelope_opts[2].c_str());

      if (!(envelope_min_value < envelope_max_value))
	{
	cerr << endl << "Error: Envelope min value must be less than max value"
	     << endl << endl;
	return -1;
	}

      use_envelope = true;
      }
    else if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output prefix requires an argument" << endl
	     << endl;
	return -1;
	}

      user_output_prefix = current_option.get_value();

      use_user_output_prefix = true;
      }
    else if (keyword == "add-comment-marker")
      {
      add_comment_marker = true;

      if (current_option.value_present())
	comment_marker = current_option.get_value() + " ";
      else
	comment_marker = default_comment_marker;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }


  // Perform some additional checks and get derived quantities

  const double sq_max_radius = max_sample_radius * max_sample_radius;

  if ((!use_dataset_min) && (!use_dataset_max))
    if (!(user_min_value < user_max_value))
      {
      cerr << endl << "Error: Min value must be less than max value" << endl 
	   << endl;
      return -1;
      }


  if ((use_user_min_x) && (use_user_max_x))
    if (!(user_min_x < user_max_x))
      {
      cerr << endl << "Error: Min x must be less than max x" << endl 
	   << endl;
      return -1;
      }


    // Run through the files

  std::list<std::string>::const_iterator datafile_iter = datafile_list.begin();

  while (datafile_iter != datafile_list.end())
    {
    const std::string & datafile_name = *datafile_iter;

    // Set up the filenames

    std::string datafile_path;
    std::string unqualified_datafile_name;

    QuickFlash::Utils::split_name_path_file(datafile_name, datafile_path,
					  unqualified_datafile_name);

    std::string output_prefix = "";
 
    if (use_user_output_prefix)
      output_prefix += user_output_prefix;

    output_prefix += unqualified_datafile_name;

    const std::string histogram_file 
      = output_prefix + "_" + dataset_name + ".csv";

    std::ofstream hist(histogram_file.c_str(), std::ios::trunc);

    hist << comment_marker << "DataFile " << unqualified_datafile_name << endl;
    hist << comment_marker << "Dataset " << dataset_name << endl;


    // Open the file

    cerr << "** Processing file " << unqualified_datafile_name << endl;

    QuickFlash::File::DataFile df(datafile_name, buffer_size, true);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double sim_time = meshinfo.get_sim_time();

    hist << comment_marker << "SimTime " << sim_time << endl;

    const unsigned int dims = meshinfo.get_dims();

    const std::vector<double> & sim_min_bounds 
      = meshinfo.get_volume_minbounds();

    const std::vector<double> & sim_max_bounds 
      = meshinfo.get_volume_maxbounds();

    const double sim_min_x = sim_min_bounds[0];
    const double sim_max_x = sim_max_bounds[0];

    std::vector<double> sim_center(dims);

    for (unsigned int i = 0; i < dims; i++)
      sim_center[i] = 0.5 * (sim_min_bounds[i] + sim_max_bounds[i]);

    double limited_min_x = use_user_min_x ? user_min_x : sim_min_x;
    double limited_max_x = use_user_max_x ? user_max_x : sim_max_x;

    if (use_envelope)
      {
      double envelope_min_x = 0.0;
      double envelope_max_x = 0.0;

      int err_state 
	= find_maximal_envelope(df, envelope_dataset, 0, envelope_min_value,
				envelope_max_value, envelope_min_x,
				envelope_max_x, limit_sample_radius,
				max_sample_radius);

      if (err_state < 0)
	{
	cerr << endl << "Error: Unable to find envelope" << endl << endl;
	return -2;
	}

      hist << comment_marker << "EnvelopeDataset " << envelope_dataset << endl;
      hist << comment_marker << "EnvelopeMinValue " << envelope_min_value
	   << endl;
      hist << comment_marker << "EnvelopeMaxValue " << envelope_max_value
	   << endl;
      hist << comment_marker << "EnvelopeMinPos " << envelope_min_x << endl;
      hist << comment_marker << "EnvelopeMaxPos " << envelope_max_x << endl;
      
      limited_min_x = std::max(limited_min_x, envelope_min_x);
      limited_max_x = std::min(limited_max_x, envelope_max_x);
      }

    const double min_x = limited_min_x;
    const double max_x = limited_max_x;

    if (!(min_x < max_x))
      {
      cerr << endl << "Error: x-bounds disagreement" << endl << endl;
      return -2;
      }

    hist << comment_marker << "Min_x " << min_x << endl;
    hist << comment_marker << "Max_x " << max_x << endl;

    if (limit_sample_radius)
      hist << comment_marker << "MaxSampleRadius " << max_sample_radius 
	   << endl;


    // Open the density and mass fraction datasets

    const QuickFlash::File::Dataset & dens_set 
      = df.get_dataset(dens_dataset_name);

    const double dens_min_value = dens_set.get_min_value();
    const double dens_max_value = dens_set.get_max_value();

    hist << comment_marker << "MinDensity " << dens_min_value << endl;
    hist << comment_marker << "MaxDensity " << dens_max_value << endl;

    // INFER CHBR FOR NOW...

    const QuickFlash::File::Dataset & poly_set
      = df.get_dataset(poly_dataset_name);
//    const QuickFlash::File::Dataset & chbr_set 
//      = df.get_dataset(chbr_dataset_name);
    const QuickFlash::File::Dataset & foam_set 
      = df.get_dataset(foam_dataset_name);

    // Open the user dataset

    const QuickFlash::File::Dataset & user_set = df.get_dataset(dataset_name);
    
    const double dataset_min_value = user_set.get_min_value();
    const double dataset_max_value = user_set.get_max_value();

    hist << comment_marker << "DatasetMinValue " << dataset_min_value << endl;
    hist << comment_marker << "DatasetMaxValue " << dataset_max_value << endl;

    const double min_value 
      = use_dataset_min ? dataset_min_value : user_min_value;

    const double max_value 
      = use_dataset_max ? dataset_max_value : user_max_value;

    if (!(min_value < max_value))
      {
      cerr << endl << "Error: Value bounds disagreement" << endl << endl;
      return -2;
      }

    if (use_exp_bin_widths && (!(min_value > 0.0)))
      {
      cerr << endl << "Error: Exponential bin widths must have postive valuess"
	   << endl << endl;
      return -1;
      }

    hist << comment_marker << "MinValue " << min_value << endl;
    hist << comment_marker << "MaxValue " << max_value << endl;


    // Set up the bins

    QuickFlash::Stats::BinArray value_bins(min_value, max_value, num_bins, 
					 use_exp_bin_widths);

    QuickFlash::Stats::BinArray poly_value_bins(min_value, max_value, num_bins, 
					      use_exp_bin_widths);

    QuickFlash::Stats::BinArray chbr_value_bins(min_value, max_value, num_bins, 
					      use_exp_bin_widths);

    QuickFlash::Stats::BinArray foam_value_bins(min_value, max_value, num_bins, 
					      use_exp_bin_widths);


    // Run through the cells

    QuickFlash::Block::BlockData<double> dens_data;

    QuickFlash::Block::BlockData<double> poly_data;
    // QuickFlash::Block::BlockData<double> chbr_data;  // INFER FOR NOW!!
    QuickFlash::Block::BlockData<double> foam_data;

    QuickFlash::Block::BlockData<double> user_data;

    std::vector<double> cell_pos;

    QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

    unsigned int counter = 0;

    while (node_iter.valid_node())
      {
      if ((counter % 1000) == 0)
	cerr << "** At block " << counter << endl;

      counter++;

      const QuickFlash::Block::BlockInfo & block_info 
	= node_iter.get_block_info();
	
      const double block_min_x = block_info.get_min_coords()[0];
      const double block_max_x = block_info.get_max_coords()[0];

      bool skip_branch = false;

      if ((block_min_x > max_x) || (block_max_x < min_x))
	skip_branch = true;

      if ((!skip_branch) && node_iter.is_leaf())
	{
	const unsigned int block_index = node_iter.get_block_index();

	const double cell_volume = block_info.get_cell_volume(); // CARTESIAN!!
	const double cell_width_x = block_info.get_cell_width()[0];

	const double cell_half_width_x = 0.5 * cell_width_x;

	const unsigned int num_cells = block_info.get_num_cells();

	bool block_loaded = false;


	// Find a cell in the desired range

	for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	  {
	  block_info.get_cell_center(cell_index, cell_pos);

	  const double cell_center_x = cell_pos[0];

	  const double cell_min_x = cell_center_x - cell_half_width_x;
	  const double cell_max_x = cell_center_x + cell_half_width_x;

	  bool use_cell = true;

	  if ((cell_min_x > max_x) || (cell_max_x < min_x))
	    use_cell = false;
	  else if (limit_sample_radius)
	    {
	    double sq_cell_pos_radius = 0.0;

	    for (unsigned int i = 1; i < dims; i++)
	      {
	      const double axis_comp = cell_pos[i] - sim_center[i];

	      sq_cell_pos_radius += axis_comp * axis_comp;
	      }

	    if (sq_cell_pos_radius > sq_max_radius)
	      use_cell = false;
	    }

	  if (use_cell)
	    {
	    if (!block_loaded)
	      {
	      dens_set.get_block_data(block_index, dens_data);

	      poly_set.get_block_data(block_index, poly_data);
	      // chbr_set.get_block_data(block_index, chbr_data);
	      foam_set.get_block_data(block_index, foam_data);

	      user_set.get_block_data(block_index, user_data);

	      block_loaded = true;
	      }

	    // SHOULD REDUCE EFFECTIVE MASS BY FRACTION OUTSIDE OF MIN/MAX X

	    const double cell_value = user_data[cell_index];

	    const double cell_dens = dens_data[cell_index];

	    const double cell_poly = poly_data[cell_index];
	    // const double cell_chbr = chbr_data[cell_index];
	    const double cell_foam = foam_data[cell_index];

	    const double cell_chbr = std::max(1.0 - (cell_poly + cell_foam),
					      0.0);

	    value_bins.add_data(cell_value, cell_dens, cell_volume);

	    poly_value_bins.add_data(cell_value, (cell_dens * cell_poly), 
				     cell_volume);

	    chbr_value_bins.add_data(cell_value, (cell_dens * cell_chbr), 
				     cell_volume);
	    foam_value_bins.add_data(cell_value, (cell_dens * cell_foam), 
				     cell_volume);
	    }
	  }
	}

      if (skip_branch)
	node_iter.next_branch();
      else
	node_iter.next();
      }

    const std::vector<double> & bin_centers = value_bins.get_bin_centers();

    std::vector<double> bin_masses(num_bins, 0.0);
    std::vector<double> bin_poly_masses(num_bins, 0.0);
    std::vector<double> bin_chbr_masses(num_bins, 0.0);
    std::vector<double> bin_foam_masses(num_bins, 0.0);

    std::vector<double> bin_mean_dens(num_bins, 0.0);
    std::vector<double> bin_mean_poly_dens(num_bins, 0.0);
    std::vector<double> bin_mean_chbr_dens(num_bins, 0.0);
    std::vector<double> bin_mean_foam_dens(num_bins, 0.0);

    double array_total_mass = 0.0;
    double array_poly_mass = 0.0;
    double array_chbr_mass = 0.0;
    double array_foam_mass = 0.0;

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      const QuickFlash::Stats::Bin & total_bin = value_bins[bin_index];

      const QuickFlash::Stats::Bin & poly_bin = poly_value_bins[bin_index];
      const QuickFlash::Stats::Bin & chbr_bin = chbr_value_bins[bin_index];
      const QuickFlash::Stats::Bin & foam_bin = foam_value_bins[bin_index];

      // The total mass is the mean value times the weight (volume) of the bin

      if (total_bin.get_weight_sum() > 0.0)
	{
	const double bin_mass = total_bin.get_weight_value_sum();

	array_total_mass += bin_mass;

	bin_masses[bin_index] = bin_mass;
	bin_mean_dens[bin_index] = total_bin.get_mean();
	}

      if (poly_bin.get_weight_sum() > 0.0)
	{
	const double bin_mass = poly_bin.get_weight_value_sum();

	array_poly_mass += bin_mass;

	bin_poly_masses[bin_index] = bin_mass;
	bin_mean_poly_dens[bin_index] = poly_bin.get_mean();
	}

      if (chbr_bin.get_weight_sum() > 0.0)
	{
	const double bin_mass = chbr_bin.get_weight_value_sum();

	array_chbr_mass += bin_mass;

	bin_chbr_masses[bin_index] = bin_mass;
	bin_mean_chbr_dens[bin_index] = chbr_bin.get_mean();
	}

      if (foam_bin.get_weight_sum() > 0.0)
	{
	const double bin_mass = foam_bin.get_weight_value_sum();

	array_foam_mass += bin_mass;

	bin_foam_masses[bin_index] = bin_mass;
	bin_mean_foam_dens[bin_index] = foam_bin.get_mean();
	}
      }

    // Output the data

    hist << comment_marker << "TotalMass " << array_total_mass << endl;
    hist << comment_marker << "PolyMass " << array_poly_mass << endl;
    hist << comment_marker << "CHBrMass " << array_chbr_mass << endl;
    hist << comment_marker << "FoamMass " << array_foam_mass << endl;

    hist << comment_marker << "HistogramData" << endl;

    hist << "Bin_Value" << "\t" << "Mass" << "\t" << "Poly_Mass" << "\t"
	 << "CHBr_Mass" << "\t" << "Foam_Mass " << "\t" << "Mean_Density"
	 << "\t" << "Mean_Poly_Density" << "\t" << "Mean_CHBr_Density" << "\t"
	 << "Mean_Foam_density"
	 << endl;

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      hist << bin_centers[bin_index] << "\t" << bin_masses[bin_index] << "\t"
	   << bin_poly_masses[bin_index] << "\t" << bin_chbr_masses[bin_index]
	   << "\t" << bin_foam_masses[bin_index] << "\t" 
	   << bin_mean_dens[bin_index] << "\t" << bin_mean_poly_dens[bin_index]
	   << "\t" << bin_mean_chbr_dens[bin_index] << "\t"
	   << bin_mean_foam_dens[bin_index]
	   << endl;
      }

    hist << endl;

    ++datafile_iter;
    }

  return 0;
  }

    
