// C++ program file omega_slicer.cpp

/*
  By Nathan C. Hearn
     November 20, 2007

  Slice and diagnostics generator for simulations of Omega Multi-mode RT
  experiments.
*/


#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <ostream>
#include <fstream>
#include <list>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_file_siminfo.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_except.hpp"
#include "quickflash_slice_slicegrid.hpp"
#include "quickflash_slice_slicer.hpp"
#include "quickflash_color_colormap.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_image_imagegrid.hpp"
#include "quickflash_image_imagemagick.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "grid_output.hpp"
#include "layer_detection.hpp"


const unsigned int default_cross_resolution = 256;

const char option_separator = ',';

const double deg_to_rad = (M_PI) / 180.0;


// 3D Rotations

// Here "xy" is really 45-degrees around x

const double xy_roll_angle = 90.0 * deg_to_rad;
const double xy_polar_angle = 45.0 * deg_to_rad;
const double xy_azim_angle = -90.0 * deg_to_rad;

// Here "xz" is really 135-degrees around x

const double xz_roll_angle = 90.0 * deg_to_rad;
const double xz_polar_angle = 135.0 * deg_to_rad;
const double xz_azim_angle = -90.0 * deg_to_rad;

// Here yz is really 45-degrees (or 90+(xy rotation) around y

const double yz_roll_angle = (45.0 + 90.0) * deg_to_rad;
const double yz_polar_angle = 90.0 * deg_to_rad;
const double yz_azim_angle = 0.0 * deg_to_rad;


const unsigned int read_buffer_size = 50;
const unsigned int read_cache_size = 3000;

const bool limit_envelope_sample_radius = false;
const double envelope_sample_radius = 0.0;

const double image_aspect_width_height = 2.0;

const double image_height = 0.09;
const double image_width = image_aspect_width_height * image_height;
const double image_half_width = 0.5 * image_width;

const double image_width_buffer = 2.0e+06;

const char image_suffix[] = "png";


const double xsect_width = 0.12;
const double xsect_height = 0.12;

const unsigned int num_xsects = 9;

const double xsect_delta_x = image_width / static_cast<double>(num_xsects - 1);

const char xsect_name[] = "xsect";

const char xsect_index_type[] = "index";

const char xsect_upper_type[] = "upper";
const char xsect_middle_type[] = "middle";
const char xsect_lower_type[] = "lower";

const char xy_plane_label[] = "xy";
const char xz_plane_label[] = "xz";

const char envelope_dataset_label[] = "foam";

const double envelope_min_value = 0.001;
const double envelope_max_value = 0.999;


void parse_dataset_name_option(const std::string & dataset_name_argument,
			       std::vector<std::string> & dataset_name_list)
  {
  dataset_name_list.clear();

  const std::string::size_type length = dataset_name_argument.size();

  std::string::size_type index = 0;

  std::string work_string;

  while (index < length)
    {
    std::string::size_type next_index 
      = dataset_name_argument.find(option_separator, index);

    if (next_index == std::string::npos)
      next_index = length;

    const std::string::size_type copy_length = next_index - index;

    work_string.assign(dataset_name_argument, index, copy_length);

    dataset_name_list.push_back(work_string);

    index = next_index + 1;
    }
  }


void filter_variable_names(const std::vector<std::string> & selected_variables,
			   const std::vector<std::string> & datafile_variables,
			   std::vector<std::string> & filtered_variables)
  {
  const unsigned int num_selected = selected_variables.size();

  const unsigned int num_datafile = datafile_variables.size();

  filtered_variables.clear();

  for (unsigned int datafile_index = 0; datafile_index < num_datafile;
       datafile_index++)
    {
    const std::string & variable_name = datafile_variables[datafile_index];

    bool found_it = false;

    unsigned int selected_index = 0;

    while ((!found_it) && (selected_index < num_selected))
      {
      if (selected_variables[selected_index] == variable_name)
	found_it = true;
      else
	selected_index++;
      }

    if (found_it)
      filtered_variables.push_back(variable_name);
    }
  }


void filter_variable_names(const std::vector<std::string> & selected_variables,
			   const std::vector<std::string> & datafile_variables,
			   std::vector<std::string> & filtered_variables,
			   std::vector<std::string> & removed_variables)
  {
  const unsigned int num_selected = selected_variables.size();

  const unsigned int num_datafile = datafile_variables.size();

  filtered_variables.clear();
  removed_variables.clear();

  for (unsigned int datafile_index = 0; datafile_index < num_datafile;
       datafile_index++)
    {
    const std::string & variable_name = datafile_variables[datafile_index];

    bool found_it = false;

    unsigned int selected_index = 0;

    while ((!found_it) && (selected_index < num_selected))
      {
      if (selected_variables[selected_index] == variable_name)
	found_it = true;
      else
	selected_index++;
      }

    if (found_it)
      filtered_variables.push_back(variable_name);
    else
      removed_variables.push_back(variable_name);
    }
  }


#ifdef USE_MAGICK
int write_image(const std::string & filename, 
		const QuickFlash::Slice::SliceGrid & data)
  {
  int ret_val = -1;  // Error state

  // Build the colormap

  double min_val = 0.0;
  double max_val = 0.0;

  data.get_minmax_values(min_val, max_val);

  if (!(min_val < max_val))
    max_val = (2.0 * std::fabs(min_val)) + 1.0;  // Make something up ...

  QuickFlash::Color::Colormap colormap(min_val, max_val);

  QuickFlash::Image::ImageGrid image(data.get_grid_data(), colormap);

  ret_val = QuickFlash::Image::ImageMagick::write_image(filename, image);

  return ret_val;
  }
#endif  // USE_MAGICK


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 1;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name << " flash_filename "
#ifdef USE_MAGICK
	 << endl << "\t"
	 << "[ --make-image ] [ --image-only ]"
#endif  // USE_MAGICK
	 << endl << "\t"
	 << "[ --cross-resolution=PIXELS ]"
	 << endl << "\t"
	 << "[ --datasets=DATASET1,DATASET2,... ]"
#ifdef USE_MAGICK
	 << endl << "\t"
	 << "[ --symm-colormap-datasets=DATASET1,DATASET2,... ]"
	 << endl << "\t"
	 << "[ --standard-colormap=COLORMAP ]"
	 << endl << "\t"
	 << "[ --symmetric-colormap=COLORMAP ]"
#endif  // USE_MAGICK
	 << endl << "\t"
	 << "  [ --output-prefix=FILENAME_PREFIX ]" 
	 << endl << endl;
    return 1;
    }


  // Read the arguments

  unsigned int argptr = 0;

  const std::string filename = argument_list[argptr++];

  
  // Process the options

#ifdef USE_MAGICK
  bool output_image = false;
#endif  // USE_MAGICK

  bool output_data = true;

  double cross_resolution = default_cross_resolution;

  std::string output_prefix = filename;  // !!!

  std::vector<std::string> dataset_name_list;

#ifdef USE_MAGICK
  std::vector<std::string> symm_cm_dataset_name_list;

  bool use_user_std_cm;
  std::string user_std_cm_file;

  bool use_user_symm_cm;
  std::string user_symm_cm_file;
#endif  // USE_MAGICK

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "make-image")
      {
#ifdef USE_MAGICK
      output_image = true;
#else
      cerr << endl << "Error: Image output support not present" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "image-only")
      {
#ifdef USE_MAGICK
      output_image = true;
      output_data = false;
#else
      cerr << endl << "Error: Image output support not present" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "cross-resolution")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Cross-resolution pixel count required" << endl
	     << endl;
	return -1;
	}

      int int_cross_res = 0;

      current_option.get_value(int_cross_res);

      if (int_cross_res < 1)
	{
	cerr << endl << "Error: Cross-resolution must be a positive integer"
	     << endl << endl;
	return -1;
	}

      cross_resolution = static_cast<unsigned int>(int_cross_res);
      }
    else if (keyword == "datasets")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Datasets option requires list of dataset names"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(dataset_name_list);
      }
    else if (keyword == "symm-colormap-datasets")
      {
#ifdef USE_MAGICK
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Symmetric colormap datasets option requires "
	     << "list of dataset names"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(symm_cm_dataset_name_list);
#else
      cerr << endl << "Error: Image output support not present" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "standard-colormap")
      {
#ifdef USE_MAGICK
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Standard colormap option requires a filename"
	     << endl << endl;
	return -1;
	}

      user_std_cm_file = current_option.get_value();

      use_user_std_cm = true;
#else
      cerr << endl << "Error: Image output support not present" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "symmetric-colormap")
      {
#ifdef USE_MAGICK
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Symmetric colormap option requires a filename"
	     << endl << endl;
	return -1;
	}

      user_symm_cm_file = current_option.get_value();

      use_user_symm_cm = true;
#else
      cerr << endl << "Error: Image output support not present" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output prefix option requires a string"
	     << endl << endl;
	return -1;
	}

      output_prefix = current_option.get_value();
      }
    else
      {
      cerr << "Error: Keyword [ " << keyword << " ] not recognized" << endl
	   << endl;
      return -1;
      }

    ++option_iter;
    }


  // Set up some additional parameters

  const std::string output_file_prefix = output_prefix + "_";

  std::string output_file_prefix_path;
  std::string output_file_prefix_name;

  QuickFlash::Utils::split_name_path_file(output_file_prefix,
					output_file_prefix_path,
					output_file_prefix_name);

  const std::string metadata_filename = output_file_prefix + "metadata.out";

  const unsigned int image_height_pixels = cross_resolution;
  const unsigned int image_width_pixels 
    = image_aspect_width_height * image_height_pixels;

  const unsigned int xsect_width_pixels = cross_resolution;
  const unsigned int xsect_height_pixels = cross_resolution;

  const std::string envelope_dataset = envelope_dataset_label;

  // Open the file

  QuickFlash::File::DataFile df(filename, read_buffer_size, true);

  const QuickFlash::File::SimInfo & siminfo = df.get_sim_info();

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (dims < 2)
    {
    cerr << endl << "Error: At least two dimensions required" << endl << endl;
    return -2;
    }

  const std::vector<double> & volume_min_bounds 
    = meshinfo.get_volume_minbounds();

  const std::vector<double> & volume_max_bounds 
    = meshinfo.get_volume_maxbounds();

  const double volume_min_x = volume_min_bounds[0];
  const double volume_max_x = volume_max_bounds[0];

  const double volume_center_y 
    = 0.5 * (volume_min_bounds[1] + volume_max_bounds[1]);

  const double volume_center_z 
    = (dims > 2) ? (0.5 * (volume_min_bounds[2] + volume_max_bounds[2])) : 0.0;

  // Open the metadata output file

  std::ofstream mdata(metadata_filename.c_str());

  // Find the approximate flame front and image bounds

  double envelope_min_x = 0.0;
  double envelope_max_x = 0.0;

  if (find_maximal_envelope(df, envelope_dataset, 0, envelope_min_value,
			    envelope_max_value, envelope_min_x,
			    envelope_max_x, limit_envelope_sample_radius,
			    envelope_sample_radius) < 0)
    {
    cerr << endl << "Error: Unable to determine mixing layer maximal envelope"
	 << endl << endl;
    return -2;
    }

  cerr << endl << "Envelope min x [ " << envelope_min_x << " ] max x [ "
       << envelope_max_x << " ]" << endl;

  const double envelope_center_x = 0.5 * (envelope_min_x + envelope_max_x);
  const double envelope_width_x = envelope_max_x - envelope_min_x;

  const double envelope_quarter_width_x = 0.25 * envelope_width_x;

  double image_x_min = envelope_center_x - image_half_width;
  double image_x_max = envelope_center_x + image_half_width;

  if (image_x_max > volume_max_x)
    {
    image_x_max = volume_max_x;
    image_x_min = image_x_max - image_width;
    }
  else if (image_x_min < volume_min_x)
    {
    image_x_min = volume_min_x;
    image_x_max = image_x_min + image_width;
    }

  const double image_y_min = -0.5 * image_height;
  const double image_y_max = 0.5 * image_height;

  std::vector<double> grid_min_coords(2);
  std::vector<double> grid_max_coords(2);

  grid_min_coords[0] = image_x_min;
  grid_min_coords[1] = image_y_min;

  grid_max_coords[0] = image_x_max;
  grid_max_coords[1] = image_y_max;

  std::vector<double> grid_center(dims, 0.0);

  grid_center[0] = 0.5 * (image_x_min + image_x_max);
  grid_center[1] = volume_center_y;

  if (dims > 2)
    grid_center[2] = volume_center_z;

  std::vector<double> grid_width(2);

  grid_width[0] = image_width;
  grid_width[1] = image_height;

  std::vector<unsigned int> grid_pixels(2);

  grid_pixels[0] = image_width_pixels;
  grid_pixels[1] = image_height_pixels;
  
  std::vector<double> xsect_grid_center(dims, 0.0);  // x-coord is variable

  xsect_grid_center[1] = volume_center_y;

  if (dims > 2)
    xsect_grid_center[2] = volume_center_z;

  std::vector<double> xsect_grid_min_coords(2);
  std::vector<double> xsect_grid_max_coords(2);

  xsect_grid_min_coords[0] = -(0.5) * xsect_width;
  xsect_grid_min_coords[1] = -(0.5) * xsect_height;

  xsect_grid_max_coords[0] = 0.5 * xsect_width;
  xsect_grid_max_coords[1] = 0.5 * xsect_height;

  std::vector<double> xsect_grid_width(2);

  xsect_grid_width[0] = xsect_width;
  xsect_grid_width[1] = xsect_height;

  std::vector<unsigned int> xsect_grid_pixels(2);

  xsect_grid_pixels[0] = xsect_width_pixels;
  xsect_grid_pixels[1] = xsect_height_pixels;

  // Find the upper, lower, and middle cross-section positions

  const double xsect_middle_x_pos = envelope_center_x;

  const double xsect_upper_x_pos 
    = envelope_center_x + envelope_quarter_width_x;

  const double xsect_lower_x_pos 
    = envelope_center_x - envelope_quarter_width_x;

  // Build the comment list

  std::list<std::string> comment_list;

  std::stringstream strstream;

  strstream << "FlashFile [ " << filename << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "Time [ " << meshinfo.get_sim_time() << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "Dimensions [ " << dims << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "EnvelopeMinPos [ " << envelope_min_x << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "EnvelopeMaxPos [ " << envelope_max_x << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "GridCenter [ " << grid_center << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "GridWidth [ " << grid_width << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "GridMinCoords [ " << grid_min_coords << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "GridMaxCoords [ " << grid_max_coords << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "GridPixels [ " << grid_pixels << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSectGridCenter [ " << xsect_grid_center << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSectGridWidth [ " << xsect_grid_width << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSectGridMinCoords [ " << xsect_grid_min_coords << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSectGridMaxCoords [ " << xsect_grid_max_coords << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSectGridPixels [ " << xsect_grid_pixels << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSectUpperPlanePos [ " << xsect_upper_x_pos << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSectLowerPlanePos [ " << xsect_lower_x_pos << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");

  strstream << "XSectMiddlePlanePos [ " << xsect_middle_x_pos << " ]";

  comment_list.push_back(strstream.str());

  strstream.str("");


  // Report comments to stdout

  std::list<std::string>::const_iterator str_iter = comment_list.begin();

  std::list<std::string>::const_iterator end_str_iter = comment_list.end();

  while (str_iter != end_str_iter)
    {
    mdata << *str_iter << endl;
    ++str_iter;
    }


  // Construct the slice grid

  QuickFlash::Slice::SliceGrid grid_xy;
  QuickFlash::Slice::SliceGrid grid_xz;  // Only for 3D

  if (dims > 2)
    {
    grid_xy.reset(grid_center, grid_width, grid_pixels, 
		  xy_azim_angle, xy_polar_angle, xy_roll_angle);

    grid_xz.reset(grid_center, grid_width, grid_pixels, 
		  xz_azim_angle, xz_polar_angle, xz_roll_angle);
    }
  else
    grid_xy.reset(grid_center, grid_width, grid_pixels);  // Default x-y plane


  QuickFlash::Slice::SliceGrid xsect_grid;

  std::vector<unsigned int> subsample_dims(dims, 1);


  // Get the variable names

  const std::vector<std::string> & datafile_var_names 
    = siminfo.get_mesh_variable_names();

  std::vector<std::string> var_names;

  if (dataset_name_list.size() > 0)
    filter_variable_names(dataset_name_list, datafile_var_names, var_names);
  else
    var_names = datafile_var_names;

  const unsigned int num_vars = var_names.size();

  for (unsigned int var_index = 0; var_index < num_vars; var_index++)
    {
    const std::string & dataset_name = var_names[var_index];

    const QuickFlash::File::Dataset & dataset = df.get_dataset(dataset_name);

    const double dataset_min_value = dataset.get_min_value();
    const double dataset_max_value = dataset.get_max_value();

    std::list<std::string> current_comments = comment_list;

    strstream << "Dataset [ " + dataset_name + " ] MinValue [ " 
	      << dataset_min_value
	      << " ] MaxValue [ " << dataset_max_value << " ]";

    const std::string dataset_comment = strstream.str();

    strstream.str("");

    current_comments.push_back(dataset_comment);

    mdata << dataset_comment << endl;
    cerr << dataset_comment << endl;

    const std::string current_output_file_prefix 
      = output_file_prefix + dataset_name;

    const std::string current_output_file_prefix_name 
      = output_file_prefix_name + dataset_name;

    // Fill the grid

    slice_simple(meshinfo, dataset, subsample_dims, grid_xy);

    cerr << "  Finished filling xy grid" << endl;

    if (dims > 2)
      {
      slice_simple(meshinfo, dataset, subsample_dims, grid_xz);

      cerr << "  Finished filling xz grid" << endl;
      }

    double grid_xy_min_value = 0.0;
    double grid_xy_max_value = 0.0;

    grid_xy.get_minmax_values(grid_xy_min_value, grid_xy_max_value);

    double grid_xz_min_value = 0.0;
    double grid_xz_max_value = 0.0;

    if (dims > 2)
      grid_xz.get_minmax_values(grid_xz_min_value, grid_xz_max_value);

    if (output_data)
      {
      const std::string xy_output_filename 
	= current_output_file_prefix + "_" + xy_plane_label + ".csv";

      const std::string xy_output_filename_simple 
	= current_output_file_prefix_name + "_" + xy_plane_label + ".csv";

      mdata << "  Data [ " << xy_plane_label 
	    << " ] File [ " << xy_output_filename_simple
	    << " ] MinCoords [ " << grid_min_coords 
	    << " ] MaxCoords [ " << grid_max_coords
	    << " ] MinValue [ " << grid_xy_min_value
	    << " ] MaxValue [ " << grid_xy_max_value << " ]" << endl;

      write_grid(xy_output_filename, grid_xy, current_comments);

      if (dims > 2)
	{
	const std::string xz_output_filename 
	  = current_output_file_prefix + "_" + xz_plane_label + ".csv";

	const std::string xz_output_filename_simple 
	  = current_output_file_prefix_name + "_" + xz_plane_label + ".csv";

	mdata << "  Data [ " << xz_plane_label 
	      << " ] File [ " << xz_output_filename_simple
	      << " ] MinCoords [ " << grid_min_coords 
	      << " ] MaxCoords [ " << grid_max_coords
	      << " ] MinValue [ " << grid_xz_min_value
	      << " ] MaxValue [ " << grid_xz_max_value << " ]" << endl;

	write_grid(xz_output_filename, grid_xz, current_comments);
	}
      }

    if (output_image)
      {
      const std::string xy_image_filename 
	= current_output_file_prefix + "_" + xy_plane_label + "." 
	  + image_suffix;

      const std::string xy_image_filename_simple 
	= current_output_file_prefix_name + "_" + xy_plane_label + "." 
  	  + image_suffix;

      mdata << "  Image [ " << xy_plane_label 
	    << " ] File [ " << xy_image_filename_simple
	    << " ] MinCoords [ " << grid_min_coords 
	    << " ] MaxCoords [ " << grid_max_coords
	    << " ] MinValue [ " << grid_xy_min_value
	    << " ] MaxValue [ " << grid_xy_max_value << " ]" << endl;

#ifdef USE_MAGICK
      int error_state = write_image(xy_image_filename, grid_xy);

      if (error_state < 0)
	{
	cerr << endl << "Error: Unable to write xy image " 
	     << xy_image_filename << endl << endl;
	return -2;
	}
#else
      cerr << endl << "Warning: ImageMagick not available" << endl << endl;
#endif  // USE_MAGICK

      if (dims > 2)
	{
	const std::string xz_image_filename 
	  = current_output_file_prefix + "_" + xz_plane_label + "." 
	    + image_suffix;

	const std::string xz_image_filename_simple 
	  = current_output_file_prefix_name + "_" + xz_plane_label + "." 
	    + image_suffix;

	mdata << "  Image [ " << xz_plane_label 
	      << " ] File [ " << xz_image_filename_simple
	      << " ] MinCoords [ " << grid_min_coords 
	      << " ] MaxCoords [ " << grid_max_coords
	      << " ] MinValue [ " << grid_xz_min_value
	      << " ] MaxValue [ " << grid_xz_max_value << " ]" << endl;

#ifdef USE_MAGICK
	error_state = write_image(xz_image_filename, grid_xz);

	if (error_state < 0)
	  {
	  cerr << endl << "Error: Unable to write xz image " 
	       << xz_image_filename << endl << endl;
	  return -2;
	  }
#else
	cerr << endl << "Warning: ImageMagick not available" << endl << endl;
#endif  // USE_MAGICK
	}
      }

    if (dims > 2)
      {
      const std::string xsect_prefix = current_output_file_prefix + "_xsect_";

      const std::string xsect_prefix_name 
	= current_output_file_prefix_name + "_xsect_";
	
#if 0
      // Get cross-sections along sliced region

      for (unsigned int xsect_index = 0; xsect_index < num_xsects; 
	   xsect_index++)
	{
	std::list<std::string> xsect_comments = current_comments;

	// Build the filename

	strstream << xsect_index;

	const std::string xsect_index_str = strstream.str();

	strstream.str("");

	const std::string xsect_output_filename_nosuffix
	  = xsect_prefix + "s" + xsect_index_str + ".";

	const std::string xsect_output_filename_nosuffix_name
	  = xsect_prefix_name + "s" + xsect_index_str + ".";

	// Find the x-position and populate the grid

	const double xsect_x_pos = image_x_min
	  + (xsect_delta_x * static_cast<double>(xsect_index));

	xsect_grid_center[0] = xsect_x_pos;

	xsect_grid.reset(xsect_grid_center, xsect_grid_width, 
			 xsect_grid_pixels,
			 yz_azim_angle, yz_polar_angle, yz_roll_angle);

	slice_simple(meshinfo, dataset, subsample_dims, xsect_grid);

	cerr << "  Finished filling cross-section " << xsect_index << endl;

	double xsect_grid_min_value = 0.0;
	double xsect_grid_max_value = 0.0;

	xsect_grid.get_minmax_values(xsect_grid_min_value, 
				     xsect_grid_max_value);

	if (output_data)
	  {
	  const std::string output_filename 
	    = xsect_output_filename_nosuffix + "csv";

	  const std::string output_filename_simple
	    = xsect_output_filename_nosuffix_name + "csv";

	  strstream << "XSect [ Data ] Index [ " << xsect_index
		    << " ] Type [ " << xsect_index_type
		    << " ] Pos [ " << xsect_x_pos 
		    << " ] File [ " << output_filename_simple
		    << " ] MinCoords [ " << xsect_grid_min_coords 
		    << " ] MaxCoords [ " << xsect_grid_max_coords
		    << " ] MinValue [ " << xsect_grid_min_value
		    << " ] MaxValue [ " << xsect_grid_max_value 
		    << " ]";

	  const std::string file_comment = strstream.str();

	  strstream.str("");

	  xsect_comments.push_back(file_comment);

	  mdata << "  " << file_comment << endl;
	
	  write_grid(output_filename, xsect_grid, xsect_comments);
	  }

	if (output_image)
	  {
	  const std::string image_filename 
	    = xsect_output_filename_nosuffix + image_suffix;

	  const std::string image_filename_simple 
	    = xsect_output_filename_nosuffix_name + image_suffix;

	  strstream << "XSect [ Image ] Index [ " << xsect_index 
		    << " ] Type [ " << xsect_index_type
		    << " ] Pos [ " << xsect_x_pos 
		    << " ] File [ " << image_filename_simple
		    << " ] MinCoords [ " << xsect_grid_min_coords 
		    << " ] MaxCoords [ " << xsect_grid_max_coords
		    << " ] MinValue [ " << xsect_grid_min_value
		    << " ] MaxValue [ " << xsect_grid_max_value 
		    << " ]";

	  const std::string file_comment = strstream.str();

	  strstream.str("");

	  mdata << "  " << file_comment << endl;

#ifdef USE_MAGICK
	  const int error_state = write_image(image_filename, xsect_grid);

	  if (error_state < 0)
	    {
	    cerr << endl << "Error: Unable to write image " << image_filename 
		 << endl << endl;
	    return -2;
	    }
#else
	  cerr << endl << "Warning: ImageMagick not available" << endl << endl;
#endif  // USE_MAGICK
	  }
	}
#endif  // 0

      // Get the upper cross-section

      std::list<std::string> xsect_comments_upper = current_comments;

      // Build the filename

      const std::string xsect_upper_output_filename_nosuffix
	= xsect_prefix + xsect_upper_type + ".";

      const std::string xsect_upper_output_filename_nosuffix_name
	= xsect_prefix_name + xsect_upper_type + ".";

      // Set the x-position and populate the grid

      xsect_grid_center[0] = xsect_upper_x_pos;

      xsect_grid.reset(xsect_grid_center, xsect_grid_width, xsect_grid_pixels,
		       yz_azim_angle, yz_polar_angle, yz_roll_angle);

      slice_simple(meshinfo, dataset, subsample_dims, xsect_grid);

      cerr << "  Finished filling upper cross-section" << endl;

      double xsect_upper_grid_min_value = 0.0;
      double xsect_upper_grid_max_value = 0.0;

      xsect_grid.get_minmax_values(xsect_upper_grid_min_value, 
				   xsect_upper_grid_max_value);

      if (output_data)
	{
	const std::string output_filename 
	  = xsect_upper_output_filename_nosuffix + "csv";

	const std::string output_filename_simple 
	  = xsect_upper_output_filename_nosuffix_name + "csv";

	strstream << "XSect [ Data ] Type [ " << xsect_upper_type
		  << " ] Pos [ " << xsect_upper_x_pos 
		  << " ] File [ " << output_filename_simple
		  << " ] MinCoords [ " << xsect_grid_min_coords 
		  << " ] MaxCoords [ " << xsect_grid_max_coords
		  << " ] MinValue [ " << xsect_upper_grid_min_value
		  << " ] MaxValue [ " << xsect_upper_grid_max_value 
		  << " ]";

	const std::string file_comment = strstream.str();

	strstream.str("");

	xsect_comments_upper.push_back(file_comment);

	mdata << "  " << file_comment << endl;

	write_grid(output_filename, xsect_grid, xsect_comments_upper);
	}

      if (output_image)
	{
	const std::string image_filename 
	  = xsect_upper_output_filename_nosuffix + image_suffix;

	const std::string image_filename_simple
	  = xsect_upper_output_filename_nosuffix_name + image_suffix;

	strstream << "XSect [ Image ] Type [ " << xsect_upper_type 
		  << " ] Pos [ " << xsect_upper_x_pos 
		  << " ] File [ " << image_filename_simple 
		  << " ] MinCoords [ " << xsect_grid_min_coords 
		  << " ] MaxCoords [ " << xsect_grid_max_coords
		  << " ] MinValue [ " << xsect_upper_grid_min_value
		  << " ] MaxValue [ " << xsect_upper_grid_max_value 
		  << " ]";

	const std::string file_comment = strstream.str();

	strstream.str("");

	mdata << "  " << file_comment << endl;

#ifdef USE_MAGICK
	const int error_state = write_image(image_filename, xsect_grid);

	if (error_state < 0)
	  {
	  cerr << endl << "Error: Unable to write image " << image_filename 
	       << endl << endl;
	  return -2;
	  }
#else
	cerr << endl << "Warning: ImageMagick not available" << endl << endl;
#endif  // USE_MAGICK
	}


      // Get the lower cross-section

      std::list<std::string> xsect_comments_lower = current_comments;

      // Build the filename

      const std::string xsect_lower_output_filename_nosuffix
	= xsect_prefix + xsect_lower_type + ".";

      const std::string xsect_lower_output_filename_nosuffix_name
	= xsect_prefix_name + xsect_lower_type + ".";

      // Set the x-position and populate the grid

      xsect_grid_center[0] = xsect_lower_x_pos;

      xsect_grid.reset(xsect_grid_center, xsect_grid_width, xsect_grid_pixels,
		       yz_azim_angle, yz_polar_angle, yz_roll_angle);

      slice_simple(meshinfo, dataset, subsample_dims, xsect_grid);

      cerr << "  Finished filling lower cross-section" << endl;

      double xsect_lower_grid_min_value = 0.0;
      double xsect_lower_grid_max_value = 0.0;

      xsect_grid.get_minmax_values(xsect_lower_grid_min_value, 
				   xsect_lower_grid_max_value);

      if (output_data)
	{
	const std::string output_filename 
	  = xsect_lower_output_filename_nosuffix + "csv";

	const std::string output_filename_simple
	  = xsect_lower_output_filename_nosuffix_name + "csv";

	strstream << "XSect [ Data ] Type [ " << xsect_lower_type
		  << " ] Pos [ " << xsect_lower_x_pos 
		  << " ] File [ " << output_filename_simple
		  << " ] MinCoords [ " << xsect_grid_min_coords 
		  << " ] MaxCoords [ " << xsect_grid_max_coords
		  << " ] MinValue [ " << xsect_lower_grid_min_value
		  << " ] MaxValue [ " << xsect_lower_grid_max_value 
		  << " ]";

	const std::string file_comment = strstream.str();

	strstream.str("");

	xsect_comments_lower.push_back(file_comment);

	mdata << "  " << file_comment << endl;

	write_grid(output_filename, xsect_grid, xsect_comments_lower);
	}
      
      if (output_image)
	{
	const std::string image_filename 
	  = xsect_lower_output_filename_nosuffix + image_suffix;

	const std::string image_filename_simple
	  = xsect_lower_output_filename_nosuffix_name + image_suffix;

	strstream << "XSect [ Image ] Type [ " << xsect_lower_type 
		  << " ] Pos [ " << xsect_lower_x_pos 
		  << " ] File [ " << image_filename_simple
		  << " ] MinCoords [ " << xsect_grid_min_coords 
		  << " ] MaxCoords [ " << xsect_grid_max_coords
		  << " ] MinValue [ " << xsect_lower_grid_min_value
		  << " ] MaxValue [ " << xsect_lower_grid_max_value 
		  << " ]";

	const std::string file_comment = strstream.str();

	strstream.str("");

	mdata << "  " << file_comment << endl;

#ifdef USE_MAGICK
	const int error_state = write_image(image_filename, xsect_grid);

	if (error_state < 0)
	  {
	  cerr << endl << "Error: Unable to write image " << image_filename 
	       << endl << endl;
	  return -2;
	  }
#else
	cerr << endl << "Warning: ImageMagick not available" << endl << endl;
#endif  // USE_MAGICK
	}


      // Get the middle cross-section

      std::list<std::string> xsect_comments_middle = current_comments;

      // Build the filename

      const std::string xsect_middle_output_filename_nosuffix
	= xsect_prefix + xsect_middle_type + ".";

      const std::string xsect_middle_output_filename_nosuffix_name
	= xsect_prefix_name + xsect_middle_type + ".";

      // Set the x-position and populate the grid

      xsect_grid_center[0] = xsect_middle_x_pos;
      
      xsect_grid.reset(xsect_grid_center, xsect_grid_width, xsect_grid_pixels,
		       yz_azim_angle, yz_polar_angle, yz_roll_angle);

      slice_simple(meshinfo, dataset, subsample_dims, xsect_grid);

      cerr << "  Finished filling middle-front cross-section" << endl;

      double xsect_middle_grid_min_value = 0.0;
      double xsect_middle_grid_max_value = 0.0;

      xsect_grid.get_minmax_values(xsect_middle_grid_min_value, 
				   xsect_middle_grid_max_value);

      if (output_data)
	{
	const std::string output_filename 
	  = xsect_middle_output_filename_nosuffix + "csv";

	const std::string output_filename_simple
	  = xsect_middle_output_filename_nosuffix_name + "csv";

	strstream << "XSect [ Data ] Type [ " << xsect_middle_type
		  << " ] Pos [ " << xsect_middle_x_pos 
		  << " ] File [ " << output_filename_simple
		  << " ] MinCoords [ " << xsect_grid_min_coords 
		  << " ] MaxCoords [ " << xsect_grid_max_coords
		  << " ] MinValue [ " << xsect_middle_grid_min_value
		  << " ] MaxValue [ " << xsect_middle_grid_max_value 
		  << " ]";

	const std::string file_comment = strstream.str();

	strstream.str("");

	xsect_comments_middle.push_back(file_comment);

	mdata << "  " << file_comment << endl;

	write_grid(output_filename, xsect_grid, xsect_comments_middle);
	}

      if (output_image)
	{
	const std::string image_filename 
	  = xsect_middle_output_filename_nosuffix + image_suffix;

	const std::string image_filename_simple 
	  = xsect_middle_output_filename_nosuffix_name + image_suffix;

	strstream << "XSect [ Image ] Type [ " << xsect_middle_type 
		  << " ] Pos [ " << xsect_middle_x_pos 
		  << " ] File [ " << image_filename_simple
		  << " ] MinCoords [ " << xsect_grid_min_coords 
		  << " ] MaxCoords [ " << xsect_grid_max_coords
		  << " ] MinValue [ " << xsect_middle_grid_min_value
		  << " ] MaxValue [ " << xsect_middle_grid_max_value 
		  << " ]";

	const std::string file_comment = strstream.str();

	strstream.str("");

	mdata << "  " << file_comment << endl;

#ifdef USE_MAGICK
	const int error_state = write_image(image_filename, xsect_grid);

	if (error_state < 0)
	  {
	  cerr << endl << "Error: Unable to write image " << image_filename 
	       << endl << endl;
	  return -2;
	  }
#else
	cerr << endl << "Warning: ImageMagick not available" << endl << endl;
#endif  // USE_MAGICK
	}
      }


    // Free some memory

    dataset.clear_cache();
    }

  cout << endl;

  return 0;
  }

