// C++ program file shock_front.cpp

/*
  By Nathan C. Hearn
     April 28, 2008

  Routines for locating the shock front in Omega RT simulations.
*/


#include "shock_front.hpp"
#include <vector>
#include <string>
#include <algorithm>
#include <cmath>
#include "quickflash_except.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_tree_block.hpp"
#include "quickflash_block_blockinfo.hpp"


bool range_radius_overlap(const std::vector<double> & min_range_coords,
			  const std::vector<double> & max_range_coords,
			  const std::vector<double> & radius_center,
			  const double max_radius)
  {
  bool ret_val = false;

  const unsigned int dims = radius_center.size();

  if (min_range_coords.size() != dims)
    throw QuickFlash::Except("Incompatible min coords vector", __FILE__, 
			   __LINE__);

  if (max_range_coords.size() != dims)
    throw QuickFlash::Except("Incompatible min coords vector", __FILE__, 
			   __LINE__);

  if (!(max_radius < 0.0))
    {
    // Calculate the minimum distance from the region to the origin

    double min_sq_dist = 0.0;

    for (unsigned int i = 0; i < dims; i++)
      {
      const double center_coord = radius_center[i];
	
      const double rel_min_coord = min_range_coords[i] - center_coord;
      const double rel_max_coord = max_range_coords[i] - center_coord;

      // Only include axis component if no overlap with origin

      if ((rel_min_coord * rel_max_coord) > 0.0)
	{
	const double axis_dist = std::min(std::fabs(rel_min_coord),
					  std::fabs(rel_max_coord));

	min_sq_dist += axis_dist * axis_dist;
	}
      }

    const double max_sqRadius = max_radius * max_radius;

    if (!(min_sq_dist > max_sqRadius))
      ret_val = true;
    }

  return ret_val;
  }


void project_vector(const unsigned int skip_axis_index,
		    const std::vector<double> & full_vector,
		    std::vector<double> & reduced_vector)
  {
  const unsigned int dims = full_vector.size();

  if (dims == 0)
    reduced_vector.clear();
  else
    {
    if (skip_axis_index >= dims)
      throw QuickFlash::Except("Axis index out of bounds", __FILE__, __LINE__);

    const unsigned int reduced_dims = dims - 1;

    reduced_vector.resize(reduced_dims);

    unsigned int reduced_index = 0;

    for (unsigned int i = 0; i < dims; i++)
      if (i != skip_axis_index)
	reduced_vector[reduced_index++] = full_vector[i];
    }
  }


int find_axis_peak(const QuickFlash::File::DataFile & df,
		   const std::string & dataset_name, 
		   const unsigned int search_axis, double & peak_coord,
		   const bool limit_sample_radius,
		   const double max_sample_radius,
		   const bool limit_axis_bounds,
		   const double min_axis_bound,
		   const double max_axis_bound)
  {
  int ret_val = -1;  // Error state

  // FOR RIGHT NOW, JUST FIND THE POSITION OF THE PEAK VALUE OF THE DATASET

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (search_axis >= dims)
    throw QuickFlash::Except("Axis index out of bounds", __FILE__, __LINE__);

  if (limit_axis_bounds)
    if (!(min_axis_bound < max_axis_bound))
      throw QuickFlash::Except("Incompatible axis bounds", __FILE__, __LINE__);

  const std::vector<double> & min_bounds = meshinfo.get_volume_minbounds();
  const std::vector<double> & max_bounds = meshinfo.get_volume_maxbounds();

  const double sq_max_sample_radius = max_sample_radius * max_sample_radius;

  std::vector<double> domain_center;
  std::vector<double> reduced_domain_center;

  if (limit_sample_radius)
    {
    if (!(max_sample_radius > 0.0))
      throw QuickFlash::Except("Sample radius must be a positive value", 
			     __FILE__, __LINE__);

    domain_center.resize(dims);

    for (unsigned int i = 0; i < dims; i++)
      domain_center[i] = 0.5 * (min_bounds[i] + max_bounds[i]);

    project_vector(search_axis, domain_center, reduced_domain_center);
    }

  const QuickFlash::File::Dataset & dataset = df.get_dataset(dataset_name);

  QuickFlash::Block::BlockData<double> block_data;

  std::vector<double> cell_pos;

  bool found_value = false;

  double peak_value = 0.0;
  double peak_value_pos = 0.0;
  
  std::vector<double> reduced_block_min_coords;
  std::vector<double> reduced_block_max_coords;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    const std::vector<double> & block_min_coords = block_info.get_min_coords();
    const std::vector<double> & block_max_coords = block_info.get_max_coords();
    
    bool skip_branch = false;

    if (limit_axis_bounds)
      {
      if ((block_min_coords[search_axis] > max_axis_bound)
	  || (block_max_coords[search_axis] < min_axis_bound))
	skip_branch = true;
      }

    if ((!skip_branch) && limit_sample_radius)
      {
      // Check the maximum radius condition

      project_vector(search_axis, block_min_coords, reduced_block_min_coords);
      project_vector(search_axis, block_max_coords, reduced_block_max_coords);
      
      // Skip branch if no overlap with sample radius

      if (!range_radius_overlap(reduced_block_min_coords, 
				reduced_block_max_coords, 
				reduced_domain_center,
				max_sample_radius))
	skip_branch = true;
      }

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      dataset.get_block_data(block_index, block_data);

      const unsigned int num_cells = block_info.get_num_cells();

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	const double cell_value = block_data[cell_index];

	if ((!found_value) || (cell_value > peak_value))
	  {
	  block_info.get_cell_center(cell_index, cell_pos);

	  bool use_cell = true;

	  if (limit_axis_bounds)
	    {
	    if ((cell_pos[search_axis] < min_axis_bound) 
		|| (cell_pos[search_axis] > max_axis_bound))
	      use_cell = false;
	    }

	  if ((!use_cell) && limit_sample_radius)
	    {
	    double sq_cell_pos_radius = 0.0;

	    for (unsigned int i = 0; i < dims; i++)
	      if (i != search_axis)
		{
		const double axis_comp = cell_pos[i] - domain_center[i];

		sq_cell_pos_radius += axis_comp * axis_comp;
		}

	    if (sq_cell_pos_radius > sq_max_sample_radius)
	      use_cell = false;
	    }

	  if (use_cell)
	    {
	    peak_value = cell_value;
	    peak_value_pos = cell_pos[search_axis];

	    found_value = true;
	    }
	  }
	}
      }
    
    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  if (found_value)
    {
    peak_coord = peak_value_pos;

    ret_val = 0;  // OK
    }

  return ret_val;
  }


int find_axis_peak(const QuickFlash::File::DataFile & df,
		   const char dataset_name[], 
		   const unsigned int search_axis, double & peak_coord,
		   const bool limit_sample_radius,
		   const double max_sample_radius,
		   const bool limit_axis_bounds,
		   const double min_axis_bound,
		   const double max_axis_bound)
  {
  const std::string dataset_name_str = dataset_name;

  return find_axis_peak(df, dataset_name_str, search_axis, peak_coord,
			limit_sample_radius, max_sample_radius,
			limit_axis_bounds, min_axis_bound, max_axis_bound);
  }
