// C++ program file traj_mpi_utils.cpp

/*
  By Nathan C. Hearn
     November 20, 2008

  MPI-related utilities for particle trajectory extraction.
*/


#ifdef USE_MPI


#include "traj_mpi_utils.hpp"
#include <mpi.h>
#include <vector>
#include <string>


const int mpi_tag_part_tag = 100;
const int mpi_tag_time_count = 200;
const int mpi_tag_num_elems = 300;
const int mpi_tag_time_array = 400;
const int mpi_tag_elem_array_base = 500;


int broadcast_name_list_send(const std::vector<std::string> & elem_name_list,
			     const MPI_Comm mpi_comm)
  {
  int ret_val = -1;  // Error state

  int mpi_rank = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);

  unsigned int num_names = elem_name_list.size();

  // Send the number of names

  if (MPI_Bcast(&num_names, 1, MPI_UNSIGNED, mpi_rank, mpi_comm) 
      == MPI_SUCCESS)
    {
    ret_val = 0;  // OK for now ...

    if (num_names > 0)
      {
      ret_val = -1;  // Error state

      unsigned int * name_len_list = new unsigned int[num_names];

      unsigned int total_name_len = 0;

      for (unsigned int name_index = 0; name_index < num_names; name_index++)
	{
	const unsigned int name_len = elem_name_list[name_index].size();

	name_len_list[name_index] = name_len;
	
	total_name_len += name_len;
	}

      if (MPI_Bcast(name_len_list, num_names, MPI_UNSIGNED, mpi_rank, 
		    mpi_comm) == MPI_SUCCESS)
	{
	ret_val = 0;  // OK for now ...

	if (total_name_len > 0)
	  {
	  ret_val = -1;  // Error state

	  char * name_char_list = new char[total_name_len];

	  unsigned int char_pos = 0;

	  for (unsigned int name_index = 0; name_index < num_names; 
	       name_index++)
	    {
	    const std::string & elem_name = elem_name_list[name_index];

	    const unsigned int name_len = elem_name.size();

	    for (unsigned int char_index = 0; char_index < name_len; 
		 char_index++)
	      name_char_list[char_pos++] = elem_name[char_index];
	    }

	  if (MPI_Bcast(name_char_list, num_names, MPI_CHAR, mpi_rank, 
			mpi_comm) == MPI_SUCCESS)
	    ret_val = 0;  // OK

	  delete[] name_char_list;
	  name_char_list = 0;
	  }
	}

      delete[] name_len_list;
      name_len_list = 0;
      }
    }

  return ret_val;
  }



int broadcast_name_list_recv(std::vector<std::string> & elem_name_list, 
			     const int root_rank, const MPI_Comm mpi_comm)
  {  
  int ret_val = -1;  // Error state

  int mpi_rank = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);

  if (root_rank != mpi_rank)
    {
    unsigned int num_names = 0;

    if (MPI_Bcast(&num_names, 1, MPI_UNSIGNED, root_rank, mpi_comm) 
	== MPI_SUCCESS)
      {
      ret_val = 0;  // OK for now ...

      elem_name_list.clear();

      if (num_names > 0)
	{
	ret_val = -1;  // Error state

	elem_name_list.resize(num_names);
	
	unsigned int * name_len_list = new unsigned int[num_names];

	if (MPI_Bcast(name_len_list, num_names, MPI_UNSIGNED, root_rank, 
		      mpi_comm) == MPI_SUCCESS)
	  {
	  ret_val = 0;  // OK for now

	  unsigned int total_name_len = 0;

	  for (unsigned int name_index = 0; name_index < num_names; 
	       name_index++)
	    total_name_len += name_len_list[name_index];

	  if (total_name_len > 0)
	    {
	    ret_val = -1;  // Error state

	    char * name_char_list = new char[total_name_len];

	    if (MPI_Bcast(name_char_list, total_name_len, MPI_CHAR, root_rank,
			  mpi_comm) == MPI_SUCCESS)
	      {
	      ret_val = 0;  // OK

	      // Assign the names

	      unsigned int char_pos = 0;

	      for (unsigned int name_index = 0; name_index < num_names; 
		   name_index++)
		{
		const unsigned int current_name_len 
		  = name_len_list[name_index];

		std::string & elem_name = elem_name_list[name_index];

		if (current_name_len > 0)
		  {
		  const char * char_ptr = &(name_char_list[char_pos]);

		  elem_name.assign(char_ptr, current_name_len);

		  char_pos += current_name_len;
		  }
		}
	      }

	    delete[] name_char_list;
	    name_char_list = 0;
	    }
	  }

	delete[] name_len_list;
	name_len_list = 0;
	}
      }
    }

  return ret_val;
  }


int send_particle_data(const int dest_rank, const unsigned int particle_tag,
		       const std::vector<double> & time_list,
		       const std::vector< std::vector<double> > & data_elems,
		       const MPI_Comm mpi_comm)
  {
  int ret_val = -1;  // Error state

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if ((dest_rank < mpi_size) && (dest_rank != mpi_rank))
    {
    unsigned int part_tag = particle_tag;

    if (MPI_Send(&part_tag, 1, MPI_UNSIGNED, dest_rank, mpi_tag_part_tag,
		 mpi_comm) == MPI_SUCCESS)
      {
      unsigned int num_times = time_list.size();

      if (MPI_Send(&num_times, 1, MPI_UNSIGNED, dest_rank, mpi_tag_time_count,
		   mpi_comm) == MPI_SUCCESS)
	{
	unsigned int num_data_elems = data_elems.size();

	if (MPI_Send(&num_data_elems, 1, MPI_UNSIGNED, dest_rank, 
		     mpi_tag_num_elems, mpi_comm) == MPI_SUCCESS)
	  {
	  ret_val = 0;  // OK for now ...

	  if (num_times > 0)
	    {
	    ret_val = -1;  // Error state

	    double * time_array = new double[num_times];

	    for (unsigned int time_index = 0; time_index < num_times; 
		 time_index++)
	      time_array[time_index] = time_list[time_index];

	    if (MPI_Send(time_array, num_times, MPI_DOUBLE, dest_rank,
			 mpi_tag_time_array, mpi_comm) == MPI_SUCCESS)
	      {
	      ret_val = 0;  // OK for now

	      if (num_data_elems > 0)
		{
		double * elem_array = new double[num_times];

		for (unsigned int elem_index = 0; elem_index < num_data_elems;
		     elem_index++)
		  {
		  const std::vector<double> & current_elem 
		    = data_elems[elem_index];

		  if (current_elem.size() != num_times)
		    ret_val = -1;  // Error!
		  else
		    {
		    for (unsigned int time_index = 0; time_index < num_times;
			 time_index++)
		      elem_array[time_index] = current_elem[time_index];

		    const int mpi_tag 
		      = mpi_tag_elem_array_base + static_cast<int>(elem_index);

		    if (MPI_Send(elem_array, num_times, MPI_DOUBLE, dest_rank,
				 mpi_tag, mpi_comm) != MPI_SUCCESS)
		      ret_val = -1;  // Error!
		    }
		  }

		delete[] elem_array;
		elem_array = 0;
		}
	      }

	    delete[] time_array;
	    time_array = 0;
	    }
	  }
	}
      }
    }

  return ret_val;
  }







int recv_particle_data(const int src_rank, unsigned int & particle_tag,
		       std::vector<double> & time_list,
		       std::vector< std::vector<double> > & data_elems,
		       const MPI_Comm mpi_comm)
  {
  int ret_val = -1;  // Error state

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  MPI_Status status;

  time_list.clear();
  data_elems.clear();

  if ((src_rank < mpi_size) && (src_rank != mpi_rank))
    {
    unsigned int part_tag = 0;

    if (MPI_Recv(&part_tag, 1, MPI_UNSIGNED, src_rank, mpi_tag_part_tag,
		 mpi_comm, &status) == MPI_SUCCESS)
      {
      particle_tag = part_tag;

      unsigned int num_times = 0;

      if (MPI_Recv(&num_times, 1, MPI_UNSIGNED, src_rank, mpi_tag_time_count,
		   mpi_comm, &status) == MPI_SUCCESS)
	{
	time_list.resize(num_times, 0.0);

	unsigned int num_data_elems = 0;

	if (MPI_Recv(&num_data_elems, 1, MPI_UNSIGNED, src_rank, 
		     mpi_tag_num_elems, mpi_comm, &status) == MPI_SUCCESS)
	  {
	  ret_val = 0;  // OK for now ...

	  data_elems.resize(num_data_elems);

	  for (unsigned int elem_index = 0; elem_index < num_data_elems;
	       elem_index++)
	    data_elems[elem_index].resize(num_times, 0.0);

	  if (num_times > 0)
	    {
	    ret_val = -1;  // Error state

	    double * time_array = new double[num_times];

	    if (MPI_Recv(time_array, num_times, MPI_DOUBLE, src_rank,
			 mpi_tag_time_array, mpi_comm, &status) == MPI_SUCCESS)
	      {
	      ret_val = 0;  // OK for now ...

	      for (unsigned int time_index = 0; time_index < num_times; 
		   time_index++)
		time_list[time_index] = time_array[time_index];

	      if (num_data_elems > 0)
		{
		double * elem_array = new double[num_times];

		for (unsigned int elem_index = 0; elem_index < num_data_elems;
		     elem_index++)
		  {
		  const int mpi_tag 
		    = mpi_tag_elem_array_base + static_cast<int>(elem_index);

		  if (MPI_Send(elem_array, num_times, MPI_DOUBLE, src_rank,
			       mpi_tag, mpi_comm) != MPI_SUCCESS)
		    ret_val = -1;  // Error!
		  else
		    {
		    std::vector<double> & current_elem 
		      = data_elems[elem_index];

		    for (unsigned int time_index = 0; time_index < num_times;
		       time_index++)
		      current_elem[time_index] = elem_array[time_index];
		    }
		  }

		delete[] elem_array;
		elem_array = 0;
		}
	      }

	    delete[] time_array;
	    time_array = 0;
	    }
	  }
	}
      }
    }

  return ret_val;
  }


#endif  // USE_MPI
