// C++ program file radial_profile.cpp

/*
  By Nathan C. Hearn
     October 12, 2008

  Program for performing simple radial profiles.
*/


#include <iostream>
#include <fstream>
#include <cmath>
#include <cstdlib>
#include <vector>
#include <list>
#include <algorithm>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_stats_bins.hpp"
#include "quickflash_except.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_option.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_utils_file.hpp"


const char default_comment_marker[] = "#";
const double default_min_radius = 0.0;

const unsigned int buffer_size = 200;


double get_max_rad(const std::vector<double> & center_coords,
		   const std::vector<double> & min_bounds,
		   const std::vector<double> & max_bounds)
  {
  const unsigned int dims = center_coords.size();

  if ((min_bounds.size() != dims) || (max_bounds.size() != dims))
    throw QuickFlash::Except("Incompatible boundary vector size", __FILE__,
			   __LINE__);

  double max_sq_rad = 0.0;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double axis_center = center_coords[i];

    const double min_center_dist = std::fabs(min_bounds[i] - axis_center);
    const double max_center_dist = std::fabs(max_bounds[i] - axis_center);

    const double axis_dist = std::max(min_center_dist, max_center_dist);

    max_sq_rad += axis_dist * axis_dist;
    }

  return std::sqrt(max_sq_rad);
  }


void fill_bins(const QuickFlash::File::DataFile & df, 
	       const std::vector<double> & center_coords, 
	       const double min_radius, const double bin_width, 
	       const unsigned int num_bins,
	       const std::vector<std::string> & dataset_name_list, 
	       QuickFlash::Stats::BinArray & dens_bin_array,
	       std::vector<QuickFlash::Stats::BinArray> & user_bin_array_list)
  {
  const QuickFlash::File::MeshInfo & mesh_info = df.get_mesh_info();

  const unsigned int dims = mesh_info.get_dims();

  if (center_coords.size() != dims)
    throw QuickFlash::Except("Incompatible dimensions for center coords array",
			   __FILE__, __LINE__);

  const unsigned int num_datasets = dataset_name_list.size();

  const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");

  std::vector<const QuickFlash::File::Dataset *> user_datasets(num_datasets, 0);

  for (unsigned int dset_index = 0; dset_index < num_datasets; dset_index++)
    user_datasets[dset_index] 
      = &(df.get_dataset(dataset_name_list[dset_index]));

  // Prepare the bin arrays

  const double max_radius = static_cast<double>(num_bins) * bin_width;

  const double min_sq_rad = min_radius * min_radius;
  const double max_sq_rad = max_radius * max_radius;

  dens_bin_array.reset(min_radius, num_bins, bin_width);

  user_bin_array_list.resize(num_datasets);

  for (unsigned int dset_index = 0; dset_index < num_datasets; dset_index++)
    user_bin_array_list[dset_index].reset(min_radius, num_bins, bin_width);

  // Run through the cells

  std::vector<double> cell_center;

  QuickFlash::Block::BlockData<double> dens_data;

  std::vector< QuickFlash::Block::BlockData<double> > 
    user_data_list(num_datasets);

  // Don't worry about being too efficient for now ...

  const unsigned int num_blocks = mesh_info.get_num_blocks();

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    if (mesh_info.is_leaf(block_index))
      {
      const QuickFlash::Block::BlockInfo & block_info 
	= mesh_info.get_block_info(block_index);

      // Assume cartesian geometry

      const double cell_vol = block_info.get_cell_volume();

      const unsigned int num_cells = block_info.get_num_cells();

      bool block_loaded = false;

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	// Don't use cell widths for now

	block_info.get_cell_center(cell_index, cell_center);

	double sq_dist = 0.0;

	for (unsigned int i = 0; i < dims; i++)
	  {
	  const double axis_disp = cell_center[i] - center_coords[i];

	  sq_dist += axis_disp * axis_disp;
	  }

	if (!((sq_dist < min_sq_rad) || (sq_dist > max_sq_rad)))
	  {
	  const double cell_rad = std::sqrt(sq_dist);

	  if (!block_loaded)
	    {
	    dens_set.get_block_data(block_index, dens_data);

	    for (unsigned int dset_index = 0; dset_index < num_datasets; 
		 dset_index++)
	      {
	      const QuickFlash::File::Dataset & dset
		= *(user_datasets[dset_index]);

	      dset.get_block_data(block_index, user_data_list[dset_index]);
	      }

	    block_loaded = true;
	    }

	  const double cell_dens_value = dens_data[cell_index];
	 
	  dens_bin_array.add_data(cell_rad, cell_dens_value, cell_vol);

	  const double cell_mass = cell_dens_value * cell_vol;

	  for (unsigned int dset_index = 0; dset_index < num_datasets; 
	       dset_index++)
	    {
	    const double user_cell_value 
	      = user_data_list[dset_index][cell_index];

	    user_bin_array_list[dset_index].add_data(cell_rad, user_cell_value,
						     cell_mass);
	    }
	  }
	}
      }
  }

  
int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  // Parse the command line arguments

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 2;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name << " bin_width"
	 << endl << "\t"
	 << "[ --output-prefix=OUTPUT_PREFIX ] "
	 << endl << "\t"
	 << "[ --datasets=DATASET1,DATASET2,... ]"
	 << endl << "\t"
	 << "[ --min-radius=MIN_RADIUS ] [ --max-radius=MAX_RADIUS ]"
	 << endl << "\t"
	 << "[ --center-pos=X,Y,... ] [ --use-comment-marker[=MARKER] ]"
	 << endl << "\t"
	 << "DATA_FILE_1 [ DATA_FILE_2 ... ]"
	 << endl << endl;

    return 1;
    }

  // Read the required arguments

  unsigned int argptr = 0;

  const double bin_width = std::atof(argument_list[argptr++].c_str());

  if (!(bin_width > 0.0))
    {
    cerr << endl << "Error: Bin width [ " << bin_width 
	 << " ] must be positive" << endl << endl;
    return -1;
    }

  std::list<std::string> datafile_name_list;

  while (argptr < num_args)
    datafile_name_list.push_back(argument_list[argptr++]);

    // Read the optional arguments

  std::string base_output_prefix = "";

  std::vector<std::string> user_dataset_list;

  double min_rad = default_min_radius;

  bool user_max_rad_given = false;
  double user_max_rad = 0.0;

  bool user_center_pos_given = false;
  std::vector<double> user_center_pos;

  bool use_comment_marks = false;
  std::string user_comment_marker = default_comment_marker;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "output-prefix")
      {
      if (!(current_option.value_present()))
	{
	cerr << endl << "Error: Output prefix is required with option"
	     << endl << endl;
	return -1;
	}

      base_output_prefix = current_option.get_value();

      // Add an underscore if prefix contains more than just a path

      std::string base_prefix_path;
      std::string base_prefix_name;

      QuickFlash::Utils::split_name_path_file(base_output_prefix, 
					    base_prefix_path, 
					    base_prefix_name);

      if (base_prefix_name.size() > 0)
	base_output_prefix += "_";
      }
    else if (keyword == "datasets")
      {
      if (!(current_option.value_present()))
	{
	cerr << endl << "Error: Dataset option requires "
	     << "comma-separated list of dataset names"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(user_dataset_list);
      }
    else if (keyword == "min-radius")
      {
      if (!(current_option.value_present()))
	{
	cerr << endl 
	     << "Error: Min radius option requires a non-negative value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(min_rad);

      if (min_rad < 0.0)
	{
	cerr << endl 
	     << "Error: Min radius option requires a non-negative value"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "max-radius")
      {
      if (!(current_option.value_present()))
	{
	cerr << endl 
	     << "Error: Max radius option requires a positive value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(user_max_rad);

      if (!(user_max_rad > 0.0))
	{
	cerr << endl 
	     << "Error: Max radius option requires a positive value"
	     << endl << endl;
	return -1;
	}

      user_max_rad_given = true;
      }
    else if (keyword == "center-pos")
      {
      if (!(current_option.value_present()))
	{
	cerr << endl 
	     << "Error: Center-position option requires a comma-separated "
	     << "coordinate vector"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(user_center_pos);

      user_center_pos_given = true;
      }
    else if (keyword == "use-comment-marker")
      {
      use_comment_marks = true;

      if (current_option.value_present())
	user_comment_marker = current_option.get_value();
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }

  // Set up any comment marks

  const std::string comment_marker 
    = use_comment_marks ? (user_comment_marker + " ") : "";

  // Perform some additional checks
    
  if (user_max_rad_given)
    if (!(user_max_rad > min_rad))
      {
      cerr << "Error: Max radius [ " << user_max_rad 
	   << " ] must be greater than min [ " << min_rad << " ]" 
	   << endl << endl;
      return -1;
      }

  const unsigned int num_datasets = user_dataset_list.size();

  // Run through the files

  std::list<std::string>::const_iterator filename_iter 
    = datafile_name_list.begin();

  while (filename_iter != datafile_name_list.end())
    {
    const std::string & filename = *filename_iter;

    // Set up the output filename prefix

    std::string datafile_path;
    std::string datafile_base_name;

    QuickFlash::Utils::split_name_path_file(filename, datafile_path, 
					  datafile_base_name);

    const std::string output_prefix = base_output_prefix + datafile_base_name;

    // Open the file

    const QuickFlash::File::DataFile df(filename, buffer_size, true);

    const QuickFlash::File::MeshInfo & mesh_info = df.get_mesh_info();

    const unsigned int dims = mesh_info.get_dims();

    const std::vector<double> & min_bounds = mesh_info.get_volume_minbounds();
    const std::vector<double> & max_bounds = mesh_info.get_volume_maxbounds();

    std::vector<double> center_coords(dims, 0.0);

    if (user_center_pos_given)
      {
      if (user_center_pos.size() != dims)
	{
	cerr << endl << "Error: Center coordinate vector [ " 
	     << user_center_pos << " ] has incompatible dimensions"
	     << endl << endl;
	return -1;
	}

      center_coords = user_center_pos;
      }
    else
      {
      // Choose the center of the volume

      for (unsigned int i = 0; i < dims; i++)
	{
	const double axis_min_coord = min_bounds[i];
	const double axis_max_coord = max_bounds[i];

	center_coords[i] = 0.5 * (axis_min_coord + axis_max_coord);
	}
      }

    // Determine the max radius

    double max_rad = user_max_rad;

    if (!user_max_rad_given)
      {
      max_rad = get_max_rad(center_coords, min_bounds, max_bounds);

      if (!(max_rad > min_rad))
	{
	cerr << "Error: Min radius [ " << min_rad 
	     << " ] is not less than computed max radius [ " << max_rad 
	     << " ]" << endl << endl;
	return -1;
	}
      }

    // Get the bin data

    const unsigned int num_bins 
      = static_cast<unsigned int>((max_rad - min_rad) / bin_width) + 1;

    QuickFlash::Stats::BinArray dens_bin_array;

    std::vector<QuickFlash::Stats::BinArray> user_bin_array_list;

    fill_bins(df, center_coords, min_rad, bin_width, num_bins, 
	      user_dataset_list, dens_bin_array, user_bin_array_list);

    // Report on the data

    const double array_min_rad = dens_bin_array.get_min_boundary();
    const double array_max_rad = dens_bin_array.get_max_boundary();

    const std::string output_filename = output_prefix + "_rad_profile.csv";

    std::ofstream f(output_filename.c_str(), std::ios::trunc);

    f << comment_marker << "DataFile " << datafile_base_name << endl;
    f << comment_marker << "CenterCoords " << center_coords << endl;
    f << comment_marker << "MinRadius " << array_min_rad << endl;
    f << comment_marker << "MaxRadius " << array_max_rad << endl;
    f << comment_marker << "BinWidth " << bin_width << endl;
    f << comment_marker << "NumBins " << num_bins << endl;
    f << comment_marker << "Datasets " << user_dataset_list << endl;

    f << endl;

    f << comment_marker << "Data" << endl;

    f << endl;

    f << "Bin_Index" << "\t" << "Bin_Center" << "\t" << "Bin_Mass" << "\t"
      << "Bin_Mean_Density" << "\t" << "Bin_Density_StdDev" << "\t"
      << "Bin_Min_Density" << "\t" << "Bin_Max_Density";

    for (unsigned int dset_index = 0; dset_index < num_datasets; dset_index++)
      {
      const std::string & dataset_prefix = user_dataset_list[dset_index] + "_";

      f << "\t" << (dataset_prefix + "Mean") 
	<< "\t" << (dataset_prefix + "StdDev") 
	<< "\t" << (dataset_prefix + "Min_Value") 
	<< "\t" << (dataset_prefix + "Max_Value");
      }

    f << endl;

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      const double bin_center = dens_bin_array.get_bin_center(bin_index);

      const QuickFlash::Stats::Bin & dens_bin = dens_bin_array[bin_index];

      const double bin_mass = dens_bin.get_weight_value_sum();

      const bool bin_ok = (bin_mass > 0.0) ? true : false;

      const double bin_dens_mean = bin_ok ? dens_bin.get_mean() : 0.0;
      const double bin_dens_stddev = bin_ok ? dens_bin.get_pop_stddev() : 0.0;
      const double bin_dens_min = bin_ok ? dens_bin.get_min_value() : 0.0;
      const double bin_dens_max = bin_ok ? dens_bin.get_max_value() : 0.0;

      f << bin_index << "\t" << bin_center << "\t" << bin_mass << "\t"
	<< bin_dens_mean << "\t" << bin_dens_stddev << "\t" << bin_dens_min
	<< "\t" << bin_dens_max;

      for (unsigned int dset_index = 0; dset_index < num_datasets; 
	   dset_index++)
	{
	const QuickFlash::Stats::Bin & dset_bin
	  = user_bin_array_list[dset_index][bin_index];

	const double dset_mean = bin_ok ? dset_bin.get_mean() : 0.0;
	const double dset_stddev = bin_ok ? dset_bin.get_pop_stddev() : 0.0;
	const double dset_min = bin_ok ? dset_bin.get_min_value() : 0.0;
	const double dset_max = bin_ok ? dset_bin.get_max_value() : 0.0;

	f << "\t" << dset_mean << "\t" << dset_stddev << "\t" << dset_min 
	  << "\t" << dset_max;
	}

      f << endl;
      }

    f << endl;
    
    ++filename_iter;
    }

  return 0;
  }
