// C++ program file detector.cpp

/*
  By Nathan C. Hearn
     November 3, 2006

  Class describing detector apparatus.
*/


#include "detector.hpp"
#include <vector>
#include <math.h>
#include "quickflash_except.hpp"


// Class Detector

const unsigned int Detector::dims = 3;
const unsigned int Detector::film_dims = 2;


void Detector::reset()
  {
  ray_gen.reset();

  film.reset();

  x_pixel_size = 0.0;
  y_pixel_size = 0.0;

  film_start_pixel_center.clear();

  film_parallel.clear();
  film_tangent.clear();
  }


void Detector::reset(const std::vector<double> & center_pos,
		     const std::vector<double> & min_domain_coords,
		     const std::vector<double> & max_domain_coords,
		     const double detector_azimuth,
		     const double source_dist, const double source_radius,
		     const double film_dist, 
		     const std::vector<double> & film_size,
		     const std::vector<unsigned int> & film_resolution)
  {
  using QuickFlash::Except;

  // Error checkin

  if (center_pos.size() != dims)
    throw Except("Incompatible center_pos vector", __FILE__, __LINE__);

  if (min_domain_coords.size() != dims)
    throw Except("Incompatible min_domain_coords vector", __FILE__,
		 __LINE__);

  if (max_domain_coords.size() != dims)
    throw Except("Incompatible max_domain_coords vector", __FILE__,
		 __LINE__);

  for (unsigned int i = 0; i < dims; i++)
    if (!(min_domain_coords[i] < max_domain_coords[i]))
      throw Except("Min-max domain coords mismatch", __FILE__, __LINE__);

  if (!(source_dist > 0.0))
    throw Except("Non-positive source distance", __FILE__, __LINE__);

  if (!(film_dist > 0.0))
    throw Except("Non-positive film distance", __FILE__, __LINE__);

  if (film_size.size() != film_dims)
    throw Except("Incompatible film_size vector", __FILE__, __LINE__);

  for (unsigned int m = 0; m < film_dims; m++)
    if (!(film_size[m] > 0.0))
      throw Except("Non-positive film_size element", __FILE__, __LINE__);

  if (film_resolution.size() != film_dims)
    throw Except("Incompatible film_resolution vector", __FILE__, __LINE__);

  for (unsigned int m = 0; m < film_dims; m++)
    if (film_resolution[m] < 1)
      throw Except("Non-positive film_resolution element", __FILE__, __LINE__);

  // Find the source position

  std::vector<double> source_pos(dims);

  const double detector_x = center_pos[0];

  const double axis_y = center_pos[1];
  const double axis_z = center_pos[2];

  source_pos[0] = detector_x;
  source_pos[1] = axis_y + (source_dist * cos(detector_azimuth));
  source_pos[2] = axis_z + (source_dist * sin(detector_azimuth));

  ray_gen.reset(source_pos, source_radius, axis_y, axis_z,
		min_domain_coords, max_domain_coords);

  // Set up the film

  const std::vector<double> film_zero(film_dims, 0.0);

  film.reset(film_resolution, film_zero, film_size);

  // Set up the film vectors

  x_pixel_size = film_size[0] / static_cast<double>(film_resolution[0]);

  y_pixel_size = film_size[1] / static_cast<double>(film_resolution[1]);

  std::vector<double> film_center(dims);

  film_center[0] = detector_x;
  film_center[1] = axis_y - (film_dist * cos(detector_azimuth));
  film_center[2] = axis_z - (film_dist * sin(detector_azimuth));

  // Get an orthogonal set of vectors parallel to the film surface

  std::vector<double> film_normal = film_center;

  for (unsigned int i = 0; i < dims; i++)
    film_normal[i] -= source_pos[i];

  film_parallel.resize(dims);

  film_parallel[0] = x_pixel_size;  // Parallel to cylinder axis (x)
  film_parallel[1] = 0.0;
  film_parallel[2] = 0.0;

  film_tangent.resize(dims);

  film_tangent[0] = 0.0;
  film_tangent[1] = film_normal[2];
  film_tangent[2] = -film_normal[1];

  double sqLen = 0.0;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double axis_comp = film_tangent[i];

    sqLen += axis_comp * axis_comp;
    }

  const double tangent_len = sqrt(sqLen);

  if (!(tangent_len > 0.0))
    throw Except("Unable to set film dimensions", __FILE__, __LINE__);

  const double y_pixel_size_over_tangent_len = y_pixel_size / tangent_len;

  for (unsigned int i = 0; i < dims; i++)
    film_tangent[i] *= y_pixel_size_over_tangent_len;

  // Find the center of the first pixel

  film_start_pixel_center = film_center;

  const double film_x_disp = static_cast<double>(film_resolution[0] - 1) * 0.5;
  const double film_y_disp = static_cast<double>(film_resolution[1] - 1) * 0.5;

  for (unsigned int i = 0; i < dims; i++)
    {
    film_start_pixel_center[i] -= (film_x_disp * film_parallel[i])
      + (film_y_disp * film_tangent[i]);
    }
  }


void Detector::get_pixel_center(const std::vector<unsigned int> & pixel_index,
				std::vector<double> & coords) const
  {
  coords = film_start_pixel_center;

  const double x_shift = static_cast<double>(pixel_index[0]);
  const double y_shift = static_cast<double>(pixel_index[1]);

  for (unsigned int i = 0; i < dims; i++)
    {
    coords[i] += (x_shift * film_parallel[i])
      + (y_shift * film_tangent[i]);
    }
  }


void Detector::get_sample_points(const unsigned int pixel_index, 
				 const double sample_spacing,
				 std::vector< std::vector<double> > & pos) 
  const
  {
  std::vector<double> pixel_center;

  get_pixel_center(pixel_index, pixel_center);

  ray_gen.get_samples(pixel_center, x_pixel_size, y_pixel_size, 
		      sample_spacing, pos);
  }

