// C++ header file detector.hpp

/*
  By Nathan C. Hearn
     November 3, 2006

  Class describing detector apparatus.
*/


#ifndef DETECTOR_HPP
#define DETECTOR_HPP


#include <vector>
#include "rays.hpp"
#include "stat_bins.hpp"


class Detector
  {
  public :
    Detector() :
      ray_gen(), film(), x_pixel_size(0.0), y_pixel_size(0.0),
      film_start_pixel_center(), film_parallel(), film_tangent()
      { }

    Detector(const std::vector<double> & center_pos,
	     const std::vector<double> & min_domain_bounds,
	     const std::vector<double> & max_domain_bounds,
	     const double detector_azimuth,
	     const double source_dist, const double source_radius,
	     const double film_dist, const std::vector<double> & film_size,
	     const std::vector<unsigned int> & film_resolution) :
      ray_gen(), film(), x_pixel_size(0.0), y_pixel_size(0.0),
      film_start_pixel_center(), film_parallel(), film_tangent()
      { 
      reset(center_pos, min_domain_bounds, max_domain_bounds, 
	    detector_azimuth, source_dist, source_radius, film_dist, film_size,
	    film_resolution);
      }
 
    ~Detector() { reset(); }

    void reset();

    void reset(const std::vector<double> & center_pos,
	       const std::vector<double> & min_domain_coords,
	       const std::vector<double> & max_domain_coords,
	       const double detector_azimuth,
	       const double source_dist, const double source_radius,
	       const double film_dist, const std::vector<double> & film_size,
	       const std::vector<unsigned int> & film_resolution);

    unsigned int get_num_pixels() const { return film.get_num_bins(); }

    void use_pixel_center(const bool state=true) 
      { ray_gen.set_use_end_point_center(state); }

    void get_pixel_center(const unsigned int pixel_index,
			  std::vector<double> & coords) const
      {
      std::vector<unsigned int> vector_index;
      get_vector_index(pixel_index, vector_index);

      get_pixel_center(vector_index, coords);
      }

    void get_pixel_center(const std::vector<unsigned int> & pixel_index,
			  std::vector<double> & coords) const;

    void get_sample_points(const unsigned int pixel_index, 
			   const double sample_spacing,
			   std::vector< std::vector<double> > & pos) const;

    void add_data(const unsigned int pixel_index, const double value)
      { film.add_data(pixel_index, value); }

    void set_data(const unsigned int pixel_index, 
		  const unsigned int num_entries,
		  const double sum_weight, const double sum_weight_values, 
		  const double sum_weight_sqValues, const double min_value, 
		  const double max_value)
      {
      film.set_data(pixel_index, num_entries, sum_weight, sum_weight_values, 
		    sum_weight_sqValues, min_value, max_value);
      }

    unsigned int get_scalar_index(const std::vector<unsigned int> 
				  & pixel_index) const
      { return film.get_scalar_index(pixel_index); }

    void get_vector_index(const unsigned int pixel_index,
			  std::vector<unsigned int> & vector_pixel_index) const
      { film.get_vector_index(pixel_index, vector_pixel_index); }

    unsigned int get_data_num_entries(const unsigned int pixel_index) const
      { return film[pixel_index].get_num_entries(); }

    double get_data_weight_sum(const unsigned int pixel_index) const
      { return film[pixel_index].get_weight_sum(); }

    double get_data_sum_weight_values(const unsigned int pixel_index) const
      { return film[pixel_index].get_sum_weight_values(); }

    double get_data_sum_weight_sqValues(const unsigned int pixel_index) const
      { return film[pixel_index].get_sum_weight_sqValues(); }

    double get_data_mean(const unsigned int pixel_index) const
      { return film[pixel_index].get_mean(); }

    double get_data_stddev(const unsigned int pixel_index) const
      {
      // Probably want a sample standard deviation, instead!!!

      return film[pixel_index].get_pop_stddev();
      }

    double get_data_min(const unsigned int pixel_index) const
      { return film[pixel_index].get_min_value(); }

    double get_data_max(const unsigned int pixel_index) const
      { return film[pixel_index].get_max_value(); }

  private:
    static const unsigned int dims;
    static const unsigned int film_dims;

    Rays ray_gen;

    Stats::BinArray film;

    double x_pixel_size;
    double y_pixel_size;

    std::vector<double> film_start_pixel_center;

    std::vector<double> film_parallel;  // Film "x" axis vector
    std::vector<double> film_tangent;  // Film "y" axis vector
  };


#endif  // DETECTOR_HPP
