// C++ header file materials.hpp

/*
  By Nathan C. Hearn
     June 7, 2007

  Absorption calculation for radiograph images.

  Absorption coefficients (in cm^2 / g) - from C. Kuranz

     Material       Backlighter       Value

       CHBr           Ti               1.00e+02
                      Sc               1.30e+02
                      V                7.85e+01  # V/Sc values switched??

       Polyimide      Ti               3.01e+01
                      Sc               3.96e+01
                      V                2.25e+01  # V/Sc values switched??

       Foam           Ti               2.48e+01
                      Sc               3.12e+01
                      V                1.63e+01

  Backlighters

     Ti  -  Used to diagnose the down-strip view
     Sc  -  Used to diagnose the across-strip view
*/


#ifndef MATERIALS_HPP
#define MATERIALS_HPP


#include <cmath>
#include <vector>
#include <string>


enum BacklighterType { Backlighter_Ti, Backlighter_Sc, Backlighter_V };


extern const char BacklighterType_Ti_label[];
extern const char BacklighterType_Sc_label[];
extern const char BacklighterType_V_label[];


int read_backlighter_type(const std::string & backlighter_label, 
			  BacklighterType & backlighter_type);

int get_backlighter_label(const BacklighterType & backlighter_type,
			   std::string & backlighter_label);


extern const double Ti_Poly_AbsCoeff;
extern const double Ti_CHBr_AbsCoeff;
extern const double Ti_Foam_AbsCoeff;

extern const double Sc_Poly_AbsCoeff;
extern const double Sc_CHBr_AbsCoeff;
extern const double Sc_Foam_AbsCoeff;

extern const double V_Poly_AbsCoeff;
extern const double V_CHBr_AbsCoeff;
extern const double V_Foam_AbsCoeff;


int get_absorb_coefficients(const BacklighterType & backlighter_type,
			    double & poly_coeff, double & chbr_coeff,
			    double & foam_coeff);

class Absorption
  {
  public:
    Absorption(const BacklighterType backlighter_type=Backlighter_Ti) : 
      backlighter(Backlighter_Ti), poly_abs_coeff(0.0), chbr_abs_coeff(0.0),
      foam_abs_coeff(0.0)
      { reset(backlighter_type); }

    ~Absorption() { }

    void reset(const BacklighterType backlighter_type);

    double get_trans_fract(const double sample_length, 
			   const std::vector<double> & poly_density,
			   const std::vector<double> & chbr_density,
			   const std::vector<double> & foam_density) const;

    double get_trans_fract(const std::vector<double> & sample_length,
			   const std::vector<double> & poly_density,
			   const std::vector<double> & chbr_density,
			   const std::vector<double> & foam_density) const;

    double get_trans_fract(const double poly_areal_density=0.0,
			   const double chbr_areal_density=0.0,
			   const double foam_areal_density=0.0) const
      {
      const double exponent = (poly_areal_density * poly_abs_coeff)
	+ (chbr_areal_density * chbr_abs_coeff)
	+ (foam_areal_density * foam_abs_coeff);

      return std::exp(-exponent);
      }

  private:

    BacklighterType backlighter;

    double poly_abs_coeff;
    double chbr_abs_coeff;
    double foam_abs_coeff;
  };


#endif  // MATERIALS_HPP
