// C++ program file rays.cpp

/*
  By Nathan C. Hearn
     November 2, 2006

  Generator for mesh sample points.
*/


#include "rays.hpp"
#include <vector>
#include <math.h>
#include <gsl/gsl_rng.h>
#include "quickflash_except.hpp"


// Class Rays

const double Rays::two_pi = 2.0 * (M_PI);

const unsigned int Rays::dims = 3;


void Rays::reset()
  {
  initialized = false;

  source_pos.clear();
  source_radius = 0.0;

  cyl_y = 0.0;
  cyl_z = 0.0;

  min_bounds.clear();
  max_bounds.clear();

  use_random_source_pos = false;

  use_end_point_center = false;

  // Free the random number generator

  if (random_gen != 0)
    {
    gsl_rng_free(random_gen);
    random_gen = 0;
    }

  source_parallel.clear();
  source_tangent.clear();
  }


void Rays::reset(const std::vector<double> & light_source_pos,
		 const double light_source_radius,
		 const double cylinder_center_y,
		 const double cylinder_center_z,
		 const std::vector<double> & min_domain_bounds,
		 const std::vector<double> & max_domain_bounds,
		 const unsigned long int random_seed)
  {
  using QuickFlash::Except;

  reset();

  source_pos = light_source_pos;

  if (source_pos.size() != dims)
    throw Except("Non-3D light source position vector", __FILE__, __LINE__);

  source_radius = light_source_radius;

  if (source_radius > 0.0)
    use_random_source_pos = true;
  else
    use_random_source_pos = false;

  use_end_point_center = false;

  cyl_y = cylinder_center_y;
  cyl_z = cylinder_center_z;

  if (min_domain_bounds.size() != dims)
    throw Except("Non-3D min domain bounds vector", __FILE__, __LINE__);

  if (max_domain_bounds.size() != dims)
    throw Except("Non-3D max domain bounds vector", __FILE__, __LINE__);

  for (unsigned int i = 0; i < dims; i++)
    if (!(min_domain_bounds[i] < max_domain_bounds[i]))
      throw Except("Min bounds -- max bounds coordinate mismatch", __FILE__,
		   __LINE__);

  min_bounds = min_domain_bounds;
  max_bounds = max_domain_bounds;

  // Build the random number generator

  random_gen = gsl_rng_alloc(gsl_rng_mt19937);

  if (random_gen == 0)
    throw Except("Unable to allocate random number generator", __FILE__, 
		 __LINE__);

  gsl_rng_set(random_gen, random_seed);

  // Set up the tangent vectors for the source

  setup_source_axes();

  initialized = true;
  }


void Rays::get_samples(const std::vector<double> & end_point_center,
		       const double end_point_box_x_width,
		       const double end_point_box_y_width,
		       const double spacing,
		       std::vector< std::vector<double> > & samples) const
  {
  using QuickFlash::Except;

  if (!initialized)
    throw Except("Ray generator not initialized", __FILE__, __LINE__);

  std::vector<double> start_point(dims);  // Source
  std::vector<double> end_point(dims);  // Film

  // Get the end point (within square around end_point_center)

  if (!use_end_point_center)
    get_random_end_pos(end_point_center, end_point_box_x_width, 
		       end_point_box_y_width, end_point);
  else
    end_point = end_point_center;

  // Get the starting point (within spacing, source_radius of source_pos)

  get_random_start_pos(end_point, spacing, start_point);

  // Get the samples along this line segment

  get_sample_points(start_point, end_point, spacing, samples);
  }


void Rays::setup_source_axes()
  {
  using QuickFlash::Except;

  /*
    Determine two orthogonal unit vectors tangent to the source.  The source
    normal is parallel to the shortest line from the center of the source
    to the axis of the cylinder.  (The cylinder runs along the x-axis.)
  */

  // Get the unit vector parallel to the cylinder

  source_parallel.resize(dims);

  source_parallel[0] = 1.0;  // Parallel to cylinder axis
  source_parallel[1] = 0.0;
  source_parallel[2] = 0.0;

  // Get a vector parallel to the source normal.

  std::vector<double> source_normal = source_pos;

  source_normal[1] -= cyl_y;
  source_normal[2] -= cyl_z;

  // The tangent is proportional to source_normal cross source_parallel

  source_tangent.resize(dims);

  source_tangent[0] = 0.0;
  source_tangent[1] = source_normal[2];
  source_tangent[2] = -1.0 * source_normal[1];

  // Now, normalize source_tangent

  double sqLen = 0.0;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double axis_len = source_tangent[i];

    sqLen += axis_len * axis_len;
    }

  const double len = sqrt(sqLen);

  if (!(len > 0.0))
    throw Except("Unable to compute source tangent vectors", __FILE__, 
		 __LINE__);

  const double oneOver_len = 1.0 / len;

  for (unsigned int i = 0; i < dims; i++)
    source_tangent[i] *= oneOver_len;
  }
  

bool Rays::in_bounds(const std::vector<double> & sample_point) const
  {
  using QuickFlash::Except;

  if (!initialized)
    throw Except("Ray generator not initialized", __FILE__, __LINE__);

  bool inside = true;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double sample_coord = sample_point[i];

    if (sample_coord < min_bounds[i])
      inside = false;
    else if (!(sample_coord < max_bounds[i]))
      inside = false;
    }

  return inside;
  }


void Rays::get_random_end_pos(const std::vector<double> & end_point_center,
			      const double end_point_box_x_width,
			      const double end_point_box_y_width,
			      std::vector<double> & end_point) const
  {
  // Determine an end-point within a square around end_point_center

  end_point = end_point_center;

  if ((end_point_box_x_width > 0.0) && (end_point_box_y_width > 0.0))
    {
    // Get random offsets along source parallel and tangent directions

    const double random_parallel 
      = end_point_box_x_width * (gsl_rng_uniform(random_gen) - 0.5);

    const double random_tangent
      = end_point_box_y_width * (gsl_rng_uniform(random_gen) - 0.5);

    // Add the offsets

    for (unsigned int i = 0; i < dims; i++)
      end_point[i] += random_parallel * source_parallel[i];

    for (unsigned int i = 0; i < dims; i++)
      end_point[i] += random_tangent * source_tangent[i];
    }
  }


void Rays::get_random_start_pos(const std::vector<double> & end_point,
				const double spacing,
				std::vector<double> & start_point) const
  {
  using QuickFlash::Except;

  // First determine the starting point within the light source

  start_point = source_pos;

  if (use_random_source_pos)
    {
    const double normalized_radius = sqrt(gsl_rng_uniform(random_gen));

    const double radius = normalized_radius * source_radius;

    const double angle = two_pi * gsl_rng_uniform(random_gen);

    const double parallel_offset = radius * cos(angle);
    const double tangent_offset = radius * sin(angle);

    for (unsigned int i = 0; i < dims; i++)
      start_point[i] += parallel_offset * source_parallel[i];

    for (unsigned int i = 0; i < dims; i++)
      start_point[i] += tangent_offset * source_tangent[i];
    }

  // Now move a random fraction of spacing from the current start_point

  std::vector<double> unit_dir = end_point;

  for (unsigned int i = 0; i < dims; i++)
    unit_dir[i] -= start_point[i];

  double sqDist = 0.0;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double axis_disp = unit_dir[i];

    sqDist += axis_disp * axis_disp;
    }

  const double distance = sqrt(sqDist);

  if (!(distance > 0.0))
    throw Except("Error computing starting point", __FILE__, __LINE__);

  const double oneOver_distance = 1.0 / distance;

  for (unsigned int i = 0; i < dims; i++)
    unit_dir[i] *= oneOver_distance;

  const double random_space = spacing * gsl_rng_uniform(random_gen);

  for (unsigned int i = 0; i < dims; i++)
    start_point[i] += random_space * unit_dir[i];
  }
  

void Rays::get_sample_points(const std::vector<double> & start_point,
			     const std::vector<double> & end_point,
			     const double spacing,
			     std::vector< std::vector<double> > & samples)
  const
  {
  // Determine the distance from start to end

  std::vector<double> pos_delta(dims);

  for (unsigned int i = 0; i < dims; i++)
    pos_delta[i] = end_point[i] - start_point[i];

  double sqDistance = 0.0;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double axis_delta = pos_delta[i];

    sqDistance += axis_delta * axis_delta;
    }

  const double distance = sqrt(sqDistance);

  if (!(distance > spacing))
    samples.push_back(start_point);  // Just one sample
  else
    {
    // Get the displacement from one sample to the next

    const double spacing_over_distance = spacing / distance;

    for (unsigned int i = 0; i < dims; i++)
      pos_delta[i] *= spacing_over_distance;

    // Collect the valid sample points

    double elapsed_distance = 0.0;

    std::vector<double> work_vect = start_point;

    while (elapsed_distance < distance)
      {
      if (in_bounds(work_vect))
	samples.push_back(work_vect);

      for (unsigned int i = 0; i < dims; i++)
	work_vect[i] += pos_delta[i];

      elapsed_distance += spacing;
      }
    }
  }
