// C++ program file raytracer.cpp

/*
  By Nathan C. Hearn
     November 3, 2006

  Ray tracer main program.

  Absorption coefficients (in cm^2 / g) - from C. Kuranz

     Material       Backlighter       Value

       CHBr           Ti               1.00e+02
                      Sc               1.30e+02
                      V                7.85e+01  # V/Sc values switched??

       Polyimide      Ti               3.01e+01
                      Sc               3.96e+01
                      V                2.25e+01  # V/Sc values switched??

       Foam           Ti               2.48e+01
                      Sc               3.12e+01
                      V                1.63e+01

  Backlighters

     Ti  -  Used to diagnose the down-strip view
     Sc  -  Used to diagnose the across-strip view
*/


#include <iostream>
#include <vector>
#include <math.h>
#include <stdlib.h>
#include <string>
#include "detector.hpp"
#include "quickflash.hpp"
#include "materials.hpp"

#ifdef USE_MPI
#include <mpi.h>
#endif  // USE_MPI



#include "debug.hpp"


const unsigned int dims = 3;
const unsigned int film_dims = 2;
const unsigned int buffer_size = 200;
const unsigned int cache_size = 4000;
const unsigned int clean_cache_size = 3800;

const std::string dens_name = "dens";
const std::string poly_name = "poly";
const std::string chbr_name = "chbr";
const std::string foam_name = "foam";

const Backlighter backlighter = Ti;


// Prototypes


void load_cache_block(const QuickFlash::File::Dataset & dataset,
		      const unsigned int block_index,
	 QuickFlash::Cache::IndexCache< QuickFlash::Block::BlockData<double> > & cache);

double get_trans_fract(const double spacing, 
		       const std::vector<double> & species_coefficients,
		       const std::vector< std::vector<double> > 
		       & species_densities);

#ifdef USE_MPI
void setup_mpi(int * argc_ptr, char *** argv_ptr);
void quit_mpi();
#endif  // USE_MPI

void get_mpi_info(unsigned int & my_rank, unsigned int & num_procs);

void determine_index_range(const unsigned int total_pixels,
			   const unsigned int my_rank, 
			   const unsigned int num_procs,
			   unsigned int & start_index, 
			   unsigned int & next_start_index);

#ifdef USE_MPI
void communicate_data(Detector & det);
#endif  // USE_MPI

#ifdef USE_MPI
void retreive_data(const unsigned int num_procs, Detector & det);
void send_data(const Detector & det);
#endif  // USE_MPI


// main

int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

#ifdef USE_MPI
  setup_mpi(&argc, &argv);
#endif  // USE_MPI

  unsigned int my_rank = 0;
  unsigned int num_procs = 0;

  get_mpi_info(my_rank, num_procs);

  const int min_args = 18;

  if (argc != min_args)
    {
    if (my_rank == 0)
      cerr << "Usage: " << argv[0] << " input_file cylinder_y cylinder_z "
	   << "cylinder_radius cylinder_x_min cylinder_x_max "
	   << "sample_spacing "
	   << "detector_x detector_aximuth source_distance source_radius "
	   << "film_dist film_size_x film_size_y pixels_x pixels_y " 
	   << "rays_per_pixel" << endl;

    return 1;
    }

  int argPtr = 1;

  const std::string input_filename = argv[argPtr++];

  const double cyl_y = atof(argv[argPtr++]);
  const double cyl_z = atof(argv[argPtr++]);

  const double cyl_rad = atof(argv[argPtr++]);

  const double cyl_x_min = atof(argv[argPtr++]);
  const double cyl_x_max = atof(argv[argPtr++]);

  const double spacing = atof(argv[argPtr++]);

  const double det_x = atof(argv[argPtr++]);
  const double det_azim = atof(argv[argPtr++]);  // radians

  const double src_dist = atof(argv[argPtr++]);
  const double src_rad = atof(argv[argPtr++]);

  const double film_dist = atof(argv[argPtr++]);

  std::vector<double> film_size(2);

  for (unsigned int i = 0; i < 2; i++)
    film_size[i] = atof(argv[argPtr++]);

  std::vector<unsigned int> pixels(2);

  for (unsigned int i = 0; i < 2; i++)
    pixels[i] = static_cast<unsigned int>(atoi(argv[argPtr++]));

  unsigned int ray_count = static_cast<unsigned int>(atoi(argv[argPtr++]));

  bool use_pixel_centers = false;

  if (ray_count < 1)
    {
    use_pixel_centers = true;
    ray_count = 1;
    }

  // Document the output

  if (my_rank == 0)
    {
    cout << "# Call [";

    for (int arg_index = 0; arg_index < argc; arg_index++)
      cout << " " << argv[arg_index];

    cout << " ]" << endl;
    }

  // Set up some helpful objects

  const Absorption absorp(backlighter);

  // Load the data

  QuickFlash::File::DataFile dfile(input_filename, buffer_size);

  const QuickFlash::File::MeshInfo & meshinfo = dfile.get_mesh_info();

  const unsigned int num_blocks = meshinfo.get_num_blocks();

  const QuickFlash::File::Dataset & dens = dfile.get_dataset(dens_name);

  // dens.set_cache_size(cache_size, clean_cache_size);
  dens.set_report_stats();

  const QuickFlash::File::Dataset & poly = dfile.get_dataset(poly_name);

  // poly.set_cache_size(cache_size, clean_cache_size);
  poly.set_report_stats();

  const QuickFlash::File::Dataset & chbr = dfile.get_dataset(chbr_name);

  // chbr.set_cache_size(cache_size, clean_cache_size);
  chbr.set_report_stats();

  const QuickFlash::File::Dataset & foam = dfile.get_dataset(foam_name);

  // foam.set_cache_size(cache_size, clean_cache_size);
  foam.set_report_stats();

  // Build the detector

  std::vector<double> center_pos(3);

  center_pos[0] = det_x;
  center_pos[1] = cyl_y;
  center_pos[2] = cyl_z;

  std::vector<double> min_bounds(3);
  std::vector<double> max_bounds(3);

  min_bounds[0] = cyl_x_min;
  max_bounds[0] = cyl_x_max;

  min_bounds[1] = cyl_y - cyl_rad;
  max_bounds[1] = cyl_y + cyl_rad;

  min_bounds[2] = cyl_z - cyl_rad;
  max_bounds[2] = cyl_z + cyl_rad;

  Detector det(center_pos, min_bounds, max_bounds, det_azim, src_dist, src_rad,
	       film_dist, film_size, pixels);

  det.use_pixel_center(use_pixel_centers);

  // Fill the detector

  const unsigned int num_pixels = det.get_num_pixels();

  unsigned int start_pixel = 0;
  unsigned int next_start_pixel = 0;

  determine_index_range(num_pixels, my_rank, num_procs, start_pixel, 
			next_start_pixel);

  const unsigned int num_local_pixels = next_start_pixel - start_pixel;

  unsigned int one_pct_local_pixels = num_local_pixels / 100;

  if (one_pct_local_pixels < 1)
    one_pct_local_pixels = 1;

  // Set up caches

  QuickFlash::Cache::IndexCache< QuickFlash::Block::BlockData<double> > 
    dens_cache(num_blocks, cache_size, clean_cache_size);

  QuickFlash::Cache::IndexCache< QuickFlash::Block::BlockData<double> > 
    poly_cache(num_blocks, cache_size, clean_cache_size);

  QuickFlash::Cache::IndexCache< QuickFlash::Block::BlockData<double> > 
    chbr_cache(num_blocks, cache_size, clean_cache_size);

  QuickFlash::Cache::IndexCache< QuickFlash::Block::BlockData<double> > 
    foam_cache(num_blocks, cache_size, clean_cache_size);

  std::vector<double> poly_density;
  std::vector<double> chbr_density;
  std::vector<double> foam_density;

  unsigned int dens_cache_hits = 0;
  unsigned int dens_cache_misses = 0;

  const QuickFlash::Block::BlockInfo * blockinfo_ptr = 0;
  unsigned int block_index = 0;

  // Run through the film pixels

  for (unsigned int pixel_index = start_pixel; pixel_index < next_start_pixel; 
       pixel_index++)
    {
    const unsigned int local_pixel_index = pixel_index - start_pixel;

    if (((local_pixel_index) % one_pct_local_pixels) == 0)
      cerr << "  Process " << my_rank << " : " 
	   << (local_pixel_index / one_pct_local_pixels) << "%" << endl;

    for (unsigned int ray_index = 0; ray_index < ray_count; ray_index++)
      {
      std::vector< std::vector<double> > sample_points;

      det.get_sample_points(pixel_index, spacing, sample_points);

      const unsigned int num_samples = sample_points.size();

      poly_density.resize(num_samples);
      chbr_density.resize(num_samples);
      foam_density.resize(num_samples);

      for (unsigned int samp = 0; samp < num_samples; samp++)
	{
	const std::vector<double> & point = sample_points[samp];

	bool block_ok = false;  // See if block exists with point

	if (blockinfo_ptr != 0)
	  if (blockinfo_ptr->in_block(point))
	    block_ok = true;

	if (!block_ok)
	  {
	  if (meshinfo.in_bounds(point))
	    {
	    block_index = meshinfo.get_block_index(point);

	    blockinfo_ptr = &(meshinfo.get_block_info(block_index));

	    if (blockinfo_ptr->in_block(point))
	      block_ok = true;
	    }
	  }

	if (block_ok)
	  {
	  const unsigned int cell_index 
	    = blockinfo_ptr->get_nearest_cell(point);

	  if (!dens_cache.data_present(block_index))
	    {
	    load_cache_block(dens, block_index, dens_cache);
	    dens_cache_misses++;
	    }
	  else
	    dens_cache_hits++;
	  
	  if (!poly_cache.data_present(block_index))
	    load_cache_block(poly, block_index, poly_cache);

	  if (!chbr_cache.data_present(block_index))
	    load_cache_block(chbr, block_index, chbr_cache);

	  if (!foam_cache.data_present(block_index))
	    load_cache_block(foam, block_index, foam_cache);

	  const QuickFlash::Block::BlockData<double> & dens_data 
	    = dens_cache[block_index];

	  const QuickFlash::Block::BlockData<double> & poly_data 
	    = poly_cache[block_index];

	  const QuickFlash::Block::BlockData<double> & chbr_data 
	    = chbr_cache[block_index];

	  const QuickFlash::Block::BlockData<double> & foam_data 
	    = foam_cache[block_index];

	  const double density = dens_data[cell_index];

	  poly_density[samp] = poly_data[cell_index] * density;
	  chbr_density[samp] = chbr_data[cell_index] * density;
	  foam_density[samp] = foam_data[cell_index] * density;
	  }
	else
	  {
	  // Point out of bounds -- use zero densities

	  poly_density[samp] = 0.0;
	  chbr_density[samp] = 0.0;
	  foam_density[samp] = 0.0;
	  }
	}

      const double trans = absorp.get_trans_fract(spacing, poly_density,
						  chbr_density, foam_density);

      det.add_data(pixel_index, trans);
      }
    }

  cerr << "Process " << my_rank << " : Finished" << endl;

  // Output the data

#ifdef USE_MPI
  communicate_data(det);
#endif  // USE_MPI

  if (my_rank == 0)
    {
    std::vector<unsigned int> film_coord(2);

    for (unsigned int i = 0; i < pixels[0]; i++)
      {
      film_coord[0] = i;

      for (unsigned int j = 0; j < pixels[1]; j++)
	{
	film_coord[1] = j;

	const unsigned int pix_index = det.get_scalar_index(film_coord);

	const double pix_val = det.get_data_mean(pix_index);

	cout << pix_val << " ";
	}

      cout << endl;
      }
    }

  cerr << "Process " << my_rank << " : Cache hits [ " 
       << dens_cache_hits << " ] misses [ " << dens_cache_misses 
       << " ] cleans [ " <<  dens_cache.get_clean_counter() << " ]"
       << endl;

#ifdef USE_MPI
  quit_mpi();
#endif  // USE_MPI

  return 0;
  }


// Function definitions


void load_cache_block(const QuickFlash::File::Dataset & dataset,
		      const unsigned int block_index,
     QuickFlash::Cache::IndexCache< QuickFlash::Block::BlockData<double> > & cache)
  {
  // Reads the specified block from dataset, and stores it in cache

  QuickFlash::Block::BlockData<double> data;

  dataset.get_block_data(block_index, data);

  cache.add_data(block_index, data);
  }


#ifdef USE_MPI
void setup_mpi(int * argc_ptr, char *** argv_ptr)
  { MPI_Init(argc_ptr, argv_ptr); }
#endif  // USE_MPI


#ifdef USE_MPI
void quit_mpi()
  {
  MPI_Barrier(MPI_COMM_WORLD);

  MPI_Finalize();
  }
#endif  // USE_MPI


void get_mpi_info(unsigned int & my_rank, unsigned int & num_procs)
  {
  my_rank = 0;
  num_procs = 1;

#ifdef USE_MPI

  int comm_rank = 0;
  int comm_size = 0;

  MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &comm_size);

  my_rank = static_cast<unsigned int>(comm_rank);
  num_procs = static_cast<unsigned int>(comm_size);

#endif  // USE_MPI
  }


void determine_index_range(const unsigned int total_pixels,
			   const unsigned int my_rank, 
			   const unsigned int num_procs,
			   unsigned int & start_index, 
			   unsigned int & next_start_index)
  {
  start_index = 0;
  next_start_index = total_pixels;

  const unsigned int max_id = num_procs - 1;

  const unsigned int pixels_per_proc 
    = total_pixels / num_procs;

  if (my_rank > 0)
    {
    // Determine start point

    start_index = my_rank * pixels_per_proc;
    }

  if (my_rank < max_id)
    {
    // Determine end point

    next_start_index = (my_rank + 1) * pixels_per_proc;
    }
  }


#ifdef USE_MPI
void communicate_data(Detector & det)
  {
  unsigned int my_rank = 0;
  unsigned int num_procs = 0;

  get_mpi_info(my_rank, num_procs);

  if (num_procs > 1)
    {
    if (my_rank == 0)
      retreive_data(num_procs, det);
    else
      send_data(det);
    }
  }
#endif  // USE_MPI


#ifdef USE_MPI
void retreive_data(const unsigned int num_procs, Detector & det)
  {
  // Used by root process

  if (num_procs > 1)
    {
    const unsigned int num_pixels = det.get_num_pixels();

    for (unsigned int proc_id = 1; proc_id < num_procs; proc_id++)
      {
      unsigned int start_index = 0;
      unsigned int next_start_index = 0;

      determine_index_range(num_pixels, proc_id, num_procs, start_index,
			    next_start_index);

      const unsigned int proc_pixels = next_start_index - start_index;

      if (proc_pixels > 0)
	{
	unsigned int * num_entries = new unsigned int[proc_pixels];
	double * sum_weight = new double[proc_pixels];
	double * sum_weight_values = new double[proc_pixels];
	double * sum_weight_sqValues = new double[proc_pixels];
	double * min_value = new double[proc_pixels];
	double * max_value = new double[proc_pixels];

	MPI_Status status;

	MPI_Recv(num_entries, static_cast<int>(proc_pixels), MPI_UNSIGNED, 
		 static_cast<int>(proc_id), 0, MPI_COMM_WORLD, &status);
	MPI_Recv(sum_weight, static_cast<int>(proc_pixels), MPI_DOUBLE, 
		 static_cast<int>(proc_id), 1, MPI_COMM_WORLD, &status);
	MPI_Recv(sum_weight_values, static_cast<int>(proc_pixels), MPI_DOUBLE, 
		 static_cast<int>(proc_id), 2, MPI_COMM_WORLD, &status);
	MPI_Recv(sum_weight_sqValues, static_cast<int>(proc_pixels), 
		 MPI_DOUBLE, static_cast<int>(proc_id), 3, MPI_COMM_WORLD, 
		 &status);
	MPI_Recv(min_value, static_cast<int>(proc_pixels), MPI_DOUBLE, 
		 static_cast<int>(proc_id), 4, MPI_COMM_WORLD, &status);
	MPI_Recv(max_value, static_cast<int>(proc_pixels), MPI_DOUBLE, 
		 static_cast<int>(proc_id), 5, MPI_COMM_WORLD, &status);

	// Store the data

	for (unsigned int remote_index = 0; remote_index < proc_pixels;
	     remote_index++)
	  {
	  const unsigned int det_pixel = remote_index + start_index;

	  det.set_data(det_pixel, num_entries[remote_index],
		       sum_weight[remote_index], 
		       sum_weight_values[remote_index],
		       sum_weight_sqValues[remote_index],
		       min_value[remote_index], max_value[remote_index]);
	  }

	delete[] num_entries;
	num_entries = 0;

	delete[] sum_weight;
	sum_weight = 0;

	delete[] sum_weight_values;
	sum_weight_values = 0;

	delete[] sum_weight_sqValues;
	sum_weight_sqValues = 0;

	delete[] min_value;
	min_value = 0;

	delete[] max_value;
	max_value = 0;
	}
      }
    }
  }
#endif  // USE_MPI


#ifdef USE_MPI
void send_data(const Detector & det)
  {
  // Used by non-root processes

  const unsigned int num_pixels = det.get_num_pixels();

  unsigned int proc_id = 0;
  unsigned int num_procs = 0;

  get_mpi_info(proc_id, num_procs);

  unsigned int start_index = 0;
  unsigned int next_start_index = 0;

  determine_index_range(num_pixels, proc_id, num_procs, start_index,
			next_start_index);

  const unsigned int proc_pixels = next_start_index - start_index;

  if (proc_pixels > 0)
    {
    unsigned int * num_entries = new unsigned int[proc_pixels];
    double * sum_weight = new double[proc_pixels];
    double * sum_weight_values = new double[proc_pixels];
    double * sum_weight_sqValues = new double[proc_pixels];
    double * min_value = new double[proc_pixels];
    double * max_value = new double[proc_pixels];

    for (unsigned int local_index = 0; local_index < proc_pixels;
	 local_index++)
      {
      const unsigned int det_pixel = local_index + start_index;

      num_entries[local_index] = det.get_data_num_entries(det_pixel);
      sum_weight[local_index] = det.get_data_weight_sum(det_pixel);
      sum_weight_values[local_index] 
	= det.get_data_sum_weight_values(det_pixel);
      sum_weight_sqValues[local_index] 
	= det.get_data_sum_weight_sqValues(det_pixel);
      min_value[local_index] = det.get_data_min(det_pixel);
      max_value[local_index] = det.get_data_max(det_pixel);
      }

    MPI_Send(num_entries, static_cast<int>(proc_pixels), MPI_UNSIGNED, 
	     0, 0, MPI_COMM_WORLD);
    MPI_Send(sum_weight, static_cast<int>(proc_pixels), MPI_DOUBLE, 
	     0, 1, MPI_COMM_WORLD);
    MPI_Send(sum_weight_values, static_cast<int>(proc_pixels), MPI_DOUBLE, 
	     0, 2, MPI_COMM_WORLD);
    MPI_Send(sum_weight_sqValues, static_cast<int>(proc_pixels), MPI_DOUBLE,
	     0, 3, MPI_COMM_WORLD);
    MPI_Send(min_value, static_cast<int>(proc_pixels), MPI_DOUBLE, 
	     0, 4, MPI_COMM_WORLD);
    MPI_Send(max_value, static_cast<int>(proc_pixels), MPI_DOUBLE, 
	     0, 5, MPI_COMM_WORLD);

    delete[] num_entries;
    num_entries = 0;

    delete[] sum_weight;
    sum_weight = 0;

    delete[] sum_weight_values;
    sum_weight_values = 0;

    delete[] sum_weight_sqValues;
    sum_weight_sqValues = 0;

    delete[] min_value;
    min_value = 0;

    delete[] max_value;
    max_value = 0;
    }
  }
#endif  // USE_MPI
