// C++ header file raytracer_core.hpp

/*
  By Nathan C. Hearn
     July 18, 2007

  Raytracing package core classes.
*/


#ifndef RAYTRACER_CORE_HPP
#define RAYTRACER_CORE_HPP


#include <string>
#include <vector>
#include "quickflash_block_blockdata.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_cache_indexcache.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "detector.hpp"


class RaytracerCore
  {
  private :

    typedef QuickFlash::Block::BlockData<double> DoubleBlock;

  public :

    RaytracerCore();

    RaytracerCore(const QuickFlash::File::DataFile & data_file,
		  const std::string & density_dataset_name,
		  const std::vector<double> & center,
		  const unsigned int cache_size=1000);


    ~RaytracerCore() { reset(); }

    void reset();

    void reset(const QuickFlash::File::DataFile & data_file,
	       const std::string & density_dataset_name,
	       const std::vector<double> & center,
	       const unsigned int cache_size=1000);

    void add_species(const std::string & dataset_name,
		     const double opacity);

    void set_source_detector(const double axis_azimuth,
			     const double source_distance,
			     const double source_radius,
			     const double sample_spacing,
			     const unsigned int rays_per_pixel,
			     const double film_distance, 
			     const double film_width,
			     const double film_height, 
			     const unsigned int width_pixels,
			     const unsigned int height_pixels);

    bool set_up() const
      {
      bool all_set = true;

      if ((dfile == 0) || (dens_set == 0) || (!detector_set_up))
	all_set = false;

      return all_set;
      }

    double get_pixel_value(const unsigned int pixel_index) const;

    void report_cache_stats() const;

  private :

    void load_data(const unsigned int block_index) const;

    double get_opacity_integral(const std::vector< std::vector<double> > 
				& sample_points) const;

    void get_data(const std::vector<double> & pos, double & density,
		  std::vector<double> & species_mass_fractions) const;

  private :

    static const unsigned int dims;

    const QuickFlash::File::DataFile * dfile;

    const QuickFlash::File::MeshInfo * meshinfo;

    unsigned int num_blocks;

    unsigned int local_cache_size;

    std::vector<double> axis_center;

    std::vector<double> min_bounds;
    std::vector<double> max_bounds;

    mutable unsigned int blockinfo_block_index;
    mutable QuickFlash::Block::BlockInfo const * blockinfo;

    const QuickFlash::File::Dataset * dens_set;

    mutable QuickFlash::Cache::IndexCache<DoubleBlock> dens_cache;

    mutable bool dens_block_read;
    mutable unsigned int dens_block_index;

    mutable DoubleBlock * dens_data;

    mutable std::vector<const QuickFlash::File::Dataset *> species_sets;

    std::vector<double> species_opacities;

    std::vector<QuickFlash::Cache::IndexCache<DoubleBlock> *> 
      species_caches;

    mutable bool species_block_read;
    mutable unsigned int species_block_index;

    mutable std::vector<DoubleBlock *> species_data;

    bool detector_set_up;

    Detector detector;

    double spacing;
    unsigned int num_rays;

    double oneOver_num_rays;

    mutable unsigned int cache_hits;
    mutable unsigned int cache_misses;

    mutable unsigned int block_reuse_count;
  };


#endif  // RAYTRACER_CORE_HPP

