// C++ program file flash_stat_bins.cpp

/*
  By Nathan C. Hearn
     September 20, 2006

  Statistical binning classes.
*/


#include "stat_bins.hpp"
#include <vector>
#include "quickflash_except.hpp"
#include "quickflash_counters.hpp"
#include "quickflash_geometry.hpp"


namespace Stats
{

// Class Bin

void Bin::reset()
  {
  num_entries = 0;

  sum_weight = 0.0;

  sum_weight_values = 0.0;
  sum_weight_sqValues = 0.0;

  min_value = 0.0;
  max_value = 0.0;
  }


void Bin::reset(const Bin & source)
  {
  if (&source != this)
    {
    num_entries = source.num_entries;

    sum_weight = source.sum_weight;

    sum_weight_values = source.sum_weight_values;
    sum_weight_sqValues = source.sum_weight_sqValues;
    }
  }


void Bin::add_bin(const Bin & source)
  {
  num_entries += source.num_entries;

  sum_weight += source.sum_weight;

  sum_weight_values += source.sum_weight_values;
  sum_weight_sqValues += source.sum_weight_sqValues;

  const double source_min_val = source.min_value;
  const double source_max_val = source.max_value;

  if (source_min_val < min_value)
    min_value = source_min_val;

  if (source_max_val > max_value)
    max_value = source_max_val;
  }


void Bin::add_entry(const double value, const double weight)
  {
  // Should we allow zero weights???

  sum_weight += weight;

  const double weight_value = weight * value;

  sum_weight_values += weight_value;

  sum_weight_sqValues += weight_value * value;

  if (num_entries > 0)
    {
    if (value < min_value)
      min_value = value;
    else if (value > max_value)
      max_value = value;
    }
  else
    {
    min_value = value;
    max_value = value;
    }

  num_entries++;
  }


double Bin::get_pop_sqStddev() const
  {
  // Implement sigma^2 = [ SUM(w_i * x_i^2) / SUM(w_i) ] - (N * MEAN(x_i)^2)

  // const double fl_num_entries = static_cast<double>(num_entries);

  // const double oneOver_mean_weight = fl_num_entries / sum_weight;

  double sqStddev = 0.0;

  if (sum_weight > 0.0)
    {
    const double oneOver_sum_weight = 1.0 / sum_weight;

    const double mean_val = get_mean();

    // double var = (sum_weight_sqValues * oneOver_mean_weight) 
    //   - (fl_num_entries * mean_val * mean_val);

    sqStddev = (sum_weight_sqValues * oneOver_sum_weight) 
      - (mean_val * mean_val);

    if (sqStddev < 0.0)
      sqStddev = 0.0;
    }

  return sqStddev;
  }


// Class BinArray


void BinArray::reset()
  {
  bin_dims.reset();
  bins.reset();
  }


void BinArray::reset(const std::vector<unsigned int> & axis_bin_count, 
		     const std::vector<double> & min_boundary, 
		     const std::vector<double> & max_boundary)
  {
  bin_dims.reset(min_boundary, max_boundary, axis_bin_count, 
		 QuickFlash::Geometry::Cartesian);
  bins.reset(axis_bin_count);
  }


void BinArray::reset(const BinArray & source)
  {
  if (&source != this)
    {
    bin_dims = source.bin_dims;
    bins = source.bins;
    }
  }


void BinArray::add_data(const std::vector<double> & coordinate, 
			const double value, const double weight)
  {
  // Find the bin for this coordinate

  if (bin_dims.in_block(coordinate))
    {
    const unsigned int bin_index = bin_dims.get_nearest_cell(coordinate);

    add_data(bin_index, value, weight);
    }
  }


}  // End namespace Stats
