#!/usr/bin/env python


try :
    import psyco
except ImportError :
    pass


class DataArray :

    def __init__(self, num_rows=int(0), num_cols=int(0),
                 default_value=float(0.0)) :

        self.__num_rows = int(0)
        self.__num_cols = int(0)

        self.__num_cells = int(0)

        self.__array = None
        self.__array_log = None

        self.reset(num_rows, num_cols, default_value)


    def reset(self, num_rows=int(0), num_cols=int(0),
              default_value=float(0.0)) :

        numeric_zeros = None
        numeric_int = None
        numeric_float = None

        try :
            import numpy
            numeric_zeros = numpy.zeros
            numeric_int = numpy.int
            numeric_float = numpy.float
        except ImportError :
            import Numeric
            numeric_zeros = Numeric.zeros
            numeric_int = Numeric.Int
            numeric_float = Numeric.Float
            

        if ((num_rows < 1) or (num_cols < 1)) :

            self.__num_rows = int(0)
            self.__num_cols = int(0)

            self.__num_cells = int(0)

            self.__array = None
            self.__array_log = None
            

        else :

            self.__num_rows = num_rows
            self.__num_cols = num_cols

            self.__num_cells = num_rows * num_cols

            self.__array = numeric_zeros(self.__num_cells, numeric_float)
            self.__array_log = numeric_zeros(self.__num_cells, numeric_int)

            self.__array[:] = default_value


    def get_index(self, i, j) :

        index = j + (self.__num_cols * i)

        return index


    def get_num_cells(self) :
        return self.__num_cells
    

    def get_num_rows(self) :
        return self.__num_rows


    def get_num_cols(self) :
        return self.__num_cols


    def __getitem__(self, index) :
        return self.__array[index]


    def __setitem__(self, index, value) :
        self.__array[index] = value
        self.__array_log[index] += 1
    

    def get_missing_data_indexes(self) :

        indexes = list()

        for i in range(self.__num_cells) :

            if (self.__array_log[i] < 1) :
                indexes.append(i)

        return indexes


    def get_duplicate_data_indexes(self) :

        indexes = list()

        for i in range(self.__num_cells) :

            if (self.__array_log[i] > 1) :
                indexes.append(i)

        return indexes



def condense_indexes(index_list) :

    from copy import copy

    # Convert a list of indexes to a list of (start_index, length) pairs

    work_list = copy(index_list)

    work_list.sort()

    pair_list = list()

    num_indexes = len(work_list)

    index_ptr = int(0)

    current_set = None
    start_new_set = True

    while (index_ptr < num_indexes) :

        current_index = work_list[index_ptr]

        if (start_new_set) :
            
            current_set = (current_index, 1)
            index_ptr += 1

            start_new_set = False
            
        elif (index_list[index_ptr - 1] == (current_index - 1)) :
            
            current_set[1] += 1
            index_ptr += 1
            
        else :
            
            start_new_set = True  # Do NOT increment index_ptr

        if (start_new_set or (index_ptr == num_indexes)) :
            pair_list.append(current_set)

    return pair_list

        

comment_marker = "#"

begin_data_marker = "["
end_data_marker = "]"

single_quote = "'"
single_quote_space = "' "
data_separator = ","

## ICK!! HARD-WIRED POSITIONS AHEAD!!!

line_start_pixel_word_index = int(6)  # The seventh element is the index
line_data_word_index = int(7)         # The data starts at the eighth element


def extract_line_data(line) :

    # Extract data from a ParTask output line

    # Each line begins with a task ID, followed by the process ID and the word
    # "Start", which is followed by the index of the first pixel, and
    # finally, the data.

    # Here, the data is formated as ['2.3', '2.7', '1.2'], brackets and all.
    # Use this formatting type to ensure that only complete items are read
    # (in case the program was killed during the write).

    start_pixel = int(-1)
    
    data_list = list()

    work_line = line.split(comment_marker)[0].strip()

    # Split off the first words, but keep the data together

    words = work_line.split(" ", line_data_word_index)  # MUST BE A BETTER WAY!

    num_words = len(words)

    if (num_words > line_data_word_index) :

        start_pixel = int(words[line_start_pixel_word_index])

        data = words[line_data_word_index]  # This is the segment with the data

        # Data starts after begin_data_marker

        start_data_pos = data.find(begin_data_marker)

        if (start_data_pos >= 0) :

            # If complete, data ends before end_data_marker

            end_data_pos = data.find(end_data_marker)

            # If no end marker is found, find to last single quote that
            # immediately follows a digit

            if (end_data_pos < 0) :

                # Find to last single quote that immediately follows a digit

                str_index = len(data) - 1

                while ((str_index > start_data_pos) and (end_data_pos < 0)) :

                    if (data[str_index] == single_quote) :

                        prev_char = data[str_index - 1]

                        if (prev_char.isdigit()) :

                            end_data_pos = str_index + 1 # Includes final quote

                        else :

                            str_index -= 1

                    else :

                        str_index -= 1

            if (end_data_pos > start_data_pos) :

                # Get the data (without the brackets), and separate by
                # commas (data_separator markers)

                data_segment = data[(start_data_pos + 1):end_data_pos]

                data_words = data_segment.split(data_separator)

                # Run through the data

                for word in data_words :

                    data_list.append(float(word.strip(single_quote_space)))

    return start_pixel, data_list


def main() :

    from sys import argv, exit, stdout, stderr

    argc = len(argv)

    min_args = int(4)

    if (argc < min_args) :
        stderr.write("Usage: " + argv[0] + " input_file num_rows num_cols"
                     + " [ empty_value ]\n\n")
        exit(1)

    argptr = int(1)

    inputfile_name = str(argv[argptr])
    argptr += 1

    num_rows = int(argv[argptr])
    argptr += 1

    if (num_rows < 1) :
        stderr.write("Error: Number of rows must be positive\n\n")
        exit(-1)

    num_cols = int(argv[argptr])
    argptr += 1

    if (num_cols < 1) :
        stderr.write("Error: Number of columns must be positive\n\n")
        exit(-1)

    empty_value = float(0.0)

    if (argptr < argc) :
        empty_value = float(argv[argptr])
        argptr += 1


    inputfile = file(inputfile_name, "r")

    data_array = DataArray(num_rows, num_cols, empty_value)

    total_pixels = data_array.get_num_cells()


    finished = False

    while (not finished) :

        line = inputfile.readline()

        if (line == "") :
            finished = True
        else :

            start_pixel, pixel_data = extract_line_data(line)

            if (start_pixel >= 0) :

                num_pixels = len(pixel_data)

                if ((start_pixel + num_pixels) > total_pixels) :
                    stderr.write("Error: Pixels index out of bounds\n\n")
                    exit(-2)

                for index in range(num_pixels) :

                    real_index = index + start_pixel
                    
                    data_array[real_index] = pixel_data[index]

            
    # Output the data

    for i in range(num_rows) :

        for j in range(num_cols) :

            index = data_array.get_index(i, j)
            
            stdout.write(str(data_array[index]) + " ")

        stdout.write("\n")

    stdout.write("\n")


    # Provide some statistics

    missing_pixels = data_array.get_missing_data_indexes()

    stderr.write("Missing pixels [ " + str(len(missing_pixels)) + " ]\n")

    duplicate_pixels = data_array.get_duplicate_data_indexes()

    stderr.write("Duplicate pixels [ " + str(len(duplicate_pixels)) + " ]\n")

    stderr.write("\n")


if (__name__ == "__main__") :
    main()

