// C++ program file build_surface.cpp

/*
  By Nathan C. Hearn
     May 6, 2008

  Program to produce a set of triangles describing an isosurface from a Flash
  dataset.
*/


#include <iostream>
#include <string>
#include <cstdlib>
#include <list>
#include <vector>
#include <fstream>
#include <cmath>
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_file_flashdefs.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_cache_indexcache.hpp"
#include "quickflash_surface_triangle.hpp"
#include "quickflash_surface_cube.hpp"
#include "quickflash_surface_marchingtetrahedrons.hpp"
#include "quickflash_surface_isosurface.hpp"


const unsigned int read_buffer_size = 100;
const unsigned int read_cache_size = 8000;

const unsigned int dims = 3;


bool block_in_bounds(const std::vector<double> & block_min_coords,
		     const std::vector<double> & block_max_coords,
		     const bool restrict_volume_min=false,
		     const std::vector<double> 
		       & volume_min_coords=std::vector<double>(),
		     const bool restrict_volume_max=false,
		     const std::vector<double> 
		       & volume_max_coords=std::vector<double>())
  {
  bool ret_val = false;  // Unless found otherwise, this is the default

  // This function does some basic checks, but does NOT check element values

  const unsigned int dims = block_min_coords.size();

  bool dims_ok = true;

  if (block_max_coords.size() != dims)
    dims_ok = false;
  else if (restrict_volume_min && (volume_min_coords.size() != dims))
    dims_ok = false;
  else if (restrict_volume_max && (volume_max_coords.size() != dims))
    dims_ok = false;

  if (dims_ok)
    {
    ret_val = true;  // Unless shown otherwise after this point

    if (restrict_volume_min)
      {
      for (unsigned int i = 0; i < dims; i++)
	{
	const double axis_block_max = block_max_coords[i];

	if (!(axis_block_max > volume_min_coords[i]))
	  ret_val = false;
	}
      }

    if (restrict_volume_max)
      {
      for (unsigned int i = 0; i < dims; i++)
	{
	const double axis_block_min = block_min_coords[i];

	if (!(axis_block_min < volume_max_coords[i]))
	  ret_val = false;
	}
      }
    }

  return ret_val;
  }
 

int separate_triangle(const QuickFlash::Surface::Triangle & orig_tri, 
		      const double separation,
		      QuickFlash::Surface::Triangle & sep_tri)
  {
  int ret_val = -1;  // Error state

  // Determine the direction of the normal vector via
  // (Vert2 - Vert1) x (Vert3 - Vert2)

  QuickFlash::Surface::Triangle::DoubleVector vert1 = orig_tri[0];
  QuickFlash::Surface::Triangle::DoubleVector vert2 = orig_tri[1];
  QuickFlash::Surface::Triangle::DoubleVector vert3 = orig_tri[2];

  QuickFlash::Surface::Triangle::DoubleVector leg1;

  for (unsigned int i = 0; i < dims; i++)
    leg1[i] = vert2[i] - vert1[i];

  QuickFlash::Surface::Triangle::DoubleVector leg2;

  for (unsigned int i = 0; i < dims; i++)
    leg2[i] = vert3[i] - vert2[i];

  QuickFlash::Surface::Triangle::DoubleVector work_vect;
    // Proporational to cross product

  work_vect[0] = (leg1[1] * leg2[2]) - (leg1[2] * leg2[1]);
  work_vect[1] = (leg1[2] * leg2[0]) - (leg1[0] * leg2[2]);
  work_vect[2] = (leg1[0] * leg2[1]) - (leg1[1] * leg2[0]);

  double sq_length = 0.0;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double elem = work_vect[i];

    sq_length += elem * elem;
    }

  const double length = std::sqrt(sq_length);

  if (length > 0.0)     
    {
    // Build the displacement vector

    const double scale = separation / length;

    for (unsigned int i = 0; i < dims; i++)
      work_vect[i] *= scale;

    // Displace the triangles

    for (unsigned int i = 0; i < dims; i++)
      {
      const double disp_elem = work_vect[i];

      vert1[i] += disp_elem;
      vert2[i] += disp_elem;
      vert3[i] += disp_elem;
      }

    // Build the new triangle, keeping the vertex order the same

    sep_tri.reset(vert1, vert2, vert3);

    ret_val = 0;  // OK
    }

  return ret_val;
  }


void build_triangle_list(const QuickFlash::File::DataFile & df,
			 const std::string & output_file_prefix,
			 const std::string & dataset_name,
			 const double isosurface_level,
			 const bool limit_refine_level=false,
			 const unsigned int max_refine_level=0,
			 const bool reverse_triangles=false,
			 const bool both_sides=false,
			 const double side_separation=0.0,
			 const bool limit_volume_min=false,
			 const std::vector<double> 
			   & volume_min_bounds=std::vector<double>(),
			 const bool limit_volume_max=false,
			 const std::vector<double> 
			   & volume_max_bounds=std::vector<double>(),
			 const std::string & obj_mtllib_file=std::string(),
			 const std::string & obj_mtl_front=std::string(),
			 const std::string & obj_mtl_back=std::string())
  {
  using std::cerr;
  using std::endl;

  const std::string & input_filename = df.get_filename();

  cerr << "Processing file [ " << input_filename << " ] isosurface level [ "
       << isosurface_level << " ]" << endl;
  
  // Get the dataset and file information

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  if (meshinfo.get_dims() != dims)
    throw QuickFlash::Except("Only three-dimensional datasets supported",
			   __FILE__, __LINE__);

  const QuickFlash::File::Dataset & dset = df.get_dataset(dataset_name);

  dset.set_cache_size(read_cache_size);
  dset.set_report_stats(true);

  // Set up the isosurface generator

  QuickFlash::Surface::IsosurfaceGen iso_gen(meshinfo, dset, limit_refine_level, 
					   max_refine_level);

  iso_gen.set_volume_limit(limit_volume_min, volume_min_bounds,
                           limit_volume_max, volume_max_bounds);

  // Open the OBJ file(s) for writing

  std::stringstream obj_file_prefix_sstream;

  obj_file_prefix_sstream << output_file_prefix;

  if (limit_refine_level)
    obj_file_prefix_sstream << "_r" << max_refine_level;

  if (reverse_triangles)
    obj_file_prefix_sstream << "_rev";

  const std::string obj_file_prefix = obj_file_prefix_sstream.str();

  std::string front_filename;
  std::string back_filename;

  if (both_sides)
    {
    front_filename = obj_file_prefix + "_front.obj";
    back_filename = obj_file_prefix + "_back.obj";
    }
  else
    front_filename = obj_file_prefix + ".obj";


  std::ofstream * ffile = 0;  // Front-side faces
  std::ofstream * bfile = 0;  // Back-side faces

  ffile = new std::ofstream(front_filename.c_str());

  if (both_sides)
    bfile = new std::ofstream(back_filename.c_str());


  // Write out the headers

  *ffile << "# DataFile " << input_filename << endl;
  *ffile << "# Dataset " << dataset_name << endl;
  *ffile << "# IsosurfaceLevel " << isosurface_level << endl;

  *ffile << endl;

  *ffile << "# Front sides" << endl;

  if (reverse_triangles)
    *ffile << "# Reveresed triangles" << endl;

  *ffile << endl;

  if (obj_mtllib_file.size() > 0)
    *ffile << "mtllib " << obj_mtllib_file << endl << endl;

  if (obj_mtl_front.size() > 0)
    *ffile << "usemtl " << obj_mtl_front << endl << endl;

  *ffile << "# Vertex data and faces " << endl;

  *ffile << endl;

  if (both_sides)
    {
    *bfile << "# DataFile " << input_filename << endl;
    *bfile << "# Dataset " << dataset_name << endl;
    *bfile << "# IsosurfaceLevel " << isosurface_level << endl;
    *bfile << "# BacksideSeparation " << side_separation << endl;

    *bfile << endl;

    *bfile << "# Back sides" << endl;

    if (reverse_triangles)
      *bfile << "# Reversed triangles" << endl;

    *bfile << endl;

    if (obj_mtllib_file.size() > 0)
      *bfile << "mtllib " << obj_mtllib_file << endl << endl;

    if (obj_mtl_back.size() > 0)
      *bfile << "usemtl " << obj_mtl_back << endl << endl;

    *bfile << "# Vertex data and faces " << endl;

    *bfile << endl;
    }


  // Run through the blocks

  double surface_area = 0.0;

  unsigned int triangle_count = 0;

  const double separation_factor = reverse_triangles ? 1.0 : -1.0;

  unsigned int block_counter = 0;

  QuickFlash::Block::BlockData<double> block_data;

  std::list<QuickFlash::Surface::Triangle> triangle_list;
  QuickFlash::Surface::Triangle backside_tri;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    if ((block_counter % 10) == 0)
      cerr << "Block " << block_counter << endl;
    block_counter++;

    const unsigned int block_index = node_iter.get_block_index();

    bool use_block = true;

    bool skip_children = false;

    if (limit_refine_level)
      {
      // NOTE: MeshInfo::get_refine_level uses Flash-style origin-one indexing

      const unsigned int refine_level = meshinfo.get_refine_level(block_index);

      if (refine_level >= max_refine_level)
	skip_children = true;
      else if (!(node_iter.is_leaf()))
	use_block = false;
      }
    else if (!(node_iter.is_leaf()))
      use_block = false;

    if (use_block && (limit_volume_min || limit_volume_max))
      {
      const QuickFlash::Block::BlockInfo & block_info 
	= node_iter.get_block_info();

      const std::vector<double> & block_min_coords 
	= block_info.get_min_coords();

      const std::vector<double> & block_max_coords 
	= block_info.get_max_coords();

      if (!(block_in_bounds(block_min_coords, block_max_coords,
			    limit_volume_min, volume_min_bounds,
			    limit_volume_max, volume_max_bounds)))
	{
	use_block = false;
	skip_children = true;
	}
      }

    if (use_block)
      {
      iso_gen.add_block_surface(block_index, isosurface_level, triangle_list);

      if (!triangle_list.empty())
	{
	// Write the triangles
	
	std::list<QuickFlash::Surface::Triangle>::const_iterator tri_iter 
	  = triangle_list.begin();

	while (tri_iter != triangle_list.end())
	  {
	  const QuickFlash::Surface::Triangle & tri = *tri_iter;

	  for (unsigned int vert_index = 0; vert_index < 3; vert_index++)
	    {
	    const QuickFlash::Surface::Triangle::DoubleVector & vert_coords 
	      = tri[vert_index];

	    *ffile << "v";

	    for (unsigned int i = 0; i < dims; i++)
	      *ffile << "\t" << vert_coords[i];

	    *ffile << endl;
	    }

	  // The face is just the last three vertices that were written

	  if (reverse_triangles)
	    *ffile << "f\t-1\t-2\t-3" << endl;
	  else
	    *ffile << "f\t-3\t-2\t-1" << endl;

	  *ffile << endl;

	  if (both_sides)
	    {
	    if (separate_triangle(tri, (separation_factor * side_separation), 
				  backside_tri) < 0)
	      throw QuickFlash::Except("Unable to flip triangle", __FILE__,
				     __LINE__);

	    for (unsigned int vert_index = 0; vert_index < 3; vert_index++)
	      {
	      const QuickFlash::Surface::Triangle::DoubleVector & vert_coords 
		= backside_tri[vert_index];

	      *bfile << "v";

	      for (unsigned int i = 0; i < dims; i++)
		*bfile << "\t" << vert_coords[i];

	      *bfile << endl;
	      }

	    // The face is just the last three vertices that were written
	    // NOTE: This triangle needs to be reversed in order!!!

	    if (!reverse_triangles)
	      *bfile << "f\t-1\t-2\t-3" << endl;
	    else
	      *bfile << "f\t-3\t-2\t-1" << endl;

	    *bfile << endl;
	    }

	  surface_area += tri_iter->get_surface_area();

	  triangle_count += 1;

	  ++tri_iter;
	  }

	triangle_list.clear();
	}
      }

    if (skip_children)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  *ffile << endl;

  *ffile << "# SurfaceArea " << surface_area << endl;
  *ffile << "# Triangles " << triangle_count << endl;

  *ffile << endl;

  if (both_sides)
    {
    *bfile << endl;

    *bfile << "# SurfaceArea " << surface_area << endl;
    *bfile << "# Triangles " << triangle_count << endl;

    *bfile << endl;
    }

  delete ffile;
  ffile = 0;

  if (bfile != 0)
    {
    delete bfile;
    bfile = 0;
    }
  }


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  const unsigned int dims = 3;  // Only valid for 3D

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 3;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name 
	 << " dataset iso_level1[,iso_level2,...]"
	 << endl << "\t" << "[ --output-prefix=FILENAME_PREFIX ]"
	 << endl << "\t" << "[ --reverse-triangles ]"
	 << endl << "\t" << "[ --both-sides=SEPARATION ]"
	 << endl << "\t" << "[ --max-refine-level=MAX_LEVEL ]"
	 << endl << "\t" << "[ --obj-mtllib=WAVEFRONT_MATERIAL_FILE ]"
	 << endl << "\t" << "[ --obj-mtl=MATERIAL[,BACK_MATERIAL] ]"
	 << endl << "\t" << "[ --min-bounds=MIN_X,MIN_Y,MIN_Z ]"
	 << endl << "\t" << "[ --max-bounds=MAX_X,MAX_Y,MAX_Z ]"
	 << endl << "\t" << "filename_1 [ filename_2 ... ]"
	 << endl << endl;
    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  const std::string dataset_name = argument_list[argptr++];

  const std::string isosurface_level_input = argument_list[argptr++];

  std::list<std::string> flash_filenames;

  while (argptr < num_args)
    flash_filenames.push_back(argument_list[argptr++]);

  // Parse the options

  std::string user_output_prefix = "";  // Empty by default

  bool limit_refine_level = false;
  unsigned int max_refine_level = 0;

  bool reverse_triangles = false;

  bool both_sides = false;
  double side_separation = 0.0;

  std::string obj_mtllib_file;

  std::string obj_mtl_front;
  std::string obj_mtl_back;

  bool restrict_volume_min = false;

  std::vector<double> volume_min_bounds;

  bool restrict_volume_max = false;

  std::vector<double> volume_max_bounds;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output prefix is required with option"
	     << endl << endl;
	return -1;
	}

      user_output_prefix = current_option.get_value();

      // If the prefix is more than just a path, add a trailing underscore

      std::string prefix_path;
      std::string prefix_name;

      QuickFlash::Utils::split_name_path_file(user_output_prefix, prefix_path, 
					    prefix_name);

      if (prefix_name.size() > 0)
	user_output_prefix += "_";
      }
    else if (keyword == "max-refine-level")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Maximum refinement level must be specified"
	     << endl << endl;
	return -1;
	}

      limit_refine_level = true;

      current_option.get_value(max_refine_level);

      if (max_refine_level < 1)
	{
	cerr << endl << "Error: Maximum refinement level must be at least one"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "reverse-triangles")
      {
      reverse_triangles = true;
      }
    else if (keyword == "both-sides")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Both-sides option requires positive separation"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(side_separation);

      if (!(side_separation > 0.0))
	{
	cerr << endl << "Error: Both-sides option requires positive separation"
	     << endl << endl;
	return -1;
	}

      both_sides = true;
      }
    else if (keyword == "obj-mtllib")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: OBJ material library option requires a filename" 
	     << endl << endl;
	return -1;
	}

      obj_mtllib_file = current_option.get_value();
      }
    else if (keyword == "obj-mtl")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: OBJ material option requires a material name"
	     << endl << endl;
	return -1;
	}
      
      std::vector<std::string> material_list;

      current_option.get_value(material_list);

      const unsigned int num_materials = material_list.size();

      if ((num_materials < 1) || (num_materials > 2))
	{
	cerr << endl 
	     << "Error: OBJ material option accepts only one or two materials"
	     << endl << endl;
	return -1;
	}

      obj_mtl_front = material_list[0];

      if (num_materials > 1)
	obj_mtl_back = material_list[1];
      else
	obj_mtl_back = material_list[0];
      }
    else if (keyword == "min-bounds")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Min bounds option requires comma-separated coordinates" 
	     << endl << endl;
	return -1;
	}

      current_option.get_value(volume_min_bounds);

      if (volume_min_bounds.size() != dims)
	{
	cerr << endl << "Min bounds must be a " << dims << "-vector" 
	     << endl << endl;
	return -1;
	}

      restrict_volume_min = true;
      }
    else if (keyword == "max-bounds")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Max bounds option requires comma-separated coordinates" 
	     << endl << endl;
	return -1;
	}

      current_option.get_value(volume_max_bounds);

      if (volume_max_bounds.size() != dims)
	{
	cerr << endl << "Max bounds must be a " << dims << "-vector" 
	     << endl << endl;
	return -1;
	}

      restrict_volume_max = true;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }

  // Run some checks

  if (restrict_volume_min && restrict_volume_max)
    {
    bool vol_bounds_ok = true;

    for (unsigned int i = 0; i < dims; i++)
      if (!(volume_min_bounds[i] < volume_max_bounds[i]))
	vol_bounds_ok = false;

    if (!vol_bounds_ok)
      {
      cerr << endl << "Error: Min bounds must be less than max bounds"
	   << endl << endl;
      return -1;
      }
    }
  
  // Perform processing on parameters

  std::vector<std::string> isosurface_level_strings;

  QuickFlash::Utils::read_vector_string(isosurface_level_input, 
				      isosurface_level_strings);

  const unsigned int num_iso_levels = isosurface_level_strings.size();

  std::vector<double> isosurface_level_list(num_iso_levels, 0.0);

  for (unsigned int iso_index = 0; iso_index < num_iso_levels; iso_index++)
    isosurface_level_list[iso_index]
      = std::atof(isosurface_level_strings[iso_index].c_str());

  // Run through the files

  std::list<std::string>::const_iterator flash_filename_iter 
    = flash_filenames.begin();

  while (flash_filename_iter != flash_filenames.end())
    {
    const std::string & filename = *flash_filename_iter;

    QuickFlash::File::DataFile df(filename, read_buffer_size, true);

    for (unsigned int iso_index = 0; iso_index < num_iso_levels; iso_index++)
      {
      const double isosurface_level = isosurface_level_list[iso_index];

      // Build the filename prefix

      std::string output_file_prefix;

      output_file_prefix = user_output_prefix + filename + "_" + dataset_name;

      // Add the isosurface level

      output_file_prefix += "_" + isosurface_level_strings[iso_index];

      // Process the file

      build_triangle_list(df, output_file_prefix, dataset_name, 
			  isosurface_level, limit_refine_level, 
			  max_refine_level, reverse_triangles, both_sides,
			  side_separation, restrict_volume_min,
			  volume_min_bounds, restrict_volume_max,
			  volume_max_bounds, obj_mtllib_file, obj_mtl_front,
			  obj_mtl_back);
      }

    ++flash_filename_iter;
    }

  return 0;
  }

