#!/usr/bin/env python

# generate_slice_svg.py

# By Nathan C. Hearn
#    March 28, 2008
#
# Script to produce axes and colormap legends for slicer images


canvas_default_width_screen_pixels = float(800.0)
canvas_default_width = float(1.0)

label_digits = int(4)


def build_scicomp_canvas(image_info, center_offset=None,
                         include_colormap_legend=False,
                         height_scale=float(1.0)) :

    import math

    import SciCompositor as scicomp

    imagefile_name = image_info.get_image_filename()

    datafile_name = image_info.get_datafile_name()
    dataset_name = image_info.get_dataset()

    colormap_imagefile_name = None
    colormap_horizontal = False
    colormap_min_value = float(0.0)
    colormap_max_value = float(0.0)
    colormap_log_scale = False
    colormap_length = int(0)
    colormap_width = int(0)

    if (include_colormap_legend) :
        
        colormap_imagefile_name = image_info.get_colormap_image_filename()

        colormap_info = image_info.get_colormap_info()
        
        colormap_horizontal = colormap_info.horizontal_colormap()

        colormap_min_value = colormap_info.get_min_value()
        colormap_max_value = colormap_info.get_max_value()

        colormap_log_scale = colormap_info.use_log_values()

        colormap_length = colormap_info.get_value_axis_length()
        colormap_width = colormap_info.get_width_pixels()  ## Not necessarily x

        if (colormap_log_scale) :
            colormap_min_value = math.log10(colormap_min_value)
            colormap_max_value = math.log10(colormap_max_value)


    image_center = None

    if (center_offset is not None) :

        if (len(center_offset) != 2) :
            raise RuntimeError("Center offset must have two elements")

        image_center = center_offset

    else :

        image_center = (0.0, 0.0)


    image_width = image_info.get_image_width()

    if (len(image_width) != 2) :
        raise RuntimeError("Image width must have two elements")

    image_coords_min_x = image_center[0] - (0.5 * image_width[0])
    image_coords_max_x = image_center[0] + (0.5 * image_width[0])

    image_coords_min_y = image_center[1] - (0.5 * image_width[1])
    image_coords_max_y = image_center[1] + (0.5 * image_width[1])

    image_aspect_height_width \
        = height_scale * (image_width[1] / image_width[0])

    image_width_ticks = int(10)
    
    image_height_ticks \
        = int(float(image_width_ticks) * image_aspect_height_width)

    if (image_height_ticks < 2) :
        image_height_ticks = 2

    long_axis = 0    ## Index of the longer axis
    short_axis = 1   ## Index of the shorter axis

    if (image_aspect_height_width > 1.0) :
        long_axis = 1
        short_axis = 0

    min_image_dimension = image_width[short_axis]
    max_image_dimension = image_width[long_axis]
    
    canvas_width = canvas_default_width
    canvas_height = canvas_width * (image_aspect_height_width + 0.05)

    if (include_colormap_legend) :
        if (colormap_horizontal) :
            canvas_height += canvas_default_width * 0.2
        else :
            canvas_width += canvas_default_width * 0.2


    image_port_width = canvas_default_width * 0.8
    image_port_height = image_port_width * image_aspect_height_width

    image_port_min_x = canvas_default_width * 0.1
    image_port_max_x = image_port_min_x + image_port_width

    image_port_max_y = canvas_height - (canvas_default_width * 0.1)
    image_port_min_y = image_port_max_y - image_port_height

    image_line_width = 0.002 * canvas_width
    image_label_font_size = 0.025 * min_image_dimension

    header_port_width = 0.9 * canvas_default_width
    header_port_height = 0.05 * canvas_height

    header_port_min_x = 0.05 * canvas_default_width
    header_port_max_x = header_port_min_x + header_port_width

    header_port_min_y = 0.95 * canvas_height
    header_port_max_y = header_port_min_y + header_port_height

    header_text_x = 0.05 * header_port_width
    header_text_y = 0.4 * header_port_height

    header_text_font_size = 0.4 * header_port_height

    colormap_port_width = 0.0
    colormap_port_height = 0.0

    colormap_port_min_x = 0.0
    colormap_port_max_x = 0.0

    colormap_port_min_y = 0.0
    colormap_port_max_y = 0.0

    colormap_coords_min_x = 0.0
    colormap_coords_max_x = 0.0

    colormap_coords_min_y = 0.0
    colormap_coords_max_y = 0.0

    colormap_line_width = 0.0
    colormap_label_font_size = 0.0

    if (include_colormap_legend) :
        
        if (colormap_horizontal) :

            colormap_aspect_height_width \
                = float(colormap_width) / float(colormap_length)

            ## To prevent round-off issues, make colormap axis limits similar

            colormap_coords_delta = colormap_max_value - colormap_min_value

            colormap_coords_min_x = colormap_min_value
            colormap_coords_max_x = colormap_max_value

            colormap_coords_min_y = 0.0
            colormap_coords_max_y \
                = colormap_coords_delta * colormap_aspect_height_width

            colormap_label_font_size = 0.02 * colormap_coords_delta

            ## Find the colormap port dimensions
            
            colormap_port_width = image_port_width
            colormap_port_height = colormap_default_width * 0.05
            
            colormap_port_min_x = image_port_min_x
            colormap_port_min_y = canvas_default_width * 0.05

        else :

            colormap_aspect_width_height \
                = float(colormap_width) / float(colormap_length)

            ## To prevent round-off issues, make colormap axis limits similar

            colormap_coords_delta = colormap_max_value - colormap_min_value

            colormap_coords_min_x = 0.0
            colormap_coords_max_x \
                = colormap_coords_delta * colormap_aspect_width_height

            colormap_coords_min_y = colormap_min_value
            colormap_coords_max_y = colormap_max_value

            colormap_label_font_size = 0.02 * colormap_coords_delta

            ## Find the colormap port dimensions
            
            colormap_port_width = canvas_default_width * 0.05
            colormap_port_height = image_port_height

            colormap_port_min_x = canvas_default_width * (1.05)
            colormap_port_min_y = image_port_min_y

        colormap_port_max_x = colormap_port_min_x + colormap_port_width
        colormap_port_max_y = colormap_port_min_y + colormap_port_height

        colormap_line_width = 0.002 * canvas_default_width


    ## Set up the canvas

    c = scicomp.Canvas(0.0, canvas_width, 0.0, canvas_height)

    # Build the header port

    header_port = c.get_new_viewport("header_port",
                                     header_port_min_x, header_port_max_x,
                                     header_port_min_y, header_port_max_y,
                                     0.0, header_port_width,
                                     0.0, header_port_height)

    header_text_str = "File " + image_info.get_datafile_name() \
                      + " - Dataset " + image_info.get_dataset()

    sim_time = image_info.get_sim_time()

    if (sim_time is not None) :
        header_text_str += " - Time " + str(sim_time)

    header_text = scicomp.TextObject("header_text", header_text_str)

    header_text.set_anchor(header_text_x, header_text_y,
                           font_size=header_text_font_size)

    header_port.add_child_object(header_text)

    # Build the image port

    image_port = c.get_new_viewport((dataset_name + "_port"),
                                    image_port_min_x, image_port_max_x,
                                    image_port_min_y, image_port_max_y,
                                    0.0, image_port_width,
                                    0.0, image_port_height)

    image_port.set_physical_coord_map(image_coords_min_x, image_coords_max_x,
                                      image_coords_min_y, image_coords_max_y)

    dataset_image = scicomp.ImageObject((dataset_name + "_image"),
                                        imagefile_name,
                                        image_coords_min_x,
                                        image_coords_max_x,
                                        image_coords_min_y,
                                        image_coords_max_y,
                                        use_physical_coords=True)

    image_port.add_child_object(dataset_image)

    dataset_axes = scicomp.AxesObject((dataset_name + "_image_axes"),
                                      image_coords_min_x, image_coords_max_x,
                                      image_coords_min_y, image_coords_max_y,
                                      use_physical_coords=True)
    
    dataset_axes.set_frame_color(red=1.0, green=0.0, blue=0.0)
    dataset_axes.set_frame_line_width(image_line_width)

    dataset_axes.set_x_major_ticks_regular(max_ticks=image_width_ticks)
    dataset_axes.set_y_major_ticks_regular(max_ticks=image_height_ticks)

    dataset_axes.set_major_ticks_color(red=0.7, green=0.0, blue=0.7,
                                       opacity=0.5)
    dataset_axes.set_major_ticks_line_width(image_line_width)

    dataset_axes.set_axis_label_font_size(image_label_font_size)
    dataset_axes.set_x_label_digits(label_digits)
    dataset_axes.set_y_label_digits(label_digits)

    image_port.add_child_object(dataset_axes)    

    ## Add the colormap

    if (include_colormap_legend) :
        
        colormap_port = c.get_new_viewport((dataset_name + "_cm_port"),
                                           colormap_port_min_x,
                                           colormap_port_max_x,
                                           colormap_port_min_y,
                                           colormap_port_max_y,
                                           0.0, colormap_port_width,
                                           0.0, colormap_port_height)

        colormap_port.set_physical_coord_map(colormap_coords_min_x,
                                             colormap_coords_max_x,
                                             colormap_coords_min_y,
                                             colormap_coords_max_y)

        colormap_image = scicomp.ImageObject((dataset_name + "_cm_image"),
                                             colormap_imagefile_name,
                                             colormap_coords_min_x,
                                             colormap_coords_max_x,
                                             colormap_coords_min_y,
                                             colormap_coords_max_y,
                                             use_physical_coords=True)

        colormap_port.add_child_object(colormap_image)

        colormap_axes = scicomp.AxesObject((dataset_name + "_cm_axes"),
                                           colormap_coords_min_x,
                                           colormap_coords_max_x,
                                           colormap_coords_min_y,
                                           colormap_coords_max_y,
                                           use_physical_coords=True)

        colormap_axes.set_frame_color(red=1.0, green=0.0, blue=0.0)
        colormap_axes.set_frame_line_width(colormap_line_width)

        if (colormap_horizontal) :

            if (colormap_log_scale) :
                colormap_axes.set_x_label_exponential_base(10.0)
                colormap_axes.set_x_ticks_logscale_regular(max_ticks=10)
            else :
                set_x_major_ticks_regular(max_ticks=10)
                
            colormap_axes.set_x_label_digits(label_digits)
            
            colormap_axes.show_y_major_labels(False)
            
        else :
            
            if (colormap_log_scale) :
                colormap_axes.set_y_label_exponential_base(10.0)
                colormap_axes.set_y_ticks_logscale_regular(max_ticks=10)
            else :
                colormap_axes.set_y_major_ticks_regular(max_ticks=10)
                
            colormap_axes.set_y_label_digits(label_digits)
            
            colormap_axes.show_x_major_labels(False)

        colormap_axes.set_major_ticks_color(red=0.7, green=0.0, blue=0.7,
                                            opacity=0.5)
        
        colormap_axes.set_major_ticks_line_width(colormap_line_width)

        colormap_axes.set_minor_ticks_color(red=0.5, green=0.5, blue=0.2,
                                            opacity=0.4)

        colormap_axes.set_minor_ticks_line_width(0.5 * colormap_line_width)

        colormap_axes.set_axis_label_font_size(colormap_label_font_size)

        colormap_port.add_child_object(colormap_axes)

    return c


def main() :

    from sys import exit, stderr
    from optparse import OptionParser

    from image_info import ImageInfo


    ## Set up the command-line parser

    usage = str("%prog [options] slice_metadata_files ...")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    canvas_width_str = str("canvas-width")
    argparser.add_option("--" + canvas_width_str, type="float",
                         dest="canvas_width",
                         help="Canvas image width (screen pixel widths)")
    argparser.set_defaults(canvas_width=canvas_default_width_screen_pixels)

    show_colormap_str = str("show-colormap")
    argparser.add_option("--" + show_colormap_str, action="store_true",
                         dest="show_colormap", help="Show the colormap legend")
    argparser.set_defaults(show_colormap=False)

    font_scale_str = str("font-scale")
    argparser.add_option("--" + font_scale_str, type="float",
                         dest="font_scale", help="Scaling factor for text")
    argparser.set_defaults(font_scale=float(1.0))

    center_coords_str = str("center-coords")
    argparser.add_option("--" + center_coords_str, type="string",
                         dest="center_coords",
                         help="Comma-separated center coords (horiz,vert) -- "
                              + "use \"x\", \"y\", or \"z\" to use the center "
                              + "of the slice grid")
    argparser.set_defaults(center_coords=None)
    
    ## Process the command-line input

    opts, args = argparser.parse_args()

    argc = len(args)

    min_args = int(1)

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n\n")
        exit(1)

    argptr = int(0)

    metadata_file_list = list()

    while (argptr < argc) :
        metadata_file_list.append(str(args[argptr]))
        argptr += 1


    canvas_width = opts.canvas_width

    if (not (canvas_width > 0.0)) :
        stderr.write("\nError: Canvas width must be positive\n\n")
        exit(-1)
        

    show_colormap = opts.show_colormap


    font_scale = opts.font_scale

    if (not (font_scale > 0.0)) :
        stderr.write("\nError: Font scale must be positive\n\n")
        exit(-1)


    horiz_center = float(0.0)
    vert_center = float(0.0)

    if (opts.center_coords is not None) :

        words = opts.center_coords.split(",")

        if (len(words) != 2) :
            stderr.write("\nError: Center coordinates requires two "
                         + "comma-separated entries\n\n")
            exit(-1)

        horiz_word = words[0]
        vert_word = words[1]

        if ((horiz_word != "x") and (horiz_word != "y")
            and (horiz_word != "z")) :
            horiz_center = float(horiz_word)
        else :
            horiz_center = horiz_word

        if ((vert_word != "x") and (vert_word != "y")
            and (vert_word != "z")) :
            vert_center = float(vert_word)
        else :
            vert_center = vert_word


    ## Build the files

    for metadata_file_name in metadata_file_list :

        image_info = ImageInfo(metadata_file_name)

        output_filename = image_info.get_image_filename() + ".svg"

        horiz_center_coord = horiz_center
        vert_center_coord = vert_center
        
        if (isinstance(horiz_center_coord, str)) :
            if (horiz_center_coord == "x") :
                horiz_center_coord = image_info.get_grid_center()[0]
            elif (horiz_center_coord == "y") :
                horiz_center_coord = image_info.get_grid_center()[1]
            elif (horiz_center_coord == "z") :
                horiz_center_coord = image_info.get_grid_center()[2]
            
        if (isinstance(vert_center_coord, str)) :
            if (vert_center_coord == "x") :
                vert_center_coord = image_info.get_grid_center()[0]
            elif (vert_center_coord == "y") :
                vert_center_coord = image_info.get_grid_center()[1]
            elif (vert_center_coord == "z") :
                vert_center_coord = image_info.get_grid_center()[2]

        center_offset = (horiz_center_coord, vert_center_coord)

        canvas = build_scicomp_canvas(image_info, center_offset,
                                      include_colormap_legend=show_colormap)

        svg_code = canvas.get_svg_code(canvas_width,
                                       font_size_factor=font_scale)

        f = open(output_filename, "w")

        for line in svg_code :

            f.write(line + "\n")

        f.write("\n")

    
if (__name__ == "__main__") :
    main()
