// C++ program file slicer.cpp

/*
  By Nathan C. Hearn
     July 14, 2007

  Takes 2D slice from a 2D or 3D Flash dataset.

  For parallel operation, define the preprocessor variable USE_MPI.
*/


#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <fstream>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_except.hpp"
#include "quickflash_slice_slicegrid.hpp"
#include "quickflash_slice_slicer.hpp"
#include "quickflash_color_colormap.hpp"
#include "quickflash_file_flashdefs.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_image_imagegrid.hpp"
#include "quickflash_image_imagemagick.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_slice_output.hpp"


#ifdef USE_MPI
#include <mpi.h>
#endif  // USE_MPI


const unsigned int default_buffer_size = 50;
const unsigned int default_cache_size = 1000;

const char default_image_extension[] = "png";


void determine_distrib(const unsigned int num_processes,
		       const unsigned int num_pixels,
		       std::vector<unsigned int> & start_indexes,
		       std::vector<unsigned int> & pixel_counts)
  {
  if (num_processes < 1)
    throw QuickFlash::Except("Process count must be positive", __FILE__, 
			   __LINE__);

  start_indexes.resize(num_processes);
  pixel_counts.resize(num_processes);

  const unsigned int mean_pixels = num_pixels / num_processes;
  const unsigned int excess_pixels = num_pixels % num_processes;

  const unsigned int mean_pixels_plusOne = mean_pixels + 1;

  unsigned int last_index = 0;

  for (unsigned int m = 0; m < excess_pixels; m++)
    {
    start_indexes[m] = last_index;
    pixel_counts[m] = mean_pixels_plusOne;

    last_index += mean_pixels_plusOne;
    }

  for (unsigned int m = excess_pixels; m < num_processes; m++)
    {
    start_indexes[m] = last_index;
    pixel_counts[m] = mean_pixels;

    last_index += mean_pixels;
    }

  if (last_index != num_pixels)
    throw QuickFlash::Except("Unable to compute distribution", __FILE__, 
			   __LINE__);
  }


void get_default_slice_bounds(const QuickFlash::File::DataFile & df,
			      std::vector<double> & slice_center,
			      std::vector<double> & slice_width,
			      std::vector<unsigned int> & slice_pixels,
			      std::vector<double> & slice_pixel_width)
  {
  // Default slice is parallel to x-y plane and intersects mesh center

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int mesh_dims = meshinfo.get_dims();

  if (mesh_dims < 2)
    throw QuickFlash::Except("Slices for one-dimensional meshes not supported",
			     __FILE__, __LINE__);

  slice_center = meshinfo.get_volume_center();

  // Set slice width to full volume width (for the x-y plane)

  const std::vector<double> & mesh_width = meshinfo.get_volume_width();
  
  slice_width.resize(2);

  for (unsigned int m = 0; m < 2; m++)
    slice_width[m] = mesh_width[m];

  for (unsigned int m = 0; m < 2; m++)
    if (!(slice_width[m] > 0.0))
      throw QuickFlash::Except("Non-positive slice width found", __FILE__, 
			       __LINE__);

  // Determine the number of pixels from the tree depth

  const std::vector<unsigned int> & base_mesh_blocks 
    = meshinfo.get_base_block_dims();

  const std::vector<unsigned int> & block_dims = meshinfo.get_block_dims();

  const unsigned int max_refine_level = meshinfo.get_max_leaf_refine_level();

  const unsigned int refine_level_factor = 1 << (max_refine_level - 1);

  slice_pixels.resize(2);

  for (unsigned int m = 0; m < 2; m++)
    slice_pixels[m] 
      = base_mesh_blocks[m] * block_dims[m] * refine_level_factor;

  slice_pixel_width.resize(2);

  for (unsigned int m = 0; m < 2; m++)
    {
    const double axis_pixel_width
      = slice_width[m] / static_cast<double>(slice_pixels[m]);

    slice_pixel_width[m] = axis_pixel_width;
    }

  // ERROR CHECKING???
  }


void populate_grid(const QuickFlash::File::DataFile & df, 
		   const std::string & dataset_name, 
		   const std::vector<unsigned int> & subsample_dims,
		   QuickFlash::Slice::SliceGrid & grid, 
		   bool quiet_console=false,
		   const double empty_value=0.0, 
		   const unsigned int cache_size=default_cache_size)
  {
  // Set up MPI-related constants

  // unsigned int proc_id = 0;
  // unsigned int num_procs = 1;

#ifdef USE_MPI
#endif  // USE_MPI

  // Open file info objects

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const QuickFlash::File::Dataset & dset = df.get_dataset(dataset_name);

  dset.set_cache_size(cache_size);

  if (!quiet_console)
    dset.set_report_stats(true);
  else
    dset.set_report_stats(false);

  QuickFlash::Slice::slice_simple(meshinfo, dset, subsample_dims, grid, 
				empty_value);
  }


#ifdef USE_MAGICK
int write_image(const std::string & filename, 
		const QuickFlash::Slice::SliceGrid & data,
		QuickFlash::Color::Colormap & colormap,
		const bool transpose_image=false, 
		const bool show_colormap=false,
		const bool use_log_scale=false,
		const bool symmetric_map_range=false,
		const double symmetric_map_range_center=0.0)
  {
  int ret_val = -1;  // Error state

  // Build the colormap

  double min_val = 0.0;
  double max_val = 0.0;

  data.get_minmax_values(min_val, max_val);

  if (!(min_val < max_val))
    max_val = (2.0 * std::fabs(min_val)) + 1.0;  // Make something up ...

  if (symmetric_map_range)
    {
    // WHAT IF WE ADD LOG-SCALING???

    const double range_half_magnitude 
      = std::max(std::fabs(min_val - symmetric_map_range_center), 
		 std::fabs(max_val - symmetric_map_range_center));

    min_val = symmetric_map_range_center - range_half_magnitude;
    max_val = symmetric_map_range_center + range_half_magnitude;
    }

  // Set up the provided Colormap object

  colormap.reset(min_val, max_val, use_log_scale);

  QuickFlash::Image::ImageGrid image(data.get_grid_data(), colormap);

  if (show_colormap)
    {
    ret_val 
      = QuickFlash::Image::ImageMagick::write_image_colormap(filename, image,
							     colormap,
							     transpose_image,
							     true);
    }
  else
    {
    ret_val = QuickFlash::Image::ImageMagick::write_image(filename, image, 
							  transpose_image);
    }

  return ret_val;
  }


int write_image(const std::string & filename, 
		const QuickFlash::Slice::SliceGrid & data,
		const bool transpose_image=false,
		const bool show_colormap=false,
		const bool use_log_scale=false, 
		const bool symmetric_map_range=false,
		const double symmetric_map_range_center=0.0)
  {
  QuickFlash::Color::Colormap colormap;

  return write_image(filename, data, colormap, transpose_image, 
		     show_colormap, use_log_scale, 
		     symmetric_map_range, symmetric_map_range_center);
  }
#endif  // USE_MAGICK


#ifdef USE_MAGICK
int write_image(const std::string & filename, 
		const std::string & colormap_filename, 
		const QuickFlash::Slice::SliceGrid & data,
		QuickFlash::Color::Colormap & colormap,
		const bool transpose_image=false, 
		const bool rescale_colormap=false,
		const bool show_colormap=false,
		const bool use_log_scale=false,
		const bool symmetric_map_range=false,
		const double symmetric_map_range_center=0.0)
  {
  using std::cerr;
  using std::endl;

  int ret_val = -1;  // Error state

  // Load the colormap into the provided Colormap object

  if (load_colormap(colormap_filename, colormap, use_log_scale) >= 0)
    {
    if (rescale_colormap)
      {
      double min_val = 0.0;
      double max_val = 0.0;

      data.get_minmax_values(min_val, max_val);

      if (symmetric_map_range)
	{
	// WHAT IF WE ADD LOG-SCALING???

	const double range_half_magnitude 
	  = std::max(std::fabs(min_val - symmetric_map_range_center), 
		     std::fabs(max_val - symmetric_map_range_center));

	min_val = symmetric_map_range_center - range_half_magnitude;
	max_val = symmetric_map_range_center + range_half_magnitude;
	}

      // WHAT IF min_val == max_val ???

      if (min_val < max_val)
	colormap.rescale_values(min_val, max_val, use_log_scale);
      }

    QuickFlash::Image::ImageGrid image(data.get_grid_data(), colormap);

    if (show_colormap)
      {
      ret_val 
	= QuickFlash::Image::ImageMagick::write_image_colormap(filename, image,
							     colormap,
							     transpose_image,
							     true);
      }
    else
      {
      ret_val = QuickFlash::Image::ImageMagick::write_image(filename, image, 
							  transpose_image);
      }
    }
  else
    cerr << endl << "Error: Unable to load colormap " << colormap_filename 
	 << endl;

  return ret_val;
  }


int write_image(const std::string & filename, 
		const std::string & colormap_filename, 
		const QuickFlash::Slice::SliceGrid & data,
		const bool transpose_image=false, 
		const bool rescale_colormap=false,
		const bool show_colormap=false,
		const bool use_log_scale=false,
		const bool symmetric_map_range=false,
		const double symmetric_map_range_center=0.0)
  {
  QuickFlash::Color::Colormap colormap;

  return write_image(filename, colormap_filename, data, colormap,
		     transpose_image, rescale_colormap, 
		     show_colormap, use_log_scale, 
		     symmetric_map_range, symmetric_map_range_center);
  }
#endif  // USE_MAGICK
    

int main(const int argc, char * const argv[])
  {
  using std::cerr;
  using std::endl;

  const double deg_to_rad = (M_PI) / 180.0;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, 
					option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 2;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name << " dataset_name "
	 << endl << "\t"
	 << "[ --slice-center=CENTER_X,CENTER_Y[,CENTER_Z] ]"
	 << endl << "\t"
	 << "[ --slice-grid-width=GRID_X_WIDTH,GRID_Y_WIDTH ]"
	 << endl << "\t"
	 << "[ --slice-grid-pixels=GRID_X_PIXELS,GRID_Y_PIXELS ]"
	 << endl << "\t"
	 << "[ --rotation=[ROLL_ANGLE,POLAR_ANGLE,]AZIMUTHAL_ANGLE ]"
	 << endl << "\t"
	 << "[ --z-samples=NUM_Z_SAMPLES,Z_WIDTH ]"
	 << endl << "\t"
	 << "[ --subsamples=SUBSAMPLES_X,SUBSAMPLES_Y ]"
	 << endl << "\t"
	 << "[ --output-prefix=OUTPUT_PREFIX ]"
	 << endl << "\t"
	 << "[ --output-suffix=DESCIPTIVE_SUFFIX ]"
	 << endl << "\t"
	 << "[ --compute-symmetry=horiz|vert ]"
         << endl << "\t"
#ifdef USE_MAGICK
	 << "[ --image-type=FILE_EXTENSION ] [ --colormap=COLORMAP_FILENAME ]"
	 << endl << "\t"
	 << "[ --rescale-colormap ] [ --symm-map-range[=CENTER_VALUE] ]"
	 << endl << "\t"
	 << "[ --logscale ]"
	 << endl << "\t"
	 << "[ --show-colormap ] [ --save-colormap[=LENGTH,WIDTH] ]"
	 << endl << "\t"
	 << "[ --horizontal-colormap ] [ --save-colormap-info ]"
	 << endl << "\t"
	 << "[ --image-only ] [ --transpose-image ] [ --save-image-metadata ]"
	 << endl << "\t"
#endif  // USE_MAGICK
	 << "[ --empty-value=VALUE ] [ --buffer-size=BUFFER_BLOCKS ]"
	 << endl << "\t"
	 << "[ --cache-size=CACHE_BLOCKS ] [ --quiet ] [ -- ]"
	 << endl << "\t"
	 << "flash_filename1 [ flash_filename2 ... ]"
	 << endl << endl;

    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  const std::string dataset_name = argument_list[argptr++];

  std::list<std::string> flash_filenames;

  while (argptr < num_args)
    flash_filenames.push_back(argument_list[argptr++]);

  // Parse the options

  bool user_slice_center = false;
  std::vector<double> slice_center;

  bool user_slice_width = false;
  std::vector<double> slice_width;

  bool user_slice_pixels = false;
  std::vector<unsigned int> slice_pixels;

  double roll_angle = 0.0;
  double polar_angle = 0.0;
  double azimuthal_angle = 0.0;

  std::vector<unsigned int> plane_subsample_dims(2, 1);

  unsigned int ortho_subsamples = 1;
  double ortho_width = 0.0;

  std::string base_output_file_prefix = "";

  std::string user_output_file_suffix = "";

  bool compute_symmetry = false;
  unsigned int symmetry_axis = 0;

#ifdef USE_MAGICK
  bool output_image = false;
  std::string image_extension = default_image_extension;

  bool use_colormap = false;
  std::string colormap_filename;

  bool rescale_colormap = false;

  bool symm_map_range = false;
  double symm_map_range_center = 0.0;

  bool use_log_scale = false;

  bool show_colormap = false;

  bool save_colormap = false;

  unsigned int colormap_length = 0;  // Value axis
  unsigned int colormap_width = 0;

  bool horizontal_colormap = false;

  bool save_colormap_info = false;
#endif  // USE_MAGICK

  bool output_data = true;

#ifdef USE_MAGICK
  bool transpose_image = false;

  bool save_image_metadata = false;
#endif  // USE_MAGICK

  double empty_value = 0.0;  // Value for cells containing no data

  unsigned int read_buffer_size = default_buffer_size;
  unsigned int read_cache_size = default_cache_size;

  bool quiet_console = false;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "slice-center")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Slice center option requires comma-separated "
	     << "coordinate elements" << endl << endl;
	return -1;
	}

      current_option.get_value(slice_center);

      const unsigned int space_dims = slice_center.size();

      if ((space_dims < 2) || (space_dims > 3))
	{
	cerr << endl << "Error: Only 2D and 3D datasets supported" << endl 
	     << endl;
	return -1;
	}

      user_slice_center = true;
      }
    else if (keyword == "slice-grid-width")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Slice width option requires comma-separated "
	     << "elements" << endl << endl;
	return -1;
	}

      current_option.get_value(slice_width);

      if (slice_width.size() != 2)
	{
	cerr << endl << "Error: Slice width requires two positive elements" 
	     << endl << endl;
	return -1;
	}

      for (unsigned int m = 0; m < 2; m++)
	if (!(slice_width[m] > 0.0))
	  {
	  cerr << endl << "Error: Slice widths must be positive" << endl 
	       << endl;
	  return -1;
	  }

      user_slice_width = true;
      }
    else if (keyword == "slice-grid-pixels")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Slice pixels option requires comma-separated "
	     << "elements" << endl << endl;
	return -1;
	}

      current_option.get_value(slice_pixels);

      if (slice_pixels.size() != 2)
	{
	cerr << endl << "Error: Slice pixels requires two positive integers" 
	     << endl << endl;
	return -1;
	}

      for (unsigned int m = 0; m < 2; m++)
	if (slice_pixels[m] < 1)
	  {
	  cerr << endl << "Error: Slice pixels must be positive" << endl 
	       << endl;
	  return -1;
	  }

      user_slice_pixels = true;
      }
    else if (keyword == "rotation")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Angle inputs are required for rotation option"
	     << endl << endl;
	return -1;
	}

      std::vector<double> rotation_vector;

      current_option.get_value(rotation_vector);

      if (rotation_vector.size() == 1)
	{
	roll_angle = 0.0;
	polar_angle = 0.0;
	azimuthal_angle = rotation_vector[0];
	}
      else if (rotation_vector.size() == 3)
	{
	roll_angle = rotation_vector[0];
	polar_angle = rotation_vector[1];
	azimuthal_angle = rotation_vector[2];
	}
      else
	{
	cerr << endl << "Error: Rotation option attributes not recognized"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "z-samples")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: z-samples option requires sample count and width"
	     << endl << endl;
	return -1;
	}

      // Well ... this isn't pretty

      std::vector<unsigned int> z_sample_vect_int;

      current_option.get_value(z_sample_vect_int);

      if (z_sample_vect_int.size() != 2)
	{
	cerr << endl 
	     << "Error: z-samples option requires sample count and width"
	     << endl << endl;
	return -1;
	}

      ortho_subsamples = z_sample_vect_int[0];

      if (ortho_subsamples < 1)
	{
	cerr << endl << "Error: At least one z-sample required" << endl 
	     << endl;
	return -1;
	}

      std::vector<double> z_sample_vect_fl;

      current_option.get_value(z_sample_vect_fl);

      ortho_width = z_sample_vect_fl[1];

      if (!(ortho_width > 0.0))
	{
	cerr << endl << "Error: z-sample width must be positive" << endl 
	     << endl;
	return -1;
	}
      }
    else if (keyword == "subsamples")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Subsamples option requires x- and y- subsample count"
	     << endl << endl;
	return -1;
	}

      std::vector<unsigned int> subsample_counts;

      current_option.get_value(subsample_counts);

      if (subsample_counts.size() != 2)
	{
	cerr << endl 
	     << "Error: Subsamples option requires x- and y- subsample count"
	     << endl << endl;
	return -1;
	}

      for (unsigned int m = 0; m < 2; m++)
	if (subsample_counts[m] < 1)
	  {
	  cerr << endl << "Error: At least one subsample required" << endl 
	       << endl;
	  return -1;
	  }

      plane_subsample_dims = subsample_counts;
      }
    else if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output prefix is required with option"
	     << endl << endl;
	return -1;
	}

      base_output_file_prefix = current_option.get_value();

      // Add an underscore if prefix contains more than just a path

      std::string base_prefix_path;
      std::string base_prefix_name;

      QuickFlash::Utils::split_name_path_file(base_output_file_prefix, 
					      base_prefix_path,
					      base_prefix_name);

      const unsigned int base_prefix_name_len = base_prefix_name.size();

      if (base_prefix_name_len > 0)
	if (base_prefix_name[base_prefix_name_len - 1] != '_')
	  base_output_file_prefix += "_";
      }
    else if (keyword == "output-suffix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output suffix is required with option"
	     << endl << endl;
	return -1;
	}

      user_output_file_suffix = current_option.get_value();

      if (user_output_file_suffix.size() > 0)
	user_output_file_suffix.insert(0, 1, '_');
      }
    else if (keyword == "compute-symmetry")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Symmetry computation option requires horiz or vert"
	     << endl << endl;
	return -1;
	}

      const std::string & symm_str = current_option.get_value();

      if (symm_str == "horiz")
	symmetry_axis = 0;
      else if (symm_str == "vert")
	symmetry_axis = 1;
      else
	{
	cerr << endl 
	     << "Error: Symmetry computation option requires horiz or vert"
	     << endl << endl;
	return -1;
	}

      compute_symmetry = true;
      }
    else if (keyword == "image-type")
      {
#ifdef USE_MAGICK
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Image type option requires image file extension"
	     << endl << endl;
	return -1;
	}

      output_image = true;
      image_extension = current_option.get_value();
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "colormap")
      {
#ifdef USE_MAGICK
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Colormap option requires colormap file name"
	     << endl << endl;
	return -1;
	}

      use_colormap = true;
      colormap_filename = current_option.get_value();
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword =="rescale-colormap")
      {
#ifdef USE_MAGICK
      rescale_colormap = true;
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword =="symm-map-range")
      {
#ifdef USE_MAGICK
      if (current_option.value_present())
	current_option.get_value(symm_map_range_center);

      symm_map_range = true;
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword =="logscale")
      {
#ifdef USE_MAGICK
      use_log_scale = true;
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "show-colormap")
      {
#ifdef USE_MAGICK
      show_colormap = true;
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "save-colormap")
      {
#ifdef USE_MAGICK
      save_colormap = true;

      if (current_option.value_present())
	{
	std::vector<unsigned int> colormap_dims;

	current_option.get_value(colormap_dims);

	if (colormap_dims.size() != 2)
	  {
	  cerr << endl << "Error: Colormap options must be length,width"
	       << endl << endl;
	  return -1;
	  }

	colormap_length = colormap_dims[0];

	if (colormap_length < 2)
	  {
	  cerr << endl << "Error: Colormap length must be at least 2 pixels" 
	       << endl << endl;
	  return -1;
	  }

	colormap_width = colormap_dims[1];

	if (colormap_width < 1)
	  {
	  cerr << endl << "Error: Colormap width must be at least 1 pixel" 
	       << endl << endl;
	  return -1;
	  }
	}
      else
	{
	colormap_length = std::max(slice_pixels[0], 
				   static_cast<unsigned int>(2));
	colormap_width = 10;
	}
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "horizontal-colormap")
      {
#ifdef USE_MAGICK
      horizontal_colormap = true;
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "save-colormap-info")
      {
#ifdef USE_MAGICK
      save_colormap_info = true;
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "image-only")
      {
#ifdef USE_MAGICK
      output_data = false;
      output_image = true;
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "transpose-image")
      {
#ifdef USE_MAGICK
      transpose_image = true;
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "save-image-metadata")
      {
#ifdef USE_MAGICK
      save_image_metadata = true;
#else
      cerr << endl << "Error: ImageMagick support not available" << endl 
	   << endl;
      return -1;
#endif  // USE_MAGICK
      }
    else if (keyword == "empty-value")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Empty value option requires a value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(empty_value);
      }
    else if (keyword == "buffer-size")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Buffer size option requires a value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(read_buffer_size);
      }
    else if (keyword == "cache-size")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Cache size option requires a value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(read_cache_size);
      }
    else if (keyword == "quiet")
      {
      quiet_console = true;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }


  // Check dependencies

#ifdef USE_MAGICK
  if (save_image_metadata && save_colormap)
    save_colormap_info = true;
#endif  // USE_MAGICK

  // Check the number of subsamples

  unsigned int num_plane_subsamples = 1;

  for (unsigned int i = 0; i < 2; i++)
    num_plane_subsamples *= plane_subsample_dims[i];

  if (num_plane_subsamples < 1)
    {
    cerr << "Error: Non-positive subsample dimension" << endl << endl;
    return -1;
    }


  // Run through the files

  std::list<std::string>::const_iterator flash_filename_iter 
    = flash_filenames.begin();

  while (flash_filename_iter != flash_filenames.end())
    {
    const std::string & filename = *flash_filename_iter;

    // Build the output file prefix

    std::string flash_filename_path;
    std::string flash_filename_name;

    QuickFlash::Utils::split_name_path_file(filename, flash_filename_path, 
					  flash_filename_name);

    const std::string output_file_prefix 
      = base_output_file_prefix + flash_filename_name + "_" + dataset_name
        + user_output_file_suffix;

    // OTHER ERROR CHECKING???

    // Open the data file and get some basic information

    QuickFlash::File::DataFile df(filename, read_buffer_size, true);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double simtime = meshinfo.get_sim_time();

    const unsigned int mesh_dims = meshinfo.get_dims();

    // Determine the slice paramters

    std::vector<double> default_slice_center;
    std::vector<double> default_slice_width;
    std::vector<unsigned int> default_slice_pixels;
    std::vector<double> default_slice_pixel_width;

    get_default_slice_bounds(df, default_slice_center,default_slice_width,
			     default_slice_pixels, default_slice_pixel_width);

    if (!user_slice_center)
      slice_center = default_slice_center;

    if (!user_slice_width)
      slice_width = default_slice_width;

    if (!user_slice_pixels)
      {
      if (!user_slice_width)
	slice_pixels = default_slice_pixels;
      else
	{
	// Use default pixel widths to determine the number of pixels

	slice_pixels.resize(2);

	for (unsigned int m = 0; m < 2; m++)
	  {
	  const double axis_width = slice_width[m];

	  const double pixel_width = default_slice_pixel_width[m];

	  const unsigned int axis_pixels 
	    = static_cast<unsigned int>(axis_width / pixel_width);

	  slice_pixels[m] = (axis_pixels > 1) ? (axis_pixels) : 1;
	  }
	}
      }

    std::vector<unsigned int> subsample_dims = plane_subsample_dims;

    if (mesh_dims > 2)
      subsample_dims.resize(mesh_dims, ortho_subsamples);
    

    // Build the report strings

    std::list<std::string> comment_list;

    std::stringstream strstream;

    strstream << "DataFile " << filename;

    comment_list.push_back(strstream.str());

    strstream.str("");

    strstream << "Dataset " << dataset_name;

    comment_list.push_back(strstream.str());

    strstream.str("");

    strstream << "SimTime " << simtime;

    comment_list.push_back(strstream.str());

    strstream.str("");

    strstream << "SliceCenter " << slice_center;

    comment_list.push_back(strstream.str());

    strstream.str("");

    strstream << "SliceWidth " << slice_width;

    comment_list.push_back(strstream.str());

    strstream.str("");

    strstream << "SlicePixels " << slice_pixels;

    comment_list.push_back(strstream.str());

    strstream.str("");

    if (mesh_dims > 2)
      {
      strstream << "RollAngle " << roll_angle;

      comment_list.push_back(strstream.str());

      strstream.str("");

      strstream << "PolarAngle " << polar_angle;

      comment_list.push_back(strstream.str());

      strstream.str("");
      }

    strstream << "AzimAngle " << azimuthal_angle;

    comment_list.push_back(strstream.str());

    strstream.str("");

    if (num_plane_subsamples > 1)
      {
      strstream << "PixelSubsamples " << subsample_dims;

      comment_list.push_back(strstream.str());

      strstream.str("");
      }

    if (mesh_dims > 2)
      if (subsample_dims[2] > 1)
	{
	strstream << "OrthoSampleParams " << subsample_dims[2] << ","
		  << ortho_width;

	comment_list.push_back(strstream.str());
      
	strstream.str("");
	}

    strstream << "EmptyValue " << empty_value;

    comment_list.push_back(strstream.str());

    strstream.str("");

    strstream << "BufferSize " << read_buffer_size;

    comment_list.push_back(strstream.str());

    strstream.str("");

    strstream << "CacheSize " << read_cache_size;

    comment_list.push_back(strstream.str());

    strstream.str("");


    // Report the setup to stderr

    if (!quiet_console)
      {
      std::list<std::string>::const_iterator str_iter = comment_list.begin();

      while (str_iter != comment_list.end())
	{
	cerr << *str_iter << endl;
	++str_iter;
	}
      }


    // Build radian versions of the angles

    const double roll_rad = roll_angle * deg_to_rad;
    const double polar_rad = polar_angle * deg_to_rad;
    const double azimuthal_rad = azimuthal_angle * deg_to_rad;


    // Construct the grid

    QuickFlash::Slice::SliceGrid grid(slice_center, slice_width, slice_pixels, 
				      azimuthal_rad, polar_rad, roll_rad, 
				      ortho_width);

    // Fill the grid

    populate_grid(df, dataset_name, subsample_dims, grid, quiet_console,
		  empty_value, read_cache_size);

    if (output_data)
      {
      const std::string output_filename = output_file_prefix + ".csv";

      QuickFlash::Slice::write_grid(output_filename, grid, comment_list);
      }

#ifdef USE_MAGICK
    if (output_image)
      {
      const std::string image_filename 
	= output_file_prefix + "." + image_extension;

      const std::string image_metadata_filename = image_filename + ".info";

      const std::string colormap_image_filename
	= output_file_prefix + "_cm." + image_extension;

      const std::string colormap_info_filename 
	= colormap_image_filename + ".info";

      int error_state = -1;  // Error state

      QuickFlash::Color::Colormap image_colormap;

      if (use_colormap)
	error_state = write_image(image_filename, colormap_filename, grid,
				  image_colormap, transpose_image,
				  rescale_colormap, show_colormap, 
				  use_log_scale, 
				  symm_map_range, symm_map_range_center);
      else
	error_state = write_image(image_filename, grid, image_colormap,
				  transpose_image,
				  show_colormap, use_log_scale, 
				  symm_map_range, symm_map_range_center);

      if (error_state < 0)
	{
	cerr << endl << "Error: Unable to write image " << image_filename 
	     << endl << endl;
	return -2;
	}

      if (save_image_metadata)
	{
	// Get unqualified names for files

	std::string datafile_path;
	std::string datafile_name;

	QuickFlash::Utils::split_name_path_file(filename, datafile_path,
					      datafile_name);

	std::string slice_image_path;
	std::string slice_image_name;

	QuickFlash::Utils::split_name_path_file(image_filename, 
						slice_image_path,
						slice_image_name);

	std::string cmap_info_path;
	std::string cmap_info_name;

	QuickFlash::Utils::split_name_path_file(colormap_info_filename,
						cmap_info_path, 
						cmap_info_name);

	std::string cmap_image_path;
	std::string cmap_image_name;

	QuickFlash::Utils::split_name_path_file(colormap_image_filename,
						cmap_image_path, 
						cmap_image_name);

	  
	std::ofstream mfile(image_metadata_filename.c_str(), std::ios::trunc);

	mfile << "DataFile " << datafile_name << endl;
	mfile << "DataFileDims " << mesh_dims << endl;
	mfile << "Dataset " << dataset_name << endl;

	mfile << "SimTime " << simtime << endl;

	mfile << "SliceCenter " << slice_center << endl;
	mfile << "SliceWidth " << slice_width << endl;
	mfile << "SlicePixels " << slice_pixels << endl;

	mfile << "TransposeImage ";

	if (transpose_image)
	  mfile << "true";
	else
	  mfile << "false";

	mfile << endl;

	if (mesh_dims > 2)
	  {
	  mfile << "RollAngle " << roll_angle << endl;
	  mfile << "PolarAngle " << polar_angle << endl;
	  }

	mfile << "AzimuthalAngle " << azimuthal_angle << endl;

	mfile << "PixelSubsamples " << subsample_dims << endl;

	if (mesh_dims > 2)
	  {
	  if (ortho_width > 1)
	    mfile << "OrthogonalSampleWidth " << ortho_width << endl;
	  }

	mfile << "EmptyValue " << empty_value << endl;

	mfile << "ImageFile " << slice_image_name << endl;

	if (save_colormap)
	  mfile << "ColormapImage " << cmap_image_name << endl;

	mfile << "ColormapInfo " << cmap_info_name << endl;

	// Write the colormap information if not saving the colormap image

	if (!save_colormap)
	  {
	  using QuickFlash::Image::ImageMagick::write_colormap_info;

	  write_colormap_info(cmap_image_name, colormap_info_filename, 
			      image_colormap, colormap_length, colormap_width,
			      horizontal_colormap);
	  }
	}

      if (save_colormap)
	{
	using QuickFlash::Image::ImageMagick::write_colormap;

	if (save_colormap_info)
	  error_state = write_colormap(colormap_image_filename, 
				       colormap_info_filename, image_colormap,
				       colormap_length, colormap_width,
				       horizontal_colormap);
	else
	  error_state = write_colormap(colormap_image_filename, image_colormap,
				       colormap_length, colormap_width,
				       horizontal_colormap);

	if (error_state < 0)
	  {
	  cerr << endl << "Error: Unable to write colormap " 
	       << colormap_filename << endl << endl;
	  return -2;
	  }
	}
      }
#endif  // USE_MAGICK

    // Print the symmetry values -- WARNING: UGLY CODE AHEAD!!!

    if (compute_symmetry)
      {
      std::list<std::string> working_comments;

      QuickFlash::Slice::SliceGrid sum_grid;

      sum_grid.set_symmetry_sum(symmetry_axis, grid);

      QuickFlash::Slice::SliceGrid diff_grid;

      diff_grid.set_symmetry_diff(symmetry_axis, grid);

      if (output_data)
	{
	const std::string output_filename = output_file_prefix + "_sum.csv";

	working_comments = comment_list;

	strstream << "Symmetry summation data for axis [ " << symmetry_axis 
		  << " ]";

	working_comments.push_back(strstream.str());

	strstream.str("");

	QuickFlash::Slice::write_grid(output_filename, sum_grid, 
				    working_comments);
	}

#ifdef USE_MAGICK
      if (output_image)
	{
	const std::string image_filename 
	  = output_file_prefix + "_sum." + image_extension;

	int error_state = -1;  // Error state

	error_state = write_image(image_filename, sum_grid, transpose_image,
				  show_colormap);

	if (error_state < 0)
	  {
	  cerr << endl << "Error: Unable to write image " << image_filename
	       << endl << endl;
	  return -2;
	  }
	}
#endif  // USE_MAGICK

      if (output_data)
	{
	const std::string output_filename = output_file_prefix + "_diff.csv";

	working_comments = comment_list;

	strstream << "Symmetry difference data for axis [ " << symmetry_axis 
		  << " ]";

	working_comments.push_back(strstream.str());

	strstream.str("");

	QuickFlash::Slice::write_grid(output_filename, diff_grid, 
				    working_comments);
	}

#ifdef USE_MAGICK
      if (output_image)
	{
	const std::string image_filename 
	  = output_file_prefix + "_diff." + image_extension;

	int error_state = -1;  // Error state

	error_state = write_image(image_filename, diff_grid, transpose_image,
				  show_colormap, false, true, 0.0);

	if (error_state < 0)
	  {
	  cerr << endl << "Error: Unable to write image " << image_filename
	       << endl << endl;
	  return -2;
	  }
	}
#endif  // USE_MAGICK
      }

    // Next file ...

    ++flash_filename_iter;
    }

  return 0;
  }
