// C++ program file symmetry_check.cpp

/*
  By Nathan C. Hearn
     July 29, 2007

  Program for determining the level of symmetry for a Flash data set.
*/


#include <stdlib.h>
#include <math.h>
#include <string>
#include <vector>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_counters.hpp"
#include "quickflash_except.hpp"


const unsigned int buffer_size = 2000;
const unsigned int cache_size = 2000;


void check_symmetry(const QuickFlash::File::DataFile & data_file, 
		    const std::string & dataset_name, 
		    const std::vector<unsigned int> & symmetry_axes,
		    const std::vector<double> & symmetry_center)
  {
  using std::cout;
  using std::endl;

  const unsigned int num_axes = symmetry_axes.size();

  const QuickFlash::File::MeshInfo & meshinfo = data_file.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (symmetry_center.size() != dims)
    throw QuickFlash::Except("Symmetry center has incompatible length");

  const unsigned int num_blocks = meshinfo.get_num_blocks();

  // Get the dataset

  const QuickFlash::File::Dataset & dataset 
    = data_file.get_dataset(dataset_name);

  dataset.set_cache_size(cache_size);

  // Reserve some memory

  QuickFlash::Block::BlockData<double> primary_data;
  QuickFlash::Block::BlockData<double> secondary_data;

  bool secondary_loaded = false;
  unsigned int current_secondary_block_index = 0;

  std::vector<double> primary_position;
  std::vector<double> primary_center_delta;
  std::vector<double> secondary_position;

  // Set up the counter for running through the symmetry axes

  std::vector<unsigned int> counter_limits(num_axes, 2);  // Check reflections

  QuickFlash::VectorCounter counter(counter_limits);

  // Run through the blocks

  unsigned int num_comparisons = 0;

  unsigned int num_out_of_bounds = 0;

  double accum_mod_error = 0.0;

  double max_mod_error = 0.0;

  unsigned int max_mod_error_primary_block = 0;
  unsigned int max_mod_error_primary_cell = 0;
  std::vector<double> max_mod_error_primary_pos;
  double max_mod_error_primary_value = 0.0;

  unsigned int max_mod_error_secondary_block = 0;
  unsigned int max_mod_error_secondary_cell = 0;
  std::vector<double> max_mod_error_secondary_pos;
  double max_mod_error_secondary_value = 0.0;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    if (meshinfo.is_leaf(block_index))
      {
      dataset.get_block_data(block_index, primary_data);

      const QuickFlash::Block::BlockInfo & primary_blockinfo 
	= meshinfo.get_block_info(block_index);

      const unsigned int num_cells = primary_blockinfo.get_num_cells();

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	primary_blockinfo.get_cell_center(cell_index, primary_position);

	// Get the distance from the center

	primary_center_delta = primary_position;

	for (unsigned int i = 0; i < dims; i++)
	  primary_center_delta[i] -= symmetry_center[i];

	const double primary_cell_data = primary_data[cell_index];

	// Run through the aliases for the secondary (skip the first one)

	counter.reset_counter();
	counter.increment();

	while (counter.in_bounds())
	  {
	  // Get the reflected position

	  secondary_position = primary_position;

	  for (unsigned int m = 0; m < num_axes; m++)
	    if (counter[m] > 0)
	      {
	      const unsigned int axis = symmetry_axes[m];

	      secondary_position[axis] 
		= symmetry_center[axis] - primary_center_delta[axis];
	      }

	  unsigned int secondary_block_index = 0;
	  unsigned int secondary_cell_index = 0;

	  if (meshinfo.in_bounds(secondary_position))
	    {
	    meshinfo.get_cell_index(secondary_position, secondary_block_index,
				   secondary_cell_index);

	    if ((!secondary_loaded) 
		|| (current_secondary_block_index != secondary_block_index))
	      {
	      dataset.get_block_data(secondary_block_index, secondary_data);

	      secondary_loaded = true;
	      current_secondary_block_index = secondary_block_index;
	      }

	    const double secondary_cell_data 
	      = secondary_data[secondary_cell_index];

	    const double mod_error 
	      = fabs(secondary_cell_data - primary_cell_data);

	    accum_mod_error += mod_error;
	    num_comparisons++;

	    if (mod_error > max_mod_error)
	      {
	      max_mod_error = mod_error;

	      max_mod_error_primary_block = block_index;
	      max_mod_error_primary_cell = cell_index;

	      max_mod_error_primary_pos = primary_position;
	      max_mod_error_primary_value = primary_cell_data;

	      max_mod_error_secondary_block = secondary_block_index;
	      max_mod_error_secondary_cell = secondary_cell_index;

	      max_mod_error_secondary_pos = secondary_position;
	      max_mod_error_secondary_value = secondary_cell_data;
	      }
	    }
	  else
	    num_out_of_bounds++;

	  counter.increment();
	  }
	}
      }

  // Report on findings

  double mean_error = 0.0;

  if (num_comparisons > 0)
    mean_error = accum_mod_error / static_cast<double>(num_comparisons);

  cout << endl << "Dataset [ " << dataset_name << " ] asymmetry report" 
       << endl << endl;

  cout << "    Comparisons [ " << num_comparisons << " ]" << endl;
  cout << "    Out of bounds [ " << num_out_of_bounds << " ]" << endl;

  cout << endl;

  cout << "    Mean error [ " << mean_error << " ]" << endl << endl;

  if (mean_error > 0.0)
    {
    cout << "    Maximum error [ " << max_mod_error << " ]" << endl << endl;

    cout << "    Primary value [ " << max_mod_error_primary_value << " ]" 
	 << endl;

    cout << "        Primary position [";

    for (unsigned int i = 0; i < dims; i++)
      cout << " " << max_mod_error_primary_pos[i];

    cout << " ]" << endl;

    cout << "        Primary block [ " << max_mod_error_primary_block 
	 << " ] cell [ " << max_mod_error_primary_cell << " ]" << endl;

    cout << endl;

    cout << "    Secondary value [ " << max_mod_error_secondary_value << " ]" 
	 << endl;

    cout << "        Secondary position [";

    for (unsigned int i = 0; i < dims; i++)
      cout << " " << max_mod_error_secondary_pos[i];

    cout << " ]" << endl;

    cout << "        Secondary block [ " << max_mod_error_secondary_block 
	 << " ] cell [ " << max_mod_error_secondary_cell << " ]" << endl;

    cout << endl;
    }

  // At some point, we will want to add the ability to delete a dataset here!
  }

  

int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  int min_args = 5;

  if (argc < min_args)
    {
    cerr << endl << "Usage: " << argv[0] << " filename dims <symmetry_center> "
	 << "num_axes symmetry_axis1 [ symmetry_axis2 ... ] "
	 << endl
	 << "dataset_name1 [ dataset2 ... ]" << endl 
	 << endl;
    return 1;
    }

  int req_args = min_args;

  int argptr = 1;

  const std::string filename = argv[argptr++];

  const unsigned int dims = static_cast<unsigned int>(atoi(argv[argptr++]));

  req_args += static_cast<int>(dims);

  if (argc < req_args)
    {
    cerr << endl << "Error: Symmetry center must have dims elements" 
	 << endl << endl;
    return -1;
    }

  std::vector<double> symmetry_center(dims);

  for (unsigned int i = 0; i < dims; i++)
    symmetry_center[i] = atof(argv[argptr++]);

  const unsigned int num_symmetry_axes 
    = static_cast<unsigned int>(atoi(argv[argptr++]));

  req_args += num_symmetry_axes;

  if (argc < req_args)
    {
    cerr << endl << "Error: All symmetry axes must be specified" << endl 
	 << endl;
    return -1;
    }

  std::vector<unsigned int> axis_list;
  axis_list.reserve(num_symmetry_axes);

  for (unsigned int m = 0; m < num_symmetry_axes; m++)
    {
    const unsigned int axis = static_cast<unsigned int>(atoi(argv[argptr++]));

    if (axis >= dims)
      {
      cerr << endl << "Error: Axis [ " << axis << " ] out of range" << endl
	   << endl;
      return -1;
      }

    axis_list.push_back(axis);
    }

  const unsigned int num_datasets = static_cast<unsigned int>(argc - argptr);

  std::vector<std::string> dataset_names;
  dataset_names.reserve(num_datasets);

  while (argptr < argc)
    dataset_names.push_back(argv[argptr++]);

  // Open the file -- skip any particles

  const QuickFlash::File::DataFile dfile(filename, buffer_size, true);

  const QuickFlash::File::MeshInfo & meshinfo = dfile.get_mesh_info();

  const unsigned int file_dims = meshinfo.get_dims();

  if (file_dims != dims)
    {
    cerr << endl << "Error: Specified dimensions [ " << dims 
	 << " ] do not match file dimensions [ " << file_dims << " ]" << endl 
	 << endl;
    return -2;
    }

  // Run through the datasets

  // SOMEDAY, MESHINFO WILL CONTAIN ALL OF THE UNKNOWN NAMES ...

  for (unsigned int m = 0; m < num_datasets; m++)
    check_symmetry(dfile, dataset_names[m], axis_list, symmetry_center);
  
  return 0;
  }
