// C++ program file analyze_volume.cpp

/*
  By Nathan C. Hearn
     May 28, 2007

  Analysis of fixed control volume.
*/


#include <iostream>
#include <string>
#include <sstream>
#include <stdlib.h>
#include "quickflash.hpp"
#include "cubic_volume.hpp"
#include "sim_1d.hpp"


const unsigned int buffer_size = 1000;


void build_datafile_name(const std::string & filename_base,
			 const unsigned int index, const unsigned int digits,
			 std::string & filename)
  {
  std::stringstream namestream;

  namestream << filename_base;

  namestream.width(digits);
  namestream.fill('0');

  namestream << index;

  filename = namestream.str();
  }


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  const int min_args = 6;

  if (argc < min_args)
    {
    cerr << "Usage: " << argv[0] << " filename_base start_index end_index" 
	 << endl;
    cerr << "dims <control_volume_center> <control_volume_width>" << endl;
    cerr << "interface_pos_sim_basename" << endl;
    cerr << endl;

    return 1;
    }

  // Read the arguments

  int req_args = min_args;

  int argptr = 1;

  const std::string filename_base = argv[argptr++];

  const unsigned int start_index 
    = static_cast<unsigned int>(atoi(argv[argptr++]));
  
  const unsigned int end_index 
    = static_cast<unsigned int>(atoi(argv[argptr++]));

  const unsigned int dims = static_cast<unsigned int>(atoi(argv[argptr++]));

  if (dims < 1)
    {
    cerr << "Error: Must have at least one spatial dimension" << endl;
    return -1;
    }

  req_args += 2 * dims;

  if (argc < req_args)
    {
    cerr << "Error: Too few arguments" << endl;
    return -1;
    }
  
  std::vector<double> control_volume_center(dims, 0.0);

  for (unsigned int i = 0; i < dims; i++)
    {
    // First element is scaled offset from mixing layer center
    control_volume_center[i] = atof(argv[argptr++]);
    }

  std::vector<double> control_volume_width(dims, 0.0);

  for (unsigned int i = 0; i < dims; i++)
    {
    // First element is length scale (relative to mixing layer depth)
    control_volume_width[i] = atof(argv[argptr++]);
    }

  const double control_volume_x_scale = control_volume_width[0];

  const std::string interface_pos_sim_basename = argv[argptr++];

  // Helpful constants

  std::vector<double> control_volume_halfwidth = control_volume_width;

  for (unsigned int i = 0; i < dims; i++)
    control_volume_halfwidth[i] *= 0.5;

  // Load the 1D simulation

  Sim1D interface_sim(interface_pos_sim_basename);

  // Perform the main loop

  const unsigned int loop_count = (end_index + 1) - start_index;

  std::vector<double> control_volume_minbounds(dims, 0.0);
  std::vector<double> control_volume_maxbounds(dims, 0.0);

  for (unsigned int loop_index = 0; loop_index < loop_count; loop_index++)
    {
    const unsigned int file_index = loop_index + start_index;

    // Load the data file

    std::string filename;

    build_datafile_name(filename_base, file_index, 4, filename);

    QuickFlash::File::DataFile df(filename, buffer_size, true);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double sim_time = meshinfo.get_sim_time();

    const unsigned int dims = meshinfo.get_dims();

    const double min_sim_x = meshinfo.get_volume_minbounds()[0];
    const double max_sim_x = meshinfo.get_volume_maxbounds()[0];

    // Set up the datasets

    const QuickFlash::File::Dataset & foam_set = df.get_dataset("foam");

    const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");

    std::vector<const QuickFlash::File::Dataset *> vel_set(dims);

    if (dims > 0)
      vel_set[0] = &(df.get_dataset("velx"));

    if (dims > 1)
      vel_set[1] = &(df.get_dataset("vely"));

    if (dims > 2)
      vel_set[2] = &(df.get_dataset("velz"));

    // Determine the depth of the mixing layer within the sample strip

    control_volume_minbounds[0] = min_sim_x;
    control_volume_maxbounds[0] = max_sim_x;

    double layer_min_x = max_sim_x;
    double layer_max_x = min_sim_x;

    QuickFlash::Tree::ConstTreeIter tree_iter = meshinfo.get_tree_iter();

    QuickFlash::Block::BlockData<double> foam;

    std::vector<double> cell_min_coords(dims);
    std::vector<double> cell_max_coords(dims);

    std::vector<double> cell_overlap_pos(dims);

    while (tree_iter.valid_node())
      {
      const QuickFlash::Block::BlockInfo & blockinfo 
	= tree_iter.get_block_info();

      // Check for overlap

      const std::vector<double> & block_min = blockinfo.get_min_coords();
      const std::vector<double> & block_max = blockinfo.get_max_coords();
      
      if (!(volumes_overlap(control_volume_minbounds, control_volume_maxbounds,
			    block_min, block_max)))
	tree_iter.next_branch();
      else
	{
	if (tree_iter.is_leaf())
	  {
	  // Read the data

	  const unsigned int block_index = tree_iter.get_block_index();

	  foam_set.get_block_data(block_index, foam);

	  // Parse the cells

	  const unsigned int num_cells = blockinfo.get_num_cells();

	  // const double cell_volume = blockinfo.get_cell_volume();

	  for (unsigned int cell_index = 0; cell_index < num_cells; 
	       cell_index++)
	    {
	    blockinfo.get_cell_bounds(cell_index, cell_min_coords, 
				      cell_max_coords);

	    double overlap_volume = 0.0;
	    double control_fract = 0.0;
	    double cell_fract = 0.0;

	    if (volumes_overlap(control_volume_minbounds, 
				control_volume_maxbounds, cell_min_coords, 
				cell_max_coords, overlap_volume, 
				cell_overlap_pos, control_fract, cell_fract))
	      {
	      const double cell_foam = foam[cell_index];

	      const double cell_x 
		= 0.5 * (cell_max_coords[0] + cell_min_coords[0]);

	      if (cell_foam < 0.5)
		{
		if (cell_x > layer_max_x)
		  layer_max_x = cell_x;
		}
	      else if (cell_x < layer_min_x)
		layer_min_x = cell_x;
	      }
	    }
	  }

	tree_iter.next();
	}
      }

    double interface_pos = 0.0;
    double interface_vel = 0.0;

    interface_sim.get_interface_pos(sim_time, interface_pos, interface_vel);

    const double layer_center_x = 0.5 * (layer_min_x + layer_max_x);

    const double layer_width 
      = (layer_max_x - layer_min_x) * control_volume_x_scale;

    const double volume_center_offset = layer_width * control_volume_center[0];

    const double volume_center_x = layer_center_x + volume_center_offset;

    const double layer_half_width = 0.5 * layer_width;

    const double scale_layer_min_x = volume_center_x - layer_half_width;
    const double scale_layer_max_x = volume_center_x + layer_half_width;

    control_volume_minbounds[0] = scale_layer_min_x;
    control_volume_maxbounds[0] = scale_layer_max_x;

    for (unsigned int i = 1; i < dims; i++)
      {
      control_volume_minbounds[i] 
	= control_volume_center[i] - control_volume_halfwidth[i];

      control_volume_maxbounds[i] 
	= control_volume_center[i] + control_volume_halfwidth[i];
      }

    // Parse the blocks

    double accum_mass = 0.0;
    std::vector<double> accum_mass_pos(dims, 0.0);
    std::vector<double> accum_mass_vel(dims, 0.0);

    tree_iter = meshinfo.get_tree_iter();

    std::vector<double> cell_vel(dims);

    QuickFlash::Block::BlockData<double> dens;
    
    std::vector< QuickFlash::Block::BlockData<double> > vel(dims);

    while (tree_iter.valid_node())
      {
      const QuickFlash::Block::BlockInfo & blockinfo 
	= tree_iter.get_block_info();

      // Check for overlap

      const std::vector<double> & block_min = blockinfo.get_min_coords();
      const std::vector<double> & block_max = blockinfo.get_max_coords();
      
      if (!(volumes_overlap(control_volume_minbounds, control_volume_maxbounds,
			    block_min, block_max)))
	tree_iter.next_branch();
      else
	{
	if (tree_iter.is_leaf())
	  {
	  // Read the data

	  const unsigned int block_index = tree_iter.get_block_index();

	  dens_set.get_block_data(block_index, dens);

	  for (unsigned int i = 0; i < dims; i++)
	    vel_set[i]->get_block_data(block_index, vel[i]);

	  // Parse the cells

	  const unsigned int num_cells = blockinfo.get_num_cells();

	  // const double cell_volume = blockinfo.get_cell_volume();

	  for (unsigned int cell_index = 0; cell_index < num_cells; 
	       cell_index++)
	    {
	    blockinfo.get_cell_bounds(cell_index, cell_min_coords, 
				      cell_max_coords);

	    double overlap_volume = 0.0;
	    double control_fract = 0.0;
	    double cell_fract = 0.0;

	    if (volumes_overlap(control_volume_minbounds, 
				control_volume_maxbounds, cell_min_coords, 
				cell_max_coords, overlap_volume, 
				cell_overlap_pos, control_fract, cell_fract))
	      {
	      const double cell_dens = dens[cell_index];

	      for (unsigned int i = 0; i < dims; i++)
		cell_vel[i] = vel[i][cell_index];

	      const double overlap_mass = cell_dens * overlap_volume;

	      accum_mass += overlap_mass;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_pos[i] += cell_overlap_pos[i] * overlap_mass;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_vel[i] += cell_vel[i] * overlap_mass;
	      }
	    }
	  }
	tree_iter.next();
	}
      }

    std::vector<double> cm(dims, 0.0);

    if (accum_mass > 0.0)
      {
      const double oneOver_mass = 1.0 / accum_mass;

      for (unsigned int i = 0; i < dims; i++)
	cm[i] = accum_mass_pos[i] * oneOver_mass;
      }

    // Write out header on first pass

    if (loop_index == 0)
      {
      cout << "# analyze_volume_layer output" << endl;
      cout << "#" << endl;

      cout << "# Command:";

      for (int idx = 0; idx < argc; idx++)
	cout << " " << argv[idx];

      cout << "#" << endl;

      cout << "#" << endl;
      cout << "# Filename base [ " << filename_base << " ]" << endl;
      cout << "# Start index [ " << start_index << " ]" << endl;
      cout << "# End index [ " << end_index << " ]" << endl;
      cout << "# Dims [ " << dims << " ]" << endl;

      cout << "# Offset scale [ " << control_volume_center[0] << " ]" << endl;

      cout << "# Control volume center [ xxx";

      for (unsigned int i = 1; i < dims; i++)
	cout << " " << control_volume_center[i];

      cout << " ]" << endl;

      cout << "# Control volume horizontal scale [ " << control_volume_width[0]
	   << " ]" << endl;

      cout << "# Control volume width [ xxx";

      for (unsigned int i = 1; i < dims; i++)
	cout << " " << control_volume_width[i];

      cout << " ]" << endl;

      cout << "# Interface position simulation filename base [ "
	   << interface_pos_sim_basename << " ]" << endl;

      cout << endl;

      cout << "File_Index" << "\t" << "Time" << "\t" << "Interface_pos" << "\t"
	   << "Interface_vel" << "\t" << "Mass";

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "cm_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "p_" << i;

      cout << "\t" << "Volume_min_x" << "\t" << "Volume_max_x" << "\t"
	   << "Layer_center_x" << "\t" << "Volume_center_x" << "\t" 
	   << "Volume_width_x";

      cout << endl;
      }

    cout << file_index << "\t" << sim_time << "\t" << interface_pos << "\t"
	 << interface_vel << "\t" << accum_mass;

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << cm[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << accum_mass_vel[i];

    cout << "\t" << scale_layer_min_x << "\t" << scale_layer_max_x << "\t"
	 << layer_center_x << "\t" << volume_center_x << "\t" << layer_width;

    cout << endl;
    }

  return 0;
  }
