#!/usr/bin/env python

# build_frames.py

# By Nathan C. Hearn
#    November 8, 2007


try :
    import psyco
except ImportError :
    pass


import random
import math


file_index_label = "File_Index"
time_label = "Time"
mass_label = "Mass"
sample_min_x_label = "Sample_Min_y"  # Really y in simulation (rotated figure)
sample_max_x_label = "Sample_Max_y"  # ...

image_file_suffix = ".png"
output_file_suffix = ".svg"

image_min_x = -3.0  # Really min_y in simulation (rotated figure)
image_max_x =  3.0  # ...
image_min_y = -0.5  # Really min_x in simulation (rotated figure)
image_max_y =  0.5  # ...


two_pi = 2.0 * math.pi

random.seed(34)


def get_random_color(intensity_scale=1.0, random_intensity=False) :

    ## Pick a uniform point on a sphere
    ## ... see http://mathworld.wolfram.com/SpherePointPicking.html

    u = -1.0 + (2.0 * random.random())
    phi = two_pi * random.random()

    intensity = intensity_scale

    if (random_intensity) :
        intensity *= random.random()

    w = math.sqrt(1.0 - (u * u)) * intensity

    red = w * math.cos(phi)
    green = w * math.sin(phi)
    blue = u * intensity

    return red, green, blue


def build_filename(prefix, index, suffix=None, digits=4) :

    index_string = ("%(idx)0" + str(digits) + "d") % { "idx" : index }

    ret_string = prefix + index_string

    if (suffix is not None) :
        ret_string += suffix

    return ret_string


def make_specific(time_list, data_list, mass_list) :

    num_elems = len(time_list)

    if (len(data_list) != num_elems) :
        raise RuntimeError("Incompatible list lengths")

    if (len(mass_list) != num_elems) :
        raise RuntimeError("Incompatible list lengths")

    spec_data_time_list = list()
    spec_data_list = list()

    for index in range(num_elems) :

        time = time_list[index]

        value = data_list[index]

        mass = mass_list[index]

        if (mass > 0.0) :
            
            spec_value = value / mass

            spec_data_time_list.append(time)
            spec_data_list.append(spec_value)

    return spec_data_time_list, spec_data_list


def get_bounds(time_list, data_list, min_time=None, max_time=None) :

    num_data = len(time_list)

    if (num_data < 1) :
        raise RuntimeError("No data present")

    if (len(data_list) != num_data) :
        raise RuntimeError("Incompatible value data")

    found_min_time = None
    found_max_time = None

    found_min_data = None
    found_max_data = None

    if ((min_time is not None) and (max_time is not None)) :
        if (min_time > max_time) :
            raise RuntimeError("Minimum time can not exceed maximum time")

    if ((min_time is None) and (max_time is None)) :

        found_min_time = time_list[0]
        found_max_time = time_list[num_data - 1]

        found_min_data = min(data_list)
        found_max_data = max(data_list)

    else :

        start_found = False
        end_found = False

        prev_time = time_list[0]

        if (min_time is None) :
            start_found = True
            found_min_time = time_list[0]

        index = 0

        while ((index < num_data) and (not end_found)) :

            current_time = time_list[index]

            if (not start_found) :
                if (not (current_time < min_time)) :
                    start_found = True
                    found_min_time = current_time

            if (start_found) :

                if (max_time is not None) :
                    if (current_time > max_time) :
                        end_found = True
                        found_max_time = prev_time

                if (not end_found) :
                    
                    current_data = data_list[index]

                    if (found_min_data is None) :
                        found_min_data = current_data
                    elif (current_data < found_min_data) :
                        found_min_data = current_data

                    if (found_max_data is None) :
                        found_max_data = current_data
                    elif (current_data > found_max_data) :
                        found_max_data = current_data

            prev_time = current_time

            index += 1

        if (not end_found) :
            found_max_time = prev_time

    return found_min_time, found_max_time, found_min_data, found_max_data


def main() :

    from optparse import OptionParser
    from sys import exit, stdout, stderr
    import math
    import scicompositor as scicomp
    
    from read_table import read_table


    ## Set up the command line parser

    usage = str("%prog [options] image_file_prefix output_file_prefix "
                + "dataset_name sim_desc_file1 [ sim_desc_file2 ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    min_time_str = str("min-time")
    argparser.add_option("--" + min_time_str, type="float",
                         dest="min_time", help="Minimum time for plot")
    argparser.set_defaults(min_time=None)

    max_time_str = str("max-time")
    argparser.add_option("--" + max_time_str, type="float",
                         dest="max_time", help="Maximum time for plot")
    argparser.set_defaults(max_time=None)

    frame_title_str = str("frame-title")
    argparser.add_option("--" + frame_title_str, type="string",
                         dest="frame_title", help="Title for frame images")
    argparser.set_defaults(frame_title=None)

    use_specific_str = str("use-specific")
    argparser.add_option("--" + use_specific_str, action="store_true",
                         dest="use_specific",
                         help="Divide quantities by mass within sample")
    argparser.set_defaults(use_specific=False)


    ## Process the command line input

    opts, args = argparser.parse_args()

    min_time = opts.min_time
    max_time = opts.max_time

    frame_title = opts.frame_title

    use_specific = opts.use_specific

    argc = len(args)

    min_args = int(4)

    if (argc < min_args) :
        argparser.print_help(file=stderr)
        exit(1)

    argptr = int(0)

    image_file_prefix = str(args[argptr])
    argptr += 1

    output_file_prefix = str(args[argptr])
    argptr += 1

    dataset_name = str(args[argptr])
    argptr += 1

    sim_desc_file_list = list()

    while (argptr < argc) :
        
        sim_desc_file_list.append(str(args[argptr]))
        argptr += 1

    check_min_time = False
    check_max_time = False

    if (min_time is not None) :
        check_min_time = True

    if (max_time is not None) :
        check_max_time = True

    if (check_min_time and check_max_time) :
        if (max_time < min_time) :
            stderr.write("\nError: Maximum time must not be less than minimum"
                         + "\n\n")
            return -1

    ## Collect the data

    image_indexes = None

    time_list = None

    plot_time_list = list()
    plot_data_list = list()

    sample_min_x_data_list = list()
    sample_max_x_data_list = list()

    plot_min_time = None
    plot_max_time = None
    plot_min_data = None
    plot_max_data = None

    plot_color_red_list = list()
    plot_color_green_list = list()
    plot_color_blue_list = list()


    for filename in sim_desc_file_list :
        
        desc_data = read_table(filename)

        ## Collect the data to be plotted

        time_data = desc_data[time_label]

        plot_times = time_data
        plot_data = desc_data[dataset_name]

        if (use_specific) :

            mass_data = desc_data[mass_label]
            
            plot_times, plot_data = make_specific(time_data, plot_data,
                                                  mass_data)

        file_indexes = desc_data[file_index_label]

        ## Use the indexes in the first file for images

        if (image_indexes is None) :
            
            image_indexes = desc_data[file_index_label]
            
        elif (image_indexes != file_indexes) :

            stderr.write("\nError: Incompatible description files\n\n")
            exit(-2)

        ## Make sure the times are consistent

        if (time_list is None) :
            
            time_list = time_data
            
        elif (time_list != time_data) :
            
            stderr.write("\nError: Incompatible description files\n\n")
            exit(-2)

        ## Parse the plot data

        plot_time_list.append(plot_times)
        plot_data_list.append(plot_data)

        set_min_time, set_max_time, set_min_data, set_max_data \
            = get_bounds(plot_times, plot_data, min_time, max_time)

        if (plot_min_time is None) :

            plot_min_time = set_min_time
            plot_max_time = set_max_time

            plot_min_data = set_min_data
            plot_max_data = set_max_data

        else :

            if (set_min_data < plot_min_data) :
                plot_min_data = set_min_data

            if (set_max_data > plot_max_data) :
                plot_max_data = set_max_data

        sample_min_x_data = desc_data[sample_min_x_label]
        sample_max_x_data = desc_data[sample_max_x_label]

        sample_min_x_data_list.append(sample_min_x_data)
        sample_max_x_data_list.append(sample_max_x_data)

        ## Get some colors

        red, green, blue = get_random_color()

        plot_color_red_list.append(red)
        plot_color_green_list.append(green)
        plot_color_blue_list.append(blue)


    if (min_time is None) :
        min_time = plot_min_time

    if (max_time is None) :
        max_time = plot_max_time


    num_times = len(time_list)

    num_datasets = len(plot_data_list)

    for time_index in range(num_times) :

        current_file_index = image_indexes[time_index]

        current_time = time_list[time_index]

        process_index = True

        if (check_min_time and (current_time < min_time)) :
            process_index = False

        if (check_max_time and (current_time > max_time)) :
            process_index = False

        if (process_index) :

            image_filename = build_filename(image_file_prefix,
                                            current_file_index,
                                            image_file_suffix)

            output_filename = build_filename(output_file_prefix,
                                             current_file_index,
                                             output_file_suffix)

            # Create the canvas

            canvas = scicomp.Canvas(0.0, 800.0, 0.0, 600.0)

            # Create the image viewport

            imageport = canvas.get_new_viewport("imageport", 50.0, 750.0,
                                                375.0,
                                                575.0, 0.0, 700.0, 0.0, 200.0)

            imageport.set_physical_coord_map(image_min_x, image_max_x,
                                             image_min_y, image_max_y)

            # Create the image

            img = scicomp.ImageObject("image", image_filename,
                                      min_x=image_min_x, max_x=image_max_x,
                                      min_y=image_min_y, max_y=image_max_y,
                                      use_physical_coords=True)

            imageport.add_child_object(img)


            # Create the sample boxes

            for dataset_index in range(num_datasets) :

                red = plot_color_red_list[dataset_index]
                green = plot_color_green_list[dataset_index]
                blue = plot_color_blue_list[dataset_index]

                sample_min_x \
                    = sample_min_x_data_list[dataset_index][time_index]

                sample_max_x \
                    = sample_max_x_data_list[dataset_index][time_index]

                if (sample_min_x < image_min_x) :
                    sample_min_x = image_min_x

                if (sample_max_x > image_max_x) :
                    sample_max_x = image_max_x

                if (sample_min_x < sample_max_x) :
                    imgbox = scicomp.BoxObject("imagebox", sample_min_x,
                                               sample_max_x,
                                               image_min_y, image_max_y,
                                               use_physical_coords=True)

                    imgbox.set_fill_color(red=red, green=green, blue=blue,
                                          opacity=0.5)
                    imgbox.set_line_opacity(0.0)
                    imgbox.set_line_width(0.0)

                    imageport.add_child_object(imgbox)


            # Create the image axes

            imgaxes = scicomp.AxesObject("imageaxes",
                                         image_min_x, image_max_x,
                                         image_min_y, image_max_y,
                                         use_physical_coords=True)

            imgaxes.set_x_major_ticks_regular(max_ticks=25)
            imgaxes.set_y_major_ticks_regular(max_ticks=7)

            imgaxes.set_frame_color(red=1.0, green=0.0, blue=0.0)
            imgaxes.set_major_ticks_color(red=1.0, green=1.0, blue=0.0,
                                          opacity=0.5)

            imgaxes.set_frame_line_width(1.0)
            imgaxes.set_major_ticks_line_width(0.8)
            imgaxes.set_axis_label_font_size(0.035)

            imageport.add_child_object(imgaxes)

            # Create the plot viewport

            plotport = canvas.get_new_viewport("plotport", 50.0, 750.0, 25.0,
                                               300.0, 0.0, 700.0, 0.0, 275.0)

            plotport.set_physical_coord_map(min_time, max_time, plot_min_data,
                                            plot_max_data)

            # Create the plot lines

            for dataset_index in range(num_datasets) :

                red = plot_color_red_list[dataset_index]
                green = plot_color_green_list[dataset_index]
                blue = plot_color_blue_list[dataset_index]

                plot_times = plot_time_list[dataset_index]
                plot_data = plot_data_list[dataset_index]

                plotline = scicomp.PolylineObject("plotline", plot_times,
                                                  plot_data,
                                                  use_physical_coords=True)
                
                plotline.set_clip_area(min_time, max_time,
                                       plot_min_data, plot_max_data)

                plotline.set_line_color(red=red, green=green, blue=blue)
                plotline.set_line_opacity(0.8)
                plotline.set_line_width(1.0)

                plotport.add_child_object(plotline)

            # Create line denoting current time

            timemarker = scicomp.LineObject("timemarker", current_time,
                                            plot_min_data, current_time,
                                            plot_max_data,
                                            use_physical_coords=True)

            timemarker.set_line_color(red=0.0, green=0.5, blue=0.0,
                                      opacity=0.5)
            timemarker.set_line_width(2.0)

            plotport.add_child_object(timemarker)

            # Create the plot axes

            plotaxes = scicomp.AxesObject("plotaxes", min_time, max_time,
                                          plot_min_data, plot_max_data,
                                          use_physical_coords=True)

            plotaxes.set_x_major_ticks_regular(max_ticks=10)
            plotaxes.set_y_major_ticks_regular(max_ticks=5)

            plotaxes.set_frame_color(red=0.0, green=0.0, blue=0.0, opacity=1.0)
            plotaxes.set_major_ticks_color(red=0.0, green=0.0, blue=1.0,
                                           opacity=0.75)

            plotaxes.set_frame_line_width(0.02)
            plotaxes.set_major_ticks_line_width(0.02)

            plot_data_delta_y = plot_max_data - plot_min_data

            plotaxes.set_axis_label_font_size(math.fabs(plot_data_delta_y
                                                        * 0.05))

            plotport.add_child_object(plotaxes)


            svg_code = canvas.get_svg_code(800)

            f = file(output_filename, "w")

            for line in svg_code :
                f.write(line + "\n")

            f.write("\n")


if (__name__ == "__main__") :
    main()
