// C++ program file layer_finder.cpp

/*
  By Nathan C. Hearn
     January 30, 2008

  Utilities for determining layer boundaries for RT setup (position along axis)
*/


#include "layer_finder.hpp"
#include <string>
#include <cmath>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_except.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_stats_bins.hpp"
#include "mixing_layer.hpp"


const char dens_name[] = "dens";


int get_maximal_envelope_range(const QuickFlash::File::DataFile & df,
			       const char dataset_name[],
			       const unsigned int axis,
			       const double min_data_value,
			       const double max_data_value,
			       double & min_pos, double & max_pos)
  {
  const std::string dataset_name_str = dataset_name;

  return get_maximal_envelope_range(df, dataset_name_str, axis, min_data_value,
				    max_data_value, min_pos, max_pos);
  }


int get_maximal_envelope_range(const QuickFlash::File::DataFile & df,
			       const std::string & dataset_name,
			       const unsigned int axis,
			       const double min_data_value,
			       const double max_data_value,
			       double & min_pos, double & max_pos)
  {
  int ret_val = -1;  // Error state

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (axis >= dims)
    throw QuickFlash::Except("Axis index out of bounds", __FILE__, __LINE__);

  const QuickFlash::File::Dataset & dataset = df.get_dataset(dataset_name);

  QuickFlash::Block::BlockData<double> block_data;

  std::vector<double> cell_pos;

  bool found_interior = false;

  double envelope_min_pos = 0.0;
  double envelope_max_pos = 0.0;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    const double block_min_pos = block_info.get_min_coords()[axis];
    const double block_max_pos = block_info.get_max_coords()[axis];

    bool skip_branch = false;

    if (found_interior)
      {
      if ((block_min_pos > envelope_min_pos) 
	  && (block_max_pos < envelope_max_pos))
	skip_branch = true;
      }

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      dataset.get_block_data(block_index, block_data);

      const unsigned int num_cells = block_info.get_num_cells();

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	const double cell_value = block_data[cell_index];

	if ((cell_value > min_data_value) && (cell_value < max_data_value))
	  {
	  block_info.get_cell_center(cell_index, cell_pos);

	  const double axis_cell_pos = cell_pos[axis];

	  if (found_interior)
	    {
	    if (axis_cell_pos < envelope_min_pos)
	      envelope_min_pos = axis_cell_pos;
	    else if (axis_cell_pos > envelope_max_pos)
	      envelope_max_pos = axis_cell_pos;
	    }
	  else
	    {
	    envelope_min_pos = axis_cell_pos;
	    envelope_max_pos = axis_cell_pos;

	    found_interior = true;
	    }
	  }
	}
      }

    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  if (found_interior)
    {
    min_pos = envelope_min_pos;
    max_pos = envelope_max_pos;

    ret_val = 0;  // OK
    }

  return ret_val;
  }


void get_mixing_layer_meanValue(const QuickFlash::File::DataFile & df,
				const char dataset_name[],
				const unsigned int axis, 
				const double min_coord, const double max_coord,
				const double bin_spacing, 
				const double min_bounds_data_value, 
				const double max_bounds_data_value,
				MixingLayer & mixing_layer,
				const bool use_abs_value)
  {
  const std::string dataset_name_str = dataset_name;

  get_mixing_layer_meanValue(df, dataset_name_str, axis, min_coord, max_coord, 
			     bin_spacing, min_bounds_data_value, 
			     max_bounds_data_value, mixing_layer, 
			     use_abs_value);
  }


void get_mixing_layer_meanValue(const QuickFlash::File::DataFile & df,
				const std::string & dataset_name,
				const unsigned int axis, 
				const double min_coord, const double max_coord,
				const double bin_spacing, 
				const double min_bounds_data_value, 
				const double max_bounds_data_value,
				MixingLayer & mixing_layer,
				const bool use_abs_value)
  {
  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (axis >= dims)
    throw QuickFlash::Except("Axis index out of bounds", __FILE__, __LINE__);

  const double time = meshinfo.get_sim_time();

  // Determine the bin layout

  const double lower_coord = min_coord;

  const double default_domain_width = max_coord - lower_coord;

  const double fl_num_bins = default_domain_width / bin_spacing;

  const unsigned int num_bins = static_cast<unsigned int>(fl_num_bins) + 1;

  const double domain_width = static_cast<double>(num_bins) * bin_spacing;
  const double upper_coord = lower_coord + domain_width;

  QuickFlash::Stats::BinArray binarray(lower_coord, num_bins, bin_spacing);

  // Load the data sets

  const QuickFlash::File::Dataset & data_set = df.get_dataset(dataset_name);

  // Run through the blocks

  const unsigned int num_blocks = meshinfo.get_num_blocks();

  std::vector<double> block_center_vect;
  std::vector<double> block_width_vect;
  
  QuickFlash::Block::BlockData<double> data_block;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    {
    if (meshinfo.is_leaf(block_index))
      {
      // See if block (x-coordinate) falls within the bin domain

      meshinfo.get_block_center(block_index, block_center_vect);
      meshinfo.get_block_width(block_index, block_width_vect);

      const double block_center = block_center_vect[axis];
      const double block_width = block_width_vect[axis];

      const double block_halfwidth = block_width * 0.5;

      const double block_min = block_center - block_halfwidth;
      const double block_max = block_center + block_halfwidth;

      if ((block_min < upper_coord) && (block_max > lower_coord))
	{
	// Load the block information

	const QuickFlash::Block::BlockInfo & block_info
	  = meshinfo.get_block_info(block_index);

	// Load the block data

	data_set.get_block_data(block_index, data_block);

	const std::vector<double> & cell_width = block_info.get_cell_width();

	const double cell_axis_width = cell_width[axis];
	const double cell_axis_halfwidth = cell_axis_width * 0.5;

	const double cell_volume = block_info.get_cell_volume();

	const unsigned int num_cells = block_info.get_num_cells();

	std::vector<double> cell_center;

	for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	  {
	  // Get the center coordinates

	  block_info.get_cell_center(cell_index, cell_center);

	  const double cell_axis_center = cell_center[axis];

	  const double cell_axis_min = cell_axis_center - cell_axis_halfwidth;
	  const double cell_axis_max = cell_axis_center + cell_axis_halfwidth;

	  if ((cell_axis_min < upper_coord) && (cell_axis_max > lower_coord))
	    {
	    double data_value = data_block[cell_index];

	    if (use_abs_value)
	      data_value = std::fabs(data_value);

	    // Weight by cell volume

	    binarray.add_data_spread(cell_axis_center, data_value, 
				     cell_axis_width, cell_volume);
	    }
	  }
	}
      }
    }

  // Find the bins where the mean data value crosses the threshholds

  const double layer_min_pos 
    = binarray.get_transition_pos(min_bounds_data_value, true);

  const double layer_max_pos 
    = binarray.get_transition_pos(max_bounds_data_value, false);

  // Reset the MixingLayer object

  mixing_layer.reset(time, layer_min_pos, layer_max_pos, 0.0);
  }






void get_mixing_layer_meanValue_peakFract(const QuickFlash::File::DataFile & df,
					  const char dataset_name[],
					  const unsigned int axis, 
					  const double min_coord, 
					  const double max_coord,
					  const double bin_spacing, 
					  const double min_bounds_peak_fract, 
					  const double max_bounds_peak_fract,
					  MixingLayer & mixing_layer,
					  const bool use_abs_value)
  {
  const std::string dataset_name_str = dataset_name;

  get_mixing_layer_meanValue_peakFract(df, dataset_name_str, axis, min_coord, 
				       max_coord, bin_spacing, 
				       min_bounds_peak_fract, 
				       max_bounds_peak_fract, mixing_layer, 
				       use_abs_value);
  }


void get_mixing_layer_meanValue_peakFract(const QuickFlash::File::DataFile & df,
					  const std::string & dataset_name,
					  const unsigned int axis, 
					  const double min_coord, 
					  const double max_coord,
					  const double bin_spacing, 
					  const double min_bounds_peak_fract, 
					  const double max_bounds_peak_fract,
					  MixingLayer & mixing_layer, 
					  const bool use_abs_value)
  {
  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (axis >= dims)
    throw QuickFlash::Except("Axis index out of bounds", __FILE__, __LINE__);

  const double time = meshinfo.get_sim_time();

  // Determine the bin layout

  const double lower_coord = min_coord;

  const double default_domain_width = max_coord - lower_coord;

  const double fl_num_bins = default_domain_width / bin_spacing;

  const unsigned int num_bins = static_cast<unsigned int>(fl_num_bins) + 1;

  const double domain_width = static_cast<double>(num_bins) * bin_spacing;
  const double upper_coord = lower_coord + domain_width;

  QuickFlash::Stats::BinArray binarray(lower_coord, num_bins, bin_spacing);

  // Load the data sets

  const QuickFlash::File::Dataset & data_set = df.get_dataset(dataset_name);

  // Run through the blocks

  const unsigned int num_blocks = meshinfo.get_num_blocks();

  std::vector<double> block_center_vect;
  std::vector<double> block_width_vect;
  
  QuickFlash::Block::BlockData<double> data_block;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    {
    if (meshinfo.is_leaf(block_index))
      {
      // See if block (x-coordinate) falls within the bin domain

      meshinfo.get_block_center(block_index, block_center_vect);
      meshinfo.get_block_width(block_index, block_width_vect);

      const double block_center = block_center_vect[axis];
      const double block_width = block_width_vect[axis];

      const double block_halfwidth = block_width * 0.5;

      const double block_min = block_center - block_halfwidth;
      const double block_max = block_center + block_halfwidth;

      if ((block_min < upper_coord) && (block_max > lower_coord))
	{
	// Load the block information

	const QuickFlash::Block::BlockInfo & block_info
	  = meshinfo.get_block_info(block_index);

	// Load the block data

	data_set.get_block_data(block_index, data_block);

	const std::vector<double> & cell_width = block_info.get_cell_width();

	const double cell_axis_width = cell_width[axis];
	const double cell_axis_halfwidth = cell_axis_width * 0.5;

	const double cell_volume = block_info.get_cell_volume();

	const unsigned int num_cells = block_info.get_num_cells();

	std::vector<double> cell_center;

	for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	  {
	  // Get the center coordinates

	  block_info.get_cell_center(cell_index, cell_center);

	  const double cell_axis_center = cell_center[axis];

	  const double cell_axis_min = cell_axis_center - cell_axis_halfwidth;
	  const double cell_axis_max = cell_axis_center + cell_axis_halfwidth;

	  if ((cell_axis_min < upper_coord) && (cell_axis_max > lower_coord))
	    {
	    double data_value = data_block[cell_index];

	    if (use_abs_value)
	      data_value = std::fabs(data_value);

	    // Weight by cell volume

	    binarray.add_data_spread(cell_axis_center, data_value, 
				     cell_axis_width, cell_volume);
	    }
	  }
	}
      }
    }

  // Get boundary threshold value

  const double peak_value = binarray.get_peak_mean();

  const double min_bounds_data_value = min_bounds_peak_fract * peak_value;
  const double max_bounds_data_value = max_bounds_peak_fract * peak_value;

  // Find the bins where the mean data value crosses the thresholds

  const double layer_min_pos 
    = binarray.get_transition_pos(min_bounds_data_value, true);

  const double layer_max_pos 
    = binarray.get_transition_pos(max_bounds_data_value, false);

  // Reset the MixingLayer object

  mixing_layer.reset(time, layer_min_pos, layer_max_pos, 0.0);
  }


void get_mixing_layer_meanValue_massWeight(const QuickFlash::File::DataFile & df,
					   const char dataset_name[],
					   const unsigned int axis, 
					   const double min_coord, 
					   const double max_coord,
					   const double bin_spacing, 
					   const double min_bounds_data_value, 
					   const double max_bounds_data_value,
					   MixingLayer & mixing_layer,
					   const bool use_abs_value)
  {
  const std::string dataset_name_str = dataset_name;

  get_mixing_layer_meanValue_massWeight(df, dataset_name_str, axis, min_coord,
					max_coord, bin_spacing, 
					min_bounds_data_value,
					max_bounds_data_value, mixing_layer,
					use_abs_value);
  }


void get_mixing_layer_meanValue_massWeight(const QuickFlash::File::DataFile & df,
					   const std::string & dataset_name,
					   const unsigned int axis, 
					   const double min_coord, 
					   const double max_coord,
					   const double bin_spacing, 
					   const double min_bounds_data_value, 
					   const double max_bounds_data_value,
					   MixingLayer & mixing_layer,
					   const bool use_abs_value)
  {
  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (axis >= dims)
    throw QuickFlash::Except("Axis index out of bounds", __FILE__, __LINE__);

  const double time = meshinfo.get_sim_time();

  // Determine the bin layout

  const double lower_coord = min_coord;

  const double default_domain_width = max_coord - lower_coord;

  const double fl_num_bins = default_domain_width / bin_spacing;

  const unsigned int num_bins = static_cast<unsigned int>(fl_num_bins) + 1;

  const double domain_width = static_cast<double>(num_bins) * bin_spacing;
  const double upper_coord = lower_coord + domain_width;

  QuickFlash::Stats::BinArray binarray(lower_coord, num_bins, bin_spacing);

  // Load the data sets

  const QuickFlash::File::Dataset & dens_set = df.get_dataset(dens_name);

  const QuickFlash::File::Dataset & data_set = df.get_dataset(dataset_name);

  // Run through the blocks

  const unsigned int num_blocks = meshinfo.get_num_blocks();
  
  std::vector<double> block_center_vect;
  std::vector<double> block_width_vect;
  
  QuickFlash::Block::BlockData<double> dens_block;
  QuickFlash::Block::BlockData<double> data_block;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    {
    if (meshinfo.is_leaf(block_index))
      {
      // See if block (x-coordinate) falls within the bin domain

      meshinfo.get_block_center(block_index, block_center_vect);
      meshinfo.get_block_width(block_index, block_width_vect);

      const double block_center = block_center_vect[axis];
      const double block_width = block_width_vect[axis];

      const double block_halfwidth = block_width * 0.5;

      const double block_min = block_center - block_halfwidth;
      const double block_max = block_center + block_halfwidth;

      if ((block_min < upper_coord) && (block_max > lower_coord))
	{
	// Load the block information

	const QuickFlash::Block::BlockInfo & block_info
	  = meshinfo.get_block_info(block_index);

	// Load the block data

	dens_set.get_block_data(block_index, dens_block);
	data_set.get_block_data(block_index, data_block);

	const std::vector<double> & cell_width = block_info.get_cell_width();

	const double cell_axis_width = cell_width[axis];
	const double cell_axis_halfwidth = cell_axis_width * 0.5;

	const double cell_volume = block_info.get_cell_volume();

	const unsigned int num_cells = block_info.get_num_cells();

	std::vector<double> cell_center;

	for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	  {
	  // Get the center coordinates

	  block_info.get_cell_center(cell_index, cell_center);

	  const double cell_axis_center = cell_center[axis];

	  const double cell_axis_min = cell_axis_center - cell_axis_halfwidth;
	  const double cell_axis_max = cell_axis_center + cell_axis_halfwidth;

	  if ((cell_axis_min < upper_coord) && (cell_axis_max > lower_coord))
	    {
	    const double dens_value = dens_block[cell_index];

	    double data_value = data_block[cell_index];

	    if (use_abs_value)
	      data_value = std::fabs(data_value);

	    const double cell_mass = cell_volume * dens_value;

	    // Weight by cell mass

 	    binarray.add_data_spread(cell_axis_center, data_value, 
				     cell_axis_width, cell_mass);
	    }
	  }
	}
      }
    }

  // Find the bins where the mean data value crosses the thresholds

  const double layer_min_pos 
    = binarray.get_transition_pos(min_bounds_data_value, true);

  const double layer_max_pos 
    = binarray.get_transition_pos(max_bounds_data_value, false);

  // Reset the MixingLayer object

  mixing_layer.reset(time, layer_min_pos, layer_max_pos, 0.0);
  }
