#!/usr/bin/env python

# anaglyph3d.py

# By Nathan C. Hearn
#    March 25, 2008
#
# Produces 3D anaglyph images from left-right stereo image pairs


## Default color scalings

red_index = int(0)
green_index = int(1)
blue_index = int(2)


rgb_scale_list = dict()

rgb_scale_list["red"] = (1.0, 0.0, 0.0)
rgb_scale_list["green"] = (0.0, 1.0, 0.0)
rgb_scale_list["blue"] = (0.0, 0.0, 1.0)
rgb_scale_list["cyan"] = (0.0, 0.5, 0.5)
rgb_scale_list["custom"] = (0.0, 0.0, 0.0)


## Image operation functions


def get_default_color_scales(image_mode_str) :

    mode_list = image_mode_str.split('-')

    if (len(mode_list) != 2) :
        raise RuntimeError("Left-Right modes required")

    left_mode_str = mode_list[0]
    right_mode_str = mode_list[1]

    if (not rgb_scale_list.has_key(left_mode_str)) :
        raise RuntimeError("Left mode not recognized")

    left_rgb_scale = list(rgb_scale_list[left_mode_str])

    if (not rgb_scale_list.has_key(right_mode_str)) :
        raise RuntimeError("Right mode not recognized")
    
    right_rgb_scale = list(rgb_scale_list[right_mode_str])

    return left_rgb_scale, right_rgb_scale
    

def make_anaglyph(left_image, right_image,
                  left_rgb_scale=rgb_scale_list["red"],
                  right_rgb_scale=rgb_scale_list["cyan"],
                  preserve_color=False, left_offset=(0, 0),
                  right_offset=(0, 0)) :

    from PIL import Image, ImageEnhance, ImageChops

    ## Find the final image size

    left_image_size = left_image.size
    right_image_size = right_image.size

    min_coords = (min(left_offset[0], right_offset[0]),
                  min(left_offset[1], right_offset[1]))

    max_coords = (max((left_image_size[0] + left_offset[0]),
                      right_image_size[0] + right_offset[0]),
                  max((left_image_size[1] + left_offset[1]),
                      right_image_size[1] + right_offset[1]))

    image_size = ((max_coords[0] - min_coords[0]),
                  (max_coords[1] - min_coords[1]))


    ## ERROR CHECKING???

    if (len(left_rgb_scale) != 3) :
        raise RuntimeError("Incompatible left RGB scale vector")

    if (len(right_rgb_scale) != 3) :
        raise RuntimeError("Incompatible right RGB scale vector")

    l_red_scale = left_rgb_scale[0]
    l_green_scale = left_rgb_scale[1]
    l_blue_scale = left_rgb_scale[2]

    r_red_scale = right_rgb_scale[0]
    r_green_scale = right_rgb_scale[1]
    r_blue_scale = right_rgb_scale[2]

    ## Get separate RGB channels for each image

    l_img = None
    r_img = None

    if (preserve_color) :
        
        l_img = left_image.convert("RGB")
        r_img = right_image.convert("RGB")
        
    else :
        
        l_img = left_image.convert("L").convert("RGB")
        r_img = right_image.convert("L").convert("RGB")

    l_red, l_green, l_blue = l_img.split()
    r_red, r_green, r_blue = r_img.split()

    ## Scale the channels

    l_red_scaled = ImageEnhance.Brightness(l_red).enhance(l_red_scale)
    l_green_scaled = ImageEnhance.Brightness(l_green).enhance(l_green_scale)
    l_blue_scaled = ImageEnhance.Brightness(l_blue).enhance(l_blue_scale)

    r_red_scaled = ImageEnhance.Brightness(r_red).enhance(r_red_scale)
    r_green_scaled = ImageEnhance.Brightness(r_green).enhance(r_green_scale)
    r_blue_scaled = ImageEnhance.Brightness(r_blue).enhance(r_blue_scale)

    ## Shift the images for the proper offset (use cropping)

    left_offset = ((left_offset[0] - min_coords[0]),
                   (left_offset[1] - min_coords[1]))

    right_offset = ((right_offset[0] - min_coords[0]),
                   (right_offset[1] - min_coords[1]))

    l_red_shifted = Image.new("L", image_size)
    l_green_shifted = Image.new("L", image_size)
    l_blue_shifted = Image.new("L", image_size)

    r_red_shifted = Image.new("L", image_size)
    r_green_shifted = Image.new("L", image_size)
    r_blue_shifted = Image.new("L", image_size)

    l_red_shifted.paste(l_red_scaled, left_offset)
    l_green_shifted.paste(l_green_scaled, left_offset)
    l_blue_shifted.paste(l_blue_scaled, left_offset)
    
    r_red_shifted.paste(r_red_scaled, right_offset)
    r_green_shifted.paste(r_green_scaled, right_offset)
    r_blue_shifted.paste(r_blue_scaled, right_offset)

    ## Create the merged image RGB channels

    merged_red = ImageChops.add(l_red_shifted, r_red_shifted)
    merged_green = ImageChops.add(l_green_shifted, r_green_shifted)
    merged_blue = ImageChops.add(l_blue_shifted, r_blue_shifted)

    ## Merge the channels

    merged_image = Image.merge("RGB", (merged_red, merged_green, merged_blue))

    return merged_image


## Main

def main() :

    from sys import exit, stderr
    from optparse import OptionParser
    
    from PIL import Image

    ## Set up the command-line parser

    usage = str("%prog [options] output_prefix image1_left image1_right "
                + "\n\t[ image2_left image2_right ] [...]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_index_start_str = str("output-index-start")
    argparser.add_option("--" + output_index_start_str, type="int",
                         dest="output_index_start",
                         help="Starting index for output filenames")
    argparser.set_defaults(output_index_start=None)

    output_index_digits_str = str("output-index-digits")
    argparser.add_option("--" + output_index_digits_str, type="int",
                         dest="output_index_digits",
                         help="Number of digits in output filename index")
    argparser.set_defaults(output_index_digits=4)

    separate_left_right_lists_str = str("separate-left-right-lists")
    argparser.add_option("--" + separate_left_right_lists_str,
                         action="store_true",
                         dest="separate_left_right_lists",
                         help="All left-side files listed before right-side")
    argparser.set_defaults(separate_left_right_lists=False)

    preserve_color_str = str("preserve-color")
    argparser.add_option("--" + preserve_color_str, action="store_true",
                         dest="preserve_color",
                         help="Try to preserve color information")
    argparser.set_defaults(preserve_color=False)

    image_mode_str = str("image-mode")
    image_mode_help = str("Image mode - Specify LEFT-RIGHT using "
                          + "red, green, blue, cyan, or custom "
                          + "(e.g., red-cyan, or custom-green)")
    argparser.add_option("--" + image_mode_str, type="string",
                         dest="image_mode",
                         help=image_mode_help)
    argparser.set_defaults(image_mode="red-cyan")

    left_red_scale_str = str("left-red-scale")
    argparser.add_option("--" + left_red_scale_str, type="float",
                         dest="left_red_scale",
                         help="Custom red channel scale for left image")
    argparser.set_defaults(left_red_scale=None)

    left_green_scale_str = str("left-green-scale")
    argparser.add_option("--" + left_green_scale_str, type="float",
                         dest="left_green_scale",
                         help="Custom green channel scale for left image")
    argparser.set_defaults(left_green_scale=None)

    left_blue_scale_str = str("left-blue-scale")
    argparser.add_option("--" + left_blue_scale_str, type="float",
                         dest="left_blue_scale",
                         help="Custom blue channel scale for left image")
    argparser.set_defaults(left_blue_scale=None)

    right_red_scale_str = str("right-red-scale")
    argparser.add_option("--" + right_red_scale_str, type="float",
                         dest="right_red_scale",
                         help="Custom red channel scale for right image")
    argparser.set_defaults(right_red_scale=None)

    right_green_scale_str = str("right-green-scale")
    argparser.add_option("--" + right_green_scale_str, type="float",
                         dest="right_green_scale",
                         help="Custom green channel scale for right image")
    argparser.set_defaults(right_green_scale=None)

    right_blue_scale_str = str("right-blue-scale")
    argparser.add_option("--" + right_blue_scale_str, type="float",
                         dest="right_blue_scale",
                         help="Custom blue channel scale for right image")
    argparser.set_defaults(right_blue_scale=None)

    left_offset_str = str("left-offset")
    argparser.add_option("--" + left_offset_str, type="string",
                         dest="left_offset", help="Left image offset as x,y")
    argparser.set_defaults(left_offset=None)

    right_offset_str = str("right-offset")
    argparser.add_option("--" + right_offset_str, type="string",
                         dest="right_offset", help="Right image offset as x,y")
    argparser.set_defaults(right_offset=None)

    quiet_str = str("quiet")
    argparser.add_option("--" + quiet_str, action="store_true",
                         dest="quiet_operation",
                         help="Suppress non-error output")
    argparser.set_defaults(quiet_operation=False)


    ## Process the command line input

    opts, args = argparser.parse_args()

    argc = len(args)

    min_args = 3

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n\n")
        exit(1)


    ## Get the optional arguments

    output_index_start = opts.output_index_start

    output_index_digits = opts.output_index_digits

    if (output_index_digits < 1) :
        stderr.write("\nError: Output index digits must be positive\n\n")
        exit(-1)

    separate_left_right_lists = opts.separate_left_right_lists

    preserve_color = opts.preserve_color

    left_rgb_scales, right_rgb_scales \
        = get_default_color_scales(opts.image_mode)

    if (opts.left_red_scale is not None) :
        left_rgb_scales[red_index] = opts.left_red_scale

    if (opts.left_green_scale is not None) :
        left_rgb_scales[green_index] = opts.left_green_scale

    if (opts.left_blue_scale is not None) :
        left_rgb_scales[blue_index] = opts.left_blue_scale

    if (opts.right_red_scale is not None) :
        right_rgb_scales[red_index] = opts.right_red_scale

    if (opts.right_green_scale is not None) :
        right_rgb_scales[green_index] = opts.right_green_scale

    if (opts.right_blue_scale is not None) :
        right_rgb_scales[blue_index] = opts.right_blue_scale

    left_offset = (0, 0)

    if (opts.left_offset is not None) :
        
        words = opts.left_offset.split(",")

        if (len(words) != 2) :
            stderr.write("\nError: Left offset must have two comma-separated "
                         + "values\n\n")
            exit(-1)

        left_offset = (int(words[0]), int(words[1]))

    right_offset = (0, 0)

    if (opts.right_offset is not None) :
        
        words = opts.right_offset.split(",")

        if (len(words) != 2) :
            stderr.write("\nError: Right offset must have two comma-separated "
                         + "values\n\n")
            exit(-1)

        right_offset = (int(words[0]), int(words[1]))

    quiet_operation = opts.quiet_operation


    ## Get the required arguments

    argptr = int(0)

    output_prefix = str(args[argptr])
    argptr += 1

    if (((argc - argptr) % 2) != 0) :
        stderr.write("\nError: Pairs of image files required\n\n")
        exit(-1)

    left_image_file_list = list()
    right_image_file_list = list()

    if (separate_left_right_lists) :

        num_pairs = (argc - argptr) / 2

        start_left_index = argptr
        start_right_index = argptr + num_pairs

        for pr_idx in range(num_pairs) :
            
            left_image_file_list.append(str(args[start_left_index + pr_idx]))
            right_image_file_list.append(str(args[start_right_index + pr_idx]))

        # Set argptr to end of file list

        argptr = start_right_index + num_pairs

    else :
        
        while (argptr < argc) :
        
            left_image_file_list.append(str(args[argptr]))
            argptr += 1

            right_image_file_list.append(str(args[argptr]))
            argptr += 1

    num_image_pairs = len(left_image_file_list)


    # Make sure there is an index if more than one image pair

    if ((output_index_start is None) and (num_image_pairs > 1)) :
        output_index_start = 0


    ## Process the files

    for file_index in range(num_image_pairs) :

        output_filename = output_prefix

        if (output_index_start is not None) :

            output_index = output_index_start + file_index

            output_filename += \
                "_" + (("%(idx)0" + str(output_index_digits) + "d")
                       % { "idx" : output_index })

        output_filename += ".png"


        left_image_filename = left_image_file_list[file_index]
        right_image_filename = right_image_file_list[file_index]

        left_image = Image.open(left_image_filename)
        right_image = Image.open(right_image_filename)

        anaglyph_image = make_anaglyph(left_image, right_image,
                                       left_rgb_scales, right_rgb_scales,
                                       preserve_color, left_offset,
                                       right_offset)

        anaglyph_image.save(output_filename)

        if (not quiet_operation) :
            stderr.write("Wrote image " + output_filename + "\n")

    if (not quiet_operation) :
        stderr.write("\nDone.\n\n")


if (__name__ == "__main__") :
    main()

