# colormap.py

# By Nathan C. Hearn
#    September 2, 2007
#
# Colormap utilities for drawimage


# Constants

color_vector_size = int(4)

R_Index = int(0)
G_Index = int(1)
B_Index = int(2)
Alpha_Index = int(3)


# Class Color

class Color :

    def __init__(self, r=float(0.0), g=float(0.0), b=float(0.0),
                 alpha=float(1.0), levels=int(256)) :

        self.__r = int(0)
        self.__g = int(0)
        self.__b = int(0)

        self.__alpha = int(0)

        if (levels < 2) :
            raise ValueError("Error: At least two color depths required")

        self.__levels = levels

        self.set_color(r, g, b, alpha)


    def reset(self, r=float(0.0), g=float(0.0), b=float(0.0), alpha=float(1.0),
              levels=int(256)) :

        if (levels < 2) :
            raise ValueError("Error: At least two color depths required")

        self.__levels = levels

        self.set_color(r, g, b, alpha)


    def get_r(self) :
        return self.__r

    def get_g(self) :
        return self.__g

    def get_b(self) :
        return self.__b

    def get_alpha(self) :
        return self.__alpha

    def get_rgb(self) :
        return (self.__r, self.__g, self.__b)

    def get_rgba(self) :
        return (self.__r, self.__g, self.__b, self.__alpha)


    def set_color(self, r=float(0.0), g=float(0.0), b=float(0.0),
                  alpha=float(1.0)) :

        max_level = self.__levels - 1

        level_scale = float(max_level)

        self.__r = int(r * level_scale)
        self.__g = int(g * level_scale)
        self.__b = int(b * level_scale)

        self.__alpha = int(alpha * level_scale)

        if (self.__r < 0) :
            self.__r = int(0)
        elif (self.__r > max_level) :
            self.__r = max_level

        if (self.__g < 0) :
            self.__g = int(0)
        elif (self.__g > max_level) :
            self.__g = max_level

        if (self.__b < 0) :
            self.__b = int(0)
        elif (self.__b > max_level) :
            self.__b = max_level

        if (self.__alpha < 0) :
            self.__alpha = int(0)
        elif (self.__alpha > max_level) :
            self.__alpha = max_level


    def set_alpha(self, alpha) :

        max_level = self__levels - 1

        level_scale = float(max_level)

        self.__alpha = int(alpha * level_scale)

        if (self.__alpha < 0) :
            self.__alpha = int(0)
        elif (self.__alpha > max_level) :
            self.__alpha = max_level


# Class ValueColor

class ValueColor(Color) :

    def __init__(self, value=float(0.0), r=float(0.0), g=float(0.0),
                 b=float(0.0), alpha=float(1.0), levels=int(256)) :

        Color.__init__(self, r, g, b, alpha, levels)

        self.__value = value


    def __cmp__(self, other) :

        ret_val = int(0)

        if (self.__value < other.__value) :
            ret_val = int(-1)
        elif (self.__value > other.__value) :
            ret_val = int(1)

        return ret_val


    def reset(self, value=float(0.0), r=float(0.0), g=float(0.0),
              b=float(0.0), alpha=float(1.0), levels=int(256)) :

        self.__value = value

        Color.reset(self, r, g, b, alpha, levels)


    def set_value(self, value) :
        self.__value = value


    def get_value(self) :
        return self.__value


def sort_value_color(obj1, obj2) :

    val1 = obj1.get_value()
    val2 = obj2.get_value()

    return cmp(val1, val2)
    


def read_colormap(filename) :

    import math

    value_color_list = list()
        
    f = file(filename, "r")

    eof = False

    while (not eof) :

        file_line = f.readline()

        if (file_line == "") :
            eof = True
        else :
            line_words = file_line.split("#")[0].strip().split()

            num_words = len(line_words)

            if (num_words > 0) :

                if (num_words == 1) :
                    raise RuntimeError("Error: No color values")
                else :
                    value = float(line_words[0])

                    color_r = float(1.0)
                    color_g = float(1.0)
                    color_b = float(1.0)
                    color_alpha = float(1.0)

                    if (num_words < 4) :

                        # Gray scale

                        gray_value = float(line_words[1])

                        color_r = gray_value
                        color_g = gray_value
                        color_b = gray_value

                        # Alpha

                        if (num_words > 2) :
                            color_alpha = float(line_words[2])

                    elif (num_words < 6) :

                        # Colors

                        color_r = float(line_words[1])
                        color_g = float(line_words[2])
                        color_b = float(line_words[3])

                        if (num_words > 4) :
                            color_alpha = float(line_words[4])

                    else :
                        raise RuntimeError("Error: Unrecongized format")

                valcolor = ValueColor(value, color_r, color_g, color_b,
                                      color_alpha)

                value_color_list.append(valcolor)

    # Sort the list

    value_color_list.sort()

    return value_color_list


# Class Colormap

class Colormap :

    def __init__(self, filename=None, logscale=False, opacity=None) :

        # Data members

        self.__num_map_values = int(0)

        self.__map_values = None
        self.__rgba_values = None

        self.__min_value = float(0.0)
        self.__max_value = float(0.0)

        self.__logscale = False

        self.reset(filename, logscale)
        

    def reset(self, filename=None, logscale=False, opacity=None) :

        self.__num_map_values = int(0)

        self.__map_values = None
        self.__rgba_values = None

        self.__min_value = float(0.0)
        self.__max_value = float(0.0)

        self.__logscale = logscale

        if (filename is not None) :
            self.__read_map(filename, opacity)


    def get_min_value(self) :
        return self.__min_value


    def get_max_value(self) :
        return self.__max_value


    def rescale_colormap(self, min_value, max_value) :

        import math

        if (self.__map_values is None) :
            raise RuntimeError("Colormap not initialized")

        ## Check the entries

        if (not (min_value < max_value)) :
            raise RuntimeError("Min value must be less than max value")

        if (self.__logscale) :
            if (not (min_value > 0.0)) :
                raise RuntimeError("Non-positive values not allowed for log "
                                   + "scale")

        ## Rescale the old values -- IS THIS CORRECT FOR NON-LOG-SCALE TO LOG??

        scale_factor = 0.0

        old_min_value = self.__min_value
        old_max_value = self.__max_value

        if (self.__logscale) :
            old_min_value = math.pow(10.0, old_min_value)
            old_max_value = math.pow(10.0, old_max_value)    

        scale_factor = (max_value - min_value) \
                       / (old_max_value - old_min_value)

        for index in range(self.__num_map_values) :

            old_value = self.__map_values[index]

            if (self.__logscale) :
                old_value = math.pow(10.0, old_value)

            new_value \
                = ((old_value - old_min_value) * scale_factor) + min_value

            if (self.__logscale) :
                new_value = math.log10(new_value)

            self.__map_values[index] = new_value

        self.__min_value = min(self.__map_values)
        self.__max_value = max(self.__map_values)


    def get_color(self, value) :

        from numeric_defs import numeric_zeros, numeric_int

        if (self.__num_map_values < 2) :
            raise RuntimeError("Error: Insufficient colormap entries")

        r = int(0)
        g = int(0)
        b = int(0)
        a = int(0)

        before_index, after_index, pos_delta, interval \
                      = self.__get_map_position(value)
        
        if ((before_index == after_index) or (not (interval > 0.0))) :

            r = self.__rgba_values[before_index][R_Index]
            g = self.__rgba_values[before_index][G_Index]
            b = self.__rgba_values[before_index][B_Index]
            a = self.__rgba_values[before_index][Alpha_Index]
            
        else :

            after_fract = pos_delta / interval
            before_fract = 1.0 - after_fract

            r = int((float(self.__rgba_values[before_index][R_Index]) *
                     before_fract) \
                    + (float(self.__rgba_values[after_index][R_Index]) *
                       after_fract))
            
            g = int((float(self.__rgba_values[before_index][G_Index]) *
                     before_fract) \
                    + (float(self.__rgba_values[after_index][G_Index]) *
                       after_fract))
            
            b = int((float(self.__rgba_values[before_index][B_Index]) *
                     before_fract) \
                    + (float(self.__rgba_values[after_index][B_Index]) *
                       after_fract))
            
            a = int((float(self.__rgba_values[before_index][Alpha_Index]) *
                     before_fract) \
                    + (float(self.__rgba_values[after_index][Alpha_Index]) *
                       after_fract))

        # print value, r, g, b, a

        return (r, g, b, a)


    def __get_map_position(self, value) :

        import math

        before_index = None
        after_index = None

        position_delta = None
        interval_size = None

        if (not (value > self.__min_value)) :
            
            before_index = 0
            after_index = before_index

            position_delta = float(0.0)
            interval_size = float(0.0)
            
        elif (not (value < self.__max_value)) :
            
            before_index = self.__num_map_values - 1
            after_index = before_index

            position_delta = float(0.0)
            interval_size = float(0.0)

        else :

            scale_value = value

            if (self.__logscale) :
                scale_value = math.log10(scale_value)

            before_index = int(0)
            after_index = self.__num_map_values - 1

            while (before_index < (after_index - 1)) :

                test_index = (before_index + after_index) / int(2)

                test_value = self.__map_values[test_index]

                if (scale_value < test_value) :
                    after_index = test_index
                else :
                    before_index = test_index

            before_value = self.__map_values[before_index]
            after_value = self.__map_values[after_index]

            position_delta = scale_value - before_value
            interval_size = after_value - before_value

        return before_index, after_index, position_delta, interval_size
        

    def __read_map(self, filename, opacity=None) :

        import math

        from numeric_defs import numeric_zeros, numeric_int, numeric_float

        value_color_list = read_colormap(filename)

        num_colors = len(value_color_list)

        if (num_colors < 2) :
            raise RuntimeError("Error: Too few entries in colormap")

        self.__min_value = value_color_list[0].get_value()
        self.__max_value = value_color_list[num_colors - 1].get_value()

        if (self.__logscale) :
            if (not (self.__min_value > 0.0)) :
                raise ValueError("Error: Non-positive antilog value")

        if (opacity is not None) :

            for list_entry in value_color_list :
                list_entry.set_alpha(opacity)

        # Build the local color representations

        self.__num_map_values = num_colors

        self.__map_values = numeric_zeros(num_colors, numeric_float)
        
        self.__rgba_values = numeric_zeros((num_colors, color_vector_size),
                                           numeric_int)

        item_ptr = int(0)

        for list_entry in value_color_list :

            value = list_entry.get_value()

            if (self.__logscale) :
                value = math.log10(value)

            self.__map_values[item_ptr] = value

            self.__rgba_values[item_ptr][R_Index] = list_entry.get_r()
            self.__rgba_values[item_ptr][G_Index] = list_entry.get_g()
            self.__rgba_values[item_ptr][B_Index] = list_entry.get_b()
            self.__rgba_values[item_ptr][Alpha_Index] = list_entry.get_alpha()
            
            item_ptr += 1

