#!/usr/bin/env python

# drawimage.py

# By Nathan C. Hearn
#    November 6, 2006

try :
    import psyco
except ImportError :
    pass  # Oh, well ... no extra runtime optimizations for me ...


# File suffixes

png_suffix = str(".png")
csv_suffix = str(".csv")


# File reading functions

def compute_symmetry(symmetry_axis, data) :

    from numeric_defs import numeric_zeros, numeric_float

    if ((symmetry_axis < 0) or (symmetry_axis > 1)) :
        raise ValueError("Improper symmetry axis")

    # Numeric package already imported

    size = data.shape

    if (len(size) != 2) :
        raise RuntimeError("Data must be two-dimensional")

    m_size = size[0]
    n_size = size[1]

    max_m = m_size - 1
    max_n = n_size - 1

    sum_data = numeric_zeros(size, numeric_float)
    diff_data = numeric_zeros(size, numeric_float)

    for m in range(m_size) :
        for n in range(n_size) :

            symm_m = m
            symm_n = n

            if (symmetry_axis == 0) :
                symm_n = max_n - n
            else :
                symm_m = max_m - m

            primary_value = data[m][n]
            reflect_value = data[symm_m][symm_n]

            sum_data[m][n] = 0.5 * (primary_value + reflect_value)
            diff_data[m][n] = 0.5 * (primary_value - reflect_value)

    return sum_data, diff_data


def main() :

    from optparse import OptionParser
    from sys import exit, stderr
    import math

    import colormap
    from read_grid_datafile import read_file
    from save_grid_image import save_data, save_image_gray, save_image_colormap

    # Set up the command line parser

    usage = str("%prog [options] input_file1 [ input_file2 ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_prefix_str = str("output-prefix")
    argparser.add_option("--" + output_prefix_str, type="string",
                         dest="output_prefix", help="Output file prefix")
    argparser.set_defaults(output_prefix=None)

    min_value_str = str("min-value")
    argparser.add_option("--" + min_value_str, type="float",
                         dest="min_value", help="Image minimum data value")
    argparser.set_defaults(min_value=None)

    max_value_str = str("max-value")
    argparser.add_option("--" + max_value_str, type="float",
                         dest="max_value", help="Image maximum data value")
    argparser.set_defaults(max_value=None)

    row_range_str = str("row-range")
    argparser.add_option("--" + row_range_str, type="string",
                         dest="row_range",
                         help="Selected range of row indexes (min,max)")
    argparser.set_defaults(row_range=None)

    col_range_str = str("col-range")
    argparser.add_option("--" + col_range_str, type="string",
                         dest="col_range",
                         help="Selected range of column indexes (min,max)")
    argparser.set_defaults(col_range=None)

    log_scale_str = str("logscale")
    argparser.add_option("--" + log_scale_str, action="store_true",
                         dest="log_scale",
                         help="Use logarithm of data values for coloring")
    argparser.set_defaults(log_scale=False)

    color_map_str = str("colormap")
    argparser.add_option("--" + color_map_str, type="string",
                         dest="color_map_file", help="Colormap file")
    argparser.set_defaults(color_map_file=None)

    rescale_color_map_str = str("rescale-colormap")
    argparser.add_option("--" + rescale_color_map_str, action="store_true",
                         dest="rescale_colormap",
                         help="Rescale the colormap for the range of values")
    argparser.set_defaults(rescale_colormap=False)

    recenter_color_map_str = str("recenter-colormap")
    argparser.add_option("--" + recenter_color_map_str, type="float",
                         dest="colormap_center",
                         help="Recenter colormap about value")
    argparser.set_defaults(colormap_center=None)

    compute_symmetry_str = str("compute-symmetry")
    argparser.add_option("--" + compute_symmetry_str, type="int",
                         dest="symmetry_axis",
                         help="Specify symmetry axis (0 or 1)")
    argparser.set_defaults(symmetry_axis=None)

    transpose_image_str = str("transpose")
    argparser.add_option("--" + transpose_image_str, action="store_true",
                         dest="transpose_image", help="Transpose images")
    argparser.set_defaults(transpose_image=False)

    write_data_str = str("write-data")
    argparser.add_option("--" + write_data_str, action="store_true",
                         dest="write_data",
                         help="Write ASCII data instead of image")
    argparser.set_defaults(write_data=False)

    data_start_marker_str = str("data-start-marker")
    argparser.add_option("--" + data_start_marker_str, type="string",
                         dest="data_start_marker",
                         help="Marker to signal beginning of image data")
    argparser.set_defaults(data_start_marker=None)
    

    # Process the command line input

    opts, args = argparser.parse_args()

    master_output_prefix = opts.output_prefix

    min_value = opts.min_value
    max_value = opts.max_value

    row_range = None

    if (opts.row_range is not None) :

        row_elems = opts.row_range.split(",")

        if (len(row_elems) != 2) :
            raise RuntimeError("Improper number of elements for row selection")

        min_row = int(row_elems[0])
        max_row = int(row_elems[1])

        if (min_row < 0) :
            raise RuntimeError("Min row index out of range")

        if (min_row > max_row) :
            raise RuntimeError("Min row index can not exceed max row index")

        row_range = [ min_row, max_row ]

    col_range = None

    if (opts.col_range is not None) :

        col_elems = opts.col_range.split(",")

        if (len(col_elems) != 2) :
            raise RuntimeError("Improper number of elements for column "
                               + "selection")

        min_col = int(col_elems[0])
        max_col = int(col_elems[1])

        if (min_col < 0) :
            raise RuntimeError("Min column index out of range")

        if (min_col > max_col) :
            raise RuntimeError("Min column index can not exceed max column "
                               + "index")

        col_range = [ min_col, max_col ]

    logscale = opts.log_scale

    rescale_colormap = opts.rescale_colormap

    colormap_center = opts.colormap_center

    recenter_colormap = False
    
    if (colormap_center is not None) :
        recenter_colormap = True

    cmap = None

    if (opts.color_map_file is not None) :
        cmap = colormap.Colormap(opts.color_map_file, logscale)

    symmetry_axis = opts.symmetry_axis

    transpose_image = opts.transpose_image

    write_data = opts.write_data

    data_start_marker = opts.data_start_marker


    min_args = int(1)

    argc = len(args)

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n")
        exit(1)

    argptr = int(0)

    input_file_list = list()

    while (argptr < argc) :
        input_file_list.append(str(args[argptr]))
        argptr += 1

    multiple_files = False

    if (len(input_file_list) > 1) :
        multiple_files = True

    # Perform some basic checks

    if ((min_value is not None) and (max_value is not None)) :
        if (not (min_value < max_value)) :
            stderr.write("Error: Min value [ " + str(min_value)
                         + " ] must be less than max [ " + str(max_value)
                         + " ]\n\n")
            exit(-1)


    for input_file in input_file_list :
        
        ## Read the data

        data = read_file(input_file, row_range, col_range, data_start_marker)

        ## Set up the output prefix

        output_prefix = ""

        if (master_output_prefix is not None) :
            output_prefix = master_output_prefix

        if (multiple_files or (master_output_prefix is None)) :
            output_prefix += input_file

        ## Output the image

        if (not write_data) :
            
            output_file = output_prefix + png_suffix

            if (cmap is not None) :

                if (rescale_colormap) :

                    new_min_value = None

                    if (min_value is None) :
                        new_min_value = data.min()
                    else :
                        new_min_value = min_value

                    new_max_value = None

                    if (max_value is None) :
                        new_max_value = data.max()
                    else :
                        new_max_value = max_value

                    if (recenter_colormap) :

                        ## IGNORE LOG SCALE FOR NOW ...

                        if (logscale) :
                            stderr.write("\nError: Colormap recentering not "
                                         + "currently supported for log "
                                         + "scaling\n\n")
                            raise RuntimeError("Log recentering not supported")

                        ## Get the maximum distance from center to min/max

                        min_dist = math.fabs(colormap_center - new_min_value)
                        max_dist = math.fabs(colormap_center - new_max_value)

                        half_width = max(min_dist, max_dist)

                        new_min_value = colormap_center - half_width
                        new_max_value = colormap_center + half_width

                    cmap.rescale_colormap(new_min_value, new_max_value)

                elif (recenter_colormap) :

                    ## IGNORE LOG SCALE FOR NOW ...

                    if (logscale) :
                        stderr.write("\nError: Colormap recentering not "
                                     + "currently supported for log "
                                     + "scaling\n\n")
                        raise RuntimeError("Log recentering not supported")

                    current_min_value = cmap.get_min_value()
                    current_max_value = cmap.get_max_value()

                    current_center = 0.5 * (current_min_value
                                            + current_max_value)

                    new_center_delta = colormap_center - current_center

                    new_min_value = current_min_value + new_center_delta
                    new_max_value = current_max_value + new_center_delta

                    cmap.rescale_colormap(new_min_value, new_max_value)
                
                save_image_colormap(output_file, data, cmap, transpose_image)
                
            else :

                ## ADD RECENTERING-CODE HERE!!!!

                if (recenter_colormap) :
                    stderr.write("\nError: Colormap recentering not currently "
                                 + "supported for implicit colormaps\n\n")
                    raise RuntimeError("Implicit colormap recentering not "
                                       + "currently supported\n\n")
                
                save_image_gray(output_file, data, min_value, max_value,
                                logscale=logscale,
                                transpose_image=transpose_image)

        else :

            output_file = None

            ## If only one input file and no output prefix, write to stdout

            if (multiple_files or (master_output_prefix is not None)) :
                output_file = output_prefix + csv_suffix

            file_comment = "Data from file [ " + input_file + " ]"

            save_data(data, filename=output_file, comment=file_comment)


        ## Output symmetry information

        if (symmetry_axis is not None) :

            sum_data, diff_data = compute_symmetry(symmetry_axis, data)

            if (write_data) :

                ## Always write to files

                sum_filename = output_prefix + "_sum" + csv_suffix
                diff_filename = output_prefix + "_diff" + csv_suffix

                sum_file_comment = "Sum data from file [ " + input_file \
                                   + " ] with symmetry axis [ " \
                                   + str(symmetry_axis) + " ]"

                diff_file_comment = "Difference data from file [ " \
                                    + input_file + " ] with symmetry axis [ " \
                                    + str(symmetry_axis) + " ]"

                save_data(sum_data, filename=sum_filename,
                          comment=sum_file_comment)
                save_data(diff_data, filename=diff_filename,
                          comment=diff_file_comment)

            else :

                sum_filename = output_prefix + "_sum" + png_suffix
                diff_filename = output_prefix + "_diff" + png_suffix

                save_image_gray(sum_filename, sum_data,
                                transpose_image=transpose_image)
                save_image_gray(diff_filename, diff_data,
                                transpose_image=transpose_image)


if (__name__ == "__main__") :
    main()
