#!/usr/bin/env python

# image_fft.py

# By Nathan C. Hearn
#    June 23, 2008
#
# Program for producing FFTs from image and grid data


try :
    import psyco
except ImportError :
    pass


png_suffix = str("png")
csv_suffix = str("csv")


def extract_line(grid_data, axis, index) :

    from numeric_defs import numeric_zeros, numeric_float

    shape = grid_data.shape

    if (len(shape) != 2) :
        raise RuntimeError("Only 2D data fields supported")

    if ((axis < 0) or (axis > 1)) :
        raise RuntimeError("Axis must be zero or one")

    line_length = shape[axis]

    cross_axis = (axis + 1) % 2

    cross_size = shape[cross_axis]

    if ((index < 0) or (index >= cross_size)) :
        raise RuntimeError("Extraction index out of range")

    line_data = None

    if (axis == 0) :
        line_data = grid_data[:, index]
    else :
        line_data = grid_data[index, :]

    return line_data


def main() :

    from optparse import OptionParser
    from sys import exit, stderr

    from read_grid_datafile import read_file
    from read_image_data import read_image_data
    from grid_fft_ops import fft_1d, split_real_imag_arrays, fft_1d_power
    from generate_plot import Plot
    from save_grid_image import save_data, save_image_gray


    ## Set up the command line parser

    usage = str("%prog [options] input_file1 [ input_file2 ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_prefix_str = str("output-prefix")
    argparser.add_option("--" + output_prefix_str, type="string",
                         dest="output_prefix", help="Output file prefix")
    argparser.set_defaults(output_prefix=None)

    row_range_str = str("row-range")
    argparser.add_option("--" + row_range_str, type="string",
                         dest="row_range",
                         help="Selected range of row indexes (min,max)")
    argparser.set_defaults(row_range=None)

    col_range_str = str("col-range")
    argparser.add_option("--" + col_range_str, type="string",
                         dest="col_range",
                         help="Selected range of column indexes (min,max)")
    argparser.set_defaults(col_range=None)

    read_image_str = str("read-image")
    argparser.add_option("--" + read_image_str, action="store_true",
                         dest="read_image", help="Read input file as image")
    argparser.set_defaults(read_image=False)

    fft_axis_str = str("fft-axis")
    argparser.add_option("--" + fft_axis_str, type="int",
                         dest="fft_axis",
                         help="Transformation axis for 1D FFTs")
    argparser.set_defaults(fft_axis=None)

    save_power_str = str("save-power")
    argparser.add_option("--" + save_power_str, action="store_true",
                         dest="save_power",
                         help="Compute and save frequency powers")
    argparser.set_defaults(save_power=False)

    plot_power_str = str("plot-power")
    argparser.add_option("--" + plot_power_str, type="string",
                         dest="plot_power",
                         help="Plot power for comma-separated list of "
                              + "row/column indexes")
    argparser.set_defaults(plot_power=None)

    save_images_str = str("save-images")
    argparser.add_option("--" + save_images_str, action="store_true",
                         dest="save_images",
                         help="Save transformed grids and powers as images")
    argparser.set_defaults(save_images=False)


    ## Process the input

    opts, args = argparser.parse_args()

    min_args = int(1)

    argc = len(args)

    if (argc < min_args) :
        argparser.print_help(file=stderr)
        exit(1)

    ## Read the required arguments

    argptr = int(0)

    input_file_list = list()

    while (argptr < argc) :
        input_file_list.append(str(args[argptr]))
        argptr += 1

    ## Read optional arguments

    user_output_prefix = opts.output_prefix

    row_range = None

    if (opts.row_range is not None) :

        row_elems = opts.row_range.split(",")

        if (len(row_elems) != 2) :
            stderr.write("\nError: Improper number of elements for row "
                         + "selection\n\n")
            exit(-1)

        min_row = int(row_elems[0])
        max_row = int(row_elems[1])

        if (min_row < 0) :
            stderr.write("\nError: Min row index out of range\n\n")
            exit(-1)

        if (min_row > max_row) :
            stderr.write("\nError: Min row index can not exceed max row "
                         + "index\n\n")
            exit(-1)

        row_range = [ min_row, max_row ]

    col_range = None

    if (opts.col_range is not None) :

        col_elems = opts.col_range.split(",")

        if (len(col_elems) != 2) :
            stderr.write("\nError: Improper number of elements for column "
                         + "selection\n\n")
            exit(-1)

        min_col = int(col_elems[0])
        max_col = int(col_elems[1])

        if (min_col < 0) :
            stderr.write("\nError: Min column index out of range\n\n")
            exit(-1)

        if (min_col > max_col) :
            stderr.write("\nError: Min column index can not exceed max column "
                         + "index\n\n")
            exit(-1)

        col_range = [ min_col, max_col ]

    read_image = opts.read_image

    compute_2D_fft = True

    fft_1D_axis = opts.fft_axis

    if (fft_1D_axis is not None) :

        if ((fft_1D_axis < 0) or (fft_1D_axis > 1)) :
            stderr.write("\nError: 1D FFT axis must be either zero or one\n\n")
            exit(-1)

        compute_2D_fft = False

    save_fft_power = opts.save_power

    if (save_fft_power and compute_2D_fft) :
        stderr.write("\nError: Currently, FFT power only available for 1D "
                     + "FFTs\n\n")
        exit(-1)

    plot_power_list = None

    if (opts.plot_power is not None) :

        if (compute_2D_fft) :
            stderr.write("\nError: Plot power option only available for 1D "
                         + "FFTs\n\n")
            exit(-1)

        if (not save_fft_power) :
            stderr.write("\nError: Plot power option requires save-power "
                         + "option\n\n")
            exit(-1)

        plot_power_list = list()

        plot_power_entries = opts.plot_power.split(",")

        for word in plot_power_entries :

            plot_power_list.append(int(word))

    save_images = opts.save_images


    ## Process the images

    for input_file in input_file_list :

        ## Read the data

        data = None

        if (read_image) :
            data = read_image_data(input_file, row_range, col_range)
        else :
            data = read_file(input_file, row_range, col_range)

        ## Set up the output prefix

        output_prefix = str()

        if (user_output_prefix is not None) :
            output_prefix += user_output_prefix + "_"

        output_prefix += input_file
        
        ## Process the data

        if (compute_2D_fft) :

            trans_data = fft_1d(1, fft_1d(0, data))

            real_trans_data, imag_trans_data \
                = split_real_imag_arrays(trans_data)

            ## Build the output prefix

            fft_output_prefix = output_prefix + "_2d_fft"

            ## Save the data

            real_output_filename = fft_output_prefix + "_real." + csv_suffix
            imag_output_filename = fft_output_prefix + "_imag." + csv_suffix

            real_file_comment = "2D FFT real values from file " + input_file
            imag_file_comment = "2D FFT imaginary values from file " \
                                + input_file

            save_data(real_trans_data, real_output_filename,
                      comment=real_file_comment)

            save_data(imag_trans_data, imag_output_filename,
                      comment=imag_file_comment)

            if (save_images) :

                real_output_filename \
                    = fft_output_prefix + "_real." + png_suffix

                imag_output_filename \
                    = fft_output_prefix + "_imag." + png_suffix

                save_image_gray(real_output_filename, real_trans_data)
                save_image_gray(imag_output_filename, imag_trans_data)

        else :

            trans_data = fft_1d(fft_1D_axis, data)

            real_trans_data, imag_trans_data \
                = split_real_imag_arrays(trans_data)

            ## Build the output prefix

            fft_output_prefix = output_prefix + "_1d_fft_" + str(fft_1D_axis)

            ## Save the FFT data

            real_output_filename = fft_output_prefix + "_real." + csv_suffix
            imag_output_filename = fft_output_prefix + "_imag." + csv_suffix

            real_file_comment = "1D FFT real values for axis " \
                + str(fft_1D_axis) + " of file " + input_file

            imag_file_comment = "1D FFT imaginary values for axis " \
                + str(fft_1D_axis) + " of file " + input_file

            save_data(real_trans_data, real_output_filename,
                      comment=real_file_comment)

            save_data(imag_trans_data, imag_output_filename,
                      comment=imag_file_comment)

            if (save_images) :

                real_output_filename \
                    = fft_output_prefix + "_real." + png_suffix

                imag_output_filename \
                    = fft_output_prefix + "_imag." + png_suffix

                save_image_gray(real_output_filename, real_trans_data)
                save_image_gray(imag_output_filename, imag_trans_data)

            ## Save the power data

            if (save_fft_power) :

                power_data = fft_1d_power(fft_1D_axis, trans_data)

                output_filename = fft_output_prefix + "_power." + csv_suffix

                file_comment = "1D FFT power for axis " + str(fft_1D_axis) \
                    + " of file " + input_file

                save_data(power_data, output_filename, comment=file_comment)

                if (save_images) :

                    output_filename \
                        = fft_output_prefix + "_power." + png_suffix

                    save_image_gray(output_filename, power_data, logscale=True)

                if (plot_power_list is not None) :

                    power_plot = Plot("FFT Power for file " + input_file
                                      + " - Axis " + str(fft_1D_axis),
                                      "k", "Spectral Power")

                    power_plot.set_logscale_x()

                    for extract_index in plot_power_list :

                        line_data = extract_line(power_data, fft_1D_axis,
                                                 extract_index)

                        ## Toss the first (index zero) element

                        line_data = line_data[1:]

                        ## Build the k array

                        k_values = range(1, len(data))

                        ## Add the data to the plot

                        power_plot.add_plot(k_values, line_data,
                                            legend_text="Line index "
                                                        + str(extract_index))

                    ## Save the plot

                    output_filename = fft_output_prefix + "_power_plot." \
                                      + png_suffix

                    power_plot.show_plot(output_filename)


if (__name__ == "__main__") :
    main()
