# save_grid_image.py

# By Nathan C. Hearn
#    June 24, 2008
#
# Routines for saving grid data as an ASCII file or an image


def save_data(data, filename=None, comment=None) :

    from sys import stdout

    size = data.shape

    if (len(size) != 2) :
        raise RuntimeError("Data must be two-dimensional")

    m_size = size[0]
    n_size = size[1]

    output = stdout

    if (filename is not None) :
        output = file(filename, "w")

    if (comment is not None) :
        output.write("# " + comment + "\n")

    for m in range(m_size) :
        
        for n in range(n_size) :
            output.write(str(data[m][n]) + " ")

        output.write("\n")


def save_image_gray(filename, data, min_value=None, max_value=None,
                    logscale=False, transpose_image=False) :

    from PIL import Image
    from sys import stderr
    import math

    from numeric_defs import numeric_zeros, numeric_int, numeric_transpose

    # NOTE
    #
    # Generally, the data read by read_file is stored with the x-
    # coordinate varying the fastest.  However, ImageMagick expects the
    # y-coordinate to be the fastest varying.  With this assumption about
    # the initial data, this function transposes the data to keep the x axis
    # aligned with the horizontal image axis.
    #
    # Furthermore, the data read by read_file starts with the smallest value
    # for y, whereas ImageMagic expects image data to start with the largest
    # value for y.  This function also accounts for this difference.
    #
    # The transpose_image flag causes this horizontal-x-axis data to be
    # tranposed (again) to yield vertical-x-axis data

    ret_val = int(-1)  ## Error state

    work_data = data

    if (transpose_image) :
        work_data = numeric_transpose(data)

    size = work_data.shape

    if (len(size) != 2) :
        raise ValueError("Data must be two-dimensional")

    i_size = size[0]
    j_size = size[1]

    # Find the max and min data values

    if ((i_size > 0) and (j_size > 0)) :

        image_min_val = min_value
        image_max_val = max_value

        single_value = False

        if ((image_min_val is None) or (image_max_val is None)) :

            data_max_val = work_data[0][0]
            data_min_val = data_max_val

            data_min_pos_val = None

            for i in range(i_size) :
                for j in range(j_size) :

                    val = work_data[i][j]

                    if (val < data_min_val) :
                        data_min_val = val
                    elif (val > data_max_val) :
                        data_max_val = val

                    if (val > 0.0) :
                        if (data_min_pos_val is None) :
                            data_min_pos_val = val
                        elif (val < data_min_pos_val) :
                            data_min_pos_val = val

            if (image_min_val is None) :

                image_min_val = data_min_val

                if (logscale) :
                    image_min_val = data_min_pos_val

            if (image_max_val is None) :
                image_max_val = data_max_val

        if ((image_min_val is None) or (image_max_val is None)) :
            
            if (image_min_val is None) :
                stderr.write("\nError: Image min value not found\n\n")

            if (image_max_val is None) :
                stderr.write("\nError: Image max value not found\n\n")
                
            raise RuntimeError("Image value limits not found")

        if (image_min_val > image_max_val) :
            raise ValueError("Error: Min value -- max value mismatch")

        if (logscale) :

            if (not (image_min_val > 0.0)) :
                raise ValueError("Error: Logscale requires positive min value")

            image_min_val = math.log(image_min_val)
            image_max_val = math.log(image_max_val)

        # Store the data in an array

        pixel_count = i_size * j_size

        pixel_array = numeric_zeros(pixel_count, numeric_int)

        if (image_min_val < image_max_val) :

            offset = image_min_val
            factor = 255.0 / (image_max_val - image_min_val)

            ptr = int(0)

            # Transpose for the image by using column-major output

            for j in range(j_size) :

                j_flip = j_size - (j + 1)  # Reverse j-values

                for i in range(i_size) :

                    # i_flip = i_size - (i + 1)  # Reverse y-values

                    raw_data = work_data[i][j_flip]

                    if (logscale) :
                        if (raw_data > 0.0) :
                            raw_data = math.log(raw_data)
                        else :
                            raw_data = image_min_val

                    pixel = int((raw_data - offset) * factor)

                    if (pixel < 0) :
                        pixel = 0
                    elif (pixel > 255) :
                        pixel = 255

                    pixel_array[ptr] = pixel
                    ptr += 1

        new_im = Image.new("L", (i_size, j_size))  # Width, Height

        new_im.putdata(pixel_array)

        new_im.save(filename)

        ret_val = int(0)  ## OK

    else :

        stderr.write("\nWarning: No data for image " + filename + "\n\n")

    return ret_val


def save_image_colormap(filename, data, colormap, transpose_image=False) :

    from PIL import Image
    import math

    from numeric_defs import numeric_zeros, numeric_int, numeric_float, \
         numeric_transpose

    # NOTE
    #
    # Generally, the data read by read_file is stored with the x-
    # coordinate varying the fastest.  However, ImageMagick expects the
    # y-coordinate to be the fastest varying.  With this assumption about
    # the initial data, this function transposes the data to keep the x axis
    # aligned with the horizontal image axis.
    #
    # Furthermore, the data read by read_file starts with the smallest value
    # for y, whereas ImageMagic expects image data to start with the largest
    # value for y.  This function also accounts for this difference.
    #
    # The transpose_image flag causes this horizontal-x-axis data to be
    # tranposed (again) to yield vertical-x-axis data

    ret_val = int(-1)  ## Error state

    work_data = data

    if (transpose_image) :
        work_data = numeric_transpose(data)

    size = work_data.shape

    if (len(size) != 2) :
        raise ValueError("Data must be two-dimensional")

    i_size = size[0]
    j_size = size[1]

    # Store the data in an array

    pixel_count = i_size * j_size

    pixel_array = list()

    ptr = int(0)

    # Transpose for the image by using column-major output

    for j in range(j_size) :

        j_flip = j_size - (j + 1)  # Reverse the x-values

        for i in range(i_size) :

            # i_flip = i_size - (i + 1)  # Reverse y-values
        
            raw_data = work_data[i][j_flip]

            color_tuple = colormap.get_color(raw_data)

            pixel_array.append(color_tuple)

    new_im = Image.new("RGBA", (i_size, j_size))  # Width, Height

    new_im.putdata(pixel_array)

    new_im.save(filename)


