// C++ program file star_generator.cpp

/*
  By Nathan C. Hearn
     June 14, 2008

  Generator for Star objects.
*/


#include "star_generator.hpp"
#include "quickflash_color_color.hpp"
#include "star.hpp"


// Class StarGenerator


StarGenerator::StarGenerator() :
  width(1000.0), height(1000.0), x_pos_gen(), y_pos_gen(), radial_scale_gen(),
  max_radius_fact(4.0), hue_gen(), saturation_gen(), brightness_gen(),
  spike_count(0), spike_scale(4.0), spike_phase(0.0), spike_power(4.0)
  { }


StarGenerator::StarGenerator(const double image_width, 
			     const double image_height, 
			     const double min_radial_scale, 
			     const double max_radial_scale,
			     const double max_radius_factor, 
			     const double min_saturation, 
			     const double max_saturation, 
			     const double min_brightness, 
			     const double max_brightness,
			     const unsigned int num_spikes, 
			     const double spike_scale_factor, 
			     const double spike_phase_deg, 
			     const double spike_steepening_power) :
  width(1000.0), height(1000.0), x_pos_gen(), y_pos_gen(), radial_scale_gen(),
  max_radius_fact(4.0), hue_gen(), saturation_gen(), brightness_gen(),
  spike_count(0), spike_scale(4.0), spike_phase(0.0), spike_power(4.0)
  { 
  reset(image_width, image_height, min_radial_scale, max_radial_scale, 
	max_radius_factor, min_saturation, max_saturation, min_brightness, 
	max_brightness, num_spikes, spike_scale_factor, spike_phase_deg,
	spike_steepening_power); 
  }


StarGenerator::StarGenerator(const StarGenerator & source) :
  width(1000.0), height(1000.0), x_pos_gen(), y_pos_gen(), radial_scale_gen(),
  max_radius_fact(4.0), hue_gen(), saturation_gen(), brightness_gen(),
  spike_count(0), spike_scale(4.0), spike_phase(0.0), spike_power(4.0)
  { reset(source); }


void StarGenerator::reset()
  {
  width = 1000.0;
  height = 1000.0;

  x_pos_gen.reset();
  y_pos_gen.reset();

  radial_scale_gen.reset();

  max_radius_fact = 4.0;

  hue_gen.reset();
  saturation_gen.reset();
  brightness_gen.reset();

  spike_count = 0;
  spike_scale = 4.0;
  spike_phase = 0.0;
  spike_power = 4.0;
  }


void StarGenerator::reset(const double image_width, 
			  const double image_height, 
			  const double min_radial_scale, 
			  const double max_radial_scale,
			  const double max_radius_factor, 
			  const double min_saturation, 
			  const double max_saturation, 
			  const double min_brightness, 
			  const double max_brightness,
			  const unsigned int num_spikes, 
			  const double spike_scale_factor, 
			  const double spike_phase_deg, 
			  const double spike_steepening_power)
  {
  // ERROR CHECKING???

  width = image_width;
  height = image_height;

  x_pos_gen.reset(0.0, image_width);
  y_pos_gen.reset(0.0, image_height);

  radial_scale_gen.reset(min_radial_scale, max_radial_scale);

  max_radius_fact = max_radius_factor;

  hue_gen.reset(0.0, 1.0);
  saturation_gen.reset(min_saturation, max_saturation);
  brightness_gen.reset(min_brightness, max_brightness);

  spike_count = num_spikes;
  spike_scale = spike_scale_factor;
  spike_phase = spike_phase_deg;
  spike_power = spike_steepening_power;
  }


void StarGenerator::reset(const StarGenerator & source)
  {
  if (&source != this)
    {
    width = source.width;
    height = source.height;

    x_pos_gen = source.x_pos_gen;
    y_pos_gen = source.y_pos_gen;

    radial_scale_gen = source.radial_scale_gen;

    max_radius_fact = source.max_radius_fact;

    hue_gen = source.hue_gen;
    saturation_gen = source.saturation_gen;
    brightness_gen = source.brightness_gen;

    spike_count = source.spike_count;
    spike_scale = source.spike_scale;
    spike_phase = source.spike_phase;
    spike_power = source.spike_power;
    }
  }


void StarGenerator::get_star(Star & star, const bool protect_border) const
  {
  const double radial_scale = radial_scale_gen.get_value();
  const double max_radius = radial_scale * max_radius_fact;

  double x = 0.0;
  double y = 0.0;

  bool x_pos_ok = false;

  while (!x_pos_ok)
    {
    x = x_pos_gen.get_value();

    if (protect_border)
      {
      if ((x > max_radius) && ((width - x) > max_radius))
	x_pos_ok = true;
      }
    else
      x_pos_ok = true;
    }

  bool y_pos_ok = false;

  while (!y_pos_ok)
    {
    y = y_pos_gen.get_value();

    if (protect_border)
      {
      if ((y > max_radius) && ((height - y) > max_radius))
	y_pos_ok = true;
      }
    else
      y_pos_ok = true;
    }

  const double hue = hue_gen.get_value();
  const double sat = saturation_gen.get_value();
  const double bright = brightness_gen.get_value();

  QuickFlash::Color::Color star_color(1.0);  // Default star is white

  star_color.set_color_hsv(hue, sat, bright);

  star.reset(x, y, radial_scale, max_radius, star_color, spike_count, 
	     spike_scale, spike_phase, spike_power);
  }
